<?php
$code = "<?php __halt_compiler();\x01?>\x02";
$tokens = token_get_all($code);

var_dump($tokens);

$code = '';
foreach ($tokens as $t)
{
	$code .= isset($t[1]) ? $t[1] : $t;
}
var_dump($code);
?>
