#include <stdlib.h>
#include <stdio.h>

char * TESTS[] =
 {
    "ext/sockets/tests/mcast_ipv4_recv.phpt",
    "ext/sockets/tests/bug46360.phpt",
    "ext/sockets/tests/socket_create_pair-wrongparams.phpt",
    "ext/sockets/tests/ipv4loop.phpt",
    "ext/sockets/tests/socket_import_stream-2.phpt",
    "ext/sockets/tests/socket_bind_params.phpt",
    "ext/sockets/tests/socket_write_params.phpt",
    "ext/sockets/tests/socket_connect_error.phpt",
    "ext/sockets/tests/socket_import_stream-3.phpt",
    "ext/sockets/tests/socket_listen_params.phpt",
    "ext/sockets/tests/socket_select-wrongparams-2.phpt",
    "ext/sockets/tests/socket_create_params.phpt",
    "ext/sockets/tests/socket_set_nonblock-wrongparams.phpt",
    "ext/sockets/tests/socket_select-wrongparams-1.phpt",
    "ext/sockets/tests/mcast_ipv4_send.phpt",
    "ext/sockets/tests/socket_create_listen_used.phpt",
    "ext/sockets/tests/socket_create_listen-wrongparams.phpt",
    "ext/sockets/tests/socket_bind.phpt",
    "ext/sockets/tests/socket_set_option_rcvtimeo.phpt",
    "ext/sockets/tests/socket_import_stream-5.phpt",
    "ext/sockets/tests/socket_create_listen-nobind.phpt",
    "ext/sockets/tests/socket_getpeername_ipv6loop.phpt",
    "ext/sockets/tests/unixloop.phpt",
    "ext/sockets/tests/socket_select.phpt",
    "ext/sockets/tests/socket_getpeername.phpt",
    "ext/sockets/tests/socket_getpeername_ipv4loop.phpt",
    "ext/sockets/tests/ipv6loop.phpt",
    "ext/sockets/tests/socket_accept-wrongparams.phpt",
    "ext/sockets/tests/socket_select-wrongparams-3.phpt",
    "ext/sockets/tests/socket_getsockname.phpt",
    "ext/sockets/tests/socket_set_block-retval.phpt",
    "ext/sockets/tests/socket_listen-wrongparams.phpt",
    "ext/sockets/tests/socket_create_listen.phpt",
    "ext/sockets/tests/mcast_ipv6_recv_limited.phpt",
    "ext/sockets/tests/socket_read_params.phpt",
    "ext/sockets/tests/socket_strerror.phpt",
    "ext/sockets/tests/socket_set_nonblock-retval.phpt",
    "ext/sockets/tests/socket_sentto_recvfrom_ipv4_udp.phpt",
    "ext/sockets/tests/socket_select-wrongparams-4.phpt",
    "ext/sockets/tests/socket_create_listen_params.phpt",
    "ext/sockets/tests/socket_set_option_sndtimeo.phpt",
    "ext/sockets/tests/socket_set_block-wrongparams.phpt",
    "ext/sockets/tests/bug51958.phpt",
    "ext/sockets/tests/mcast_ipv6_send.phpt",
    "ext/sockets/tests/socket_set_option_seolinger.phpt",
    "ext/sockets/tests/mcast_ipv6_recv.phpt",
    "ext/sockets/tests/socket_close_params.phpt",
    "ext/sockets/tests/socket_import_stream-1.phpt",
    "ext/sockets/tests/socket_sentto_recvfrom_ipv6_udp.phpt",
    "ext/sockets/tests/socket_create_pair.phpt",
    "ext/sockets/tests/socket_import_stream-4.phpt",
    "ext/sockets/tests/socket_connect_params.phpt",
    "ext/sockets/tests/socket_sentto_recvfrom_unix.phpt",
    "ext/sockets/tests/socket_set_nonblock.phpt",
    "ext/sockets/tests/socket_set_option_error_socket_option.phpt",
    "ext/enchant/tests/broker_init.phpt",
    "ext/enchant/tests/broker_free.phpt",
    "ext/enchant/tests/dict_quick_check.phpt",
    "ext/enchant/tests/broker_describe.phpt",
    "ext/enchant/tests/broker_request_dict.phpt",
    "ext/enchant/tests/bug13181.phpt",
    "ext/sybase_ct/tests/bug28354.phpt",
    "ext/sybase_ct/tests/test_close_notopen.phpt",
    "ext/sybase_ct/tests/bug30312-withfree.phpt",
    "ext/sybase_ct/tests/bug30312.phpt",
    "ext/sybase_ct/tests/test_msghandler_handled.phpt",
    "ext/sybase_ct/tests/test_unbuffered_no_full_fetch.phpt",
    "ext/sybase_ct/tests/test_connection_caching.phpt",
    "ext/sybase_ct/tests/test_unbuffered_query.phpt",
    "ext/sybase_ct/tests/test_query_nostore.phpt",
    "ext/sybase_ct/tests/bug27843.phpt",
    "ext/sybase_ct/tests/bug43578.phpt",
    "ext/sybase_ct/tests/bug29064.phpt",
    "ext/sybase_ct/tests/test_close_default.phpt",
    "ext/sybase_ct/tests/test_long.phpt",
    "ext/sybase_ct/tests/test_fetch_object.phpt",
    "ext/sybase_ct/tests/bug22403.phpt",
    "ext/sybase_ct/tests/test_msghandler.phpt",
    "ext/sybase_ct/tests/test_close.phpt",
    "ext/sybase_ct/tests/test_fields.phpt",
    "ext/sybase_ct/tests/bug26407.phpt",
    "ext/sybase_ct/tests/test_connectionbased_msghandler.phpt",
    "ext/sybase_ct/tests/test_connect.phpt",
    "ext/sybase_ct/tests/test_types.phpt",
    "ext/sybase_ct/tests/test_appname.phpt",
    "ext/sybase_ct/tests/bug6339.phpt",
    "ext/skeleton/tests/001.phpt",
    "ext/sqlite3/tests/sqlite3_32_createAggregate_paramCount.phpt",
    "ext/sqlite3/tests/sqlite3_close_with_params.phpt",
    "ext/sqlite3/tests/sqlite3_22_loadextension.phpt",
    "ext/sqlite3/tests/sqlite3_33_reset.phpt",
    "ext/sqlite3/tests/sqlite3_19_columninfo.phpt",
    "ext/sqlite3/tests/sqlite3_32_last_insert_rowid_param.phpt",
    "ext/sqlite3/tests/sqlite3_close_error.phpt",
    "ext/sqlite3/tests/bug45798.phpt",
    "ext/sqlite3/tests/sqlite3_27_reset_prepared_stmt_result.phpt",
    "ext/sqlite3/tests/sqlite3stmt_reset_params.phpt",
    "ext/sqlite3/tests/sqlite3_loadextension_with_wrong_param.phpt",
    "ext/sqlite3/tests/sqlite3_21_security.phpt",
    "ext/sqlite3/tests/sqlite3_06_prepared_stmt.phpt",
    "ext/sqlite3/tests/sqlite3result_fetcharray_with_two_params_fails.phpt",
    "ext/sqlite3/tests/sqlite3_prepare_with_empty_string.phpt",
    "ext/sqlite3/tests/sqlite3_version_noparam.phpt",
    "ext/sqlite3/tests/sqlite3_31_open.phpt",
    "ext/sqlite3/tests/sqlite3_02_create.phpt",
    "ext/sqlite3/tests/sqlite3_11_numrows.phpt",
    "ext/sqlite3/tests/sqlite3_07_prepared_stmt.phpt",
    "ext/sqlite3/tests/sqlite3_12_unfinalized_stmt_cleanup.phpt",
    "ext/sqlite3/tests/sqlite3_01_open.phpt",
    "ext/sqlite3/tests/sqlite3_18_changes.phpt",
    "ext/sqlite3/tests/sqlite3_13_skip_all_cleanup.phpt",
    "ext/sqlite3/tests/sqlite3_14_querysingle.phpt",
    "ext/sqlite3/tests/sqlite3_26_reset_prepared_stmt.phpt",
    "ext/sqlite3/tests/sqlite3_enable_exceptions.phpt",
    "ext/sqlite3/tests/sqlite3_15_open_error.phpt",
    "ext/sqlite3/tests/sqlite3_02_open.phpt",
    "ext/sqlite3/tests/sqlite3_32_changes.phpt",
    "ext/sqlite3/tests/sqlite3_31_changes.phpt",
    "ext/sqlite3/tests/sqlite3_lasterrormsg_with_params.phpt",
    "ext/sqlite3/tests/bug47159.phpt",
    "ext/sqlite3/tests/sqlite3stmt_paramCount_basic.phpt",
    "ext/sqlite3/tests/sqlite3_34_load_extension_ext_dir.phpt",
    "ext/sqlite3/tests/sqlite3_17_version.phpt",
    "ext/sqlite3/tests/sqlite3_04_update.phpt",
    "ext/sqlite3/tests/sqlite3_open_empty_string.phpt",
    "ext/sqlite3/tests/sqlite3_exec_wrongparams.phpt",
    "ext/sqlite3/tests/sqlite3_prepare_faultystmt.phpt",
    "ext/sqlite3/tests/sqlite3_30_blobopen.phpt",
    "ext/sqlite3/tests/sqlite3stmt_paramCount_error.phpt",
    "ext/sqlite3/tests/sqlite3_16_select_no_results.phpt",
    "ext/sqlite3/tests/sqlite3_05_delete.phpt",
    "ext/sqlite3/tests/sqlite3_24_last_insert_rowid.phpt",
    "ext/sqlite3/tests/sqlite3_28_clear_bindings.phpt",
    "ext/sqlite3/tests/sqlite3_29_createfunction.phpt",
    "ext/sqlite3/tests/bug53463.phpt",
    "ext/sqlite3/tests/sqlite3_querysingle_error.phpt",
    "ext/sqlite3/tests/sqlite3_03_insert.phpt",
    "ext/sqlite3/tests/sqlite3result_reset_with_params_fails.phpt",
    "ext/sqlite3/tests/sqlite3_10_bound_value_name.phpt",
    "ext/sqlite3/tests/sqlite3_25_create_aggregate.phpt",
    "ext/sqlite3/tests/sqlite3_openblob_wrongparams.phpt",
    "ext/sqlite3/tests/sqlite3result_numcolumns_error.phpt",
    "ext/sqlite3/tests/sqlite3_lasterrorcode_with_params.phpt",
    "ext/sqlite3/tests/sqlite3_query_error.phpt",
    "ext/sqlite3/tests/sqlite3_08_udf.phpt",
    "ext/sqlite3/tests/sqlite3_prepared_stmt_clear_with_params.phpt",
    "ext/sqlite3/tests/sqlite3_09_blob_bound_param.phpt",
    "ext/sqlite3/tests/sqlite3_20_error.phpt",
    "ext/sqlite3/tests/sqlite3_33_createAggregate_notcallable.phpt",
    "ext/sqlite3/tests/sqlite3_prepare_wrongparams.phpt",
    "ext/sqlite3/tests/sqlite3_prepare_001.phpt",
    "ext/sqlite3/tests/sqlite3_35_stmt_readonly.phpt",
    "ext/sqlite3/tests/sqlite3_33_load_extension_param.phpt",
    "ext/sqlite3/tests/sqlite3_23_escape_string.phpt",
    "ext/xmlrpc/tests/bug45555.phpt",
    "ext/xmlrpc/tests/bug42189.phpt",
    "ext/xmlrpc/tests/bug51288.phpt",
    "ext/xmlrpc/tests/bug18916.phpt",
    "ext/xmlrpc/tests/bug50761.phpt",
    "ext/xmlrpc/tests/bug40576_64bit.phpt",
    "ext/xmlrpc/tests/bug47818.phpt",
    "ext/xmlrpc/tests/bug40576.phpt",
    "ext/xmlrpc/tests/001.phpt",
    "ext/xmlrpc/tests/bug50282.phpt",
    "ext/xmlrpc/tests/bug44996.phpt",
    "ext/xmlrpc/tests/bug42736.phpt",
    "ext/xmlrpc/tests/bug38431.phpt",
    "ext/xmlrpc/tests/bug37057.phpt",
    "ext/xmlrpc/tests/bug45556.phpt",
    "ext/xmlrpc/tests/002.phpt",
    "ext/xmlrpc/tests/bug50285.phpt",
    "ext/xmlrpc/tests/bug45226.phpt",
    "ext/pdo_oci/tests/pdo_oci_quote1.phpt",
    "ext/pdo_oci/tests/pdo_oci_attr_client.phpt",
    "ext/pdo_oci/tests/pdo_oci_attr_prefetch_1.phpt",
    "ext/pdo_oci/tests/pdo_oci_attr_prefetch_2.phpt",
    "ext/pdo_oci/tests/bug_33707.phpt",
    "ext/pdo_oci/tests/bug41996.phpt",
    "ext/pdo_oci/tests/pdo_oci_stream_1.phpt",
    "ext/pdo_oci/tests/bug46274_2.phpt",
    "ext/pdo_oci/tests/pdo_oci_attr_autocommit_2.phpt",
    "ext/pdo_oci/tests/pecl_bug_6364.phpt",
    "ext/pdo_oci/tests/pdo_oci_attr_autocommit_1.phpt",
    "ext/pdo_oci/tests/bug46274.phpt",
    "ext/pdo_oci/tests/common.phpt",
    "ext/pdo_oci/tests/pdo_oci_attr_autocommit_3.phpt",
    "ext/pdo_oci/tests/pdo_oci_attr_server.phpt",
    "ext/pdo_oci/tests/bug44301.phpt",
    "ext/pdo_oci/tests/pdo_oci_attr_nulls_1.phpt",
    "ext/pdo_oci/tests/pdo_oci_attr_case.phpt",
    "ext/pdo_oci/tests/pdo_oci_fread_1.phpt",
    "ext/pdo_oci/tests/pecl_bug_11345.phpt",
    "ext/pdo_oci/tests/pdo_oci_attr_drivername.phpt",
    "ext/date/tests/bug32086.phpt",
    "ext/date/tests/bug34087.phpt",
    "ext/date/tests/DateTime_setTime_variation2.phpt",
    "ext/date/tests/DateTime_diff_add_sub-spring-type3-type2.phpt",
    "ext/date/tests/bug36988.phpt",
    "ext/date/tests/gmstrftime_variation22.phpt",
    "ext/date/tests/date_timezone_set_variation2.phpt",
    "ext/date/tests/checkdate_variation2.phpt",
    "ext/date/tests/DateTime_format_error.phpt",
    "ext/date/tests/oo_001.phpt",
    "ext/date/tests/date_create_from_format_basic2.phpt",
    "ext/date/tests/DateTime_format_basic1.phpt",
    "ext/date/tests/bug49059.phpt",
    "ext/date/tests/010.phpt",
    "ext/date/tests/bug33536.phpt",
    "ext/date/tests/date-parse-by-format001.phpt",
    "ext/date/tests/DateTimeZone_getOffset_variation1.phpt",
    "ext/date/tests/bug33578.phpt",
    "ext/date/tests/bug43003.phpt",
    "ext/date/tests/gmstrftime_basic.phpt",
    "ext/date/tests/bug27780.phpt",
    "ext/date/tests/DatePeriod_wrong_constructor.phpt",
    "ext/date/tests/DateTime_setDate_variation3.phpt",
    "ext/date/tests/date_sunrise_variation4.phpt",
    "ext/date/tests/bug41599.phpt",
    "ext/date/tests/oo_002.phpt",
    "ext/date/tests/DateTime_diff_add_sub-spring-type3-type3.phpt",
    "ext/date/tests/time_error.phpt",
    "ext/date/tests/strtotime2.phpt",
    "ext/date/tests/005.phpt",
    "ext/date/tests/DateTimeZone_getTransitions_basic1.phpt",
    "ext/date/tests/date_isodate_set_variation2.phpt",
    "ext/date/tests/timezone_offset_get_error.phpt",
    "ext/date/tests/gmmktime_variation5.phpt",
    "ext/date/tests/DateTime_setDate_basic1.phpt",
    "ext/date/tests/bug48058.phpt",
    "ext/date/tests/gmstrftime_variation7.phpt",
    "ext/date/tests/getdate_basic.phpt",
    "ext/date/tests/bug52454.phpt",
    "ext/date/tests/gmstrftime_variation11.phpt",
    "ext/date/tests/timezone_name_from_abbr_basic1.phpt",
    "ext/date/tests/014.phpt",
    "ext/date/tests/gmmktime_variation6.phpt",
    "ext/date/tests/bug43808.phpt",
    "ext/date/tests/DateTime_clone_basic4.phpt",
    "ext/date/tests/date_time_set_variation3.phpt",
    "ext/date/tests/bug29595.phpt",
    "ext/date/tests/bug41523-64bit.phpt",
    "ext/date/tests/DateTime_diff_add_sub-fall-type3-type3.phpt",
    "ext/date/tests/getdate_error.phpt",
    "ext/date/tests/gmstrftime_variation8.phpt",
    "ext/date/tests/localtime_basic.phpt",
    "ext/date/tests/gmstrftime_variation18.phpt",
    "ext/date/tests/bug41842.phpt",
    "ext/date/tests/bug46108.phpt",
    "ext/date/tests/date_parse_basic1.phpt",
    "ext/date/tests/DateTimeZone_clone_basic3.phpt",
    "ext/date/tests/011.phpt",
    "ext/date/tests/DateTime_modify_variation1.phpt",
    "ext/date/tests/bug29150.phpt",
    "ext/date/tests/date_date_set_variation3.phpt",
    "ext/date/tests/strtotime.phpt",
    "ext/date/tests/DateTime_diff_add_sub-fall-type2-type3.phpt",
    "ext/date/tests/bug43075.phpt",
    "ext/date/tests/strtotime_basic.phpt",
    "ext/date/tests/DateTime_setISODate_variation3.phpt",
    "ext/date/tests/default-timezone-1.phpt",
    "ext/date/tests/test-parse-from-format.phpt",
    "ext/date/tests/localtime_variation3.phpt",
    "ext/date/tests/getdate_variation5.phpt",
    "ext/date/tests/gmstrftime_error.phpt",
    "ext/date/tests/bug21399.phpt",
    "ext/date/tests/gmstrftime_variation3.phpt",
    "ext/date/tests/DateTimeZone_compare_basic1.phpt",
    "ext/date/tests/gmdate_variation4.phpt",
    "ext/date/tests/date_create-2.phpt",
    "ext/date/tests/strftime_error.phpt",
    "ext/date/tests/date_variation1.phpt",
    "ext/date/tests/DateTime_modify_basic1.phpt",
    "ext/date/tests/gmdate_variation10.phpt",
    "ext/date/tests/bug35624.phpt",
    "ext/date/tests/bug30096.phpt",
    "ext/date/tests/gmstrftime_variation5.phpt",
    "ext/date/tests/gmdate_variation2.phpt",
    "ext/date/tests/timezone_offset_get_variation2.phpt",
    "ext/date/tests/DateTime_diff_add_sub-massive.phpt",
    "ext/date/tests/date_diff.phpt",
    "ext/date/tests/strtotime-mysql-64bit.phpt",
    "ext/date/tests/date_sunrise_variation7.phpt",
    "ext/date/tests/DateTime_getOffset_basic1.phpt",
    "ext/date/tests/timezone_transitions_get_basic1.phpt",
    "ext/date/tests/gmstrftime_variation17.phpt",
    "ext/date/tests/strftime_variation12.phpt",
    "ext/date/tests/DateTimeZone_getTransitions_error.phpt",
    "ext/date/tests/timezone_offset_get_variation1.phpt",
    "ext/date/tests/gmdate_variation12.phpt",
    "ext/date/tests/bug51393.phpt",
    "ext/date/tests/strtotime_variation_scottish.phpt",
    "ext/date/tests/DateTime_setTime_variation3.phpt",
    "ext/date/tests/date_create_basic.phpt",
    "ext/date/tests/date_sunrise_variation5.phpt",
    "ext/date/tests/bug46874.phpt",
    "ext/date/tests/bug43960.phpt",
    "ext/date/tests/date_isodate_set_variation3.phpt",
    "ext/date/tests/gmdate_error.phpt",
    "ext/date/tests/format-negative-timestamp.phpt",
    "ext/date/tests/009_win32.phpt",
    "ext/date/tests/getdate_variation7.phpt",
    "ext/date/tests/date_sun_info_variation3.phpt",
    "ext/date/tests/idate_error.phpt",
    "ext/date/tests/DateTime_format_variation1.phpt",
    "ext/date/tests/bug34771.phpt",
    "ext/date/tests/bug36510.phpt",
    "ext/date/tests/bug35885.phpt",
    "ext/date/tests/date_sunset_variation4.phpt",
    "ext/date/tests/gmdate_variation14.phpt",
    "ext/date/tests/idate_variation3.phpt",
    "ext/date/tests/gmdate_variation3.phpt",
    "ext/date/tests/date_timezone_get_variation1.phpt",
    "ext/date/tests/bug33452.phpt",
    "ext/date/tests/date_date_set_variation1.phpt",
    "ext/date/tests/bug43527.phpt",
    "ext/date/tests/bug45554.phpt",
    "ext/date/tests/strtotime-relative.phpt",
    "ext/date/tests/DateTimeZone_extends_basic1.phpt",
    "ext/date/tests/bug33415-2.phpt",
    "ext/date/tests/date_sunrise_variation2.phpt",
    "ext/date/tests/date_timestamp_get.phpt",
    "ext/date/tests/015.phpt",
    "ext/date/tests/bug35887.phpt",
    "ext/date/tests/date_timezone_set_error.phpt",
    "ext/date/tests/DateTimeZone_getName_basic1.phpt",
    "ext/date/tests/date_variation2.phpt",
    "ext/date/tests/bug35218.phpt",
    "ext/date/tests/DateTime_format_basic2.phpt",
    "ext/date/tests/date_sub_basic.phpt",
    "ext/date/tests/DateTimeZone_construct_basic.phpt",
    "ext/date/tests/date_add_basic.phpt",
    "ext/date/tests/strftime_variation22.phpt",
    "ext/date/tests/date_sunset_variation7.phpt",
    "ext/date/tests/bug51819.phpt",
    "ext/date/tests/mktime_variation1.phpt",
    "ext/date/tests/004.phpt",
    "ext/date/tests/gmmktime_variation9.phpt",
    "ext/date/tests/DateTime_compare_basic1.phpt",
    "ext/date/tests/gmstrftime_variation2.phpt",
    "ext/date/tests/mktime_variation2.phpt",
    "ext/date/tests/bug30532.phpt",
    "ext/date/tests/strftime_variation7.phpt",
    "ext/date/tests/DateTime_serialize.phpt",
    "ext/date/tests/date_modify_variation1.phpt",
    "ext/date/tests/idate_variation4.phpt",
    "ext/date/tests/date_sunrise_variation1.phpt",
    "ext/date/tests/date_format_variation2.phpt",
    "ext/date/tests/default-timezone-2.phpt",
    "ext/date/tests/date_offset_get_variation1.phpt",
    "ext/date/tests/bug27719.phpt",
    "ext/date/tests/DateTimeZone_construct_variation1.phpt",
    "ext/date/tests/timezones.phpt",
    "ext/date/tests/DateTime_diff_add_sub-dates.phpt",
    "ext/date/tests/bug52808.phpt",
    "ext/date/tests/date_default_timezone_get_error.phpt",
    "ext/date/tests/date_sun_info_001.phpt",
    "ext/date/tests/date_constants.phpt",
    "ext/date/tests/DateTimeZone_getName_error.phpt",
    "ext/date/tests/bug49081.phpt",
    "ext/date/tests/date_create_variation1.phpt",
    "ext/date/tests/bug33415-1.phpt",
    "ext/date/tests/DateTimeZone_construct_error.phpt",
    "ext/date/tests/date_parse_001.phpt",
    "ext/date/tests/bug33869.phpt",
    "ext/date/tests/strftime_variation6.phpt",
    "ext/date/tests/getdate_variation1.phpt",
    "ext/date/tests/strftime_variation4.phpt",
    "ext/date/tests/date-lenient-create.phpt",
    "ext/date/tests/bug35705.phpt",
    "ext/date/tests/DateTime_verify.phpt",
    "ext/date/tests/gmdate_variation7.phpt",
    "ext/date/tests/bug38229.phpt",
    "ext/date/tests/timezones-list.phpt",
    "ext/date/tests/bug48276.phpt",
    "ext/date/tests/bug26320.phpt",
    "ext/date/tests/date_timezone_set_variation1.phpt",
    "ext/date/tests/date_isodate_set_variation1.phpt",
    "ext/date/tests/date_create-1.phpt",
    "ext/date/tests/DateTimeZone_getOffset_basic1.phpt",
    "ext/date/tests/mktime-3-64bit.phpt",
    "ext/date/tests/bug28088.phpt",
    "ext/date/tests/date_time_set_basic1.phpt",
    "ext/date/tests/bug45529.phpt",
    "ext/date/tests/date_sunset_variation3.phpt",
    "ext/date/tests/DateTime_extends_basic1.phpt",
    "ext/date/tests/bug35630.phpt",
    "ext/date/tests/bug44742.phpt",
    "ext/date/tests/gmdate_variation13.phpt",
    "ext/date/tests/date_sunrise_variation8.phpt",
    "ext/date/tests/003.phpt",
    "ext/date/tests/timezone_name_get_error.phpt",
    "ext/date/tests/bug35143.phpt",
    "ext/date/tests/date_get_last_errors_basic.phpt",
    "ext/date/tests/gmmktime_variation7.phpt",
    "ext/date/tests/DateTime_diff_add_sub-fall-type2-type2.phpt",
    "ext/date/tests/DateTime_setTime_basic1.phpt",
    "ext/date/tests/bug42910.phpt",
    "ext/date/tests/bug28599.phpt",
    "ext/date/tests/gmstrftime_variation21.phpt",
    "ext/date/tests/date_modify-2.phpt",
    "ext/date/tests/DateTime_clone_basic3.phpt",
    "ext/date/tests/bug45682.phpt",
    "ext/date/tests/date_modify_error.phpt",
    "ext/date/tests/bug52290.phpt",
    "ext/date/tests/date_sunrise_variation3.phpt",
    "ext/date/tests/gmdate_variation9.phpt",
    "ext/date/tests/gmstrftime_variation16.phpt",
    "ext/date/tests/date_sunset_variation9.phpt",
    "ext/date/tests/strftime_variation1.phpt",
    "ext/date/tests/strftime_variation14.phpt",
    "ext/date/tests/date_basic1.phpt",
    "ext/date/tests/date_sunset_variation8.phpt",
    "ext/date/tests/DateTime_setDate_variation1.phpt",
    "ext/date/tests/date_modify_variation2.phpt",
    "ext/date/tests/DateTimeZone_listAbbreviations_basic1.phpt",
    "ext/date/tests/mktime_variation4.phpt",
    "ext/date/tests/DateTime_diff_add_sub-spring-type2-type3.phpt",
    "ext/date/tests/strftime_variation17.phpt",
    "ext/date/tests/bug50055.phpt",
    "ext/date/tests/bug34676.phpt",
    "ext/date/tests/timezone_abbreviations_list_basic1.phpt",
    "ext/date/tests/gmdate_variation11.phpt",
    "ext/date/tests/DateTime_setTimezone_error.phpt",
    "ext/date/tests/timezone_name_from_abbr_error.phpt",
    "ext/date/tests/bug26090.phpt",
    "ext/date/tests/mktime_variation5.phpt",
    "ext/date/tests/mktime_variation7.phpt",
    "ext/date/tests/date_time_set_variation1.phpt",
    "ext/date/tests/date_sunrise_error.phpt",
    "ext/date/tests/006.phpt",
    "ext/date/tests/date_isodate_set_basic1.phpt",
    "ext/date/tests/date_timezone_get_basic1.phpt",
    "ext/date/tests/date_sunset_variation1.phpt",
    "ext/date/tests/bug33562.phpt",
    "ext/date/tests/bug49585.phpt",
    "ext/date/tests/strftime_variation16.phpt",
    "ext/date/tests/bug49778.phpt",
    "ext/date/tests/strftime_variation2.phpt",
    "ext/date/tests/timezone_open_error.phpt",
    "ext/date/tests/gmstrftime_variation1.phpt",
    "ext/date/tests/date_sunset_variation6.phpt",
    "ext/date/tests/getdate_variation4.phpt",
    "ext/date/tests/strftime_variation5.phpt",
    "ext/date/tests/localtime_variation5.phpt",
    "ext/date/tests/date_sun_info_variation2.phpt",
    "ext/date/tests/timezone_name_from_abbr_variation3.phpt",
    "ext/date/tests/date-time-modify-times.phpt",
    "ext/date/tests/strftime_variation10.phpt",
    "ext/date/tests/bug52668.phpt",
    "ext/date/tests/date_create_error.phpt",
    "ext/date/tests/date_parse_from_format_basic.phpt",
    "ext/date/tests/bug14561.phpt",
    "ext/date/tests/idate_variation1.phpt",
    "ext/date/tests/DateTime_construct_basic1.phpt",
    "ext/date/tests/date_default_timezone_set-1.phpt",
    "ext/date/tests/strftime_variation11.phpt",
    "ext/date/tests/gettimeofday_variation1.phpt",
    "ext/date/tests/date_sun_info_002.phpt",
    "ext/date/tests/gmstrftime_variation15.phpt",
    "ext/date/tests/getdate_variation6.phpt",
    "ext/date/tests/cal_days_in_month_invalid_date.phpt",
    "ext/date/tests/DateTime_setTimezone_basic1.phpt",
    "ext/date/tests/DateTime_diff_add_sub-fall-type3-type2.phpt",
    "ext/date/tests/bug37368.phpt",
    "ext/date/tests/DateTime_extends_basic3.phpt",
    "ext/date/tests/date_format_error.phpt",
    "ext/date/tests/bug46268.phpt",
    "ext/date/tests/checkdate_error.phpt",
    "ext/date/tests/mktime_basic1.phpt",
    "ext/date/tests/bug48097.phpt",
    "ext/date/tests/DateTime_setISODate_error.phpt",
    "ext/date/tests/bug40743.phpt",
    "ext/date/tests/DateTime_modify_error.phpt",
    "ext/date/tests/date_offset_get_error.phpt",
    "ext/date/tests/strftime_variation23.phpt",
    "ext/date/tests/mktime-1.phpt",
    "ext/date/tests/microtime_basic.phpt",
    "ext/date/tests/bug33532.phpt",
    "ext/date/tests/bug21966.phpt",
    "ext/date/tests/bug54316.phpt",
    "ext/date/tests/bug13142.phpt",
    "ext/date/tests/timestamp-in-dst.phpt",
    "ext/date/tests/timezone_location_get.phpt",
    "ext/date/tests/microtime_error.phpt",
    "ext/date/tests/date_isodate_set_variation4.phpt",
    "ext/date/tests/gmstrftime_variation13.phpt",
    "ext/date/tests/idate_variation2.phpt",
    "ext/date/tests/timezone_open_variation1.phpt",
    "ext/date/tests/strftime_variation20.phpt",
    "ext/date/tests/007.phpt",
    "ext/date/tests/bug26694.phpt",
    "ext/date/tests/sunfuncts.phpt",
    "ext/date/tests/gmstrftime_variation12.phpt",
    "ext/date/tests/bug20382-1.phpt",
    "ext/date/tests/DateTime_setISODate_variation2.phpt",
    "ext/date/tests/gmmktime_variation3.phpt",
    "ext/date/tests/DateTime_clone_basic2.phpt",
    "ext/date/tests/bug33414-1.phpt",
    "ext/date/tests/DateInterval_format.phpt",
    "ext/date/tests/bug37747.phpt",
    "ext/date/tests/bug20382-2.phpt",
    "ext/date/tests/idate_variation5.phpt",
    "ext/date/tests/strftime_variation9.phpt",
    "ext/date/tests/gmmktime_basic.phpt",
    "ext/date/tests/timezone_open_basic1.phpt",
    "ext/date/tests/checkdate_variation3.phpt",
    "ext/date/tests/DateTimeZone_clone_basic4.phpt",
    "ext/date/tests/bug37616.phpt",
    "ext/date/tests/DateTime_getOffset_error.phpt",
    "ext/date/tests/strtotime3-64bit.phpt",
    "ext/date/tests/date_date_set_basic1.phpt",
    "ext/date/tests/strftime_variation15.phpt",
    "ext/date/tests/bug35456.phpt",
    "ext/date/tests/strtotime-mysql.phpt",
    "ext/date/tests/bug50680.phpt",
    "ext/date/tests/mktime_variation3.phpt",
    "ext/date/tests/DateTime_construct_variation2.phpt",
    "ext/date/tests/bug32270.phpt",
    "ext/date/tests/002.phpt",
    "ext/date/tests/gmstrftime_variation19.phpt",
    "ext/date/tests/bug26198.phpt",
    "ext/date/tests/008.phpt",
    "ext/date/tests/date_create_from_format_basic.phpt",
    "ext/date/tests/DateTime_construct_error.phpt",
    "ext/date/tests/date_sunset_error.phpt",
    "ext/date/tests/mktime_variation6.phpt",
    "ext/date/tests/bug35699.phpt",
    "ext/date/tests/date_date_set_error.phpt",
    "ext/date/tests/date_add_basic2.phpt",
    "ext/date/tests/strftime_variation13.phpt",
    "ext/date/tests/date_sunset_variation2.phpt",
    "ext/date/tests/date_sun_info_variation1.phpt",
    "ext/date/tests/date_timezone_get_error.phpt",
    "ext/date/tests/date_parse_variation1.phpt",
    "ext/date/tests/bug52342.phpt",
    "ext/date/tests/timezone_transitions_get_error.phpt",
    "ext/date/tests/date_create-relative.phpt",
    "ext/date/tests/mktime_error.phpt",
    "ext/date/tests/date_parse_error.phpt",
    "ext/date/tests/bug50475.phpt",
    "ext/date/tests/date_date_set_variation2.phpt",
    "ext/date/tests/date_error.phpt",
    "ext/date/tests/bug41523.phpt",
    "ext/date/tests/date_date_set_variation4.phpt",
    "ext/date/tests/timezone_transitions_get_variation2.phpt",
    "ext/date/tests/DateTime_setTimezone_variation1.phpt",
    "ext/date/tests/DateTime_diff_add_sub-february.phpt",
    "ext/date/tests/gmstrftime_variation6.phpt",
    "ext/date/tests/bug50392.phpt",
    "ext/date/tests/bug41709.phpt",
    "ext/date/tests/DateTime_clone_basic1.phpt",
    "ext/date/tests/bug33414-2.phpt",
    "ext/date/tests/013.phpt",
    "ext/date/tests/checkdate_basic1.phpt",
    "ext/date/tests/gmmktime_error.phpt",
    "ext/date/tests/DateTimeZone_clone_basic1.phpt",
    "ext/date/tests/gmstrftime_variation4.phpt",
    "ext/date/tests/bug40861.phpt",
    "ext/date/tests/date_default_timezone_set_error.phpt",
    "ext/date/tests/idate_variation6.phpt",
    "ext/date/tests/gettimeofday_basic.phpt",
    "ext/date/tests/bug33563.phpt",
    "ext/date/tests/date_format_variation1.phpt",
    "ext/date/tests/date-lenient.phpt",
    "ext/date/tests/bug52577.phpt",
    "ext/date/tests/bug35414.phpt",
    "ext/date/tests/strftime_variation19.phpt",
    "ext/date/tests/getdate_variation3.phpt",
    "ext/date/tests/bug36224.phpt",
    "ext/date/tests/gmdate_variation5.phpt",
    "ext/date/tests/DateTime_getTimeZone_basic1.phpt",
    "ext/date/tests/bug34304.phpt",
    "ext/date/tests/date_offset_get_basic1.phpt",
    "ext/date/tests/timezone_transitions_get_variation3.phpt",
    "ext/date/tests/DateTime_setISODate_basic1.phpt",
    "ext/date/tests/DateTime_setDate_error.phpt",
    "ext/date/tests/date_sunrise_variation9.phpt",
    "ext/date/tests/bug49700.phpt",
    "ext/date/tests/localtime_variation1.phpt",
    "ext/date/tests/timezone_identifiers_list_basic1.phpt",
    "ext/date/tests/strtotime3.phpt",
    "ext/date/tests/getdate_variation2.phpt",
    "ext/date/tests/DateTimeZone_getOffset_error.phpt",
    "ext/date/tests/date_sunrise_variation6.phpt",
    "ext/date/tests/date.phpt",
    "ext/date/tests/bug45866.phpt",
    "ext/date/tests/012.phpt",
    "ext/date/tests/bug32555.phpt",
    "ext/date/tests/date_time_set_variation4.phpt",
    "ext/date/tests/DateTimeZone_verify.phpt",
    "ext/date/tests/gmdate_variation6.phpt",
    "ext/date/tests/bug32588.phpt",
    "ext/date/tests/gmdate_variation1.phpt",
    "ext/date/tests/bug36599.phpt",
    "ext/date/tests/strftime_variation18.phpt",
    "ext/date/tests/date_default_timezone_set_variation1.phpt",
    "ext/date/tests/bug35422.phpt",
    "ext/date/tests/mktime-2.phpt",
    "ext/date/tests/bug41844.phpt",
    "ext/date/tests/bug33056.phpt",
    "ext/date/tests/time_basic.phpt",
    "ext/date/tests/DateTime_construct_variation1.phpt",
    "ext/date/tests/gmstrftime_variation14.phpt",
    "ext/date/tests/bug17988.phpt",
    "ext/date/tests/date_timezone_set_basic1.phpt",
    "ext/date/tests/gmstrftime_variation9.phpt",
    "ext/date/tests/strftime_variation8.phpt",
    "ext/date/tests/bug46111.phpt",
    "ext/date/tests/DateTime_setTime_variation1.phpt",
    "ext/date/tests/timezone_transitions_get_variation1.phpt",
    "ext/date/tests/bug51994.phpt",
    "ext/date/tests/bug54340.phpt",
    "ext/date/tests/bug43452.phpt",
    "ext/date/tests/bug28024.phpt",
    "ext/date/tests/DateTimeZone_listIdentifiers_basic1.phpt",
    "ext/date/tests/localtime_variation4.phpt",
    "ext/date/tests/gmmktime_variation1.phpt",
    "ext/date/tests/DateTime_setTime_error.phpt",
    "ext/date/tests/DateTime_setDate_variation2.phpt",
    "ext/date/tests/DateTime_extends_basic2.phpt",
    "ext/date/tests/timezone_offset_get_basic1.phpt",
    "ext/date/tests/date_create_variation2.phpt",
    "ext/date/tests/checkdate_variation1.phpt",
    "ext/date/tests/date_time_set_error.phpt",
    "ext/date/tests/bug37514.phpt",
    "ext/date/tests/date_sunset_variation5.phpt",
    "ext/date/tests/localtime_error.phpt",
    "ext/date/tests/date_sun_info_error.phpt",
    "ext/date/tests/bug35425.phpt",
    "ext/date/tests/date-set-timestamp.phpt",
    "ext/date/tests/cal_days_in_month_invalid_calendar.phpt",
    "ext/date/tests/gmdate_basic.phpt",
    "ext/date/tests/idate_basic.phpt",
    "ext/date/tests/gmmktime_variation2.phpt",
    "ext/date/tests/DateTime_diff_add_sub-spring-type2-type2.phpt",
    "ext/date/tests/timezone_name_from_abbr_variation1.phpt",
    "ext/date/tests/DateTimeZone_serialize.phpt",
    "ext/date/tests/bug48187.phpt",
    "ext/date/tests/bug33957.phpt",
    "ext/date/tests/DateTime_setISODate_variation1.phpt",
    "ext/date/tests/bug35499.phpt",
    "ext/date/tests/strtotime_basic2.phpt",
    "ext/date/tests/DateInterval_format_a.phpt",
    "ext/date/tests/DateTimeZone_clone_basic2.phpt",
    "ext/date/tests/date_sunrise_and_sunset_basic.phpt",
    "ext/date/tests/bug44562.phpt",
    "ext/date/tests/bug41964.phpt",
    "ext/date/tests/date_time_set_variation2.phpt",
    "ext/date/tests/date_modify-1.phpt",
    "ext/date/tests/date_format_basic1.phpt",
    "ext/date/tests/date_default_timezone_get-2.phpt",
    "ext/date/tests/bug29585.phpt",
    "ext/date/tests/date_isodate_set_error.phpt",
    "ext/date/tests/strftime_variation3.phpt",
    "ext/date/tests/timezone_identifiers_list_wrong_constructor.phpt",
    "ext/date/tests/date_period.phpt",
    "ext/date/tests/bug52063.phpt",
    "ext/date/tests/gmdate_variation8.phpt",
    "ext/date/tests/bug37017.phpt",
    "ext/date/tests/big_year.phpt",
    "ext/date/tests/gmstrftime_variation20.phpt",
    "ext/date/tests/strftime_basic.phpt",
    "ext/date/tests/localtime_variation2.phpt",
    "ext/date/tests/bug51096.phpt",
    "ext/date/tests/009.phpt",
    "ext/date/tests/DateTime_diff_add_sub-absolute.phpt",
    "ext/date/tests/bug51866.phpt",
    "ext/date/tests/bug39782.phpt",
    "ext/date/tests/timezone-configuration.phpt",
    "ext/date/tests/date_modify_basic1.phpt",
    "ext/date/tests/bug54283.phpt",
    "ext/date/tests/date_default_timezone_get-3.phpt",
    "ext/date/tests/date_default_timezone_get-1.phpt",
    "ext/date/tests/gmmktime_variation8.phpt",
    "ext/date/tests/gettimeofday_error.phpt",
    "ext/date/tests/mktime-3.phpt",
    "ext/date/tests/strftime_variation21.phpt",
    "ext/date/tests/bug52738.phpt",
    "ext/date/tests/gmmktime_variation4.phpt",
    "ext/date/tests/gmstrftime_variation10.phpt",
    "ext/date/tests/bug52430.phpt",
    "ext/date/tests/bug48678.phpt",
    "ext/date/tests/timezone_name_from_abbr_variation2.phpt",
    "ext/date/tests/bug26317.phpt",
    "ext/ldap/tests/ldap_error_basic.phpt",
    "ext/ldap/tests/ldap_mod_add_error.phpt",
    "ext/ldap/tests/ldap_get_entries_variation.phpt",
    "ext/ldap/tests/ldap_modify_error.phpt",
    "ext/ldap/tests/ldap_get_entries_basic.phpt",
    "ext/ldap/tests/ldap_connect_basic.phpt",
    "ext/ldap/tests/ldap_set_rebind_proc_error.phpt",
    "ext/ldap/tests/ldap_add_basic.phpt",
    "ext/ldap/tests/ldap_delete_error.phpt",
    "ext/ldap/tests/ldap_errno_error.phpt",
    "ext/ldap/tests/ldap_control_paged_results_variation3.phpt",
    "ext/ldap/tests/ldap_rename_error.phpt",
    "ext/ldap/tests/ldap_read_error.phpt",
    "ext/ldap/tests/ldap_search_error.phpt",
    "ext/ldap/tests/ldap_get_option_basic.phpt",
    "ext/ldap/tests/ldap_mod_replace_error.phpt",
    "ext/ldap/tests/ldap_error_error.phpt",
    "ext/ldap/tests/ldap_modify_basic.phpt",
    "ext/ldap/tests/ldap_unbind_error.phpt",
    "ext/ldap/tests/ldap_get_values_len_error.phpt",
    "ext/ldap/tests/ldap_free_result_basic.phpt",
    "ext/ldap/tests/ldap_dn2ufn.phpt",
    "ext/ldap/tests/ldap_next_reference_basic.phpt",
    "ext/ldap/tests/ldap_first_attribute_basic.phpt",
    "ext/ldap/tests/ldap_count_entries_basic.phpt",
    "ext/ldap/tests/ldap_compare_basic.phpt",
    "ext/ldap/tests/ldap_parse_result_basic.phpt",
    "ext/ldap/tests/ldap_search_variation1.phpt",
    "ext/ldap/tests/ldap_get_entries_error.phpt",
    "ext/ldap/tests/ldap_explode_dn.phpt",
    "ext/ldap/tests/ldap_free_result_error.phpt",
    "ext/ldap/tests/ldap_unbind_basic.phpt",
    "ext/ldap/tests/ldap_search_variation3.phpt",
    "ext/ldap/tests/ldap_set_rebind_proc_basic.phpt",
    "ext/ldap/tests/ldap_search_variation6.phpt",
    "ext/ldap/tests/ldap_get_dn_error.phpt",
    "ext/ldap/tests/ldap_bind_error.phpt",
    "ext/ldap/tests/ldap_parse_result_error.phpt",
    "ext/ldap/tests/ldap_search_basic.phpt",
    "ext/ldap/tests/ldap_connect_error.phpt",
    "ext/ldap/tests/ldap_mod_replace_basic.phpt",
    "ext/ldap/tests/ldap_first_entry_basic.phpt",
    "ext/ldap/tests/ldap_first_attribute_error.phpt",
    "ext/ldap/tests/ldap_err2str_error.phpt",
    "ext/ldap/tests/ldap_get_attributes_error.phpt",
    "ext/ldap/tests/ldap_get_option_error.phpt",
    "ext/ldap/tests/ldap_rename_basic.phpt",
    "ext/ldap/tests/ldap_add_error.phpt",
    "ext/ldap/tests/ldap_set_option_basic.phpt",
    "ext/ldap/tests/ldap_unbind_variation.phpt",
    "ext/ldap/tests/bug48441.phpt",
    "ext/ldap/tests/ldap_next_entry_error.phpt",
    "ext/ldap/tests/ldap_count_entries_error.phpt",
    "ext/ldap/tests/ldap_control_paged_results_variation1.phpt",
    "ext/ldap/tests/ldap_compare_error.phpt",
    "ext/ldap/tests/ldap_connect_variation.phpt",
    "ext/ldap/tests/ldap_read_basic.phpt",
    "ext/ldap/tests/ldap_search_variation2.phpt",
    "ext/ldap/tests/ldap_mod_del_error.phpt",
    "ext/ldap/tests/ldap_next_attribute_error.phpt",
    "ext/ldap/tests/ldap_sasl_bind_error.phpt",
    "ext/ldap/tests/bug48696.phpt",
    "ext/ldap/tests/ldap_get_dn_basic.phpt",
    "ext/ldap/tests/ldap_set_option_error.phpt",
    "ext/ldap/tests/ldap_search_variation4.phpt",
    "ext/ldap/tests/ldap_get_option_variation.phpt",
    "ext/ldap/tests/ldap_sort_error.phpt",
    "ext/ldap/tests/ldap_sasl_bind_basic.phpt",
    "ext/ldap/tests/ldap_start_tls_error.phpt",
    "ext/ldap/tests/ldap_set_option_variation.phpt",
    "ext/ldap/tests/ldap_err2str_basic.phpt",
    "ext/ldap/tests/ldap_mod_add_basic.phpt",
    "ext/ldap/tests/ldap_control_paged_results_variation2.phpt",
    "ext/ldap/tests/ldap_get_attributes_basic.phpt",
    "ext/ldap/tests/ldap_bind_variation.phpt",
    "ext/ldap/tests/ldap_get_values_len_basic.phpt",
    "ext/ldap/tests/ldap_bind_basic.phpt",
    "ext/ldap/tests/ldap_mod_del_basic.phpt",
    "ext/ldap/tests/ldap_next_attribute_basic.phpt",
    "ext/ldap/tests/ldap_first_entry_error.phpt",
    "ext/ldap/tests/ldap_first_reference_error.phpt",
    "ext/ldap/tests/ldap_list_error.phpt",
    "ext/ldap/tests/ldap_sort_variation.phpt",
    "ext/ldap/tests/ldap_parse_reference_basic.phpt",
    "ext/ldap/tests/ldap_search_variation5.phpt",
    "ext/ldap/tests/ldap_sort_basic.phpt",
    "ext/ldap/tests/ldap_parse_reference_error.phpt",
    "ext/ldap/tests/ldap_first_reference_basic.phpt",
    "ext/ldap/tests/ldap_next_reference_error.phpt",
    "ext/ldap/tests/ldap_next_entry_basic.phpt",
    "ext/ldap/tests/ldap_start_tls_basic.phpt",
    "ext/ldap/tests/ldap_list_basic.phpt",
    "ext/ldap/tests/ldap_delete_basic.phpt",
    "ext/ldap/tests/ldap_errno_basic.phpt",
    "ext/imap/tests/imap_fetch_overview_variation3.phpt",
    "ext/imap/tests/imap_lsub_basic.phpt",
    "ext/imap/tests/imap_fetch_overview_variation5.phpt",
    "ext/imap/tests/imap_clearflag_full_basic.phpt",
    "ext/imap/tests/imap_close_error.phpt",
    "ext/imap/tests/imap_fetchbody_variation4.phpt",
    "ext/imap/tests/imap_fetchbody_basic.phpt",
    "ext/imap/tests/imap_num_msg_error.phpt",
    "ext/imap/tests/imap_timeout_basic.phpt",
    "ext/imap/tests/imap_mail_copy_basic.phpt",
    "ext/imap/tests/imap_fetchbody_variation1.phpt",
    "ext/imap/tests/imap_fetch_overview_variation4.phpt",
    "ext/imap/tests/bug31142_1.phpt",
    "ext/imap/tests/imap_createmailbox_basic.phpt",
    "ext/imap/tests/bug44098.phpt",
    "ext/imap/tests/imap_fetchheader_variation2.phpt",
    "ext/imap/tests/imap_fetchbody_variation5.phpt",
    "ext/imap/tests/imap_headerinfo_error.phpt",
    "ext/imap/tests/bug46918.phpt",
    "ext/imap/tests/bug45705_1.phpt",
    "ext/imap/tests/imap_append_basic.phpt",
    "ext/imap/tests/imap_rfc822_parse_headers_basic.phpt",
    "ext/imap/tests/imap_errors_basic.phpt",
    "ext/imap/tests/imap_fetchbody_variation2.phpt",
    "ext/imap/tests/imap_close_variation2.phpt",
    "ext/imap/tests/imap_body.phpt",
    "ext/imap/tests/bug45705_2.phpt",
    "ext/imap/tests/imap_fetchheader_variation3.phpt",
    "ext/imap/tests/imap_close_variation3.phpt",
    "ext/imap/tests/imap_close_basic.phpt",
    "ext/imap/tests/imap_alerts_error.phpt",
    "ext/imap/tests/imap_utf8.phpt",
    "ext/imap/tests/imap_mutf7_to_utf8.phpt",
    "ext/imap/tests/imap_undelete_error.phpt",
    "ext/imap/tests/bug53377.phpt",
    "ext/imap/tests/imap_expunge_error.phpt",
    "ext/imap/tests/imap_gc_error.phpt",
    "ext/imap/tests/imap_fetchbody_error.phpt",
    "ext/imap/tests/imap_mail_move.phpt",
    "ext/imap/tests/imap_base64_basic.phpt",
    "ext/imap/tests/imap_close_variation1.phpt",
    "ext/imap/tests/imap_body_basic.phpt",
    "ext/imap/tests/imap_undelete_basic.phpt",
    "ext/imap/tests/imap_bodystruct_basic.phpt",
    "ext/imap/tests/imap_savebody_basic.phpt",
    "ext/imap/tests/imap_fetch_overview_error.phpt",
    "ext/imap/tests/imap_fetch_overview_variation2.phpt",
    "ext/imap/tests/imap_close_variation4.phpt",
    "ext/imap/tests/bug35669.phpt",
    "ext/imap/tests/imap_binary_basic.phpt",
    "ext/imap/tests/imap_headers.phpt",
    "ext/imap/tests/imap_ping_error.phpt",
    "ext/imap/tests/imap_fetchbody_variation3.phpt",
    "ext/imap/tests/imap_mail_copy.phpt",
    "ext/imap/tests/imap_fetch_overview_variation6.phpt",
    "ext/imap/tests/imap_fetchheader_variation4.phpt",
    "ext/imap/tests/imap_mail_move_basic.phpt",
    "ext/imap/tests/imap_8bit_basic.phpt",
    "ext/imap/tests/imap_open_error.phpt",
    "ext/imap/tests/imap_headerinfo_basic.phpt",
    "ext/imap/tests/imap_fetch_overview_basic.phpt",
    "ext/imap/tests/imap_list_basic.phpt",
    "ext/imap/tests/imap_fetchstructure_basic.phpt",
    "ext/imap/tests/imap_fetchheader_variation1.phpt",
    "ext/imap/tests/imap_fetchbody_variation6.phpt",
    "ext/imap/tests/imap_fetchheader_error.phpt",
    "ext/imap/tests/imap_fetchheader_basic.phpt",
    "ext/imap/tests/bug32589.phpt",
    "ext/imap/tests/imap_num_recent_error.phpt",
    "ext/imap/tests/imap_fetchheader_variation5.phpt",
    "ext/imap/tests/imap_utf8_to_mutf7_basic.phpt",
    "ext/imap/tests/imap_fetch_overview_variation1.phpt",
    "ext/imap/tests/bug31142_2.phpt",
    "ext/imap/tests/imap_renamemailbox_basic.phpt",
    "ext/imap/tests/bug40854.phpt",
    "ext/imap/tests/imap_getsubscribed_basic.phpt",
    "ext/pdo_odbc/tests/common.phpt",
    "ext/pdo_odbc/tests/long_columns.phpt",
    "ext/zlib/tests/gzfile_variation12.phpt",
    "ext/zlib/tests/zlib_scheme_unlink_basic.phpt",
    "ext/zlib/tests/readgzfile_variation14.phpt",
    "ext/zlib/tests/readgzfile_variation7.phpt",
    "ext/zlib/tests/gzfile_variation1.phpt",
    "ext/zlib/tests/gzencode_variation2.phpt",
    "ext/zlib/tests/gzfile_basic.phpt",
    "ext/zlib/tests/005.phpt",
    "ext/zlib/tests/gzuncompress_error1.phpt",
    "ext/zlib/tests/readgzfile_basic2.phpt",
    "ext/zlib/tests/gzinflate_length.phpt",
    "ext/zlib/tests/gzinflate-bug42663.phpt",
    "ext/zlib/tests/gztell_error.phpt",
    "ext/zlib/tests/gzfile_variation2.phpt",
    "ext/zlib/tests/zlib_scheme_file_basic.phpt",
    "ext/zlib/tests/gzread_error2.phpt",
    "ext/zlib/tests/readgzfile_variation8.phpt",
    "ext/zlib/tests/zlib_scheme_fopen_variation1.phpt",
    "ext/zlib/tests/gzopen_basic.phpt",
    "ext/zlib/tests/gzfile_error.phpt",
    "ext/zlib/tests/gzfile_variation8.phpt",
    "ext/zlib/tests/zlib_scheme_stat_basic.phpt",
    "ext/zlib/tests/gzfile_variation4.phpt",
    "ext/zlib/tests/zlib_filter_inflate2.phpt",
    "ext/zlib/tests/bug51269.phpt",
    "ext/zlib/tests/bug_40189.phpt",
    "ext/zlib/tests/bug_52944.phpt",
    "ext/zlib/tests/zlib_scheme_file_read_file_basic.phpt",
    "ext/zlib/tests/gzseek_basic.phpt",
    "ext/zlib/tests/readgzfile_variation2.phpt",
    "ext/zlib/tests/zlib_wrapper_ftruncate_basic.phpt",
    "ext/zlib/tests/gzeof_basic.phpt",
    "ext/zlib/tests/gzcompress_variation1.phpt",
    "ext/zlib/tests/gzgetc_error.phpt",
    "ext/zlib/tests/001.phpt",
    "ext/zlib/tests/readgzfile_variation13.phpt",
    "ext/zlib/tests/gzopen_variation9.phpt",
    "ext/zlib/tests/004.phpt",
    "ext/zlib/tests/gzseek_variation5.phpt",
    "ext/zlib/tests/readgzfile_basic.phpt",
    "ext/zlib/tests/readgzfile_variation12.phpt",
    "ext/zlib/tests/gztell_basic2.phpt",
    "ext/zlib/tests/readgzfile_error.phpt",
    "ext/zlib/tests/gzopen_variation8.phpt",
    "ext/zlib/tests/gzopen_variation6.phpt",
    "ext/zlib/tests/gzclose_error.phpt",
    "ext/zlib/tests/readgzfile_variation1.phpt",
    "ext/zlib/tests/gzfile_variation13.phpt",
    "ext/zlib/tests/gzfile_variation11.phpt",
    "ext/zlib/tests/gzopen_error.phpt",
    "ext/zlib/tests/gzclose_basic.phpt",
    "ext/zlib/tests/gzrewind_error.phpt",
    "ext/zlib/tests/gzseek_error.phpt",
    "ext/zlib/tests/gzputs_basic.phpt",
    "ext/zlib/tests/003.phpt",
    "ext/zlib/tests/gzopen_basic2.phpt",
    "ext/zlib/tests/gzgets_basic.phpt",
    "ext/zlib/tests/gzdeflate_variation1.phpt",
    "ext/zlib/tests/gzcompress_error1.phpt",
    "ext/zlib/tests/gzopen_variation1.phpt",
    "ext/zlib/tests/gzfile_basic2.phpt",
    "ext/zlib/tests/gzrewind_basic.phpt",
    "ext/zlib/tests/gzreadgzwriteplain.phpt",
    "ext/zlib/tests/gzcompress_basic1.phpt",
    "ext/zlib/tests/006.phpt",
    "ext/zlib/tests/gzeof_variation1.phpt",
    "ext/zlib/tests/gzseek_variation7.phpt",
    "ext/zlib/tests/gzfile_variation7.phpt",
    "ext/zlib/tests/gzgets_error.phpt",
    "ext/zlib/tests/ob_002.phpt",
    "ext/zlib/tests/compress_zlib_wrapper.phpt",
    "ext/zlib/tests/gzopen_variation4.phpt",
    "ext/zlib/tests/gzseek_variation1.phpt",
    "ext/zlib/tests/readgzfile_variation9.phpt",
    "ext/zlib/tests/zlib_scheme_copy_basic.phpt",
    "ext/zlib/tests/gzpassthru_basic.phpt",
    "ext/zlib/tests/007.phpt",
    "ext/zlib/tests/gzfile_variation3.phpt",
    "ext/zlib/tests/readgzfile_variation3.phpt",
    "ext/zlib/tests/gzrewind_basic2.phpt",
    "ext/zlib/tests/gzread_error.phpt",
    "ext/zlib/tests/gzfile_variation10.phpt",
    "ext/zlib/tests/gzopen_variation2.phpt",
    "ext/zlib/tests/gzgetc_basic.phpt",
    "ext/zlib/tests/gzencode_variation1.phpt",
    "ext/zlib/tests/gzseek_basic2.phpt",
    "ext/zlib/tests/gztell_basic.phpt",
    "ext/zlib/tests/zlib_filter_deflate.phpt",
    "ext/zlib/tests/ob_005.phpt",
    "ext/zlib/tests/002.phpt",
    "ext/zlib/tests/readgzfile_variation10.phpt",
    "ext/zlib/tests/008.phpt",
    "ext/zlib/tests/readgzfile_variation5.phpt",
    "ext/zlib/tests/gzinflate_error1.phpt",
    "ext/zlib/tests/gzread_basic.phpt",
    "ext/zlib/tests/gzseek_variation2.phpt",
    "ext/zlib/tests/readgzfile_variation11.phpt",
    "ext/zlib/tests/zlib_wrapper_meta_data_basic.phpt",
    "ext/zlib/tests/gzfilegzreadfile.phpt",
    "ext/zlib/tests/gzpassthru_error.phpt",
    "ext/zlib/tests/zlib_wrapper_fstat_basic.phpt",
    "ext/zlib/tests/zlib_scheme_copy_variation2.phpt",
    "ext/zlib/tests/gzdeflate_error1.phpt",
    "ext/zlib/tests/zlib_scheme_file_get_contents_basic.phpt",
    "ext/zlib/tests/zlib_wrapper_fflush_basic.phpt",
    "ext/zlib/tests/gzeof_error.phpt",
    "ext/zlib/tests/zlib_filter_inflate.phpt",
    "ext/zlib/tests/gzencode_variation2-win32.phpt",
    "ext/zlib/tests/gzencode_error1.phpt",
    "ext/zlib/tests/gzseek_variation6.phpt",
    "ext/zlib/tests/gzfile_variation6.phpt",
    "ext/zlib/tests/gzreadgzwrite.phpt",
    "ext/zlib/tests/gzuncompress_basic1.phpt",
    "ext/zlib/tests/zlib_scheme_dir_basic.phpt",
    "ext/zlib/tests/zlib_scheme_file_put_contents_basic.phpt",
    "ext/zlib/tests/gzseek_variation4.phpt",
    "ext/zlib/tests/readgzfile_variation4.phpt",
    "ext/zlib/tests/gzseek_variation3.phpt",
    "ext/zlib/tests/gzdeflate_basic1.phpt",
    "ext/zlib/tests/gzwrite_basic.phpt",
    "ext/zlib/tests/ob_001.phpt",
    "ext/zlib/tests/zlib_scheme_rename_basic.phpt",
    "ext/zlib/tests/zlib_wrapper_flock_basic.phpt",
    "ext/zlib/tests/zlib_filter_deflate2.phpt",
    "ext/zlib/tests/gzfile_variation15.phpt",
    "ext/zlib/tests/gzopen_variation7.phpt",
    "ext/zlib/tests/ob_003.phpt",
    "ext/zlib/tests/readgzfile_variation6.phpt",
    "ext/zlib/tests/gzwrite_error.phpt",
    "ext/zlib/tests/gzopen_variation5.phpt",
    "ext/zlib/tests/gzwrite_error2.phpt",
    "ext/zlib/tests/gzwrite_variation1.phpt",
    "ext/zlib/tests/gzfile_variation9.phpt",
    "ext/zlib/tests/gzread_variation1.phpt",
    "ext/zlib/tests/zlib_scheme_fopen_basic.phpt",
    "ext/zlib/tests/gzopen_variation3.phpt",
    "ext/zlib/tests/gzencode_basic1.phpt",
    "ext/zlib/tests/gzfile_variation14.phpt",
    "ext/zlib/tests/readgzfile_variation15.phpt",
    "ext/zlib/tests/ob_004.phpt",
    "ext/zlib/tests/bug_34821.phpt",
    "ext/zlib/tests/gzrewind_variation1.phpt",
    "ext/zlib/tests/zlib_scheme_copy_variation1.phpt",
    "ext/zlib/tests/zlib_scheme_stat_basic2.phpt",
    "ext/zlib/tests/gzencode_variation1-win32.phpt",
    "ext/zlib/tests/gzfile_variation5.phpt",
    "ext/pdo/tests/pdo_025.phpt",
    "ext/pdo/tests/pdo_033.phpt",
    "ext/pdo/tests/pdo_017.phpt",
    "ext/pdo/tests/pdo_028.phpt",
    "ext/pdo/tests/bug_40285.phpt",
    "ext/pdo/tests/pdo_027.phpt",
    "ext/pdo/tests/bug_50458.phpt",
    "ext/pdo/tests/bug_44173.phpt",
    "ext/pdo/tests/bug_34630.phpt",
    "ext/pdo/tests/pdo_003.phpt",
    "ext/pdo/tests/bug_39656.phpt",
    "ext/pdo/tests/pdo_037.phpt",
    "ext/pdo/tests/pdo_026.phpt",
    "ext/pdo/tests/pdo_009.phpt",
    "ext/pdo/tests/pdo_021.phpt",
    "ext/pdo/tests/pdo_035.phpt",
    "ext/pdo/tests/pdo_031.phpt",
    "ext/pdo/tests/bug_36428.phpt",
    "ext/pdo/tests/pdo_011.phpt",
    "ext/pdo/tests/pdo_007.phpt",
    "ext/pdo/tests/pdo_019.phpt",
    "ext/pdo/tests/bug_44861.phpt",
    "ext/pdo/tests/pdo_023.phpt",
    "ext/pdo/tests/bug_42917.phpt",
    "ext/pdo/tests/pdo_029.phpt",
    "ext/pdo/tests/pdo_013.phpt",
    "ext/pdo/tests/pecl_bug_5217.phpt",
    "ext/pdo/tests/pdo_022.phpt",
    "ext/pdo/tests/pdo_002.phpt",
    "ext/pdo/tests/pdo_034.phpt",
    "ext/pdo/tests/pdo_030.phpt",
    "ext/pdo/tests/pecl_bug_5772.phpt",
    "ext/pdo/tests/bug_38253.phpt",
    "ext/pdo/tests/pdo_001.phpt",
    "ext/pdo/tests/bug47769.phpt",
    "ext/pdo/tests/bug_34687.phpt",
    "ext/pdo/tests/pdo_032.phpt",
    "ext/pdo/tests/bug_43130.phpt",
    "ext/pdo/tests/bug_44159.phpt",
    "ext/pdo/tests/pdo_016a.phpt",
    "ext/pdo/tests/pdo_018.phpt",
    "ext/pdo/tests/bug_35671.phpt",
    "ext/pdo/tests/bug_38394.phpt",
    "ext/pdo/tests/pdo_014.phpt",
    "ext/pdo/tests/pdo_010.phpt",
    "ext/pdo/tests/pdo_004.phpt",
    "ext/pdo/tests/bug_36798.phpt",
    "ext/pdo/tests/bug_43663.phpt",
    "ext/pdo/tests/pecl_bug_5809.phpt",
    "ext/pdo/tests/pdo_012.phpt",
    "ext/pdo/tests/bug_44409.phpt",
    "ext/pdo/tests/pdo_006.phpt",
    "ext/pdo/tests/pdo_016.phpt",
    "ext/pdo/tests/pdo_020.phpt",
    "ext/pdo/tests/bug_39398.phpt",
    "ext/pdo/tests/pdo_024.phpt",
    "ext/pdo/tests/pdorow.phpt",
    "ext/pdo/tests/pdo_015.phpt",
    "ext/pdo/tests/pdo_008.phpt",
    "ext/pdo/tests/pdo_005.phpt",
    "ext/pdo/tests/bug_43139.phpt",
    "ext/pdo/tests/pdo_036.phpt",
    "ext/sysvmsg/tests/005.phpt",
    "ext/sysvmsg/tests/001.phpt",
    "ext/sysvmsg/tests/004.phpt",
    "ext/sysvmsg/tests/003.phpt",
    "ext/sysvmsg/tests/006.phpt",
    "ext/sysvmsg/tests/002.phpt",
    "ext/mysqli/tests/050.phpt",
    "ext/mysqli/tests/027.phpt",
    "ext/mysqli/tests/mysqli_fork.phpt",
    "ext/mysqli/tests/mysqli_get_server_version.phpt",
    "ext/mysqli/tests/mysqli_send_query.phpt",
    "ext/mysqli/tests/bug31141.phpt",
    "ext/mysqli/tests/mysqli_debug_ini.phpt",
    "ext/mysqli/tests/mysqli_stmt_close.phpt",
    "ext/mysqli/tests/mysqli_pconn_disabled.phpt",
    "ext/mysqli/tests/033.phpt",
    "ext/mysqli/tests/bug46109.phpt",
    "ext/mysqli/tests/mysqli_get_connection_stats.phpt",
    "ext/mysqli/tests/mysqli_get_connection_stats_off.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_short_len.phpt",
    "ext/mysqli/tests/mysqli_real_connect_pconn.phpt",
    "ext/mysqli/tests/mysqli_stmt_bind_param_many_columns.phpt",
    "ext/mysqli/tests/mysqli_stmt_get_result_bit.phpt",
    "ext/mysqli/tests/010.phpt",
    "ext/mysqli/tests/mysqli_stmt_get_result2.phpt",
    "ext/mysqli/tests/mysqli_stmt_bind_param_references.phpt",
    "ext/mysqli/tests/066.phpt",
    "ext/mysqli/tests/mysqli_real_escape_string_euckr.phpt",
    "ext/mysqli/tests/mysqli_real_escape_string_gbk.phpt",
    "ext/mysqli/tests/045.phpt",
    "ext/mysqli/tests/mysqli_connect_oo_warnings.phpt",
    "ext/mysqli/tests/bug36745.phpt",
    "ext/mysqli/tests/005.phpt",
    "ext/mysqli/tests/bug35759.phpt",
    "ext/mysqli/tests/mysqli_ps_select_union.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_negative_len.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_replace_buffer.phpt",
    "ext/mysqli/tests/mysqli_errno_oo.phpt",
    "ext/mysqli/tests/mysqli_use_result.phpt",
    "ext/mysqli/tests/mysqli_num_fields.phpt",
    "ext/mysqli/tests/054.phpt",
    "ext/mysqli/tests/mysqli_connect.phpt",
    "ext/mysqli/tests/mysqli_stmt_result_metadata.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_closures.phpt",
    "ext/mysqli/tests/mysqli_stmt_bind_param_call_user_func.phpt",
    "ext/mysqli/tests/mysqli_get_client_stats_skipped.phpt",
    "ext/mysqli/tests/bug54221.phpt",
    "ext/mysqli/tests/014.phpt",
    "ext/mysqli/tests/mysqli_stmt_bind_param.phpt",
    "ext/mysqli/tests/044.phpt",
    "ext/mysqli/tests/029.phpt",
    "ext/mysqli/tests/mysqli_warning_unclonable.phpt",
    "ext/mysqli/tests/mysqli_stmt_get_result_types.phpt",
    "ext/mysqli/tests/mysqli_warning_count.phpt",
    "ext/mysqli/tests/bug51647.phpt",
    "ext/mysqli/tests/073.phpt",
    "ext/mysqli/tests/mysqli_fetch_object_oo.phpt",
    "ext/mysqli/tests/mysqli_get_client_stats.phpt",
    "ext/mysqli/tests/072.phpt",
    "ext/mysqli/tests/mysqli_options.phpt",
    "ext/mysqli/tests/mysqli_fetch_assoc_zerofill.phpt",
    "ext/mysqli/tests/mysqli_stmt_error.phpt",
    "ext/mysqli/tests/mysqli_insert_packet_overflow.phpt",
    "ext/mysqli/tests/mysqli_sqlstate.phpt",
    "ext/mysqli/tests/011.phpt",
    "ext/mysqli/tests/mysqli_get_server_info.phpt",
    "ext/mysqli/tests/mysqli_get_proto_info.phpt",
    "ext/mysqli/tests/bug50772.phpt",
    "ext/mysqli/tests/mysqli_fetch_assoc_oo.phpt",
    "ext/mysqli/tests/mysqli_get_host_info.phpt",
    "ext/mysqli/tests/mysqli_num_rows.phpt",
    "ext/mysqli/tests/062.phpt",
    "ext/mysqli/tests/mysqli_driver.phpt",
    "ext/mysqli/tests/mysqli_poll_reference.phpt",
    "ext/mysqli/tests/mysqli_stmt_bind_result_bit.phpt",
    "ext/mysqli/tests/mysqli_fetch_fields.phpt",
    "ext/mysqli/tests/017.phpt",
    "ext/mysqli/tests/mysqli_class_mysqli_warning_reflection.phpt",
    "ext/mysqli/tests/mysqli_error_unicode.phpt",
    "ext/mysqli/tests/mysqli_kill.phpt",
    "ext/mysqli/tests/mysqli_stmt_get_result_geom.phpt",
    "ext/mysqli/tests/mysqli_result_references.phpt",
    "ext/mysqli/tests/mysqli_class_mysqli_properties_no_conn.phpt",
    "ext/mysqli/tests/036.phpt",
    "ext/mysqli/tests/mysqli_stmt_attr_get_prefetch.phpt",
    "ext/mysqli/tests/bug36420.phpt",
    "ext/mysqli/tests/mysqli_class_mysqli_interface.phpt",
    "ext/mysqli/tests/mysqli_result_references_mysqlnd.phpt",
    "ext/mysqli/tests/mysqli_stmt_execute_stored_proc.phpt",
    "ext/mysqli/tests/mysqli_stmt_bind_limits.phpt",
    "ext/mysqli/tests/mysqli_stmt_send_long_data_packet_size_libmysql.phpt",
    "ext/mysqli/tests/mysqli_connect_error.phpt",
    "ext/mysqli/tests/mysqli_errno.phpt",
    "ext/mysqli/tests/mysqli_dump_debug_info.phpt",
    "ext/mysqli/tests/bug54674.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_buffer_overflow.phpt",
    "ext/mysqli/tests/mysqli_connect_twice.phpt",
    "ext/mysqli/tests/mysqli_stat.phpt",
    "ext/mysqli/tests/mysqli_debug_mysqlnd_only.phpt",
    "ext/mysqli/tests/bug38710.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_nofileop.phpt",
    "ext/mysqli/tests/mysqli_change_user_set_names.phpt",
    "ext/mysqli/tests/mysqli_get_warnings.phpt",
    "ext/mysqli/tests/mysqli_debug_append.phpt",
    "ext/mysqli/tests/mysqli_fetch_assoc_no_alias_utf8.phpt",
    "ext/mysqli/tests/mysqli_constants_categories.phpt",
    "ext/mysqli/tests/mysqli_field_tell.phpt",
    "ext/mysqli/tests/043.phpt",
    "ext/mysqli/tests/028.phpt",
    "ext/mysqli/tests/060.phpt",
    "ext/mysqli/tests/068.phpt",
    "ext/mysqli/tests/001.phpt",
    "ext/mysqli/tests/041.phpt",
    "ext/mysqli/tests/mysqli_pconnect.phpt",
    "ext/mysqli/tests/mysqli_fetch_array_oo.phpt",
    "ext/mysqli/tests/bug33491.phpt",
    "ext/mysqli/tests/mysqli_class_mysqli_result_reflection.phpt",
    "ext/mysqli/tests/mysqli_stmt_attr_get.phpt",
    "ext/mysqli/tests/mysqli_affected_rows_oo.phpt",
    "ext/mysqli/tests/mysqli_change_user_insert_id.phpt",
    "ext/mysqli/tests/bug_mysql_49406.phpt",
    "ext/mysqli/tests/015.phpt",
    "ext/mysqli/tests/mysqli_fetch_field_oo.phpt",
    "ext/mysqli/tests/026.phpt",
    "ext/mysqli/tests/bug35103.phpt",
    "ext/mysqli/tests/004.phpt",
    "ext/mysqli/tests/mysqli_fetch_field_types.phpt",
    "ext/mysqli/tests/mysqli_stmt_insert_id.phpt",
    "ext/mysqli/tests/mysqli_stmt_free_result.phpt",
    "ext/mysqli/tests/mysqli_stmt_store_result.phpt",
    "ext/mysqli/tests/mysqli_pconn_limits.phpt",
    "ext/mysqli/tests/mysqli_info.phpt",
    "ext/mysqli/tests/bug33090.phpt",
    "ext/mysqli/tests/mysqli_stmt_fetch_fields_win32_unicode.phpt",
    "ext/mysqli/tests/mysqli_real_escape_string_unicode.phpt",
    "ext/mysqli/tests/mysqli_multi_query.phpt",
    "ext/mysqli/tests/bug47050.phpt",
    "ext/mysqli/tests/mysqli_get_client_version.phpt",
    "ext/mysqli/tests/mysqli_free_result.phpt",
    "ext/mysqli/tests/mysqli_get_client_stats_implicit_free.phpt",
    "ext/mysqli/tests/mysqli_debug_control_string.phpt",
    "ext/mysqli/tests/mysqli_connect_oo_defaults.phpt",
    "ext/mysqli/tests/mysqli_prepare.phpt",
    "ext/mysqli/tests/mysqli_mysqli_result_invalid_mode.phpt",
    "ext/mysqli/tests/mysqli_options_openbasedir.phpt",
    "ext/mysqli/tests/061.phpt",
    "ext/mysqli/tests/048.phpt",
    "ext/mysqli/tests/mysqli_enable_reads_from_master.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_unregister.phpt",
    "ext/mysqli/tests/mysqli_last_insert_id.phpt",
    "ext/mysqli/tests/mysqli_field_count.phpt",
    "ext/mysqli/tests/mysqli_character_set_name.phpt",
    "ext/mysqli/tests/bug31668.phpt",
    "ext/mysqli/tests/mysqli_driver_unclonable.phpt",
    "ext/mysqli/tests/mysqli_stmt_datatype_change.phpt",
    "ext/mysqli/tests/mysqli_stmt_fetch_geom.phpt",
    "ext/mysqli/tests/bug49027.phpt",
    "ext/mysqli/tests/003.phpt",
    "ext/mysqli/tests/bug45019.phpt",
    "ext/mysqli/tests/mysqli_connect_errno.phpt",
    "ext/mysqli/tests/mysqli_fetch_row.phpt",
    "ext/mysqli/tests/mysqli_stmt_get_warnings.phpt",
    "ext/mysqli/tests/mysqli_pconn_max_links.phpt",
    "ext/mysqli/tests/mysqli_poll_kill.phpt",
    "ext/mysqli/tests/bug52891.phpt",
    "ext/mysqli/tests/049.phpt",
    "ext/mysqli/tests/mysqli_init.phpt",
    "ext/mysqli/tests/mysqli_stmt_send_long_data.phpt",
    "ext/mysqli/tests/mysqli_stmt_prepare.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_new_query.phpt",
    "ext/mysqli/tests/064.phpt",
    "ext/mysqli/tests/mysqli_embedded_connect.phpt",
    "ext/mysqli/tests/020.phpt",
    "ext/mysqli/tests/mysqli_no_reconnect.phpt",
    "ext/mysqli/tests/022.phpt",
    "ext/mysqli/tests/mysqli_stmt_bind_result_format.phpt",
    "ext/mysqli/tests/mysqli_error.phpt",
    "ext/mysqli/tests/006.phpt",
    "ext/mysqli/tests/031.phpt",
    "ext/mysqli/tests/mysqli_insert_id.phpt",
    "ext/mysqli/tests/071.phpt",
    "ext/mysqli/tests/mysqli_result_unclonable.phpt",
    "ext/mysqli/tests/023.phpt",
    "ext/mysqli/tests/mysqli_stmt_errno.phpt",
    "ext/mysqli/tests/058.phpt",
    "ext/mysqli/tests/046.phpt",
    "ext/mysqli/tests/mysqli_real_escape_string.phpt",
    "ext/mysqli/tests/mysqli_max_links.phpt",
    "ext/mysqli/tests/mysqli_fetch_field_direct.phpt",
    "ext/mysqli/tests/bug45289.phpt",
    "ext/mysqli/tests/mysqli_disable_reads_from_master.phpt",
    "ext/mysqli/tests/mysqli_fetch_assoc_no_alias.phpt",
    "ext/mysqli/tests/mysqli_change_user_rollback.phpt",
    "ext/mysqli/tests/mysqli_autocommit_oo.phpt",
    "ext/mysqli/tests/mysqli_stmt_field_count.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_default.phpt",
    "ext/mysqli/tests/mysqli_fetch_object_no_constructor.phpt",
    "ext/mysqli/tests/mysqli_stmt_get_result_seek.phpt",
    "ext/mysqli/tests/mysqli_options_int_and_float_native.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_openbasedir.phpt",
    "ext/mysqli/tests/mysqli_stmt_num_rows.phpt",
    "ext/mysqli/tests/mysqli_stmt_get_result_metadata_fetch_field.phpt",
    "ext/mysqli/tests/mysqli_fetch_array_large.phpt",
    "ext/mysqli/tests/mysqli_query_unicode.phpt",
    "ext/mysqli/tests/bug34810.phpt",
    "ext/mysqli/tests/bug38003.phpt",
    "ext/mysqli/tests/mysqli_query_local_infile_large.phpt",
    "ext/mysqli/tests/069.phpt",
    "ext/mysqli/tests/007.phpt",
    "ext/mysqli/tests/mysqli_select_db.phpt",
    "ext/mysqli/tests/mysqli_stmt_bind_param_type_juggling.phpt",
    "ext/mysqli/tests/mysqli_report_wo_ps.phpt",
    "ext/mysqli/tests/mysqli_change_user_oo.phpt",
    "ext/mysqli/tests/bug37090.phpt",
    "ext/mysqli/tests/mysqli_stmt_reset.phpt",
    "ext/mysqli/tests/074.phpt",
    "ext/mysqli/tests/mysqli_pconn_kill.phpt",
    "ext/mysqli/tests/mysqli_stmt_sqlstate.phpt",
    "ext/mysqli/tests/mysqli_real_escape_string_nobackslash.phpt",
    "ext/mysqli/tests/mysqli_change_user.phpt",
    "ext/mysqli/tests/mysqli_fetch_array_assoc.phpt",
    "ext/mysqli/tests/mysqli_fetch_all_oo.phpt",
    "ext/mysqli/tests/mysqli_fetch_array.phpt",
    "ext/mysqli/tests/mysqli_change_user_prepared_statements.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_nested_call.phpt",
    "ext/mysqli/tests/mysqli_mysqlnd_read_timeout_zero.phpt",
    "ext/mysqli/tests/mysqli_thread_id.phpt",
    "ext/mysqli/tests/mysqli_poll.phpt",
    "ext/mysqli/tests/bug32405.phpt",
    "ext/mysqli/tests/002.phpt",
    "ext/mysqli/tests/mysqli_poll_mixing_insert_select.phpt",
    "ext/mysqli/tests/bug44897.phpt",
    "ext/mysqli/tests/008.phpt",
    "ext/mysqli/tests/018.phpt",
    "ext/mysqli/tests/mysqli_query_stored_proc.phpt",
    "ext/mysqli/tests/mysqli_character_set_name_oo.phpt",
    "ext/mysqli/tests/mysqli_affected_rows.phpt",
    "ext/mysqli/tests/040.phpt",
    "ext/mysqli/tests/mysqli_real_escape_string_gb2312.phpt",
    "ext/mysqli/tests/mysqli_stmt_get_result.phpt",
    "ext/mysqli/tests/056.phpt",
    "ext/mysqli/tests/mysqli_rollback.phpt",
    "ext/mysqli/tests/mysqli_insert_id_variation.phpt",
    "ext/mysqli/tests/034.phpt",
    "ext/mysqli/tests/bug33263.phpt",
    "ext/mysqli/tests/039.phpt",
    "ext/mysqli/tests/035.phpt",
    "ext/mysqli/tests/070.phpt",
    "ext/mysqli/tests/mysqli_dump_debug_info_oo.phpt",
    "ext/mysqli/tests/mysqli_phpinfo.phpt",
    "ext/mysqli/tests/mysqli_fetch_array_many_rows.phpt",
    "ext/mysqli/tests/mysqli_class_mysqli_warning.phpt",
    "ext/mysqli/tests/mysqli_commit_oo.phpt",
    "ext/mysqli/tests/013.phpt",
    "ext/mysqli/tests/mysqli_real_query.phpt",
    "ext/mysqli/tests/037.phpt",
    "ext/mysqli/tests/016.phpt",
    "ext/mysqli/tests/mysqli_commit.phpt",
    "ext/mysqli/tests/mysqli_store_result.phpt",
    "ext/mysqli/tests/mysqli_change_user_get_lock.phpt",
    "ext/mysqli/tests/bug29311.phpt",
    "ext/mysqli/tests/mysqli_stmt_unclonable.phpt",
    "ext/mysqli/tests/mysqli_stmt_bind_result.phpt",
    "ext/mysqli/tests/mysqli_unclonable.phpt",
    "ext/mysqli/tests/055.phpt",
    "ext/mysqli/tests/mysqli_class_mysqli_stmt_interface.phpt",
    "ext/mysqli/tests/bug53503.phpt",
    "ext/mysqli/tests/047.phpt",
    "ext/mysqli/tests/bug49442.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_close_link.phpt",
    "ext/mysqli/tests/mysqli_constants.phpt",
    "ext/mysqli/tests/bug48909.phpt",
    "ext/mysqli/tests/mysqli_stmt_data_seek.phpt",
    "ext/mysqli/tests/mysqli_class_mysqli_driver_interface.phpt",
    "ext/mysqli/tests/mysqli_stmt_get_result_non_select.phpt",
    "ext/mysqli/tests/059.phpt",
    "ext/mysqli/tests/mysqli_ping.phpt",
    "ext/mysqli/tests/mysqli_real_escape_string_eucjpms.phpt",
    "ext/mysqli/tests/025.phpt",
    "ext/mysqli/tests/mysqli_query_iterators.phpt",
    "ext/mysqli/tests/mysqli_data_seek.phpt",
    "ext/mysqli/tests/mysqli_stmt_get_result_field_count.phpt",
    "ext/mysqli/tests/012.phpt",
    "ext/mysqli/tests/038.phpt",
    "ext/mysqli/tests/mysqli_fetch_lengths.phpt",
    "ext/mysqli/tests/bug46614.phpt",
    "ext/mysqli/tests/mysqli_class_mysqli_driver_reflection.phpt",
    "ext/mysqli/tests/mysqli_ssl_set.phpt",
    "ext/mysqli/tests/mysqli_fetch_lengths_oo.phpt",
    "ext/mysqli/tests/021.phpt",
    "ext/mysqli/tests/mysqli_stmt_bind_result_references.phpt",
    "ext/mysqli/tests/bug36949.phpt",
    "ext/mysqli/tests/052.phpt",
    "ext/mysqli/tests/mysqli_fetch_assoc_bit.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_kill_link.phpt",
    "ext/mysqli/tests/mysqli_mysqlnd_read_timeout.phpt",
    "ext/mysqli/tests/mysqli_stmt_attr_set.phpt",
    "ext/mysqli/tests/mysqli_pconn_twice.phpt",
    "ext/mysqli/tests/mysqli_query.phpt",
    "ext/mysqli/tests/mysqli_report.phpt",
    "ext/mysqli/tests/bug30967.phpt",
    "ext/mysqli/tests/bug52082.phpt",
    "ext/mysqli/tests/067.phpt",
    "ext/mysqli/tests/mysqli_autocommit.phpt",
    "ext/mysqli/tests/mysqli_options_init_command.phpt",
    "ext/mysqli/tests/mysqli_error_oo.phpt",
    "ext/mysqli/tests/mysqli_more_results.phpt",
    "ext/mysqli/tests/mysqli_get_client_info.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_closefile.phpt",
    "ext/mysqli/tests/mysqli_stmt_bind_result_zerofill.phpt",
    "ext/mysqli/tests/bug36802.phpt",
    "ext/mysqli/tests/mysqli_set_opt.phpt",
    "ext/mysqli/tests/057.phpt",
    "ext/mysqli/tests/mysqli_get_client_stats_ps.phpt",
    "ext/mysqli/tests/mysqli_get_client_stats_off.phpt",
    "ext/mysqli/tests/mysqli_fetch_object.phpt",
    "ext/mysqli/tests/030.phpt",
    "ext/mysqli/tests/mysqli_fetch_field_direct_oo.phpt",
    "ext/mysqli/tests/mysqli_fetch_all.phpt",
    "ext/mysqli/tests/mysqli_stmt_fetch.phpt",
    "ext/mysqli/tests/mysqli_stmt_send_long_data_packet_size_mysqlnd.phpt",
    "ext/mysqli/tests/mysqli_data_seek_oo.phpt",
    "ext/mysqli/tests/bug39457.phpt",
    "ext/mysqli/tests/mysqli_stmt_fetch_bit.phpt",
    "ext/mysqli/tests/bug34785.phpt",
    "ext/mysqli/tests/mysqli_debug_mysqlnd_control_string.phpt",
    "ext/mysqli/tests/bug42378.phpt",
    "ext/mysqli/tests/bug35517.phpt",
    "ext/mysqli/tests/mysqli_next_result.phpt",
    "ext/mysqli/tests/mysqli_stmt_get_result_metadata.phpt",
    "ext/mysqli/tests/mysqli_fetch_object_no_object.phpt",
    "ext/mysqli/tests/063.phpt",
    "ext/mysqli/tests/032.phpt",
    "ext/mysqli/tests/mysqli_prepare_no_object.phpt",
    "ext/mysqli/tests/bug42548.phpt",
    "ext/mysqli/tests/mysqli_stmt_affected_rows.phpt",
    "ext/mysqli/tests/mysqli_stmt_result_metadata_sqltests.phpt",
    "ext/mysqli/tests/mysqli_stmt_bind_param_check_param_no_change.phpt",
    "ext/mysqli/tests/019.phpt",
    "ext/mysqli/tests/mysqli_field_seek.phpt",
    "ext/mysqli/tests/mysqli_class_mysqli_result_interface.phpt",
    "ext/mysqli/tests/mysqli_set_local_infile_handler_bad_character.phpt",
    "ext/mysqli/tests/bug51605.phpt",
    "ext/mysqli/tests/mysqli_fetch_assoc.phpt",
    "ext/mysqli/tests/mysqli_set_charset.phpt",
    "ext/mysqli/tests/mysqli_mysqlnd_read_timeout_long.phpt",
    "ext/mysqli/tests/mysqli_explain_metadata.phpt",
    "ext/mysqli/tests/mysqli_thread_safe.phpt",
    "ext/mysqli/tests/mysqli_real_escape_string_sjis.phpt",
    "ext/mysqli/tests/mysqli_connect_oo.phpt",
    "ext/mysqli/tests/053.phpt",
    "ext/mysqli/tests/051.phpt",
    "ext/mysqli/tests/mysqli_pconn_conn_multiple.phpt",
    "ext/mysqli/tests/mysqli_fetch_field.phpt",
    "ext/mysqli/tests/024.phpt",
    "ext/mysqli/tests/mysqli_auth_pam.phpt",
    "ext/mysqli/tests/mysqli_fetch_field_flags.phpt",
    "ext/mysqli/tests/065.phpt",
    "ext/mysqli/tests/mysqli_change_user_locks_temporary.phpt",
    "ext/mysqli/tests/009.phpt",
    "ext/mysqli/tests/mysqli_real_connect.phpt",
    "ext/mysqli/tests/mysqli_class_mysqli_reflection.phpt",
    "ext/mysqli/tests/mysqli_stmt_execute.phpt",
    "ext/mysqli/tests/mysqli_pconn_reuse.phpt",
    "ext/mysqli/tests/mysqli_reconnect.phpt",
    "ext/mysqli/tests/mysqli_close_oo.phpt",
    "ext/mysqli/tests/042.phpt",
    "ext/mysqli/tests/mysqli_debug.phpt",
    "ext/mysqli/tests/mysqli_stmt_param_count.phpt",
    "ext/mysqli/tests/mysqli_close.phpt",
    "ext/mysqli/tests/mysqli_get_charset.phpt",
    "ext/mysqli/tests/mysqli_character_set.phpt",
    "ext/mysqli/tests/mysqli_real_escape_string_big5.phpt",
    "ext/mysqli/tests/mysqli_stmt_init.phpt",
    "ext/mysqli/tests/bug28817.phpt",
    "ext/gettext/tests/dcngettext.phpt",
    "ext/gettext/tests/gettext_gettext_error_wrongparams.phpt",
    "ext/gettext/tests/gettext_bindtextdomain-emptydomain.phpt",
    "ext/gettext/tests/gettext_basic.phpt",
    "ext/gettext/tests/gettext_dcgettext.phpt",
    "ext/gettext/tests/gettext_textdomain-retval.phpt",
    "ext/gettext/tests/gettext_bindtextdomain-cwd.phpt",
    "ext/gettext/tests/gettext_dngettext-plural.phpt",
    "ext/gettext/tests/gettext_dngettext-wrongparams.phpt",
    "ext/gettext/tests/gettext_dgettext.phpt",
    "ext/gettext/tests/gettext_bindtextdomain-wrongparams.phpt",
    "ext/gettext/tests/gettext_dgettext_error_wrongparams.phpt",
    "ext/gettext/tests/gettext_ngettext-wrongparams.phpt",
    "ext/gettext/tests/gettext_bind_textdomain_codeset-retval.phpt",
    "ext/gettext/tests/gettext_ngettext.phpt",
    "ext/gettext/tests/gettext_bindtextdomain-path.phpt",
    "ext/gettext/tests/44938.phpt",
    "ext/gettext/tests/gettext_phpinfo.phpt",
    "ext/gettext/tests/gettext_dcgettext-wrongparams.phpt",
    "ext/gettext/tests/gettext_bind_textdomain_codeset-wrongparams.phpt",
    "ext/gettext/tests/gettext_basic-enus.phpt",
    "ext/gettext/tests/gettext_textdomain-wrongparams.phpt",
    "ext/pgsql/tests/pg_delete_001.phpt",
    "ext/pgsql/tests/80_bug42783.phpt",
    "ext/pgsql/tests/07optional.phpt",
    "ext/pgsql/tests/80_bug14383.phpt",
    "ext/pgsql/tests/16pg_result_status.phpt",
    "ext/pgsql/tests/28large_object_import_oid.phpt",
    "ext/pgsql/tests/80_bug36625.phpt",
    "ext/pgsql/tests/bug37100.phpt",
    "ext/pgsql/tests/20pg_get_pid.phpt",
    "ext/pgsql/tests/12pg_insert.phpt",
    "ext/pgsql/tests/80_bug32223.phpt",
    "ext/pgsql/tests/01createdb.phpt",
    "ext/pgsql/tests/03sync_query.phpt",
    "ext/pgsql/tests/80_bug39971.phpt",
    "ext/pgsql/tests/22pg_fetch_object.phpt",
    "ext/pgsql/tests/80_bug24499.phpt",
    "ext/pgsql/tests/17result.phpt",
    "ext/pgsql/tests/12pg_insert_85.phpt",
    "ext/pgsql/tests/pg_insert_001.phpt",
    "ext/pgsql/tests/21pg_get_notify.phpt",
    "ext/pgsql/tests/18pg_escape_bytea.phpt",
    "ext/pgsql/tests/98old_api.phpt",
    "ext/pgsql/tests/13pg_select_85.phpt",
    "ext/pgsql/tests/24sync_query_prepared.phpt",
    "ext/pgsql/tests/09notice.phpt",
    "ext/pgsql/tests/15pg_delete.phpt",
    "ext/pgsql/tests/pg_select_001.phpt",
    "ext/pgsql/tests/14pg_update_85.phpt",
    "ext/pgsql/tests/26async_query_prepared.phpt",
    "ext/pgsql/tests/pg_update_001.phpt",
    "ext/pgsql/tests/80_bug27597.phpt",
    "ext/pgsql/tests/23sync_query_params.phpt",
    "ext/pgsql/tests/06copy.phpt",
    "ext/pgsql/tests/13pg_select.phpt",
    "ext/pgsql/tests/14pg_update.phpt",
    "ext/pgsql/tests/11pg_meta_data.phpt",
    "ext/pgsql/tests/08escape.phpt",
    "ext/pgsql/tests/04async_query.phpt",
    "ext/pgsql/tests/05large_object.phpt",
    "ext/pgsql/tests/pg_meta_data_001.phpt",
    "ext/pgsql/tests/10pg_convert_85.phpt",
    "ext/pgsql/tests/bug37100_85.phpt",
    "ext/pgsql/tests/80_bug32223b.phpt",
    "ext/pgsql/tests/25async_query_params.phpt",
    "ext/pgsql/tests/02connection.phpt",
    "ext/pgsql/tests/19pg_ping.phpt",
    "ext/pgsql/tests/9999dropdb.phpt",
    "ext/pgsql/tests/bug47199.phpt",
    "ext/pgsql/tests/10pg_convert.phpt",
    "ext/pgsql/tests/27large_object_oid.phpt",
    "ext/spl/tests/bug48493.phpt",
    "ext/spl/tests/bug45622.phpt",
    "ext/spl/tests/multiple_iterator_001.phpt",
    "ext/spl/tests/iterator_035.phpt",
    "ext/spl/tests/SplDoublyLinkedList_offsetGet_param_string.phpt",
    "ext/spl/tests/dit_005.phpt",
    "ext/spl/tests/SplFixedArray_offsetSet_one_invalid_parameter.phpt",
    "ext/spl/tests/bug53071.phpt",
    "ext/spl/tests/spl_iterator_apply_error.phpt",
    "ext/spl/tests/fixedarray_019.phpt",
    "ext/spl/tests/SplDoublyLinkedList_current.phpt",
    "ext/spl/tests/DirectoryIterator_getBasename_basic_test.phpt",
    "ext/spl/tests/arrayObject_getFlags_basic2.phpt",
    "ext/spl/tests/SplDoublyLinkedList_setIteratorMode_param_SplDoublyLinkedList.phpt",
    "ext/spl/tests/SplDoublylinkedlist_offsetunset_first002.phpt",
    "ext/spl/tests/recursiveIteratorIterator_endchildren_error.phpt",
    "ext/spl/tests/fileobject_001.phpt",
    "ext/spl/tests/iterator_049b.phpt",
    "ext/spl/tests/SplDoublyLinkedList_isEmpty_empty.phpt",
    "ext/spl/tests/bug41692.phpt",
    "ext/spl/tests/spl_iterator_apply_error_001.phpt",
    "ext/spl/tests/SplDoublyLinkedList_getIteratorMode.phpt",
    "ext/spl/tests/iterator_067.phpt",
    "ext/spl/tests/array_007.phpt",
    "ext/spl/tests/observer_003.phpt",
    "ext/spl/tests/pqueue_002.phpt",
    "ext/spl/tests/limititerator_seek.phpt",
    "ext/spl/tests/bug46053.phpt",
    "ext/spl/tests/pqueue_001.phpt",
    "ext/spl/tests/iterator_002.phpt",
    "ext/spl/tests/bug33136.phpt",
    "ext/spl/tests/bug48361.phpt",
    "ext/spl/tests/bug31926.phpt",
    "ext/spl/tests/SplDoublyLinkedList_top_pass_array.phpt",
    "ext/spl/tests/dit_001.phpt",
    "ext/spl/tests/fixedarray_008.phpt",
    "ext/spl/tests/iterator_004.phpt",
    "ext/spl/tests/pqueue_003.phpt",
    "ext/spl/tests/SplFileObject_fputcsv_variation7.phpt",
    "ext/spl/tests/arrayObject_magicMethods3.phpt",
    "ext/spl/tests/iterator_017.phpt",
    "ext/spl/tests/SplObjectStorage_contains_invalid_parameter.phpt",
    "ext/spl/tests/SplDoublyLinkedList_isEmpty_not-empty-with-parameter.phpt",
    "ext/spl/tests/SplFixedArray_fromarray_non_indexes.phpt",
    "ext/spl/tests/DirectoryIterator_by_reference.phpt",
    "ext/spl/tests/bug31346.phpt",
    "ext/spl/tests/spl_heap_iteration_error.phpt",
    "ext/spl/tests/array_026.phpt",
    "ext/spl/tests/iterator_038.phpt",
    "ext/spl/tests/spl_cachingiterator___toString_basic.phpt",
    "ext/spl/tests/SplFileObject_fputcsv_variation8.phpt",
    "ext/spl/tests/SplHeap_extract_invalid_parameter.phpt",
    "ext/spl/tests/SplFileObject_fputcsv.phpt",
    "ext/spl/tests/heap_004.phpt",
    "ext/spl/tests/array_002.phpt",
    "ext/spl/tests/SplObjectStorage_getHash.phpt",
    "ext/spl/tests/SplFixedArray_construct_param_SplFixedArray.phpt",
    "ext/spl/tests/spl_iterator_to_array_basic.phpt",
    "ext/spl/tests/fixedarray_012.phpt",
    "ext/spl/tests/heap_top_variation_002.phpt",
    "ext/spl/tests/iterator_065.phpt",
    "ext/spl/tests/SplFileObject_fseek_error_001.phpt",
    "ext/spl/tests/iterator_032.phpt",
    "ext/spl/tests/recursiveiteratoriterator_enditeration_basic.phpt",
    "ext/spl/tests/bug51532.phpt",
    "ext/spl/tests/spl_iterator_to_array_error.phpt",
    "ext/spl/tests/bug38618.phpt",
    "ext/spl/tests/spl_autoload_010.phpt",
    "ext/spl/tests/SplObjectStorage_removeAllExcept_basic.phpt",
    "ext/spl/tests/SplFixedArray_setsize_shrink.phpt",
    "ext/spl/tests/SplTempFileObject_constructor_memory_lt1_variation.phpt",
    "ext/spl/tests/SplFixedArray_getSize_pass_param.phpt",
    "ext/spl/tests/iterator_031.phpt",
    "ext/spl/tests/bug41828.phpt",
    "ext/spl/tests/SplFixedArray_setSize_reduce.phpt",
    "ext/spl/tests/iterator_025.phpt",
    "ext/spl/tests/SplObjectStorage_removeAllExcept_invalid_parameter.phpt",
    "ext/spl/tests/recursive_tree_iterator_006.phpt",
    "ext/spl/tests/SplFixedArray_setSize_param_float.phpt",
    "ext/spl/tests/pqueue_compare_basic.phpt",
    "ext/spl/tests/arrayObject_ksort_basic2.phpt",
    "ext/spl/tests/spl_pq_top_error_corrupt.phpt",
    "ext/spl/tests/iterator_011.phpt",
    "ext/spl/tests/spl_004.phpt",
    "ext/spl/tests/heap_005.phpt",
    "ext/spl/tests/SplObjectStorage_offsetGet.phpt",
    "ext/spl/tests/regexiterator_setpregflags_exception.phpt",
    "ext/spl/tests/SplFileObject_fgetcsv_enclosure_error.phpt",
    "ext/spl/tests/spl_006.phpt",
    "ext/spl/tests/SplObjectStorage_addAll_invalid_parameter.phpt",
    "ext/spl/tests/iterator_063.phpt",
    "ext/spl/tests/array_019.phpt",
    "ext/spl/tests/bug40036.phpt",
    "ext/spl/tests/arrayObject_magicMethods5.phpt",
    "ext/spl/tests/iterator_030.phpt",
    "ext/spl/tests/arrayObject___construct_basic5.phpt",
    "ext/spl/tests/fileobject_003.phpt",
    "ext/spl/tests/bug52238.phpt",
    "ext/spl/tests/arrayObject_uasort_basic1.phpt",
    "ext/spl/tests/fixedarray_003.phpt",
    "ext/spl/tests/iterator_033.phpt",
    "ext/spl/tests/arrayObject_uksort_error1.phpt",
    "ext/spl/tests/SplFixedArray_rewind_param.phpt",
    "ext/spl/tests/dit_004.phpt",
    "ext/spl/tests/spl_pq_top_basic.phpt",
    "ext/spl/tests/bug54323.phpt",
    "ext/spl/tests/array_009.phpt",
    "ext/spl/tests/SplObjectStorage_getInfo_empty_storage.phpt",
    "ext/spl/tests/fixedarray_004.phpt",
    "ext/spl/tests/iterator_029.phpt",
    "ext/spl/tests/SplObjectStorage_detach_invalid_parameter.phpt",
    "ext/spl/tests/SplFixedArray_next_param.phpt",
    "ext/spl/tests/array_021.phpt",
    "ext/spl/tests/SplObjectStorage_current_empty_storage.phpt",
    "ext/spl/tests/SplFixedArray_setSize_filled_to_smaller.phpt",
    "ext/spl/tests/dllist_006.phpt",
    "ext/spl/tests/dit_006.phpt",
    "ext/spl/tests/heap_011.phpt",
    "ext/spl/tests/heap_corruption.phpt",
    "ext/spl/tests/arrayObject___construct_basic6.phpt",
    "ext/spl/tests/bug54291.phpt",
    "ext/spl/tests/regexIterator_flags_basic.phpt",
    "ext/spl/tests/recursiveiteratoriterator_nextelement_basic.phpt",
    "ext/spl/tests/fixedarray_006.phpt",
    "ext/spl/tests/iterator_005.phpt",
    "ext/spl/tests/bug48023.phpt",
    "ext/spl/tests/iterator_061.phpt",
    "ext/spl/tests/bug45826.phpt",
    "ext/spl/tests/recursive_tree_iterator_005.phpt",
    "ext/spl/tests/observer_007.phpt",
    "ext/spl/tests/heap_009.phpt",
    "ext/spl/tests/bug49723.phpt",
    "ext/spl/tests/spl_iterator_getcallchildren.phpt",
    "ext/spl/tests/SplFixedArray_offsetGet_invalid_parameter.phpt",
    "ext/spl/tests/observer_005.phpt",
    "ext/spl/tests/fixedarray_018.phpt",
    "ext/spl/tests/spl_heap_count_error.phpt",
    "ext/spl/tests/array_015.phpt",
    "ext/spl/tests/observer_006.phpt",
    "ext/spl/tests/heap_isempty_variation_001.phpt",
    "ext/spl/tests/SplFixedArray_setSize_param_array.phpt",
    "ext/spl/tests/SplDoublyLinkedList_count_param_SplDoublyLinkedList.phpt",
    "ext/spl/tests/SplFixedArray_fromArray_invalid_parameter_002.phpt",
    "ext/spl/tests/recursive_tree_iterator_002.phpt",
    "ext/spl/tests/SplDoublyLinkedList_top_pass_float.phpt",
    "ext/spl/tests/SplFileObject_fwrite_variation_001.phpt",
    "ext/spl/tests/SplPriorityQueue_insert_invalid_parameter.phpt",
    "ext/spl/tests/dit_001_noglob.phpt",
    "ext/spl/tests/iterator_006.phpt",
    "ext/spl/tests/SplDoublyLinkedList_top_pass_null.phpt",
    "ext/spl/tests/SplFileObject_fwrite_variation_002.phpt",
    "ext/spl/tests/SplDoublyLinkedList_top_pass_integer.phpt",
    "ext/spl/tests/heap_it_current_empty.phpt",
    "ext/spl/tests/SplFixedArray_fromArray_invalid_parameter_001.phpt",
    "ext/spl/tests/array_023.phpt",
    "ext/spl/tests/SplFixedArray_fromarray_indexes.phpt",
    "ext/spl/tests/iterator_012.phpt",
    "ext/spl/tests/bug47534.phpt",
    "ext/spl/tests/SplFixedArray_offsetExists_invalid_parameter.phpt",
    "ext/spl/tests/arrayObject_count_basic1.phpt",
    "ext/spl/tests/bug37457.phpt",
    "ext/spl/tests/spl_heap_count_basic.phpt",
    "ext/spl/tests/fileobject_getsize_basic.phpt",
    "ext/spl/tests/arrayObject_exchangeArray_basic2.phpt",
    "ext/spl/tests/observer_002.phpt",
    "ext/spl/tests/spl_minheap_compare_error.phpt",
    "ext/spl/tests/bug36941.phpt",
    "ext/spl/tests/bug42364.phpt",
    "ext/spl/tests/regexiterator_setpregflags.phpt",
    "ext/spl/tests/recursiveiteratoriterator_getsubiterator_basic.phpt",
    "ext/spl/tests/SplFileInfo_getOwner_error.phpt",
    "ext/spl/tests/array_001.phpt",
    "ext/spl/tests/spl_005.phpt",
    "ext/spl/tests/SplFileObject_fputcsv_variation5.phpt",
    "ext/spl/tests/observer_009.phpt",
    "ext/spl/tests/bug42654.phpt",
    "ext/spl/tests/SplDoublyLinkedList_offsetUnset_negative-parameter.phpt",
    "ext/spl/tests/dllist_007.phpt",
    "ext/spl/tests/DirectoryIterator_getInode_basic.phpt",
    "ext/spl/tests/recursive_tree_iterator_008.phpt",
    "ext/spl/tests/dllist_002.phpt",
    "ext/spl/tests/SplFixedArray__construct_param_string.phpt",
    "ext/spl/tests/DirectoryIterator_getExtension_basic.phpt",
    "ext/spl/tests/dllist_010.phpt",
    "ext/spl/tests/iterator_050.phpt",
    "ext/spl/tests/heap_002.phpt",
    "ext/spl/tests/arrayObject___construct_basic4.phpt",
    "ext/spl/tests/iterator_040.phpt",
    "ext/spl/tests/SplFileObject_fgetcsv_delimiter_error.phpt",
    "ext/spl/tests/SplObjectStorage_attach_invalid_parameter.phpt",
    "ext/spl/tests/recursiveiteratoriterator_getsubiterator_variation_002.phpt",
    "ext/spl/tests/array_024.phpt",
    "ext/spl/tests/bug45216.phpt",
    "ext/spl/tests/iterator_047.phpt",
    "ext/spl/tests/spl_007.phpt",
    "ext/spl/tests/bug42703.phpt",
    "ext/spl/tests/bug46115.phpt",
    "ext/spl/tests/iterator_008.phpt",
    "ext/spl/tests/SplFixedArray_setSize_param_null.phpt",
    "ext/spl/tests/fixedarray_010.phpt",
    "ext/spl/tests/SplFixedArray_current_param.phpt",
    "ext/spl/tests/SplObjectStorage_unserialize_nested.phpt",
    "ext/spl/tests/iterator_041.phpt",
    "ext/spl/tests/iterator_036.phpt",
    "ext/spl/tests/spl_fileinfo_getlinktarget_basic.phpt",
    "ext/spl/tests/fixedarray_021.phpt",
    "ext/spl/tests/iterator_018.phpt",
    "ext/spl/tests/iterator_042.phpt",
    "ext/spl/tests/iterator_052.phpt",
    "ext/spl/tests/spl_003.phpt",
    "ext/spl/tests/arrayObject_getIteratorClass_basic1.phpt",
    "ext/spl/tests/SplObjectStorage_unserialize_invalid_parameter3.phpt",
    "ext/spl/tests/spl_heap_extract_parameter_error.phpt",
    "ext/spl/tests/bug53144.phpt",
    "ext/spl/tests/arrayObject_setIteratorClass_error1.phpt",
    "ext/spl/tests/bug50579.phpt",
    "ext/spl/tests/bug49263.phpt",
    "ext/spl/tests/countable_count_variation1.phpt",
    "ext/spl/tests/SplFileObject_fgetcsv_delimiter_basic.phpt",
    "ext/spl/tests/SplObjectStorage_removeAll_invalid_parameter.phpt",
    "ext/spl/tests/bug41691.phpt",
    "ext/spl/tests/iterator_037.phpt",
    "ext/spl/tests/recursiveIteratorIterator_beginchildren_error.phpt",
    "ext/spl/tests/iterator_055.phpt",
    "ext/spl/tests/array_017.phpt",
    "ext/spl/tests/iterator_041a.phpt",
    "ext/spl/tests/spl_heap_isempty.phpt",
    "ext/spl/tests/iterator_066.phpt",
    "ext/spl/tests/spl_heap_recoverfromcorruption_arguments.phpt",
    "ext/spl/tests/array_010.phpt",
    "ext/spl/tests/iterator_014.phpt",
    "ext/spl/tests/SplFileObject_fgetcsv_enclosure_basic.phpt",
    "ext/spl/tests/iterator_022.phpt",
    "ext/spl/tests/SplFileInfo_getInode_error.phpt",
    "ext/spl/tests/SplFixedArray_setsize_grow.phpt",
    "ext/spl/tests/SplFixedArray_key_setsize.phpt",
    "ext/spl/tests/regexIterator_mode_basic.phpt",
    "ext/spl/tests/array_013.phpt",
    "ext/spl/tests/bug31185.phpt",
    "ext/spl/tests/bug54281.phpt",
    "ext/spl/tests/iterator_044.phpt",
    "ext/spl/tests/iterator_013.phpt",
    "ext/spl/tests/SplFileInfo_getPerms_error.phpt",
    "ext/spl/tests/SplHeap_insert_invalid_parameter.phpt",
    "ext/spl/tests/SplFileInfo_getInode_basic.phpt",
    "ext/spl/tests/arrayObject_magicMethods4.phpt",
    "ext/spl/tests/bug31348.phpt",
    "ext/spl/tests/class_implements_variation1.phpt",
    "ext/spl/tests/iterator_048.phpt",
    "ext/spl/tests/fixedarray_017.phpt",
    "ext/spl/tests/SplDoublyLinkedList_lifoMode.phpt",
    "ext/spl/tests/spl_maxheap_compare_basic.phpt",
    "ext/spl/tests/array_022.phpt",
    "ext/spl/tests/SplQueue_setIteratorMode_param_lifo.phpt",
    "ext/spl/tests/splDoublyLinkedList_shift_noParams.phpt",
    "ext/spl/tests/iterator_016.phpt",
    "ext/spl/tests/spl_heap_is_empty_basic.phpt",
    "ext/spl/tests/dit_002.phpt",
    "ext/spl/tests/SplFixedArray_setsize_001.phpt",
    "ext/spl/tests/fixedarray_009.phpt",
    "ext/spl/tests/DirectoryIterator_getBasename_pass_array.phpt",
    "ext/spl/tests/bug32134.phpt",
    "ext/spl/tests/SplTempFileObject_constructor_error.phpt",
    "ext/spl/tests/SplFileObject_fputcsv_variation14.phpt",
    "ext/spl/tests/class_implements_variation.phpt",
    "ext/spl/tests/iterator_057.phpt",
    "ext/spl/tests/array_003.phpt",
    "ext/spl/tests/SplDoublyLinkedList_offsetGet_param_array.phpt",
    "ext/spl/tests/spl_caching_iterator_constructor_flags.phpt",
    "ext/spl/tests/iterator_059.phpt",
    "ext/spl/tests/SplObjectStorage_unserialize_invalid_parameter2.phpt",
    "ext/spl/tests/SplFixedArray_count_param_int.phpt",
    "ext/spl/tests/SplFixedArray_offsetUnset_invalid_parameter.phpt",
    "ext/spl/tests/fixedarray_015.phpt",
    "ext/spl/tests/fixedarray_020.phpt",
    "ext/spl/tests/DirectoryIterator_empty_constructor.phpt",
    "ext/spl/tests/recursive_tree_iterator_001.phpt",
    "ext/spl/tests/iterator_071.phpt",
    "ext/spl/tests/SplPriorityQueue_extract_invalid_parameter.phpt",
    "ext/spl/tests/array_011.phpt",
    "ext/spl/tests/SplFileObject_fgetcsv_escape_default.phpt",
    "ext/spl/tests/array_016.phpt",
    "ext/spl/tests/bug28822.phpt",
    "ext/spl/tests/SplDoublylinkedlist_offsetunset_first.phpt",
    "ext/spl/tests/SplFileObject_fputcsv_variation10.phpt",
    "ext/spl/tests/bug44615.phpt",
    "ext/spl/tests/SplFixedArray__construct_param_float.phpt",
    "ext/spl/tests/SplFileObject_fputcsv_variation12.phpt",
    "ext/spl/tests/spl_recursive_iterator_iterator_key_case.phpt",
    "ext/spl/tests/spl_autoload_011.phpt",
    "ext/spl/tests/SplFixedArray_toArray_with-params.phpt",
    "ext/spl/tests/SplFixedArray_count_checkParams.phpt",
    "ext/spl/tests/spl_iterator_caching_count_error.phpt",
    "ext/spl/tests/iterator_001.phpt",
    "ext/spl/tests/SplObjectStorage_unserialize_invalid_parameter1.phpt",
    "ext/spl/tests/iterator_023.phpt",
    "ext/spl/tests/SplFileObject_ftruncate_error_001.phpt",
    "ext/spl/tests/CallbackFilterIteratorTest.phpt",
    "ext/spl/tests/fileobject_getcurrentline_basic.phpt",
    "ext/spl/tests/SplObjectStorage_unserialize_bad.phpt",
    "ext/spl/tests/bug51119.phpt",
    "ext/spl/tests/recursive_tree_iterator_003.phpt",
    "ext/spl/tests/observer_008.phpt",
    "ext/spl/tests/spl_limit_iterator_check_limits.phpt",
    "ext/spl/tests/bug49972.phpt",
    "ext/spl/tests/regexIterator_setMode_error.phpt",
    "ext/spl/tests/recursiveiteratoriterator_beginiteration_basic.phpt",
    "ext/spl/tests/spl_autoload_001.phpt",
    "ext/spl/tests/SplFileObject_fgetcsv_escape_error.phpt",
    "ext/spl/tests/array_025.phpt",
    "ext/spl/tests/arrayObject_magicMethods2.phpt",
    "ext/spl/tests/bug40091.phpt",
    "ext/spl/tests/recursive_tree_iterator_007.phpt",
    "ext/spl/tests/SplTempFileObject_constructor_maxmemory_basic.phpt",
    "ext/spl/tests/fileobject_getmaxlinelen_basic.phpt",
    "ext/spl/tests/bug36825.phpt",
    "ext/spl/tests/bug53362.phpt",
    "ext/spl/tests/DirectoryIterator_getInode_error.phpt",
    "ext/spl/tests/iterator_043.phpt",
    "ext/spl/tests/iterator_068.phpt",
    "ext/spl/tests/SplFileObject_fwrite_variation_003.phpt",
    "ext/spl/tests/iterator_051.phpt",
    "ext/spl/tests/fileobject_setmaxlinelen_basic.phpt",
    "ext/spl/tests/pqueue_current_error.phpt",
    "ext/spl/tests/ArrayObject_unserialize_empty_string.phpt",
    "ext/spl/tests/arrayObject_uksort_basic1.phpt",
    "ext/spl/tests/SplObjectStorage_var_dump.phpt",
    "ext/spl/tests/dit_003.phpt",
    "ext/spl/tests/spl_autoload_002.phpt",
    "ext/spl/tests/bug32394.phpt",
    "ext/spl/tests/array_012.phpt",
    "ext/spl/tests/SplFixedArray_offsetUnset_string.phpt",
    "ext/spl/tests/SplFileObject_seek_error_001.phpt",
    "ext/spl/tests/arrayObject___construct_error2.phpt",
    "ext/spl/tests/fileobject_002.phpt",
    "ext/spl/tests/heap_012.phpt",
    "ext/spl/tests/fixedarray_001.phpt",
    "ext/spl/tests/iterator_028.phpt",
    "ext/spl/tests/iterator_046.phpt",
    "ext/spl/tests/iterator_069.phpt",
    "ext/spl/tests/iterator_count.phpt",
    "ext/spl/tests/arrayObject_natsort_basic1.phpt",
    "ext/spl/tests/spl_pq_top_error_empty.phpt",
    "ext/spl/tests/heap_007.phpt",
    "ext/spl/tests/arrayObject___construct_basic1.phpt",
    "ext/spl/tests/iterator_053.phpt",
    "ext/spl/tests/SplDoublyLinkedList_isEmpty_not-empty.phpt",
    "ext/spl/tests/arrayObject_natcasesort_basic1.phpt",
    "ext/spl/tests/bug51374.phpt",
    "ext/spl/tests/arrayObject_uasort_error1.phpt",
    "ext/spl/tests/spl_autoload_007.phpt",
    "ext/spl/tests/fixedarray_013.phpt",
    "ext/spl/tests/SplFileObject_fputcsv_error.phpt",
    "ext/spl/tests/bug46051.phpt",
    "ext/spl/tests/fixedarray_011.phpt",
    "ext/spl/tests/SplFileObject_fputcsv_variation13.phpt",
    "ext/spl/tests/bug45614.phpt",
    "ext/spl/tests/SplDoublyLinkedList_offsetGet_empty.phpt",
    "ext/spl/tests/SplFileObject_fputcsv_variation6.phpt",
    "ext/spl/tests/spl_heap_insert_basic.phpt",
    "ext/spl/tests/SplDoublyLinkedList_bottom_pass_integer.phpt",
    "ext/spl/tests/spl_autoload_009.phpt",
    "ext/spl/tests/SplDoublyLinkedList_current_empty.phpt",
    "ext/spl/tests/SplDoublyLinkedList_bottom_pass_array.phpt",
    "ext/spl/tests/spl_autoload_005.phpt",
    "ext/spl/tests/heap_001.phpt",
    "ext/spl/tests/CallbackFilterIteratorTest-002.phpt",
    "ext/spl/tests/fileobject_checktype_basic.phpt",
    "ext/spl/tests/SplFixedArray_key_param.phpt",
    "ext/spl/tests/SplFixedArray_fromarray_param_boolean.phpt",
    "ext/spl/tests/arrayObject_asort_basic1.phpt",
    "ext/spl/tests/heap_top_variation_001.phpt",
    "ext/spl/tests/SplDoublyLinkedList_push_missing_parameter.phpt",
    "ext/spl/tests/spl_iterator_iterator_constructor.phpt",
    "ext/spl/tests/spl_autoload_008.phpt",
    "ext/spl/tests/arrayObject_magicMethods1.phpt",
    "ext/spl/tests/SplFixedArray__construct_param_array.phpt",
    "ext/spl/tests/SplDoublyLinkedList_bottom_pass_null.phpt",
    "ext/spl/tests/SplFileObject_fgetcsv_escape_basic.phpt",
    "ext/spl/tests/array_004.phpt",
    "ext/spl/tests/SplDoublyLinkedList_offsetExists_success.phpt",
    "ext/spl/tests/dllist_008.phpt",
    "ext/spl/tests/arrayObject_exchangeArray_basic1.phpt",
    "ext/spl/tests/spl_autoload_003.phpt",
    "ext/spl/tests/iterator_045.phpt",
    "ext/spl/tests/spl_recursiveIteratorIterator_setMaxDepth_parameter_count.phpt",
    "ext/spl/tests/heap_008.phpt",
    "ext/spl/tests/SplDoublyLinkedList_pop_noParams.phpt",
    "ext/spl/tests/SplObjectStorage_offsetGet_missing_object.phpt",
    "ext/spl/tests/iterator_058.phpt",
    "ext/spl/tests/spl_autoload_004.phpt",
    "ext/spl/tests/SplArray_fromArray.phpt",
    "ext/spl/tests/recursiveIteratorIterator_nextelement_error.phpt",
    "ext/spl/tests/SplQueue_setIteratorMode.phpt",
    "ext/spl/tests/array_006.phpt",
    "ext/spl/tests/spl_autoload_bug48541.phpt",
    "ext/spl/tests/array_020.phpt",
    "ext/spl/tests/SplFixedArray__construct_param_null.phpt",
    "ext/spl/tests/splpriorityqueue_setextractflags.phpt",
    "ext/spl/tests/bug45622b.phpt",
    "ext/spl/tests/dllist_001.phpt",
    "ext/spl/tests/bug46031.phpt",
    "ext/spl/tests/bug46088.phpt",
    "ext/spl/tests/recursive_tree_iterator_004.phpt",
    "ext/spl/tests/arrayObject_clone_basic3.phpt",
    "ext/spl/tests/SplFileObject_getCsvControl_basic_001.phpt",
    "ext/spl/tests/array_027.phpt",
    "ext/spl/tests/dllist_011.phpt",
    "ext/spl/tests/spl_autoload_014.phpt",
    "ext/spl/tests/spl_autoload_013.phpt",
    "ext/spl/tests/recursiveiteratoriterator_getsubiterator_error.phpt",
    "ext/spl/tests/iterator_024.phpt",
    "ext/spl/tests/splpriorityqueue_extract.phpt",
    "ext/spl/tests/SplObjectStorage_offsetGet_invalid_parameter.phpt",
    "ext/spl/tests/SplFileObject_fflush_basic_001.phpt",
    "ext/spl/tests/SplFileObject_fputcsv_variation1.phpt",
    "ext/spl/tests/SplDoublyLinkedList_pop_params.phpt",
    "ext/spl/tests/countable_class_basic1.phpt",
    "ext/spl/tests/bug36258.phpt",
    "ext/spl/tests/splfixedarray_offsetExists_larger.phpt",
    "ext/spl/tests/iterator_060.phpt",
    "ext/spl/tests/regexiterator_setflags_exception.phpt",
    "ext/spl/tests/heap_003.phpt",
    "ext/spl/tests/iterator_049.phpt",
    "ext/spl/tests/iterator_054.phpt",
    "ext/spl/tests/bug53515.phpt",
    "ext/spl/tests/class_implements_basic.phpt",
    "ext/spl/tests/recursiveiteratoriterator_getsubiterator_variation.phpt",
    "ext/spl/tests/array_014.phpt",
    "ext/spl/tests/spl_autoload_006.phpt",
    "ext/spl/tests/bug38325.phpt",
    "ext/spl/tests/bug52573.phpt",
    "ext/spl/tests/array_018.phpt",
    "ext/spl/tests/spl_pq_top_error_args.phpt",
    "ext/spl/tests/fixedarray_005.phpt",
    "ext/spl/tests/SplTempFileObject_constructor_basic.phpt",
    "ext/spl/tests/heap_current_variation_001.phpt",
    "ext/spl/tests/iterator_034.phpt",
    "ext/spl/tests/pqueue_compare_error.phpt",
    "ext/spl/tests/bug40442.phpt",
    "ext/spl/tests/observer_001.phpt",
    "ext/spl/tests/SplFileInfo_getOwner_basic.phpt",
    "ext/spl/tests/SplHeap_isEmpty_invalid_parameter.phpt",
    "ext/spl/tests/SplFixedArray_offsetSet_invalid_parameter.phpt",
    "ext/spl/tests/class_implements_variation2.phpt",
    "ext/spl/tests/pqueue_004.phpt",
    "ext/spl/tests/SplDoublyLinkedList_offsetSet_invalid_parameter.phpt",
    "ext/spl/tests/SplHeap_count_invalid_parameter.phpt",
    "ext/spl/tests/iterator_to_array.phpt",
    "ext/spl/tests/SplDoublyLinkedList_unshift_missing_parameter.phpt",
    "ext/spl/tests/spl_iterator_recursive_getiterator_error.phpt",
    "ext/spl/tests/iterator_010.phpt",
    "ext/spl/tests/SplObjectStorage_setInfo_invalid_parameter.phpt",
    "ext/spl/tests/arrayObject_clone_basic1.phpt",
    "ext/spl/tests/fixedarray_002.phpt",
    "ext/spl/tests/arrayObject___construct_error1.phpt",
    "ext/spl/tests/SplHeap_isEmpty.phpt",
    "ext/spl/tests/SplDoublyLinkedList_bottom_pass_float.phpt",
    "ext/spl/tests/SplDoublyLinkedList_debug-info.phpt",
    "ext/spl/tests/heap_top_variation_003.phpt",
    "ext/spl/tests/SplDoublyLinkedList_offsetSet_one_invalid_parameter.phpt",
    "ext/spl/tests/fixedarray_016.phpt",
    "ext/spl/tests/SplFileObject_fputcsv_variation11.phpt",
    "ext/spl/tests/dllist_003.phpt",
    "ext/spl/tests/iterator_070.phpt",
    "ext/spl/tests/arrayObject_setFlags_basic2.phpt",
    "ext/spl/tests/arrayObject_clone_basic2.phpt",
    "ext/spl/tests/arrayObject___construct_basic3.phpt",
    "ext/spl/tests/array_005.phpt",
    "ext/spl/tests/bug41528.phpt",
    "ext/spl/tests/SplFileObject_fgetcsv_basic.phpt",
    "ext/spl/tests/iterator_009.phpt",
    "ext/spl/tests/fixedarray_007.phpt",
    "ext/spl/tests/SplDoublylinkedlist_offsetunset_last.phpt",
    "ext/spl/tests/SplFileInfo_getGroup_error.phpt",
    "ext/spl/tests/spl_iterator_caching_getcache_error.phpt",
    "ext/spl/tests/recursive_tree_iterator_setprefixpart.phpt",
    "ext/spl/tests/regexiterator_getregex.phpt",
    "ext/spl/tests/iterator_019.phpt",
    "ext/spl/tests/bug40872.phpt",
    "ext/spl/tests/fileobject_005.phpt",
    "ext/spl/tests/arrayObject___construct_basic2.phpt",
    "ext/spl/tests/arrayObject_asort_basic2.phpt",
    "ext/spl/tests/fixedarray_014.phpt",
    "ext/spl/tests/SplFileObject_fwrite_error_001.phpt",
    "ext/spl/tests/iterator_039.phpt",
    "ext/spl/tests/recursiveIteratorIterator_callHasChildren_error.phpt",
    "ext/spl/tests/iterator_056.phpt",
    "ext/spl/tests/SplFileInfo_getPerms_basic.phpt",
    "ext/spl/tests/dllist_004.phpt",
    "ext/spl/tests/bug36287.phpt",
    "ext/spl/tests/SplFileObject_fpassthru_basic.phpt",
    "ext/spl/tests/arrayObject_magicMethods6.phpt",
    "ext/spl/tests/class_implements_basic2.phpt",
    "ext/spl/tests/heap_010.phpt",
    "ext/spl/tests/SplFileInfo_getExtension_basic.phpt",
    "ext/spl/tests/arrayObject_ksort_basic1.phpt",
    "ext/spl/tests/iterator_026.phpt",
    "ext/spl/tests/recursiveiteratoriterator_getsubiterator_variation_003.phpt",
    "ext/spl/tests/spl_iterator_caching_count_basic.phpt",
    "ext/spl/tests/bug46160.phpt",
    "ext/spl/tests/SplFixedArray_offsetExists_less_than_zero.phpt",
    "ext/spl/tests/regexiterator_getpregflags.phpt",
    "ext/spl/tests/iterator_007.phpt",
    "ext/spl/tests/SplFixedArray_fromarray_param_multiarray.phpt",
    "ext/spl/tests/RecursiveCallbackFilterIteratorTest.phpt",
    "ext/spl/tests/array_009a.phpt",
    "ext/spl/tests/SplFileObject_fscanf_basic.phpt",
    "ext/spl/tests/heap_006.phpt",
    "ext/spl/tests/arrayObject_setFlags_basic1.phpt",
    "ext/spl/tests/fileobject_getbasename_basic.phpt",
    "ext/spl/tests/iterator_003.phpt",
    "ext/spl/tests/bug44144.phpt",
    "ext/spl/tests/spl_002.phpt",
    "ext/spl/tests/iterator_062.phpt",
    "ext/spl/tests/iterator_015.phpt",
    "ext/spl/tests/spl_autoload_012.phpt",
    "ext/spl/tests/iterator_021.phpt",
    "ext/spl/tests/array_008.phpt",
    "ext/spl/tests/dllist_memleak.phpt",
    "ext/spl/tests/SplDoublyLinkedList_offsetUnset_parameter-larger-num-elements.phpt",
    "ext/spl/tests/fileobject_getfileinfo_basic.phpt",
    "ext/spl/tests/iterator_064.phpt",
    "ext/spl/tests/fileobject_004.phpt",
    "ext/spl/tests/spl_001.phpt",
    "ext/spl/tests/dllist_012.phpt",
    "ext/spl/tests/SplDoublyLinkedList_offsetExists_invalid_parameter.phpt",
    "ext/spl/tests/SplDoublyLinkedList_offsetGet_missing_param.phpt",
    "ext/spl/tests/observer_004.phpt",
    "ext/spl/tests/spl_priorityqeue_insert_two_params_error.phpt",
    "ext/spl/tests/SplFileInfo_getGroup_basic.phpt",
    "ext/spl/tests/SplFileObject_fputcsv_002.phpt",
    "ext/spl/tests/spl_classes.phpt",
    "ext/spl/tests/SplDoublyLinkedList_isEmpty_empty-with-parameter.phpt",
    "ext/spl/tests/SplStack_setIteratorMode.phpt",
    "ext/spl/tests/bug54384.phpt",
    "ext/spl/tests/iterator_027.phpt",
    "ext/spl/tests/SplDoublyLinkedList_getIteratorMode_error.phpt",
    "ext/spl/tests/arrayObject_getFlags_basic1.phpt",
    "ext/spl/tests/bug34548.phpt",
    "ext/spl/tests/SplDoublyLinkedList_count.phpt",
    "ext/spl/tests/spl_cachingiterator_setFlags_basic.phpt",
    "ext/spl/tests/iterator_020.phpt",
    "ext/spl/tests/iterator_041b.phpt",
    "ext/spl/tests/arrayObject_exchangeArray_basic3.phpt",
    "ext/spl/tests/dllist_005.phpt",
    "ext/spl/tests/bug54292.phpt",
    "ext/spl/tests/SplDoublyLinkedList_offsetUnset_greater_than_elements.phpt",
    "ext/spl/tests/SplObjectStorage_setInfo_empty_storage.phpt",
    "ext/spl/examples/tests/dualiterator_001.phpt",
    "ext/mysql/tests/mysql_fetch_row.phpt",
    "ext/mysql/tests/mysql_field_len.phpt",
    "ext/mysql/tests/mysql_info.phpt",
    "ext/mysql/tests/mysql_escape_string.phpt",
    "ext/mysql/tests/mysql_trace_mode.phpt",
    "ext/mysql/tests/mysql_list_processes.phpt",
    "ext/mysql/tests/mysql_fetch_object.phpt",
    "ext/mysql/tests/bug51242.phpt",
    "ext/mysql/tests/mysql_affected_rows.phpt",
    "ext/mysql/tests/mysql_mysqlnd_read_timeout_long.phpt",
    "ext/mysql/tests/mysql_phpinfo.phpt",
    "ext/mysql/tests/mysql_real_escape_string.phpt",
    "ext/mysql/tests/mysql_fetch_lengths.phpt",
    "ext/mysql/tests/mysql_pconn_reuse.phpt",
    "ext/mysql/tests/001.phpt",
    "ext/mysql/tests/mysql_data_seek.phpt",
    "ext/mysql/tests/mysql_max_links.phpt",
    "ext/mysql/tests/mysql_select_db.phpt",
    "ext/mysql/tests/mysql_list_tables.phpt",
    "ext/mysql/tests/mysql_query.phpt",
    "ext/mysql/tests/mysql_fetch_field.phpt",
    "ext/mysql/tests/mysql_fetch_array.phpt",
    "ext/mysql/tests/mysql_max_persistent.phpt",
    "ext/mysql/tests/mysql_insert_id.phpt",
    "ext/mysql/tests/bug53649.phpt",
    "ext/mysql/tests/003.phpt",
    "ext/mysql/tests/mysql_fetch_assoc.phpt",
    "ext/mysql/tests/mysql_ping.phpt",
    "ext/mysql/tests/mysql_get_client_info.phpt",
    "ext/mysql/tests/mysql_free_result.phpt",
    "ext/mysql/tests/mysql_errno.phpt",
    "ext/mysql/tests/mysql_pconnect.phpt",
    "ext/mysql/tests/bug47438.phpt",
    "ext/mysql/tests/mysql_thread_id.phpt",
    "ext/mysql/tests/mysql_get_server_info.phpt",
    "ext/mysql/tests/mysql_db_name.phpt",
    "ext/mysql/tests/mysql_client_encoding.phpt",
    "ext/mysql/tests/mysql_pconn_disable.phpt",
    "ext/mysql/tests/mysql_result.phpt",
    "ext/mysql/tests/mysql_unbuffered_query.phpt",
    "ext/mysql/tests/mysql_get_host_info.phpt",
    "ext/mysql/tests/mysql_deprecated_api.phpt",
    "ext/mysql/tests/mysql_set_charset.phpt",
    "ext/mysql/tests/mysql_error.phpt",
    "ext/mysql/tests/002.phpt",
    "ext/mysql/tests/mysql_pconn_kill.phpt",
    "ext/mysql/tests/mysql_constants.phpt",
    "ext/mysql/tests/mysql_query_load_data_openbasedir.phpt",
    "ext/mysql/tests/mysql_field_table.phpt",
    "ext/mysql/tests/mysql_stat.phpt",
    "ext/mysql/tests/mysql_sql_safe_mode.phpt",
    "ext/mysql/tests/mysql_field_type.phpt",
    "ext/mysql/tests/mysql_pconn_max_links.phpt",
    "ext/mysql/tests/mysql_list_dbs.phpt",
    "ext/mysql/tests/mysql_num_rows.phpt",
    "ext/mysql/tests/mysql_field_seek.phpt",
    "ext/mysql/tests/mysql_connect.phpt",
    "ext/mysql/tests/mysql_list_fields.phpt",
    "ext/mysql/tests/mysql_drop_db.phpt",
    "ext/mysql/tests/mysql_tablename.phpt",
    "ext/mysql/tests/mysql_get_proto_info.phpt",
    "ext/mysql/tests/mysql_close.phpt",
    "ext/mysql/tests/mysql_num_fields.phpt",
    "ext/mysql/tests/mysql_db_query.phpt",
    "ext/mysql/tests/mysql_field_name.phpt",
    "ext/mysql/tests/mysql_field_flags.phpt",
    "ext/mysql/tests/bug48754.phpt",
    "ext/mysql/tests/mysql_create_db.phpt",
    "ext/pdo_pgsql/tests/large_objects.phpt",
    "ext/pdo_pgsql/tests/bug46274_2.phpt",
    "ext/pdo_pgsql/tests/is_in_transaction.phpt",
    "ext/pdo_pgsql/tests/bug43925.phpt",
    "ext/pdo_pgsql/tests/bug36727.phpt",
    "ext/pdo_pgsql/tests/bug48764.phpt",
    "ext/pdo_pgsql/tests/bug46274.phpt",
    "ext/pdo_pgsql/tests/common.phpt",
    "ext/pdo_pgsql/tests/bug_33876.phpt",
    "ext/pdo_pgsql/tests/bug_49985.phpt",
    "ext/pdo_pgsql/tests/copy_from.phpt",
    "ext/pdo_pgsql/tests/copy_to.phpt",
    "ext/session/tests/027.phpt",
    "ext/session/tests/rfc1867_sid_get.phpt",
    "ext/session/tests/session_save_path_variation5.phpt",
    "ext/session/tests/session_set_cookie_params_error.phpt",
    "ext/session/tests/session_save_path_basic.phpt",
    "ext/session/tests/session_unset_error.phpt",
    "ext/session/tests/rfc1867_sid_only_cookie_2.phpt",
    "ext/session/tests/session_commit_variation1.phpt",
    "ext/session/tests/session_module_name_variation2.phpt",
    "ext/session/tests/bug26862.phpt",
    "ext/session/tests/010.phpt",
    "ext/session/tests/rfc1867.phpt",
    "ext/session/tests/session_is_registered_variation2.phpt",
    "ext/session/tests/rfc1867_sid_cookie.phpt",
    "ext/session/tests/session_encode_error2.phpt",
    "ext/session/tests/005.phpt",
    "ext/session/tests/session_start_variation1.phpt",
    "ext/session/tests/rfc1867_disabled_2.phpt",
    "ext/session/tests/session_encode_variation4.phpt",
    "ext/session/tests/session_save_path_variation2.phpt",
    "ext/session/tests/014.phpt",
    "ext/session/tests/029.phpt",
    "ext/session/tests/session_id_error4.phpt",
    "ext/session/tests/session_register_variation1.phpt",
    "ext/session/tests/session_is_registered_basic.phpt",
    "ext/session/tests/011.phpt",
    "ext/session/tests/session_encode_variation7.phpt",
    "ext/session/tests/017.phpt",
    "ext/session/tests/session_decode_variation3.phpt",
    "ext/session/tests/session_set_save_handler_error.phpt",
    "ext/session/tests/session_unregister_basic.phpt",
    "ext/session/tests/session_module_name_error.phpt",
    "ext/session/tests/session_cache_limiter_variation2.phpt",
    "ext/session/tests/session_encode_basic.phpt",
    "ext/session/tests/session_module_name_variation3.phpt",
    "ext/session/tests/session_decode_basic.phpt",
    "ext/session/tests/session_id_error.phpt",
    "ext/session/tests/session_start_variation7.phpt",
    "ext/session/tests/028.phpt",
    "ext/session/tests/001.phpt",
    "ext/session/tests/session_decode_error.phpt",
    "ext/session/tests/session_set_save_handler_error2.phpt",
    "ext/session/tests/015.phpt",
    "ext/session/tests/session_encode_variation6.phpt",
    "ext/session/tests/026.phpt",
    "ext/session/tests/session_name_variation2.phpt",
    "ext/session/tests/004.phpt",
    "ext/session/tests/session_save_path_variation1.phpt",
    "ext/session/tests/session_set_save_handler_closures.phpt",
    "ext/session/tests/session_name_basic.phpt",
    "ext/session/tests/session_id_error3.phpt",
    "ext/session/tests/session_unregister_variation1.phpt",
    "ext/session/tests/session_cache_limiter_basic.phpt",
    "ext/session/tests/session_name_error.phpt",
    "ext/session/tests/session_register_variation2.phpt",
    "ext/session/tests/session_name_variation1.phpt",
    "ext/session/tests/session_decode_variation2.phpt",
    "ext/session/tests/session_unregister_variation3.phpt",
    "ext/session/tests/session_set_cookie_params_variation1.phpt",
    "ext/session/tests/session_id_error2.phpt",
    "ext/session/tests/session_write_close_variation4.phpt",
    "ext/session/tests/session_set_save_handler_basic.phpt",
    "ext/session/tests/session_start_variation5.phpt",
    "ext/session/tests/session_encode_error.phpt",
    "ext/session/tests/session_set_save_handler_variation3.phpt",
    "ext/session/tests/rfc1867_disabled.phpt",
    "ext/session/tests/bug51338.phpt",
    "ext/session/tests/session_cache_expire_variation2.phpt",
    "ext/session/tests/session_save_path_error.phpt",
    "ext/session/tests/rfc1867_sid_invalid.phpt",
    "ext/session/tests/bug24592.phpt",
    "ext/session/tests/020.phpt",
    "ext/session/tests/rfc1867_invalid_settings.phpt",
    "ext/session/tests/022.phpt",
    "ext/session/tests/session_regenerate_id_error.phpt",
    "ext/session/tests/006.phpt",
    "ext/session/tests/031.phpt",
    "ext/session/tests/bug41600.phpt",
    "ext/session/tests/session_cache_expire_variation1.phpt",
    "ext/session/tests/session_commit_variation4.phpt",
    "ext/session/tests/023.phpt",
    "ext/session/tests/session_cache_limiter_variation3.phpt",
    "ext/session/tests/session_cache_expire_variation3.phpt",
    "ext/session/tests/session_destroy_variation3.phpt",
    "ext/session/tests/session_unset_variation1.phpt",
    "ext/session/tests/bug38377.phpt",
    "ext/session/tests/session_commit_error.phpt",
    "ext/session/tests/session_write_close_variation1.phpt",
    "ext/session/tests/session_register_variation3.phpt",
    "ext/session/tests/rfc1867_sid_get_2.phpt",
    "ext/session/tests/session_set_cookie_params_variation3.phpt",
    "ext/session/tests/007.phpt",
    "ext/session/tests/session_write_close_variation3.phpt",
    "ext/session/tests/session_cache_limiter_variation1.phpt",
    "ext/session/tests/session_start_variation9.phpt",
    "ext/session/tests/session_commit_variation3.phpt",
    "ext/session/tests/session_encode_variation8.phpt",
    "ext/session/tests/session_register_basic.phpt",
    "ext/session/tests/session_unregister_variation2.phpt",
    "ext/session/tests/session_cache_expire_error.phpt",
    "ext/session/tests/bug42596.phpt",
    "ext/session/tests/session_decode_variation1.phpt",
    "ext/session/tests/session_set_cookie_params_variation4.phpt",
    "ext/session/tests/session_decode_variation4.phpt",
    "ext/session/tests/session_start_variation3.phpt",
    "ext/session/tests/session_encode_variation3.phpt",
    "ext/session/tests/002.phpt",
    "ext/session/tests/session_commit_variation2.phpt",
    "ext/session/tests/session_unset_basic.phpt",
    "ext/session/tests/session_save_path_variation4.phpt",
    "ext/session/tests/018.phpt",
    "ext/session/tests/session_save_path_variation3.phpt",
    "ext/session/tests/session_encode_variation5.phpt",
    "ext/session/tests/session_start_variation2.phpt",
    "ext/session/tests/rfc1867_sid_only_cookie.phpt",
    "ext/session/tests/session_module_name_variation1.phpt",
    "ext/session/tests/session_is_registered_variation3.phpt",
    "ext/session/tests/session_start_variation4.phpt",
    "ext/session/tests/013.phpt",
    "ext/session/tests/bug36459.phpt",
    "ext/session/tests/016.phpt",
    "ext/session/tests/bug32330.phpt",
    "ext/session/tests/session_get_cookie_params_error.phpt",
    "ext/session/tests/rfc1867_cleanup.phpt",
    "ext/session/tests/rfc1867_inter.phpt",
    "ext/session/tests/session_set_save_handler_variation2.phpt",
    "ext/session/tests/025.phpt",
    "ext/session/tests/session_id_variation1.phpt",
    "ext/session/tests/session_set_save_handler_variation1.phpt",
    "ext/session/tests/012.phpt",
    "ext/session/tests/session_module_name_basic.phpt",
    "ext/session/tests/session_start_error.phpt",
    "ext/session/tests/rfc1867_no_name.phpt",
    "ext/session/tests/session_set_cookie_params_variation2.phpt",
    "ext/session/tests/021.phpt",
    "ext/session/tests/bug53141.phpt",
    "ext/session/tests/session_cache_limiter_error.phpt",
    "ext/session/tests/session_is_registered_variation1.phpt",
    "ext/session/tests/session_set_save_handler_variation4.phpt",
    "ext/session/tests/session_id_basic.phpt",
    "ext/session/tests/session_destroy_variation2.phpt",
    "ext/session/tests/session_start_variation6.phpt",
    "ext/session/tests/session_destroy_error.phpt",
    "ext/session/tests/rfc1867_invalid_settings_2.phpt",
    "ext/session/tests/030.phpt",
    "ext/session/tests/session_write_close_basic.phpt",
    "ext/session/tests/session_destroy_variation1.phpt",
    "ext/session/tests/bug31454.phpt",
    "ext/session/tests/session_set_save_handler_error3.phpt",
    "ext/session/tests/session_start_variation8.phpt",
    "ext/session/tests/session_id_variation2.phpt",
    "ext/session/tests/session_regenerate_id_basic.phpt",
    "ext/session/tests/rfc1867_sid_post.phpt",
    "ext/session/tests/session_encode_variation1.phpt",
    "ext/session/tests/019.phpt",
    "ext/session/tests/session_module_name_variation4.phpt",
    "ext/session/tests/session_cache_expire_basic.phpt",
    "ext/session/tests/session_get_cookie_params_basic.phpt",
    "ext/session/tests/session_set_cookie_params_variation5.phpt",
    "ext/session/tests/session_register_variation4.phpt",
    "ext/session/tests/024.phpt",
    "ext/session/tests/009.phpt",
    "ext/session/tests/session_decode_error2.phpt",
    "ext/session/tests/session_write_close_error.phpt",
    "ext/session/tests/session_regenerate_id_variation1.phpt",
    "ext/session/tests/session_write_close_variation2.phpt",
    "ext/session/tests/session_commit_basic.phpt",
    "ext/session/tests/session_encode_variation2.phpt",
    "ext/session/tests/session_set_save_handler_error4.phpt",
    "ext/session/tests/session_get_cookie_params_variation1.phpt",
    "ext/session/tests/session_set_cookie_params_basic.phpt",
    "ext/hash/tests/md4.phpt",
    "ext/hash/tests/whirlpool.phpt",
    "ext/hash/tests/sha1.phpt",
    "ext/hash/tests/fnv132.phpt",
    "ext/hash/tests/hash_file_error.phpt",
    "ext/hash/tests/ripemd160.phpt",
    "ext/hash/tests/md2.phpt",
    "ext/hash/tests/sha384.phpt",
    "ext/hash/tests/gost.phpt",
    "ext/hash/tests/hash_hmac_error.phpt",
    "ext/hash/tests/hash_hmac_file_basic.phpt",
    "ext/hash/tests/bug52240.phpt",
    "ext/hash/tests/tiger.phpt",
    "ext/hash/tests/hash_hmac_file_error.phpt",
    "ext/hash/tests/ripemd256.phpt",
    "ext/hash/tests/hash_error.phpt",
    "ext/hash/tests/haval.phpt",
    "ext/hash/tests/fnv164.phpt",
    "ext/hash/tests/hash_file_basic1.phpt",
    "ext/hash/tests/hash_file_basic.phpt",
    "ext/hash/tests/mhash_003.phpt",
    "ext/hash/tests/hmac-md5.phpt",
    "ext/hash/tests/mhash_002.phpt",
    "ext/hash/tests/joaat.phpt",
    "ext/hash/tests/hash_algos.phpt",
    "ext/hash/tests/mhash_004.phpt",
    "ext/hash/tests/hash_copy_001.phpt",
    "ext/hash/tests/sha512.phpt",
    "ext/hash/tests/md5.phpt",
    "ext/hash/tests/snefru.phpt",
    "ext/hash/tests/hash_hmac_basic.phpt",
    "ext/hash/tests/crc32.phpt",
    "ext/hash/tests/ripemd128.phpt",
    "ext/hash/tests/adler32.phpt",
    "ext/hash/tests/mhash_001.phpt",
    "ext/hash/tests/ripemd320.phpt",
    "ext/hash/tests/sha224.phpt",
    "ext/hash/tests/hash_copy_002.phpt",
    "ext/hash/tests/sha256.phpt",
    "ext/pdo_dblib/tests/bug_38955.phpt",
    "ext/pdo_dblib/tests/bug_47588.phpt",
    "ext/pdo_dblib/tests/bug_45876.phpt",
    "ext/pdo_dblib/tests/bug_50755.phpt",
    "ext/mcrypt/tests/mcrypt_ofb.phpt",
    "ext/mcrypt/tests/mcrypt_decrypt_error.phpt",
    "ext/mcrypt/tests/mcrypt_list_modes.phpt",
    "ext/mcrypt/tests/bug35496.phpt",
    "ext/mcrypt/tests/mcrypt_module_is_block_algorithm.phpt",
    "ext/mcrypt/tests/bug8040.phpt",
    "ext/mcrypt/tests/mcrypt_enc_get_iv_size.phpt",
    "ext/mcrypt/tests/mcrypt_encrypt_variation2.phpt",
    "ext/mcrypt/tests/mcrypt_module_get_algo_key_size.phpt",
    "ext/mcrypt/tests/mcrypt_ecb_variation5.phpt",
    "ext/mcrypt/tests/mcrypt_enc_is_block_algorithm.phpt",
    "ext/mcrypt/tests/mcrypt_get_key_size.phpt",
    "ext/mcrypt/tests/mcrypt_ecb_3des_encrypt.phpt",
    "ext/mcrypt/tests/mcrypt_rijndael128_128BitKey.phpt",
    "ext/mcrypt/tests/mcrypt_module_get_supported_key_sizes.phpt",
    "ext/mcrypt/tests/mcrypt_module_is_block_mode.phpt",
    "ext/mcrypt/tests/mcrypt_ecb_variation2.phpt",
    "ext/mcrypt/tests/mcrypt_get_cipher_name.phpt",
    "ext/mcrypt/tests/mcrypt_cbc_variation5.phpt",
    "ext/mcrypt/tests/mcrypt_encrypt_variation3.phpt",
    "ext/mcrypt/tests/mcrypt_module_open.phpt",
    "ext/mcrypt/tests/mcrypt_encrypt_variation5.phpt",
    "ext/mcrypt/tests/mcrypt_enc_get_supported_key_sizes.phpt",
    "ext/mcrypt/tests/mcrypt_ecb_variation3.phpt",
    "ext/mcrypt/tests/mcrypt_encrypt_variation1.phpt",
    "ext/mcrypt/tests/mcrypt_get_iv_size.phpt",
    "ext/mcrypt/tests/mcrypt_cbc_variation2.phpt",
    "ext/mcrypt/tests/mcrypt_enc_get_key_size.phpt",
    "ext/mcrypt/tests/mcrypt_ecb_error.phpt",
    "ext/mcrypt/tests/mcrypt_cbc_3des_encrypt.phpt",
    "ext/mcrypt/tests/mcrypt_decrypt_variation5.phpt",
    "ext/mcrypt/tests/mcrypt_ecb_variation1.phpt",
    "ext/mcrypt/tests/mcrypt_list_algorithms.phpt",
    "ext/mcrypt/tests/mcrypt_decrypt_variation4.phpt",
    "ext/mcrypt/tests/mcrypt_module_self_test.phpt",
    "ext/mcrypt/tests/mcrypt_decrypt.phpt",
    "ext/mcrypt/tests/mcrypt_module_is_block_algorithm_mode.phpt",
    "ext/mcrypt/tests/mcrypt_cbc_3des_decrypt.phpt",
    "ext/mcrypt/tests/mcrypt_cbc_error.phpt",
    "ext/mcrypt/tests/mcrypt_ecb_3des_decrypt.phpt",
    "ext/mcrypt/tests/blowfish.phpt",
    "ext/mcrypt/tests/mcrypt_enc_get_block_size.phpt",
    "ext/mcrypt/tests/mcrypt_enc_get_mode_name.phpt",
    "ext/mcrypt/tests/mcrypt_rijndael128_256BitKey.phpt",
    "ext/mcrypt/tests/mcrypt_decrypt_variation1.phpt",
    "ext/mcrypt/tests/mcrypt_module_get_algo_block_size.phpt",
    "ext/mcrypt/tests/mcrypt_cbf.phpt",
    "ext/mcrypt/tests/mcrypt_ecb_variation4.phpt",
    "ext/mcrypt/tests/bug43143.phpt",
    "ext/mcrypt/tests/mcrypt_decrypt_3des_ecb.phpt",
    "ext/mcrypt/tests/mcrypt_create_iv.phpt",
    "ext/mcrypt/tests/mcrypt_cbc_variation3.phpt",
    "ext/mcrypt/tests/mcrypt_enc_get_algorithms_name.phpt",
    "ext/mcrypt/tests/mcrypt_enc_is_block_algorithm_mode.phpt",
    "ext/mcrypt/tests/mcrypt_cbc_variation4.phpt",
    "ext/mcrypt/tests/mcrypt_enc_is_block_mode.phpt",
    "ext/mcrypt/tests/mcrypt_cbc.phpt",
    "ext/mcrypt/tests/bug46010.phpt",
    "ext/mcrypt/tests/mcrypt_encrypt_3des_ecb.phpt",
    "ext/mcrypt/tests/mcrypt_encrypt_error.phpt",
    "ext/mcrypt/tests/bug49738.phpt",
    "ext/mcrypt/tests/mcrypt_encrypt_3des_cbc.phpt",
    "ext/mcrypt/tests/mcrypt_ecb.phpt",
    "ext/mcrypt/tests/mcrypt_encrypt_variation4.phpt",
    "ext/mcrypt/tests/mcrypt_cbc_variation1.phpt",
    "ext/mcrypt/tests/mcrypt_enc_self_test.phpt",
    "ext/mcrypt/tests/mcrypt_filters.phpt",
    "ext/mcrypt/tests/mcrypt_decrypt_3des_cbc.phpt",
    "ext/mcrypt/tests/mcrypt_decrypt_variation2.phpt",
    "ext/mcrypt/tests/mcrypt_get_block_size.phpt",
    "ext/mcrypt/tests/mcrypt_decrypt_variation3.phpt",
    "ext/mcrypt/tests/bug37595.phpt",
    "ext/mcrypt/tests/bug41252.phpt",
    "ext/mbstring/tests/zend_multibyte-04.phpt",
    "ext/mbstring/tests/mb_stristr_variation1.phpt",
    "ext/mbstring/tests/mb_output_handler_pattern-12.phpt",
    "ext/mbstring/tests/mb_stripos_variation1.phpt",
    "ext/mbstring/tests/zend_multibyte-12.phpt",
    "ext/mbstring/tests/mb_eregi.phpt",
    "ext/mbstring/tests/overload02.phpt",
    "ext/mbstring/tests/mb_strrchr_error2.phpt",
    "ext/mbstring/tests/mb_strrchr_variation6.phpt",
    "ext/mbstring/tests/mb_strstr_basic.phpt",
    "ext/mbstring/tests/bug25140.phpt",
    "ext/mbstring/tests/mb_convert_encoding.phpt",
    "ext/mbstring/tests/bug54494.phpt",
    "ext/mbstring/tests/mb_substr_count_variation4.phpt",
    "ext/mbstring/tests/mb_ereg_replace_variation4.phpt",
    "ext/mbstring/tests/zend_multibyte-11.phpt",
    "ext/mbstring/tests/mb_ereg_variation3.phpt",
    "ext/mbstring/tests/mb_regex_encoding_variation1.phpt",
    "ext/mbstring/tests/mb_internal_encoding_ini_basic2.phpt",
    "ext/mbstring/tests/mb_substr_variation1.phpt",
    "ext/mbstring/tests/mb_stristr_variation5.phpt",
    "ext/mbstring/tests/mb_parse_str.phpt",
    "ext/mbstring/tests/bug31911.phpt",
    "ext/mbstring/tests/mb_strstr_error1.phpt",
    "ext/mbstring/tests/mb_strtolower_error1.phpt",
    "ext/mbstring/tests/mb_decode_mimeheader_variation2.phpt",
    "ext/mbstring/tests/mb_substitute_character_variation2.phpt",
    "ext/mbstring/tests/mb_stristr_error2.phpt",
    "ext/mbstring/tests/mb_strrichr_variation5.phpt",
    "ext/mbstring/tests/mb_substitute_character.phpt",
    "ext/mbstring/tests/mb_substitute_character_basic.phpt",
    "ext/mbstring/tests/mb_ereg_error.phpt",
    "ext/mbstring/tests/mb_strpos_basic.phpt",
    "ext/mbstring/tests/mb_substr_count_basic.phpt",
    "ext/mbstring/tests/mb_output_handler_pattern-01.phpt",
    "ext/mbstring/tests/mb_strtolower_error2.phpt",
    "ext/mbstring/tests/bug52981.phpt",
    "ext/mbstring/tests/mb_internal_encoding_variation2.phpt",
    "ext/mbstring/tests/mb_ereg_replace-compat-08.phpt",
    "ext/mbstring/tests/mb_encode_mimeheader_basic2.phpt",
    "ext/mbstring/tests/mb_internal_encoding_error2.phpt",
    "ext/mbstring/tests/mb_send_mail01.phpt",
    "ext/mbstring/tests/mb_split_error.phpt",
    "ext/mbstring/tests/php_gr_jp_dev_884-1.phpt",
    "ext/mbstring/tests/mb_substr_count.phpt",
    "ext/mbstring/tests/mb_strrichr_error2.phpt",
    "ext/mbstring/tests/mb_substr_variation7.phpt",
    "ext/mbstring/tests/mb_ereg_replace.phpt",
    "ext/mbstring/tests/mb_internal_encoding_error1.phpt",
    "ext/mbstring/tests/mb_ereg.phpt",
    "ext/mbstring/tests/mb_regex_set_options.phpt",
    "ext/mbstring/tests/mb_stripos_variation5_Bug45923.phpt",
    "ext/mbstring/tests/mb_send_mail03.phpt",
    "ext/mbstring/tests/mb_strrichr_variation3.phpt",
    "ext/mbstring/tests/mb_strrchr_variation4.phpt",
    "ext/mbstring/tests/mb_strrichr_error1.phpt",
    "ext/mbstring/tests/mb_substr_count_error1.phpt",
    "ext/mbstring/tests/mb_stristr_variation4.phpt",
    "ext/mbstring/tests/mb_strpos_variation5.phpt",
    "ext/mbstring/tests/mb_strlen_error1.phpt",
    "ext/mbstring/tests/ini_language.phpt",
    "ext/mbstring/tests/mb_substr_count_variation1.phpt",
    "ext/mbstring/tests/mb_output_handler_pattern-04.phpt",
    "ext/mbstring/tests/mb_substr_count_variation2.phpt",
    "ext/mbstring/tests/mb_strtolower_variation1.phpt",
    "ext/mbstring/tests/mb_ereg_replace_variation1.phpt",
    "ext/mbstring/tests/mb_ereg3.phpt",
    "ext/mbstring/tests/mb_strlen_basic.phpt",
    "ext/mbstring/tests/mb_strripos_variation2.phpt",
    "ext/mbstring/tests/bug43993.phpt",
    "ext/mbstring/tests/mb_convert_kana.phpt",
    "ext/mbstring/tests/mb_send_mail02.phpt",
    "ext/mbstring/tests/mb_strripos_basic.phpt",
    "ext/mbstring/tests/mb_stristr_basic.phpt",
    "ext/mbstring/tests/mb_decode_mimeheader_variation1.phpt",
    "ext/mbstring/tests/mb_stripos_error2.phpt",
    "ext/mbstring/tests/mb_ereg_variation2.phpt",
    "ext/mbstring/tests/mb_strlen_variation2.phpt",
    "ext/mbstring/tests/zend_multibyte-03.phpt",
    "ext/mbstring/tests/mb_strpos_error1.phpt",
    "ext/mbstring/tests/mb_send_mail07.phpt",
    "ext/mbstring/tests/mb_strrchr_basic.phpt",
    "ext/mbstring/tests/mb_ereg_replace-compat-05.phpt",
    "ext/mbstring/tests/mb_strstr_error2.phpt",
    "ext/mbstring/tests/mb_internal_encoding_basic.phpt",
    "ext/mbstring/tests/mb_ereg_search_pos.phpt",
    "ext/mbstring/tests/mb_strripos_variation3_Bug45923.phpt",
    "ext/mbstring/tests/bug46843.phpt",
    "ext/mbstring/tests/mb_strtoupper_variation2.phpt",
    "ext/mbstring/tests/mb_detect_encoding.phpt",
    "ext/mbstring/tests/bug20087.phpt",
    "ext/mbstring/tests/mb_ereg_search_regs.phpt",
    "ext/mbstring/tests/mb_eregi_replace.phpt",
    "ext/mbstring/tests/mb_strpos_error2.phpt",
    "ext/mbstring/tests/mb_ereg_match_error1.phpt",
    "ext/mbstring/tests/bug52931.phpt",
    "ext/mbstring/tests/zend_multibyte-02.phpt",
    "ext/mbstring/tests/mb_ereg_variation7.phpt",
    "ext/mbstring/tests/mb_ereg-compat-01.phpt",
    "ext/mbstring/tests/mb_substr_variation3.phpt",
    "ext/mbstring/tests/mb_strtolower_basic.phpt",
    "ext/mbstring/tests/bug48645.phpt",
    "ext/mbstring/tests/php_gr_jp_16242.phpt",
    "ext/mbstring/tests/bug28220.phpt",
    "ext/mbstring/tests/mb_substitute_character_variation1.phpt",
    "ext/mbstring/tests/mb_strripos_error1.phpt",
    "ext/mbstring/tests/mb_ereg_replace_variation3.phpt",
    "ext/mbstring/tests/mb_substr_variation2.phpt",
    "ext/mbstring/tests/mb_split_variation3.phpt",
    "ext/mbstring/tests/zend_multibyte-08.phpt",
    "ext/mbstring/tests/mb_strrchr_variation1.phpt",
    "ext/mbstring/tests/mb_substr_basic.phpt",
    "ext/mbstring/tests/bug49354.phpt",
    "ext/mbstring/tests/mb_output_handler_shift_jis.phpt",
    "ext/mbstring/tests/mb_substr_error2.phpt",
    "ext/mbstring/tests/mb_strimwidth.phpt",
    "ext/mbstring/tests/mb_output_handler_pattern-10.phpt",
    "ext/mbstring/tests/mb_encode_mimeheader_variation5.phpt",
    "ext/mbstring/tests/mb_strstr_variation1.phpt",
    "ext/mbstring/tests/mb_substr_error1.phpt",
    "ext/mbstring/tests/mb_stripos_variation2.phpt",
    "ext/mbstring/tests/mb_output_handler_runtime_ini_alteration-02.phpt",
    "ext/mbstring/tests/mb_strstr_variation3.phpt",
    "ext/mbstring/tests/mb_ereg_variation5.phpt",
    "ext/mbstring/tests/bug49528.phpt",
    "ext/mbstring/tests/mb_ereg_replace_basic.phpt",
    "ext/mbstring/tests/mb_encode_mimeheader_variation3.phpt",
    "ext/mbstring/tests/mb_strtolower_variation4.phpt",
    "ext/mbstring/tests/bug43841.phpt",
    "ext/mbstring/tests/mb_strlen_variation1.phpt",
    "ext/mbstring/tests/mb_stripos_basic2.phpt",
    "ext/mbstring/tests/bug30549.phpt",
    "ext/mbstring/tests/zend_multibyte-13.phpt",
    "ext/mbstring/tests/mb_strtoupper_variation3.phpt",
    "ext/mbstring/tests/mb_ereg_variation6.phpt",
    "ext/mbstring/tests/mb_encode_mimeheader_variation1.phpt",
    "ext/mbstring/tests/mb_ereg2.phpt",
    "ext/mbstring/tests/mb_strrpos_variation4.phpt",
    "ext/mbstring/tests/mb_strtoupper_error2.phpt",
    "ext/mbstring/tests/mb_output_handler_pattern-09.phpt",
    "ext/mbstring/tests/mb_send_mail06.phpt",
    "ext/mbstring/tests/mb_strripos_variation4.phpt",
    "ext/mbstring/tests/bug52861.phpt",
    "ext/mbstring/tests/mb_regex_encoding_variation2.phpt",
    "ext/mbstring/tests/mb_substr_variation4.phpt",
    "ext/mbstring/tests/bug43840.phpt",
    "ext/mbstring/tests/mb_strripos_variation1.phpt",
    "ext/mbstring/tests/mb_output_handler_pattern-08.phpt",
    "ext/mbstring/tests/zend_multibyte-09.phpt",
    "ext/mbstring/tests/mb_substr_variation6.phpt",
    "ext/mbstring/tests/mb_ereg_replace-compat-03.phpt",
    "ext/mbstring/tests/mb_convert_encoding_basic.phpt",
    "ext/mbstring/tests/mb_strrpos_variation3.phpt",
    "ext/mbstring/tests/mb_ereg_replace-compat-01.phpt",
    "ext/mbstring/tests/zend_multibyte-05.phpt",
    "ext/mbstring/tests/mb_ereg_replace_variation2.phpt",
    "ext/mbstring/tests/mb_encoding_aliases.phpt",
    "ext/mbstring/tests/mb_substr_count_error2.phpt",
    "ext/mbstring/tests/mb_parse_str02.phpt",
    "ext/mbstring/tests/mb_ereg_replace-compat-04.phpt",
    "ext/mbstring/tests/mb_substr_variation5.phpt",
    "ext/mbstring/tests/mb_strrpos_error2.phpt",
    "ext/mbstring/tests/mb_regex_encoding_error2.phpt",
    "ext/mbstring/tests/mb_strpos_variation4.phpt",
    "ext/mbstring/tests/mb_ereg_replace-compat-02.phpt",
    "ext/mbstring/tests/mb_list_encodings.phpt",
    "ext/mbstring/tests/mb_encode_mimeheader_basic3.phpt",
    "ext/mbstring/tests/mb_preferred_mime_name.phpt",
    "ext/mbstring/tests/bug47399.phpt",
    "ext/mbstring/tests/mb_strtoupper_basic.phpt",
    "ext/mbstring/tests/mb_strrchr_variation5.phpt",
    "ext/mbstring/tests/mb_get_info.phpt",
    "ext/mbstring/tests/mb_substitute_character_error.phpt",
    "ext/mbstring/tests/mb_strstr.phpt",
    "ext/mbstring/tests/mb_strrichr_variation2.phpt",
    "ext/mbstring/tests/mb_ereg_replace-compat-10.phpt",
    "ext/mbstring/tests/mb_strrichr_variation4.phpt",
    "ext/mbstring/tests/mb_ereg_basic.phpt",
    "ext/mbstring/tests/mb_ereg-compat-02.phpt",
    "ext/mbstring/tests/mb_ereg1.phpt",
    "ext/mbstring/tests/mb_ereg_variation4.phpt",
    "ext/mbstring/tests/bug45722.phpt",
    "ext/mbstring/tests/mb_substr_count_variation3.phpt",
    "ext/mbstring/tests/mb_convert_encoding_stateful.phpt",
    "ext/mbstring/tests/mb_strpos_variation3.phpt",
    "ext/mbstring/tests/mb_stristr_variation3.phpt",
    "ext/mbstring/tests/mb_regex_encoding_basic.phpt",
    "ext/mbstring/tests/mb_strpos.phpt",
    "ext/mbstring/tests/zend_multibyte-07.phpt",
    "ext/mbstring/tests/mb_output_handler_pattern-02.phpt",
    "ext/mbstring/tests/mb_http_output.phpt",
    "ext/mbstring/tests/mb_output_handler_runtime_ini_alteration-01.phpt",
    "ext/mbstring/tests/bug43998.phpt",
    "ext/mbstring/tests/mb_ereg_search.phpt",
    "ext/mbstring/tests/zend_multibyte-10.phpt",
    "ext/mbstring/tests/mb_strrpos_variation1.phpt",
    "ext/mbstring/tests/mb_strtoupper_variation4.phpt",
    "ext/mbstring/tests/zend_multibyte-14.phpt",
    "ext/mbstring/tests/simpletest.phpt",
    "ext/mbstring/tests/mb_strtolower_variation2.phpt",
    "ext/mbstring/tests/mb_stristr_error1.phpt",
    "ext/mbstring/tests/mb_internal_encoding_variation1.phpt",
    "ext/mbstring/tests/mb_output_handler_euc_jp.phpt",
    "ext/mbstring/tests/mb_encode_numericentity.phpt",
    "ext/mbstring/tests/php_gr_jp_10830.phpt",
    "ext/mbstring/tests/mb_stripos_variation3.phpt",
    "ext/mbstring/tests/mb_split.phpt",
    "ext/mbstring/tests/mb_strtoupper_error1.phpt",
    "ext/mbstring/tests/mb_strlen_error2.phpt",
    "ext/mbstring/tests/bug48697.phpt",
    "ext/mbstring/tests/mb_ereg_replace-compat-11.phpt",
    "ext/mbstring/tests/mb_internal_encoding.phpt",
    "ext/mbstring/tests/mb_strlen.phpt",
    "ext/mbstring/tests/mb_strcut_missing_boundary_check.phpt",
    "ext/mbstring/tests/bug43994.phpt",
    "ext/mbstring/tests/mb_strrchr_variation3.phpt",
    "ext/mbstring/tests/mb_ereg_search_xxx.phpt",
    "ext/mbstring/tests/bug46806.phpt",
    "ext/mbstring/tests/mb_ereg_replace-compat-13.phpt",
    "ext/mbstring/tests/mb_http_input.phpt",
    "ext/mbstring/tests/mb_strrichr_basic.phpt",
    "ext/mbstring/tests/mb_strrpos_variation2.phpt",
    "ext/mbstring/tests/bug49536.phpt",
    "ext/mbstring/tests/mb_strpos_variation1.phpt",
    "ext/mbstring/tests/htmlent.phpt",
    "ext/mbstring/tests/illformed_utf_sequences.phpt",
    "ext/mbstring/tests/mb_split-compat-01.phpt",
    "ext/mbstring/tests/mb_convert_variables.phpt",
    "ext/mbstring/tests/mb_encode_mimeheader_basic.phpt",
    "ext/mbstring/tests/overload01.phpt",
    "ext/mbstring/tests/mb_output_handler_pattern-11.phpt",
    "ext/mbstring/tests/mb_stripos_error1.phpt",
    "ext/mbstring/tests/mb_split_variation2.phpt",
    "ext/mbstring/tests/mb_eregi_invalid_arguments.phpt",
    "ext/mbstring/tests/mb_encode_mimeheader_variation4.phpt",
    "ext/mbstring/tests/mb_encode_mimeheader_indent.phpt",
    "ext/mbstring/tests/zend_multibyte-06.phpt",
    "ext/mbstring/tests/mb_strrpos_basic.phpt",
    "ext/mbstring/tests/bug26639.phpt",
    "ext/mbstring/tests/mb_ereg_match_basic.phpt",
    "ext/mbstring/tests/mb_stripos_variation4.phpt",
    "ext/mbstring/tests/zend_multibyte-01.phpt",
    "ext/mbstring/tests/mb_stripos.phpt",
    "ext/mbstring/tests/mb_strstr_variation6.phpt",
    "ext/mbstring/tests/mb_stristr_variation2.phpt",
    "ext/mbstring/tests/mb_strripos_basic2.phpt",
    "ext/mbstring/tests/mb_decode_numericentity.phpt",
    "ext/mbstring/tests/mb_strwidth.phpt",
    "ext/mbstring/tests/mb_strstr_variation5.phpt",
    "ext/mbstring/tests/mb_ereg_replace-compat-07.phpt",
    "ext/mbstring/tests/mb_regex_encoding_error1.phpt",
    "ext/mbstring/tests/mb_strrchr_variation2.phpt",
    "ext/mbstring/tests/mb_output_handler_pattern-03.phpt",
    "ext/mbstring/tests/mb_encode_mimeheader_variation2.phpt",
    "ext/mbstring/tests/mb_ereg4.phpt",
    "ext/mbstring/tests/mb_strcut.phpt",
    "ext/mbstring/tests/mb_decode_mimeheader_basic.phpt",
    "ext/mbstring/tests/mb_decode_mimeheader_error.phpt",
    "ext/mbstring/tests/mb_strripos_variation5_Bug45923.phpt",
    "ext/mbstring/tests/mb_detect_order.phpt",
    "ext/mbstring/tests/mb_ereg_replace_error.phpt",
    "ext/mbstring/tests/mb_strstr_variation2.phpt",
    "ext/mbstring/tests/casefold.phpt",
    "ext/mbstring/tests/mb_strpos_variation2.phpt",
    "ext/mbstring/tests/mb_stripos_basic.phpt",
    "ext/mbstring/tests/mb_output_handler_pattern-05.phpt",
    "ext/mbstring/tests/mb_strstr_variation4.phpt",
    "ext/mbstring/tests/mb_send_mail05.phpt",
    "ext/mbstring/tests/mb_strrchr_error1.phpt",
    "ext/mbstring/tests/mb_strrpos_variation5.phpt",
    "ext/mbstring/tests/mb_strtolower_variation3.phpt",
    "ext/mbstring/tests/mb_encode_mimeheader_variation6.phpt",
    "ext/mbstring/tests/mb_send_mail04.phpt",
    "ext/mbstring/tests/mb_ereg_replace-compat-09.phpt",
    "ext/mbstring/tests/mb_ereg_replace-compat-12.phpt",
    "ext/mbstring/tests/mb_strtoupper_variation1.phpt",
    "ext/mbstring/tests/mb_ereg_variation1.phpt",
    "ext/mbstring/tests/mb_strlen_variation3.phpt",
    "ext/mbstring/tests/mb_strripos_error2.phpt",
    "ext/mbstring/tests/php_gr_jp_dev_884-2.phpt",
    "ext/mbstring/tests/bug45923.phpt",
    "ext/mbstring/tests/mb_decode_mimeheader_variation3.phpt",
    "ext/mbstring/tests/bug43301.phpt",
    "ext/mbstring/tests/mb_substr.phpt",
    "ext/mbstring/tests/bug45239.phpt",
    "ext/mbstring/tests/mb_split_variation1.phpt",
    "ext/mbstring/tests/mb_strrpos_error1.phpt",
    "ext/mbstring/tests/mb_strrichr_variation1.phpt",
    "ext/mbstring/tests/mb_ereg_replace-compat-06.phpt",
    "ext/mbstring/tests/mb_output_handler_pattern-06.phpt",
    "ext/mbstring/tests/mb_output_handler_pattern-07.phpt",
    "ext/mbstring/tests/mb_convert_encoding_variation1.phpt",
    "ext/readline/tests/readline_write_history_001.phpt",
    "ext/readline/tests/readline_add_history_001.phpt",
    "ext/readline/tests/readline_clear_history_001.phpt",
    "ext/readline/tests/readline_completion_function_001.phpt",
    "ext/readline/tests/readline_list_history_001.phpt",
    "ext/readline/tests/readline_read_history_001.phpt",
    "ext/readline/tests/readline_callback_handler_remove_001.phpt",
    "ext/readline/tests/readline_callback_handler_install_001.phpt",
    "ext/readline/tests/readline_info_001.phpt",
    "ext/curl/tests/curl_copy_handle_basic_001.phpt",
    "ext/curl/tests/curl_multi_getcontent_basic3.phpt",
    "ext/curl/tests/curl_multi_getcontent_error3.phpt",
    "ext/curl/tests/bug52202.phpt",
    "ext/curl/tests/curl_multi_select_basic1.phpt",
    "ext/curl/tests/curl_basic_001.phpt",
    "ext/curl/tests/curl_basic_006.phpt",
    "ext/curl/tests/curl_setopt_CURLOPT_READFUNCTION.phpt",
    "ext/curl/tests/bug46739.phpt",
    "ext/curl/tests/curl_writeheader_callback.phpt",
    "ext/curl/tests/curl_basic_007.phpt",
    "ext/curl/tests/curl_setopt_error.phpt",
    "ext/curl/tests/curl_setopt_array_basic.phpt",
    "ext/curl/tests/curl_basic_018.phpt",
    "ext/curl/tests/curl_copy_handle_basic_002.phpt",
    "ext/curl/tests/bug46711.phpt",
    "ext/curl/tests/curl_write_file.phpt",
    "ext/curl/tests/curl_basic_015.phpt",
    "ext/curl/tests/curl_basic_005.phpt",
    "ext/curl/tests/curl_copy_handle_variation1.phpt",
    "ext/curl/tests/curl_copy_handle_variation2.phpt",
    "ext/curl/tests/curl_setopt_basic004.phpt",
    "ext/curl/tests/curl_basic_004.phpt",
    "ext/curl/tests/curl_version_variation1.phpt",
    "ext/curl/tests/bug48203.phpt",
    "ext/curl/tests/curl_basic_020.phpt",
    "ext/curl/tests/curl_basic_009.phpt",
    "ext/curl/tests/curl_basic_002.phpt",
    "ext/curl/tests/curl_multi_getcontent_error2.phpt",
    "ext/curl/tests/curl_copy_handle_basic.phpt",
    "ext/curl/tests/curl_basic_008.phpt",
    "ext/curl/tests/curl_basic_016.phpt",
    "ext/curl/tests/curl_copy_handle_basic_004.phpt",
    "ext/curl/tests/curl_multi_init_basic.phpt",
    "ext/curl/tests/curl_multi_getcontent_error1.phpt",
    "ext/curl/tests/curl_multi_getcontent_error4.phpt",
    "ext/curl/tests/curl_copy_handle_basic_007.phpt",
    "ext/curl/tests/curl_setopt_basic003.phpt",
    "ext/curl/tests/curl_copy_handle_basic_006.phpt",
    "ext/curl/tests/curl_basic_014.phpt",
    "ext/curl/tests/curl_copy_handle_basic_005.phpt",
    "ext/curl/tests/curl_basic_019.phpt",
    "ext/curl/tests/curl_basic_017.phpt",
    "ext/curl/tests/curl_setopt_CURLOPT_FOLLOWLOCATION_open_basedir.phpt",
    "ext/curl/tests/curl_basic_012.phpt",
    "ext/curl/tests/curl_CURLOPT_READDATA.phpt",
    "ext/curl/tests/curl_setopt_basic001.phpt",
    "ext/curl/tests/curl_write_callback.phpt",
    "ext/curl/tests/curl_basic_010.phpt",
    "ext/curl/tests/bug48514.phpt",
    "ext/curl/tests/curl_ftp_pasv.phpt",
    "ext/curl/tests/curl_setopt_basic002.phpt",
    "ext/curl/tests/curl_basic_013.phpt",
    "ext/curl/tests/curl_basic_011.phpt",
    "ext/curl/tests/curl_basic_003.phpt",
    "ext/curl/tests/bug52827.phpt",
    "ext/curl/tests/curl_multi_close_basic.phpt",
    "ext/curl/tests/curl_close_basic.phpt",
    "ext/curl/tests/bug48207.phpt",
    "ext/curl/tests/bug45161.phpt",
    "ext/curl/tests/curl_write_stdout.phpt",
    "ext/curl/tests/curl_error_basic.phpt",
    "ext/curl/tests/curl_basic_021.phpt",
    "ext/curl/tests/curl_version_error.phpt",
    "ext/curl/tests/curl_write_return.phpt",
    "ext/dba/tests/bug36436.phpt",
    "ext/dba/tests/dba_qdbm.phpt",
    "ext/dba/tests/dba_db4_006.phpt",
    "ext/dba/tests/dba_ndbm.phpt",
    "ext/dba/tests/dba_db4_011.phpt",
    "ext/dba/tests/bug49125.phpt",
    "ext/dba/tests/dba_db4_sync.phpt",
    "ext/dba/tests/dba_db4_001.phpt",
    "ext/dba/tests/dba_db4_004.phpt",
    "ext/dba/tests/dba012.phpt",
    "ext/dba/tests/dba_db4_013.phpt",
    "ext/dba/tests/dba015.phpt",
    "ext/dba/tests/dba006.phpt",
    "ext/dba/tests/dba010.phpt",
    "ext/dba/tests/dba_db4_017.phpt",
    "ext/dba/tests/dba007.phpt",
    "ext/dba/tests/bug38698.phpt",
    "ext/dba/tests/dba_db4_018.phpt",
    "ext/dba/tests/dba_db4_015.phpt",
    "ext/dba/tests/dba_db4_handlers.phpt",
    "ext/dba/tests/dba_db3.phpt",
    "ext/dba/tests/dba_db4_010.phpt",
    "ext/dba/tests/dba_cdb_read.phpt",
    "ext/dba/tests/dba_gdbm.phpt",
    "ext/dba/tests/dba_db4_007.phpt",
    "ext/dba/tests/dba_inifile.phpt",
    "ext/dba/tests/dba004.phpt",
    "ext/dba/tests/dba_optimize.phpt",
    "ext/dba/tests/dba_handlers.phpt",
    "ext/dba/tests/dba_cdb_001.phpt",
    "ext/dba/tests/dba_dbm.phpt",
    "ext/dba/tests/dba_db1.phpt",
    "ext/dba/tests/dba_db4_003.phpt",
    "ext/dba/tests/dba_sync.phpt",
    "ext/dba/tests/dba_db4_002.phpt",
    "ext/dba/tests/dba_db4_016.phpt",
    "ext/dba/tests/dba_db4_000.phpt",
    "ext/dba/tests/dba_db4_012.phpt",
    "ext/dba/tests/dba_db4_009.phpt",
    "ext/dba/tests/bug48240.phpt",
    "ext/dba/tests/dba002.phpt",
    "ext/dba/tests/dba011.phpt",
    "ext/dba/tests/dba_flatfile.phpt",
    "ext/dba/tests/dba_db4_optimize.phpt",
    "ext/dba/tests/dba_db4_008.phpt",
    "ext/dba/tests/dba014.phpt",
    "ext/dba/tests/dba_cdb_make.phpt",
    "ext/dba/tests/dba005.phpt",
    "ext/dba/tests/dba_tcadb.phpt",
    "ext/dba/tests/dba016.phpt",
    "ext/dba/tests/dba013.phpt",
    "ext/dba/tests/dba_db4_014.phpt",
    "ext/dba/tests/dba008.phpt",
    "ext/dba/tests/dba_split.phpt",
    "ext/dba/tests/dba_cdb.phpt",
    "ext/dba/tests/dba001.phpt",
    "ext/dba/tests/dba009.phpt",
    "ext/dba/tests/dba_db2.phpt",
    "ext/dba/tests/dba_db4_005.phpt",
    "ext/dba/tests/dba003.phpt",
    "ext/ftp/tests/005.phpt",
    "ext/ftp/tests/ftp_chmod_basic.phpt",
    "ext/ftp/tests/bug39583-2.phpt",
    "ext/ftp/tests/ftp_exec_basic.phpt",
    "ext/ftp/tests/001.phpt",
    "ext/ftp/tests/ftp_fget_basic1.phpt",
    "ext/ftp/tests/004.phpt",
    "ext/ftp/tests/bug37799.phpt",
    "ext/ftp/tests/ftp_get_basic.phpt",
    "ext/ftp/tests/ftp_nb_fget_basic1.phpt",
    "ext/ftp/tests/003.phpt",
    "ext/ftp/tests/ftp_rmdir_basic.phpt",
    "ext/ftp/tests/006.phpt",
    "ext/ftp/tests/bug39583.phpt",
    "ext/ftp/tests/ftp_fget_basic.phpt",
    "ext/ftp/tests/ftp_nb_fget_basic2.phpt",
    "ext/ftp/tests/ftp_rawlist_basic2.phpt",
    "ext/ftp/tests/ftp_nb_fget_basic3.phpt",
    "ext/ftp/tests/bug7216.phpt",
    "ext/ftp/tests/002.phpt",
    "ext/ftp/tests/bug7216-2.phpt",
    "ext/ftp/tests/bug27809.phpt",
    "ext/ftp/tests/ftp_alloc_basic1.phpt",
    "ext/ftp/tests/ftp_mdtm_basic.phpt",
    "ext/ftp/tests/ftp_fget_basic2.phpt",
    "ext/ftp/tests/bug39458-2.phpt",
    "ext/ftp/tests/ftp_alloc_basic2.phpt",
    "ext/ftp/tests/bug39458.phpt",
    "ext/ftp/tests/ftp_fget_basic3.phpt",
    "ext/pdo_sqlite/tests/pdo_fetch_func_001.phpt",
    "ext/pdo_sqlite/tests/bug46542.phpt",
    "ext/pdo_sqlite/tests/debugdumpparams_001.phpt",
    "ext/pdo_sqlite/tests/bug46139.phpt",
    "ext/pdo_sqlite/tests/pdo_sqlite_transaction.phpt",
    "ext/pdo_sqlite/tests/bug33841.phpt",
    "ext/pdo_sqlite/tests/bug44327_3.phpt",
    "ext/pdo_sqlite/tests/common.phpt",
    "ext/pdo_sqlite/tests/bug43831.phpt",
    "ext/pdo_sqlite/tests/pdo_sqlite_lastinsertid.phpt",
    "ext/pdo_sqlite/tests/bug52487.phpt",
    "ext/pdo_sqlite/tests/pdo_sqlite_createaggregate.phpt",
    "ext/pdo_sqlite/tests/pdo_sqlite_createfunction.phpt",
    "ext/pdo_sqlite/tests/bug48773.phpt",
    "ext/pdo_sqlite/tests/bug_42589.phpt",
    "ext/pdo_sqlite/tests/bug44327_2.phpt",
    "ext/pdo_sqlite/tests/bug50728.phpt",
    "ext/pdo_sqlite/tests/bug35336.phpt",
    "ext/openssl/tests/bug25614.phpt",
    "ext/openssl/tests/005.phpt",
    "ext/openssl/tests/bug41353.phpt",
    "ext/openssl/tests/014.phpt",
    "ext/openssl/tests/011.phpt",
    "ext/openssl/tests/017.phpt",
    "ext/openssl/tests/openssl_x509_parse_basic.phpt",
    "ext/openssl/tests/001.phpt",
    "ext/openssl/tests/openssl_encrypt_error.phpt",
    "ext/openssl/tests/015.phpt",
    "ext/openssl/tests/004.phpt",
    "ext/openssl/tests/openssl_random_pseudo_bytes.phpt",
    "ext/openssl/tests/sni_001.phpt",
    "ext/openssl/tests/003.phpt",
    "ext/openssl/tests/022.phpt",
    "ext/openssl/tests/006.phpt",
    "ext/openssl/tests/023.phpt",
    "ext/openssl/tests/bug54060.phpt",
    "ext/openssl/tests/007.phpt",
    "ext/openssl/tests/openssl_digest_basic.phpt",
    "ext/openssl/tests/bug47828.phpt",
    "ext/openssl/tests/002.phpt",
    "ext/openssl/tests/bug41033.phpt",
    "ext/openssl/tests/008.phpt",
    "ext/openssl/tests/018.phpt",
    "ext/openssl/tests/bug36732.phpt",
    "ext/openssl/tests/openssl_decrypt_error.phpt",
    "ext/openssl/tests/bug38261.phpt",
    "ext/openssl/tests/013.phpt",
    "ext/openssl/tests/016.phpt",
    "ext/openssl/tests/bug39217.phpt",
    "ext/openssl/tests/bug37820.phpt",
    "ext/openssl/tests/bug46127.phpt",
    "ext/openssl/tests/025.phpt",
    "ext/openssl/tests/012.phpt",
    "ext/openssl/tests/021.phpt",
    "ext/openssl/tests/bug48182.phpt",
    "ext/openssl/tests/bug54061.phpt",
    "ext/openssl/tests/019.phpt",
    "ext/openssl/tests/024.phpt",
    "ext/openssl/tests/bug38255.phpt",
    "ext/openssl/tests/009.phpt",
    "ext/openssl/tests/bug28382.phpt",
    "ext/pdo_firebird/tests/common.phpt",
    "ext/pdo_firebird/tests/connect.phpt",
    "ext/pdo_firebird/tests/rowCount.phpt",
    "ext/pdo_firebird/tests/execute.phpt",
    "ext/pdo_firebird/tests/ddl.phpt",
    "ext/com_dotnet/tests/bug45280.phpt",
    "ext/com_dotnet/tests/bug39596.phpt",
    "ext/com_dotnet/tests/bug34272.phpt",
    "ext/com_dotnet/tests/variants.phpt",
    "ext/com_dotnet/tests/bug39606.phpt",
    "ext/com_dotnet/tests/bug49192.phpt",
    "ext/com_dotnet/tests/bug33386.phpt",
    "ext/com_dotnet/tests/27974.phpt",
    "ext/gd/tests/imagefilltoborder_error6.phpt",
    "ext/gd/tests/image_type_to_mime_type_variation4.phpt",
    "ext/gd/tests/imageinterlace_variation2.phpt",
    "ext/gd/tests/imagestringup_basic.phpt",
    "ext/gd/tests/imagechar_error4.phpt",
    "ext/gd/tests/imagedashedline_basic.phpt",
    "ext/gd/tests/imageellipse_error6.phpt",
    "ext/gd/tests/imagestring_basic.phpt",
    "ext/gd/tests/image_type_to_mime_type_variation3.phpt",
    "ext/gd/tests/imagecreatetruecolor_error2.phpt",
    "ext/gd/tests/bug48801.phpt",
    "ext/gd/tests/bug48555.phpt",
    "ext/gd/tests/bug24594.phpt",
    "ext/gd/tests/imagerectangle_error8.phpt",
    "ext/gd/tests/imagefilter_error9.phpt",
    "ext/gd/tests/imagestring_error2.phpt",
    "ext/gd/tests/imagefilter_error2.phpt",
    "ext/gd/tests/imageinterlace_error2.phpt",
    "ext/gd/tests/imagestring_error4.phpt",
    "ext/gd/tests/imagefilter_error3.phpt",
    "ext/gd/tests/libgd00086_extern.phpt",
    "ext/gd/tests/imageellipse_error2.phpt",
    "ext/gd/tests/crafted_gd2.phpt",
    "ext/gd/tests/gif2jpg.phpt",
    "ext/gd/tests/bug39508.phpt",
    "ext/gd/tests/imagesetbrush_basic.phpt",
    "ext/gd/tests/bug28147.phpt",
    "ext/gd/tests/bug44849.phpt",
    "ext/gd/tests/imagecolorallocatealpha_error4.phpt",
    "ext/gd/tests/gd_info_error.phpt",
    "ext/gd/tests/imageellipse_basic.phpt",
    "ext/gd/tests/imagefilter_error15.phpt",
    "ext/gd/tests/imagefilter.phpt",
    "ext/gd/tests/imagetruecolortopalette_basic.phpt",
    "ext/gd/tests/imagecolorclosesthwb.phpt",
    "ext/gd/tests/imagecharup_error2.phpt",
    "ext/gd/tests/imagestringup_error5.phpt",
    "ext/gd/tests/image_type_to_mime_type_basic.phpt",
    "ext/gd/tests/imagechar_basic.phpt",
    "ext/gd/tests/imagecolorallocatealpha_error1.phpt",
    "ext/gd/tests/imagerectangle_error7.phpt",
    "ext/gd/tests/imagearc_basic.phpt",
    "ext/gd/tests/bug43073.phpt",
    "ext/gd/tests/bug45799.phpt",
    "ext/gd/tests/imagestring_error5.phpt",
    "ext/gd/tests/createfromwbmp2.phpt",
    "ext/gd/tests/imagefilter_error12.phpt",
    "ext/gd/tests/bug37360.phpt",
    "ext/gd/tests/bug38112.phpt",
    "ext/gd/tests/imagerectangle_error6.phpt",
    "ext/gd/tests/imagecolorallocatealpha_error2.phpt",
    "ext/gd/tests/imageconvolution_error2.phpt",
    "ext/gd/tests/imagegammacorrect_basic.phpt",
    "ext/gd/tests/png2wbmp_error2.phpt",
    "ext/gd/tests/imagecolordeallocate_basic.phpt",
    "ext/gd/tests/xbm2png.phpt",
    "ext/gd/tests/bug39366.phpt",
    "ext/gd/tests/imageellipse_error7.phpt",
    "ext/gd/tests/image_type_to_mime_type_error.phpt",
    "ext/gd/tests/image_type_to_mime_type_variation2.phpt",
    "ext/gd/tests/imagecolorallocate_variation4.phpt",
    "ext/gd/tests/imagecolorset_basic.phpt",
    "ext/gd/tests/bug43121.phpt",
    "ext/gd/tests/bug38212.phpt",
    "ext/gd/tests/imageellipse_error1.phpt",
    "ext/gd/tests/bug19366.phpt",
    "ext/gd/tests/imageellipse_error8.phpt",
    "ext/gd/tests/imagefilter_error1.phpt",
    "ext/gd/tests/bug39780_extern.phpt",
    "ext/gd/tests/imageinterlace_basic.phpt",
    "ext/gd/tests/types.phpt",
    "ext/gd/tests/libgd00094.phpt",
    "ext/gd/tests/001.phpt",
    "ext/gd/tests/imagecolordeallocate_error3.phpt",
    "ext/gd/tests/imagefilledarc_variation2.phpt",
    "ext/gd/tests/imagestringup_error2.phpt",
    "ext/gd/tests/imagecolormatch_basic.phpt",
    "ext/gd/tests/imagearc_variation2.phpt",
    "ext/gd/tests/libgd00100.phpt",
    "ext/gd/tests/imagefilter_error10.phpt",
    "ext/gd/tests/colorat.phpt",
    "ext/gd/tests/bug48732.phpt",
    "ext/gd/tests/bug36697.phpt",
    "ext/gd/tests/imagefilledarc_basic.phpt",
    "ext/gd/tests/imagestringup_error3.phpt",
    "ext/gd/tests/imagefilter_error4.phpt",
    "ext/gd/tests/imagetruecolortopalette_error4.phpt",
    "ext/gd/tests/imagesetthickness_error2.phpt",
    "ext/gd/tests/imagelayereffect_error3.phpt",
    "ext/gd/tests/colorexact.phpt",
    "ext/gd/tests/imagefilter_error13.phpt",
    "ext/gd/tests/imagecharup_basic.phpt",
    "ext/gd/tests/imagefilter_error16.phpt",
    "ext/gd/tests/bug51671.phpt",
    "ext/gd/tests/imagecolordeallocate_error4.phpt",
    "ext/gd/tests/libgd00086.phpt",
    "ext/gd/tests/bug49600.phpt",
    "ext/gd/tests/imageistruecolor_error1.phpt",
    "ext/gd/tests/imageconvolution_basic.phpt",
    "ext/gd/tests/gd_info_variation1.phpt",
    "ext/gd/tests/bug39273.phpt",
    "ext/gd/tests/copyresized.phpt",
    "ext/gd/tests/imagefttext.phpt",
    "ext/gd/tests/imagefilter_error20.phpt",
    "ext/gd/tests/imagefilltoborder_basic.phpt",
    "ext/gd/tests/imagelayereffect_error1.phpt",
    "ext/gd/tests/imagefontwidth_basic.phpt",
    "ext/gd/tests/image_type_to_mime_type_variation1.phpt",
    "ext/gd/tests/bug39780.phpt",
    "ext/gd/tests/imageantialias_error2.phpt",
    "ext/gd/tests/jpeg2wbmp_error3.phpt",
    "ext/gd/tests/imageellipse_error5.phpt",
    "ext/gd/tests/imagecolorstotal_basic.phpt",
    "ext/gd/tests/createfromwbmp.phpt",
    "ext/gd/tests/imagechar_error1.phpt",
    "ext/gd/tests/imagerectangle_error2.phpt",
    "ext/gd/tests/imagerectangle_error1.phpt",
    "ext/gd/tests/bug39286.phpt",
    "ext/gd/tests/imagecolorallocate_variation1.phpt",
    "ext/gd/tests/imagestring_error6.phpt",
    "ext/gd/tests/imagecolordeallocate_error2.phpt",
    "ext/gd/tests/imagefilter_error7.phpt",
    "ext/gd/tests/imageellipse_error3.phpt",
    "ext/gd/tests/imagecharup_error5.phpt",
    "ext/gd/tests/imagestringup_error4.phpt",
    "ext/gd/tests/imagefilltoborder_error5.phpt",
    "ext/gd/tests/imagelayereffect_basic.phpt",
    "ext/gd/tests/imagepolygon_negative.phpt",
    "ext/gd/tests/imagecolorallocate_variation3.phpt",
    "ext/gd/tests/imagegammacorrect_error1.phpt",
    "ext/gd/tests/lines.phpt",
    "ext/gd/tests/imagecopyresampled_basic.phpt",
    "ext/gd/tests/imagecharup_error3.phpt",
    "ext/gd/tests/jpeg2png.phpt",
    "ext/gd/tests/imagechar_error6.phpt",
    "ext/gd/tests/png2wbmp_error3.phpt",
    "ext/gd/tests/imagecharup_error4.phpt",
    "ext/gd/tests/imagechar_error5.phpt",
    "ext/gd/tests/imagerectangle_error4.phpt",
    "ext/gd/tests/imagecolorallocate_error.phpt",
    "ext/gd/tests/imagefontheight_error1.phpt",
    "ext/gd/tests/imagegammacorrect_error4.phpt",
    "ext/gd/tests/bug51263.phpt",
    "ext/gd/tests/imagecharup_error6.phpt",
    "ext/gd/tests/imagecolourstotal_basic.phpt",
    "ext/gd/tests/colorresolve.phpt",
    "ext/gd/tests/imagefilledarc_variation1.phpt",
    "ext/gd/tests/libgd00101.phpt",
    "ext/gd/tests/imagefilter_error17.phpt",
    "ext/gd/tests/bug41442.phpt",
    "ext/gd/tests/imagefilltoborder_error1.phpt",
    "ext/gd/tests/imagearc_error1.phpt",
    "ext/gd/tests/imageloadfont_invalid.phpt",
    "ext/gd/tests/imagesetthickness_error1.phpt",
    "ext/gd/tests/imagecolorallocatealpha_error3.phpt",
    "ext/gd/tests/colormatch.phpt",
    "ext/gd/tests/imagechar_error7.phpt",
    "ext/gd/tests/imagewbmp.phpt",
    "ext/gd/tests/libgd00186.phpt",
    "ext/gd/tests/imageantialias_error1.phpt",
    "ext/gd/tests/imagearc_variation1.phpt",
    "ext/gd/tests/bug38179.phpt",
    "ext/gd/tests/dashedlines.phpt",
    "ext/gd/tests/bug24155.phpt",
    "ext/gd/tests/imagecolormatch_error1.phpt",
    "ext/gd/tests/gif2png.phpt",
    "ext/gd/tests/bug27582_1.phpt",
    "ext/gd/tests/imagegammacorrect_error3.phpt",
    "ext/gd/tests/bug37346.phpt",
    "ext/gd/tests/imagefilter_error5.phpt",
    "ext/gd/tests/colorclosest.phpt",
    "ext/gd/tests/imagerectangle_error3.phpt",
    "ext/gd/tests/imagetruecolortopalette_error3.phpt",
    "ext/gd/tests/imagecolorallocate_variation5.phpt",
    "ext/gd/tests/imagecolorallocate_variation6.phpt",
    "ext/gd/tests/imagefilltoborder_error7.phpt",
    "ext/gd/tests/imagecopymerge_basic.phpt",
    "ext/gd/tests/imagecopymerge_error.phpt",
    "ext/gd/tests/imagestring_error3.phpt",
    "ext/gd/tests/gif2gd.phpt",
    "ext/gd/tests/imagecharup_error7.phpt",
    "ext/gd/tests/jpeg2wbmp_error1.phpt",
    "ext/gd/tests/imagefilter_error19.phpt",
    "ext/gd/tests/bug40764.phpt",
    "ext/gd/tests/libgd00106.phpt",
    "ext/gd/tests/png2wbmp_error1.phpt",
    "ext/gd/tests/imagetruecolortopalette_error2.phpt",
    "ext/gd/tests/imageinterlace_variation1.phpt",
    "ext/gd/tests/imagecreatetruecolor_error1.phpt",
    "ext/gd/tests/imagefilltoborder_error3.phpt",
    "ext/gd/tests/imagefilltoborder_error4.phpt",
    "ext/gd/tests/png2gd.phpt",
    "ext/gd/tests/gif.phpt",
    "ext/gd/tests/bug39082.phpt",
    "ext/gd/tests/jpeg2wbmp_error2.phpt",
    "ext/gd/tests/imagefilter_error6.phpt",
    "ext/gd/tests/imagecreatetruecolor_error3.phpt",
    "ext/gd/tests/imagecolordeallocate_error1.phpt",
    "ext/gd/tests/imagestringup_error6.phpt",
    "ext/gd/tests/copypalette.phpt",
    "ext/gd/tests/libgd00191.phpt",
    "ext/gd/tests/imagefilledpolygon_basic.phpt",
    "ext/gd/tests/pngcomp.phpt",
    "ext/gd/tests/xpm2gd.phpt",
    "ext/gd/tests/imagecreatetruecolor_basic.phpt",
    "ext/gd/tests/imagecolorallocate_basic.phpt",
    "ext/gd/tests/imagerotate_overflow.phpt",
    "ext/gd/tests/imagestring_error7.phpt",
    "ext/gd/tests/imageistruecolor_basic.phpt",
    "ext/gd/tests/createfromwbmp2_extern.phpt",
    "ext/gd/tests/imagecolormatch_error3.phpt",
    "ext/gd/tests/imagerectangle_basic.phpt",
    "ext/gd/tests/imagestringup_error7.phpt",
    "ext/gd/tests/imagefilter_error8.phpt",
    "ext/gd/tests/imagegammacorrect_variation1.phpt",
    "ext/gd/tests/imagetruecolortopalette_error1.phpt",
    "ext/gd/tests/bug42434.phpt",
    "ext/gd/tests/imagecolormatch_error2.phpt",
    "ext/gd/tests/imageellipse_error4.phpt",
    "ext/gd/tests/imagecolorallocatealpha_error5.phpt",
    "ext/gd/tests/imagecolormatch_error4.phpt",
    "ext/gd/tests/bug22544.phpt",
    "ext/gd/tests/imagesetthickness_basic.phpt",
    "ext/gd/tests/xpm2jpg.phpt",
    "ext/gd/tests/copy.phpt",
    "ext/gd/tests/imageconvolution_error3.phpt",
    "ext/gd/tests/imagefontwidth_error1.phpt",
    "ext/gd/tests/imageconvolution_error1.phpt",
    "ext/gd/tests/imagecharup_error1.phpt",
    "ext/gd/tests/imagestringup_error1.phpt",
    "ext/gd/tests/imagestring_error1.phpt",
    "ext/gd/tests/imagefilter_error14.phpt",
    "ext/gd/tests/imagechar_error3.phpt",
    "ext/gd/tests/imageinterlace_error1.phpt",
    "ext/gd/tests/imagefilltoborder_error2.phpt",
    "ext/gd/tests/imagefontheight_basic.phpt",
    "ext/gd/tests/imagelayereffect_error2.phpt",
    "ext/gd/tests/imagefilter_error11.phpt",
    "ext/gd/tests/xpm2png.phpt",
    "ext/gd/tests/imagefilledarc_error1.phpt",
    "ext/gd/tests/createfromstring.phpt",
    "ext/gd/tests/imagepolygon_basic.phpt",
    "ext/gd/tests/imagecolorallocatealpha_basic.phpt",
    "ext/gd/tests/imagegammacorrect_error2.phpt",
    "ext/gd/tests/imagerectangle_error5.phpt",
    "ext/gd/tests/imagechar_error2.phpt",
    "ext/gd/tests/imagecolorstotal_error.phpt",
    "ext/gd/tests/imagecolorallocate_variation2.phpt",
    "ext/gd/tests/imagefilter_error18.phpt",
    "ext/gd/tests/imagefilledpolygon_negative.phpt",
    "ext/gd/tests/gd_info_basic.phpt",
    "ext/gd/tests/imagefill_1.phpt",
    "ext/gd/tests/createfromgd2.phpt",
    "ext/gd/tests/jpg2gd.phpt",
    "ext/wddx/tests/005.phpt",
    "ext/wddx/tests/bug41283.phpt",
    "ext/wddx/tests/001-64bit.phpt",
    "ext/wddx/tests/bug48562.phpt",
    "ext/wddx/tests/bug35410_64bit.phpt",
    "ext/wddx/tests/001.phpt",
    "ext/wddx/tests/bug27287.phpt",
    "ext/wddx/tests/004.phpt",
    "ext/wddx/tests/003.phpt",
    "ext/wddx/tests/bug35410.phpt",
    "ext/wddx/tests/bug37587.phpt",
    "ext/wddx/tests/bug37569.phpt",
    "ext/wddx/tests/bug34306.phpt",
    "ext/wddx/tests/002.phpt",
    "ext/wddx/tests/bug45901.phpt",
    "ext/wddx/tests/bug41527.phpt",
    "ext/wddx/tests/bug52468.phpt",
    "ext/standard/tests/http/bug48929.phpt",
    "ext/standard/tests/http/ignore_errors.phpt",
    "ext/standard/tests/http/bug38802.phpt",
    "ext/standard/tests/http/bug43510.phpt",
    "ext/standard/tests/http/bug53198.phpt",
    "ext/standard/tests/php_real_logo_guid.phpt",
    "ext/standard/tests/mail/mail_variation_alt1-win32.phpt",
    "ext/standard/tests/mail/mail_basic_alt2-win32.phpt",
    "ext/standard/tests/mail/mail_basic4.phpt",
    "ext/standard/tests/mail/mail_basic_alt3-win32.phpt",
    "ext/standard/tests/mail/mail_variation_alt3-win32.phpt",
    "ext/standard/tests/mail/mail_basic3.phpt",
    "ext/standard/tests/mail/mail_basic.phpt",
    "ext/standard/tests/mail/ezmlm_hash_variation1.phpt",
    "ext/standard/tests/mail/mail_basic2.phpt",
    "ext/standard/tests/mail/mail_variation2.phpt",
    "ext/standard/tests/mail/mail_variation1.phpt",
    "ext/standard/tests/mail/ezmlm_hash_basic.phpt",
    "ext/standard/tests/mail/mail_basic_alt4-win32.phpt",
    "ext/standard/tests/mail/ezmlm_hash_basic_64bit.phpt",
    "ext/standard/tests/mail/mail_variation_alt2-win32.phpt",
    "ext/standard/tests/mail/mail_error.phpt",
    "ext/standard/tests/mail/mail_basic_alt1-win32.phpt",
    "ext/standard/tests/mail/ezmlm_hash_error.phpt",
    "ext/standard/tests/mail/mail_basic5.phpt",
    "ext/standard/tests/mail/bug51604.phpt",
    "ext/standard/tests/directory/DirectoryClass_error_001.phpt",
    "ext/standard/tests/directory/DirectoryClass_basic_001.phpt",
    "ext/standard/tests/directory/directory_constants-win32.phpt",
    "ext/standard/tests/directory/open_basedir_001.phpt",
    "ext/standard/tests/directory/directory_constants.phpt",
    "ext/standard/tests/php_ini_loaded_file.phpt",
    "ext/standard/tests/bug49244.phpt",
    "ext/standard/tests/file/glob_error_002-win32.phpt",
    "ext/standard/tests/file/touch_basic-win32.phpt",
    "ext/standard/tests/file/include_streams.phpt",
    "ext/standard/tests/file/copy_variation15.phpt",
    "ext/standard/tests/file/rename_variation2.phpt",
    "ext/standard/tests/file/fscanf_variation14.phpt",
    "ext/standard/tests/file/fgetss_error.phpt",
    "ext/standard/tests/file/disk_free_space_variation.phpt",
    "ext/standard/tests/file/fscanf_variation24.phpt",
    "ext/standard/tests/file/fopen_variation9.phpt",
    "ext/standard/tests/file/fscanf.phpt",
    "ext/standard/tests/file/file_put_contents_variation8.phpt",
    "ext/standard/tests/file/005_variation-win32.phpt",
    "ext/standard/tests/file/lstat_stat_variation3.phpt",
    "ext/standard/tests/file/glob_basic.phpt",
    "ext/standard/tests/file/bug45181.phpt",
    "ext/standard/tests/file/fscanf_variation1.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_variation8-win32.phpt",
    "ext/standard/tests/file/fputcsv_variation8.phpt",
    "ext/standard/tests/file/stat_variation6-win32.phpt",
    "ext/standard/tests/file/copy_variation7.phpt",
    "ext/standard/tests/file/touch_variation3.phpt",
    "ext/standard/tests/file/disk_free_space_error.phpt",
    "ext/standard/tests/file/rename_variation3-win32.phpt",
    "ext/standard/tests/file/readfile_variation9.phpt",
    "ext/standard/tests/file/007_variation6.phpt",
    "ext/standard/tests/file/popen_pclose_error-win32.phpt",
    "ext/standard/tests/file/copy_variation14.phpt",
    "ext/standard/tests/file/fputcsv_variation4.phpt",
    "ext/standard/tests/file/fopen_variation15.phpt",
    "ext/standard/tests/file/file_get_contents_variation3.phpt",
    "ext/standard/tests/file/file_exists_variation1.phpt",
    "ext/standard/tests/file/copy_variation6.phpt",
    "ext/standard/tests/file/is_executable_basic.phpt",
    "ext/standard/tests/file/move_uploaded_file_basic.phpt",
    "ext/standard/tests/file/ftruncate_variation7.phpt",
    "ext/standard/tests/file/fscanf_variation23.phpt",
    "ext/standard/tests/file/file_get_contents_file_put_contents_variation1.phpt",
    "ext/standard/tests/file/flock_basic.phpt",
    "ext/standard/tests/file/fgets_variation2.phpt",
    "ext/standard/tests/file/fwrite_variation5.phpt",
    "ext/standard/tests/file/fwrite_variation3.phpt",
    "ext/standard/tests/file/readlink_realpath_variation3.phpt",
    "ext/standard/tests/file/unlink_variation8.phpt",
    "ext/standard/tests/file/bug12556.phpt",
    "ext/standard/tests/file/bug44034.phpt",
    "ext/standard/tests/file/fpassthru_variation1.phpt",
    "ext/standard/tests/file/fopen_variation5.phpt",
    "ext/standard/tests/file/file_basic.phpt",
    "ext/standard/tests/file/fgetss_variation4.phpt",
    "ext/standard/tests/file/touch_variation5-win32.phpt",
    "ext/standard/tests/file/fgets_variation5.phpt",
    "ext/standard/tests/file/fscanf_variation8.phpt",
    "ext/standard/tests/file/bug38450_1.phpt",
    "ext/standard/tests/file/disk_free_space_error-win32.phpt",
    "ext/standard/tests/file/userstreams_005.phpt",
    "ext/standard/tests/file/readfile_variation3.phpt",
    "ext/standard/tests/file/rmdir_variation2.phpt",
    "ext/standard/tests/file/mkdir_variation5-win32.phpt",
    "ext/standard/tests/file/stream_get_line.phpt",
    "ext/standard/tests/file/lstat_stat_variation13.phpt",
    "ext/standard/tests/file/fgets_socket_variation2.phpt",
    "ext/standard/tests/file/fscanf_variation15.phpt",
    "ext/standard/tests/file/stream_002.phpt",
    "ext/standard/tests/file/umask_basic.phpt",
    "ext/standard/tests/file/file_get_contents_variation7.phpt",
    "ext/standard/tests/file/copy_variation16-win32.phpt",
    "ext/standard/tests/file/parse_ini_file_variation2.phpt",
    "ext/standard/tests/file/glob_variation2.phpt",
    "ext/standard/tests/file/filetype_variation.phpt",
    "ext/standard/tests/file/fclose_variation1.phpt",
    "ext/standard/tests/file/is_uploaded_file_basic.phpt",
    "ext/standard/tests/file/readlink_realpath_variation1.phpt",
    "ext/standard/tests/file/bug26938.phpt",
    "ext/standard/tests/file/chmod_error.phpt",
    "ext/standard/tests/file/disk_total_space_error-win32.phpt",
    "ext/standard/tests/file/file_get_contents_error001.phpt",
    "ext/standard/tests/file/007_variation16.phpt",
    "ext/standard/tests/file/fgetcsv_error.phpt",
    "ext/standard/tests/file/ftruncate.phpt",
    "ext/standard/tests/file/rename_variation1.phpt",
    "ext/standard/tests/file/001-win32.phpt",
    "ext/standard/tests/file/is_readable_error.phpt",
    "ext/standard/tests/file/file_error.phpt",
    "ext/standard/tests/file/readlink_realpath_variation2.phpt",
    "ext/standard/tests/file/copy_variation4.phpt",
    "ext/standard/tests/file/rename_variation12-win32.phpt",
    "ext/standard/tests/file/007_variation12-win32.phpt",
    "ext/standard/tests/file/bug44607.phpt",
    "ext/standard/tests/file/fscanf_variation4.phpt",
    "ext/standard/tests/file/copy_variation5.phpt",
    "ext/standard/tests/file/chmod_basic.phpt",
    "ext/standard/tests/file/file_put_contents_variation4.phpt",
    "ext/standard/tests/file/fgetcsv_variation31.phpt",
    "ext/standard/tests/file/fscanf_variation19.phpt",
    "ext/standard/tests/file/file_get_contents_variation8-win32.phpt",
    "ext/standard/tests/file/readlink_realpath_basic1.phpt",
    "ext/standard/tests/file/fscanf_variation35.phpt",
    "ext/standard/tests/file/fscanf_variation32.phpt",
    "ext/standard/tests/file/stat_basic-win32.phpt",
    "ext/standard/tests/file/umask_variation1.phpt",
    "ext/standard/tests/file/file_get_contents_variation6.phpt",
    "ext/standard/tests/file/stat_variation1-win32.phpt",
    "ext/standard/tests/file/file_get_contents_variation2.phpt",
    "ext/standard/tests/file/stream_rfc2397_007.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_basic1.phpt",
    "ext/standard/tests/file/fgetc_variation1.phpt",
    "ext/standard/tests/file/rename_basic.phpt",
    "ext/standard/tests/file/pathinfo_basic1-win32.phpt",
    "ext/standard/tests/file/symlink_link_linkinfo_is_link_variation3.phpt",
    "ext/standard/tests/file/is_file_basic.phpt",
    "ext/standard/tests/file/rename_variation2-win32.phpt",
    "ext/standard/tests/file/fstat_variation6.phpt",
    "ext/standard/tests/file/flock_error.phpt",
    "ext/standard/tests/file/parse_ini_file_variation5.phpt",
    "ext/standard/tests/file/fgetss_variation3.phpt",
    "ext/standard/tests/file/007_variation14.phpt",
    "ext/standard/tests/file/rename_variation12.phpt",
    "ext/standard/tests/file/get_current_user.phpt",
    "ext/standard/tests/file/fgetss_basic2.phpt",
    "ext/standard/tests/file/ftruncate_variation7-win32.phpt",
    "ext/standard/tests/file/unlink_variation7.phpt",
    "ext/standard/tests/file/filetype_variation3.phpt",
    "ext/standard/tests/file/rename_variation13.phpt",
    "ext/standard/tests/file/005_error.phpt",
    "ext/standard/tests/file/symlink_link_linkinfo_is_link_variation7.phpt",
    "ext/standard/tests/file/file_put_contents_variation9.phpt",
    "ext/standard/tests/file/fscanf_variation22.phpt",
    "ext/standard/tests/file/fscanf_variation54.phpt",
    "ext/standard/tests/file/symlink_to_symlink.phpt",
    "ext/standard/tests/file/fopen_variation14.phpt",
    "ext/standard/tests/file/readfile_basic.phpt",
    "ext/standard/tests/file/file_put_contents_variation8-win32.phpt",
    "ext/standard/tests/file/file_put_contents_variation1.phpt",
    "ext/standard/tests/file/bug45303.phpt",
    "ext/standard/tests/file/006_basic.phpt",
    "ext/standard/tests/file/lstat_stat_variation10.phpt",
    "ext/standard/tests/file/bug39538.phpt",
    "ext/standard/tests/file/file_variation2.phpt",
    "ext/standard/tests/file/rename_variation1-win32.phpt",
    "ext/standard/tests/file/symlink_link_linkinfo_is_link_variation9.phpt",
    "ext/standard/tests/file/fgetcsv_variation21.phpt",
    "ext/standard/tests/file/mkdir-001.phpt",
    "ext/standard/tests/file/bug41693.phpt",
    "ext/standard/tests/file/file_put_contents_variation3.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_error1.phpt",
    "ext/standard/tests/file/fgetss1.phpt",
    "ext/standard/tests/file/popen_pclose_basic.phpt",
    "ext/standard/tests/file/fputcsv_variation3.phpt",
    "ext/standard/tests/file/fscanf_variation51.phpt",
    "ext/standard/tests/file/filesize_variation1.phpt",
    "ext/standard/tests/file/007_variation18.phpt",
    "ext/standard/tests/file/lstat_stat_variation21.phpt",
    "ext/standard/tests/file/tempnam_variation1-win32.phpt",
    "ext/standard/tests/file/copy_variation1.phpt",
    "ext/standard/tests/file/fopen_variation13.phpt",
    "ext/standard/tests/file/symlink.phpt",
    "ext/standard/tests/file/fgetcsv_variation15.phpt",
    "ext/standard/tests/file/statcache-corruption.phpt",
    "ext/standard/tests/file/fgetcsv_variation18.phpt",
    "ext/standard/tests/file/lstat_stat_variation11.phpt",
    "ext/standard/tests/file/dirname_variation1.phpt",
    "ext/standard/tests/file/is_readable_variation3.phpt",
    "ext/standard/tests/file/windows_acls/bug44859_2.phpt",
    "ext/standard/tests/file/windows_acls/bug44859_4.phpt",
    "ext/standard/tests/file/windows_acls/bug44859_3.phpt",
    "ext/standard/tests/file/windows_acls/bug44859.phpt",
    "ext/standard/tests/file/rmdir_variation1-win32.phpt",
    "ext/standard/tests/file/fgetcsv_variation9.phpt",
    "ext/standard/tests/file/copy_variation3-win32.phpt",
    "ext/standard/tests/file/file_get_contents_variation7-win32.phpt",
    "ext/standard/tests/file/bug27619.phpt",
    "ext/standard/tests/file/readfile_basic-win32.phpt",
    "ext/standard/tests/file/007_variation11.phpt",
    "ext/standard/tests/file/popen_pclose_error.phpt",
    "ext/standard/tests/file/file_put_contents_variation7.phpt",
    "ext/standard/tests/file/copy_variation12.phpt",
    "ext/standard/tests/file/file_get_contents_file_put_contents_basic.phpt",
    "ext/standard/tests/file/mkdir_rmdir_variation2.phpt",
    "ext/standard/tests/file/fgetss.phpt",
    "ext/standard/tests/file/fopen_variation7.phpt",
    "ext/standard/tests/file/readfile_variation2.phpt",
    "ext/standard/tests/file/unlink_variation6.phpt",
    "ext/standard/tests/file/file_get_contents_variation8.phpt",
    "ext/standard/tests/file/001.phpt",
    "ext/standard/tests/file/fputcsv.phpt",
    "ext/standard/tests/file/fgetss_basic2-win32.phpt",
    "ext/standard/tests/file/lstat_stat_variation12.phpt",
    "ext/standard/tests/file/file_get_contents_variation5.phpt",
    "ext/standard/tests/file/ftruncate_variation5-win32.phpt",
    "ext/standard/tests/file/fopen_variation11-win32.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_variation1.phpt",
    "ext/standard/tests/file/parse_ini_file_variation6-win32.phpt",
    "ext/standard/tests/file/bug41815.phpt",
    "ext/standard/tests/file/file_get_contents_error002.phpt",
    "ext/standard/tests/file/glob_error.phpt",
    "ext/standard/tests/file/is_writable_error.phpt",
    "ext/standard/tests/file/rename_variation13-win32.phpt",
    "ext/standard/tests/file/fwrite_variation3-win32.phpt",
    "ext/standard/tests/file/unlink_variation1.phpt",
    "ext/standard/tests/file/disk_free_space_basic.phpt",
    "ext/standard/tests/file/fflush_variation1.phpt",
    "ext/standard/tests/file/fopen_variation16.phpt",
    "ext/standard/tests/file/chroot_001.phpt",
    "ext/standard/tests/file/file_put_contents_variation6.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_variation8.phpt",
    "ext/standard/tests/file/mkdir-002.phpt",
    "ext/standard/tests/file/unlink_variation10.phpt",
    "ext/standard/tests/file/proc_open01.phpt",
    "ext/standard/tests/file/basename_variation2.phpt",
    "ext/standard/tests/file/fgetcsv_variation14.phpt",
    "ext/standard/tests/file/readfile_variation7.phpt",
    "ext/standard/tests/file/rename_error.phpt",
    "ext/standard/tests/file/004.phpt",
    "ext/standard/tests/file/ftruncate_variation6-win32.phpt",
    "ext/standard/tests/file/007_variation21.phpt",
    "ext/standard/tests/file/realpath_variation-win32.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_variation6.phpt",
    "ext/standard/tests/file/userstreams_004.phpt",
    "ext/standard/tests/file/ftruncate_variation3.phpt",
    "ext/standard/tests/file/007_variation23.phpt",
    "ext/standard/tests/file/fopen_variation14-win32.phpt",
    "ext/standard/tests/file/mkdir_variation1-win32.phpt",
    "ext/standard/tests/file/005_variation2-win32.phpt",
    "ext/standard/tests/file/fopen_variation15-win32.phpt",
    "ext/standard/tests/file/realpath_basic2.phpt",
    "ext/standard/tests/file/readfile_variation10-win32.phpt",
    "ext/standard/tests/file/lchown_basic.phpt",
    "ext/standard/tests/file/bug43137.phpt",
    "ext/standard/tests/file/userstreams_006.phpt",
    "ext/standard/tests/file/file_variation6.phpt",
    "ext/standard/tests/file/bug41655_2.phpt",
    "ext/standard/tests/file/chmod_variation3.phpt",
    "ext/standard/tests/file/ftruncate_variation5.phpt",
    "ext/standard/tests/file/bug43216.phpt",
    "ext/standard/tests/file/fwrite_variation1.phpt",
    "ext/standard/tests/file/copy_variation13.phpt",
    "ext/standard/tests/file/file_variation8-win32.phpt",
    "ext/standard/tests/file/stream_rfc2397_001.phpt",
    "ext/standard/tests/file/fputcsv_002.phpt",
    "ext/standard/tests/file/chown.phpt",
    "ext/standard/tests/file/filesize_variation2-win32.phpt",
    "ext/standard/tests/file/copy_variation6-win32.phpt",
    "ext/standard/tests/file/fstat_variation1.phpt",
    "ext/standard/tests/file/fnmatch_basic.phpt",
    "ext/standard/tests/file/stream_rfc2397_002.phpt",
    "ext/standard/tests/file/unlink_variation5.phpt",
    "ext/standard/tests/file/bug38086.phpt",
    "ext/standard/tests/file/file_put_contents_variation2.phpt",
    "ext/standard/tests/file/userdirstream.phpt",
    "ext/standard/tests/file/fileinode_variation1.phpt",
    "ext/standard/tests/file/fgetcsv_variation7.phpt",
    "ext/standard/tests/file/file_variation5-win32.phpt",
    "ext/standard/tests/file/basename_variation1-win32.phpt",
    "ext/standard/tests/file/realpath_variation2.phpt",
    "ext/standard/tests/file/bug47767.phpt",
    "ext/standard/tests/file/unlink_variation9-win32.phpt",
    "ext/standard/tests/file/stat_error-win32.phpt",
    "ext/standard/tests/file/bug37158.phpt",
    "ext/standard/tests/file/stream_rfc2397_006.phpt",
    "ext/standard/tests/file/stream_001.phpt",
    "ext/standard/tests/file/fflush_variation1-win32.phpt",
    "ext/standard/tests/file/is_writable_variation1.phpt",
    "ext/standard/tests/file/filetype_variation2.phpt",
    "ext/standard/tests/file/php_fd_wrapper_01.phpt",
    "ext/standard/tests/file/fscanf_variation3.phpt",
    "ext/standard/tests/file/stat_variation8-win32.phpt",
    "ext/standard/tests/file/copy_basic.phpt",
    "ext/standard/tests/file/bug43522.phpt",
    "ext/standard/tests/file/006_variation2.phpt",
    "ext/standard/tests/file/basename-win32.phpt",
    "ext/standard/tests/file/rename_variation.phpt",
    "ext/standard/tests/file/copy_variation16.phpt",
    "ext/standard/tests/file/filesize_basic.phpt",
    "ext/standard/tests/file/stream_copy_to_stream.phpt",
    "ext/standard/tests/file/fgetcsv.phpt",
    "ext/standard/tests/file/file_variation8.phpt",
    "ext/standard/tests/file/fgetcsv_variation19.phpt",
    "ext/standard/tests/file/symlink_link_linkinfo_is_link_basic1.phpt",
    "ext/standard/tests/file/realpath_cache_win32.phpt",
    "ext/standard/tests/file/bug20424.phpt",
    "ext/standard/tests/file/dirname_basic.phpt",
    "ext/standard/tests/file/symlink_link_linkinfo_is_link_variation4.phpt",
    "ext/standard/tests/file/fscanf_variation43.phpt",
    "ext/standard/tests/file/ftruncate_variation1.phpt",
    "ext/standard/tests/file/fgetcsv_variation27.phpt",
    "ext/standard/tests/file/bug41655_1.phpt",
    "ext/standard/tests/file/pathinfo_variation1.phpt",
    "ext/standard/tests/file/fgetss_variation3-win32.phpt",
    "ext/standard/tests/file/fscanf_variation47.phpt",
    "ext/standard/tests/file/fwrite_basic.phpt",
    "ext/standard/tests/file/fscanf_variation13.phpt",
    "ext/standard/tests/file/pathinfo_basic1.phpt",
    "ext/standard/tests/file/php_fd_wrapper_04.phpt",
    "ext/standard/tests/file/rename_variation11-win32.phpt",
    "ext/standard/tests/file/unlink_basic.phpt",
    "ext/standard/tests/file/copy_variation5-win32.phpt",
    "ext/standard/tests/file/stream_rfc2397_005.phpt",
    "ext/standard/tests/file/003.phpt",
    "ext/standard/tests/file/007_variation20.phpt",
    "ext/standard/tests/file/fileperms_variation1.phpt",
    "ext/standard/tests/file/touch_basic.phpt",
    "ext/standard/tests/file/chmod_variation1.phpt",
    "ext/standard/tests/file/ftruncate_variation3-win32.phpt",
    "ext/standard/tests/file/tempnam_variation2-win32.phpt",
    "ext/standard/tests/file/symlink_link_linkinfo_is_link_variation8.phpt",
    "ext/standard/tests/file/file_put_contents_variation5.phpt",
    "ext/standard/tests/file/fgetc_variation3.phpt",
    "ext/standard/tests/file/ftruncate_error.phpt",
    "ext/standard/tests/file/popen_pclose_error-win32-debug.phpt",
    "ext/standard/tests/file/lstat_stat_variation18.phpt",
    "ext/standard/tests/file/fopen_variation19.phpt",
    "ext/standard/tests/file/fscanf_variation21.phpt",
    "ext/standard/tests/file/fscanf_variation50.phpt",
    "ext/standard/tests/file/fgetcsv_variation25.phpt",
    "ext/standard/tests/file/fstat_variation8.phpt",
    "ext/standard/tests/file/fscanf_variation31.phpt",
    "ext/standard/tests/file/is_file_variation3.phpt",
    "ext/standard/tests/file/fscanf_variation34.phpt",
    "ext/standard/tests/file/lstat_stat_variation8.phpt",
    "ext/standard/tests/file/tempnam_error.phpt",
    "ext/standard/tests/file/fscanf_variation26.phpt",
    "ext/standard/tests/file/feof_basic.phpt",
    "ext/standard/tests/file/filetype_basic.phpt",
    "ext/standard/tests/file/fscanf_variation11.phpt",
    "ext/standard/tests/file/fflush_variation2.phpt",
    "ext/standard/tests/file/fread_error.phpt",
    "ext/standard/tests/file/007_variation4.phpt",
    "ext/standard/tests/file/fgetss_basic1.phpt",
    "ext/standard/tests/file/bug32160.phpt",
    "ext/standard/tests/file/readfile_variation1.phpt",
    "ext/standard/tests/file/007_variation7.phpt",
    "ext/standard/tests/file/fgetcsv_variation10.phpt",
    "ext/standard/tests/file/bug35781.phpt",
    "ext/standard/tests/file/readfile_variation10.phpt",
    "ext/standard/tests/file/rename_variation7.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_variation6-win32.phpt",
    "ext/standard/tests/file/fileinode_error.phpt",
    "ext/standard/tests/file/fscanf_variation18.phpt",
    "ext/standard/tests/file/symlink_link_linkinfo_is_link_error2.phpt",
    "ext/standard/tests/file/fscanf_variation41.phpt",
    "ext/standard/tests/file/touch_variation2.phpt",
    "ext/standard/tests/file/basename_variation4.phpt",
    "ext/standard/tests/file/bug39673.phpt",
    "ext/standard/tests/file/fscanf_variation30.phpt",
    "ext/standard/tests/file/unlink_variation2-win32.phpt",
    "ext/standard/tests/file/unlink_variation2.phpt",
    "ext/standard/tests/file/stat_variation3-win32.phpt",
    "ext/standard/tests/file/fgetcsv_variation16.phpt",
    "ext/standard/tests/file/fgetss_variation5.phpt",
    "ext/standard/tests/file/fscanf_variation17.phpt",
    "ext/standard/tests/file/lstat_stat_variation6.phpt",
    "ext/standard/tests/file/fscanf_variation33.phpt",
    "ext/standard/tests/file/readfile_variation5.phpt",
    "ext/standard/tests/file/fscanf_variation36.phpt",
    "ext/standard/tests/file/fputcsv_variation10.phpt",
    "ext/standard/tests/file/is_readable_basic.phpt",
    "ext/standard/tests/file/bug40374.phpt",
    "ext/standard/tests/file/fileinode_basic.phpt",
    "ext/standard/tests/file/is_writable_variation2.phpt",
    "ext/standard/tests/file/tempnam_variation7.phpt",
    "ext/standard/tests/file/copy_variation8.phpt",
    "ext/standard/tests/file/touch_variation1.phpt",
    "ext/standard/tests/file/readfile_variation4.phpt",
    "ext/standard/tests/file/fwrite_basic-win32.phpt",
    "ext/standard/tests/file/tempnam_variation3-win32.phpt",
    "ext/standard/tests/file/fgetcsv_variation8.phpt",
    "ext/standard/tests/file/bug22382.phpt",
    "ext/standard/tests/file/fopen_variation3.phpt",
    "ext/standard/tests/file/fwrite_variation2.phpt",
    "ext/standard/tests/file/touch_variation3-win32.phpt",
    "ext/standard/tests/file/php_fd_wrapper_02.phpt",
    "ext/standard/tests/file/file_get_contents_basic001.phpt",
    "ext/standard/tests/file/file_get_contents_variation4.phpt",
    "ext/standard/tests/file/parse_ini_file_error.phpt",
    "ext/standard/tests/file/fstat_variation5.phpt",
    "ext/standard/tests/file/fscanf_variation7.phpt",
    "ext/standard/tests/file/fscanf_variation2.phpt",
    "ext/standard/tests/file/readlink_realpath_basic2.phpt",
    "ext/standard/tests/file/005_basic.phpt",
    "ext/standard/tests/file/stream_enclosed.phpt",
    "ext/standard/tests/file/fgetcsv_variation23.phpt",
    "ext/standard/tests/file/007_variation5.phpt",
    "ext/standard/tests/file/fgetcsv_variation12.phpt",
    "ext/standard/tests/file/fscanf_variation45.phpt",
    "ext/standard/tests/file/ftruncate_variation1-win32.phpt",
    "ext/standard/tests/file/fscanf_variation25.phpt",
    "ext/standard/tests/file/lstat_stat_variation22.phpt",
    "ext/standard/tests/file/fscanf_variation10.phpt",
    "ext/standard/tests/file/link_win32.phpt",
    "ext/standard/tests/file/include_userstream_002.phpt",
    "ext/standard/tests/file/fscanf_variation29.phpt",
    "ext/standard/tests/file/fwrite_error.phpt",
    "ext/standard/tests/file/chmod_variation2-win32.phpt",
    "ext/standard/tests/file/unlink_variation3.phpt",
    "ext/standard/tests/file/007_variation24.phpt",
    "ext/standard/tests/file/fileperms_variation2.phpt",
    "ext/standard/tests/file/fstat_variation3.phpt",
    "ext/standard/tests/file/fopencookie.phpt",
    "ext/standard/tests/file/fgetss_variation2.phpt",
    "ext/standard/tests/file/filesize_variation2.phpt",
    "ext/standard/tests/file/file_variation5.phpt",
    "ext/standard/tests/file/unlink_error-win32.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_variation2-win32.phpt",
    "ext/standard/tests/file/is_executable_variation1.phpt",
    "ext/standard/tests/file/mkdir_rmdir_variation1.phpt",
    "ext/standard/tests/file/fgetcsv_variation29.phpt",
    "ext/standard/tests/file/fread_variation2.phpt",
    "ext/standard/tests/file/include_userstream_001.phpt",
    "ext/standard/tests/file/bug52820.phpt",
    "ext/standard/tests/file/touch.phpt",
    "ext/standard/tests/file/filegroup_variation2.phpt",
    "ext/standard/tests/file/007_variation11-win32.phpt",
    "ext/standard/tests/file/fscanf_variation16.phpt",
    "ext/standard/tests/file/is_file_variation4.phpt",
    "ext/standard/tests/file/bug22414.phpt",
    "ext/standard/tests/file/rmdir_variation3.phpt",
    "ext/standard/tests/file/rename_variation8-win32.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_basic2.phpt",
    "ext/standard/tests/file/fileowner_error.phpt",
    "ext/standard/tests/file/dirname_basic-win32.phpt",
    "ext/standard/tests/file/lstat_stat_variation9.phpt",
    "ext/standard/tests/file/fflush_variation3.phpt",
    "ext/standard/tests/file/lstat_stat_error.phpt",
    "ext/standard/tests/file/006_error.phpt",
    "ext/standard/tests/file/tempnam_variation5.phpt",
    "ext/standard/tests/file/mkdir_variation4.phpt",
    "ext/standard/tests/file/fgetcsv_variation2.phpt",
    "ext/standard/tests/file/fseek_variation2.phpt",
    "ext/standard/tests/file/lstat_stat_variation2.phpt",
    "ext/standard/tests/file/fwrite_variation4-win32.phpt",
    "ext/standard/tests/file/lstat_stat_variation1.phpt",
    "ext/standard/tests/file/fread_variation4.phpt",
    "ext/standard/tests/file/fileowner_basic.phpt",
    "ext/standard/tests/file/mkdir_rmdir_variation-win32.phpt",
    "ext/standard/tests/file/fgetc_basic.phpt",
    "ext/standard/tests/file/fgetss_variation1.phpt",
    "ext/standard/tests/file/fopen_variation8.phpt",
    "ext/standard/tests/file/mkdir_rmdir_error.phpt",
    "ext/standard/tests/file/bug39367.phpt",
    "ext/standard/tests/file/parse_ini_file_variation1.phpt",
    "ext/standard/tests/file/fopen_variation1.phpt",
    "ext/standard/tests/file/file_variation4.phpt",
    "ext/standard/tests/file/ftruncate_variation6.phpt",
    "ext/standard/tests/file/fileinode_variation2.phpt",
    "ext/standard/tests/file/filegroup_variation1.phpt",
    "ext/standard/tests/file/fgetcsv_variation26.phpt",
    "ext/standard/tests/file/umask_variation3.phpt",
    "ext/standard/tests/file/parse_ini_file_variation6.phpt",
    "ext/standard/tests/file/symlink_link_linkinfo_is_link_variation5.phpt",
    "ext/standard/tests/file/filesize_variation5.phpt",
    "ext/standard/tests/file/bug43353-win32.phpt",
    "ext/standard/tests/file/fgetc_variation4.phpt",
    "ext/standard/tests/file/fputcsv_variation14.phpt",
    "ext/standard/tests/file/fread_variation3-win32.phpt",
    "ext/standard/tests/file/rename_variation5.phpt",
    "ext/standard/tests/file/is_executable_variation2.phpt",
    "ext/standard/tests/file/ftruncate_variation4-win32.phpt",
    "ext/standard/tests/file/fscanf_variation37.phpt",
    "ext/standard/tests/file/mkdir_variation2.phpt",
    "ext/standard/tests/file/lstat_stat_variation19.phpt",
    "ext/standard/tests/file/symlink_link_linkinfo_is_link_error1.phpt",
    "ext/standard/tests/file/007_variation1.phpt",
    "ext/standard/tests/file/fputcsv_variation1.phpt",
    "ext/standard/tests/file/fwrite_variation2-win32.phpt",
    "ext/standard/tests/file/directory_wrapper_fstat_basic.phpt",
    "ext/standard/tests/file/fgetss_variation1-win32.phpt",
    "ext/standard/tests/file/fstat.phpt",
    "ext/standard/tests/file/filegroup_error.phpt",
    "ext/standard/tests/file/007_variation8.phpt",
    "ext/standard/tests/file/parse_ini_file_variation4.phpt",
    "ext/standard/tests/file/rename_variation6.phpt",
    "ext/standard/tests/file/copy_variation10.phpt",
    "ext/standard/tests/file/flock_variation.phpt",
    "ext/standard/tests/file/stat_variation4-win32.phpt",
    "ext/standard/tests/file/bug39551.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_variation4.phpt",
    "ext/standard/tests/file/symlink_link_linkinfo_is_link_basic2.phpt",
    "ext/standard/tests/file/filesize_variation1-win32.phpt",
    "ext/standard/tests/file/popen_pclose_error-sunos.phpt",
    "ext/standard/tests/file/file_put_contents_variation7-win32.phpt",
    "ext/standard/tests/file/fputcsv_variation5.phpt",
    "ext/standard/tests/file/bug38450.phpt",
    "ext/standard/tests/file/is_readable_basic-win32.phpt",
    "ext/standard/tests/file/fgetss_variation5-win32.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_error3.phpt",
    "ext/standard/tests/file/bug24482.phpt",
    "ext/standard/tests/file/007_variation13.phpt",
    "ext/standard/tests/file/filesize_variation3-win32.phpt",
    "ext/standard/tests/file/tempnam_variation3.phpt",
    "ext/standard/tests/file/chmod_basic-win32.phpt",
    "ext/standard/tests/file/umask_error.phpt",
    "ext/standard/tests/file/fread_socket_variation1.phpt",
    "ext/standard/tests/file/fopen_variation6.phpt",
    "ext/standard/tests/file/file_get_contents_error.phpt",
    "ext/standard/tests/file/basename_variation3.phpt",
    "ext/standard/tests/file/statpage.phpt",
    "ext/standard/tests/file/002.phpt",
    "ext/standard/tests/file/fileowner_variation2.phpt",
    "ext/standard/tests/file/filegroup_variation3.phpt",
    "ext/standard/tests/file/fwrite.phpt",
    "ext/standard/tests/file/file_variation7.phpt",
    "ext/standard/tests/file/fputcsv_variation9.phpt",
    "ext/standard/tests/file/fputcsv_variation6.phpt",
    "ext/standard/tests/file/tempnam_variation5-win32.phpt",
    "ext/standard/tests/file/file_put_contents.phpt",
    "ext/standard/tests/file/filegroup_basic.phpt",
    "ext/standard/tests/file/bug46347.phpt",
    "ext/standard/tests/file/is_file_variation2.phpt",
    "ext/standard/tests/file/filesize_variation4-win32.phpt",
    "ext/standard/tests/file/is_readable_variation1.phpt",
    "ext/standard/tests/file/copy_variation18.phpt",
    "ext/standard/tests/file/bug37864.phpt",
    "ext/standard/tests/file/fputcsv_variation2.phpt",
    "ext/standard/tests/file/clearstatcache_001.phpt",
    "ext/standard/tests/file/unlink_variation8-win32.phpt",
    "ext/standard/tests/file/disk_total_space_error.phpt",
    "ext/standard/tests/file/copy_variation11.phpt",
    "ext/standard/tests/file/bug41874.phpt",
    "ext/standard/tests/file/basename_variation2-win32.phpt",
    "ext/standard/tests/file/fread_variation4-win32.phpt",
    "ext/standard/tests/file/fgetcsv_variation30.phpt",
    "ext/standard/tests/file/fgetcsv_variation11.phpt",
    "ext/standard/tests/file/007_error.phpt",
    "ext/standard/tests/file/disk.phpt",
    "ext/standard/tests/file/realpath_basic-win32.phpt",
    "ext/standard/tests/file/filestat.phpt",
    "ext/standard/tests/file/readlink_realpath_error.phpt",
    "ext/standard/tests/file/file_get_contents_file_put_contents_variation2.phpt",
    "ext/standard/tests/file/magic_quotes.phpt",
    "ext/standard/tests/file/stream_rfc2397_003.phpt",
    "ext/standard/tests/file/bug38450_2.phpt",
    "ext/standard/tests/file/bug43353.phpt",
    "ext/standard/tests/file/mkdir_variation3.phpt",
    "ext/standard/tests/file/007_variation13-win32.phpt",
    "ext/standard/tests/file/007_basic.phpt",
    "ext/standard/tests/file/touch_variation4-win32.phpt",
    "ext/standard/tests/file/fileperms_variation3.phpt",
    "ext/standard/tests/file/fgetcsv_variation5.phpt",
    "ext/standard/tests/file/copy_variation3.phpt",
    "ext/standard/tests/file/unlink_variation1-win32.phpt",
    "ext/standard/tests/file/rename_variation7-win32.phpt",
    "ext/standard/tests/file/005_variation2.phpt",
    "ext/standard/tests/file/fstat_variation2.phpt",
    "ext/standard/tests/file/fgetc_error.phpt",
    "ext/standard/tests/file/mkdir-004.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_basic2-win32.phpt",
    "ext/standard/tests/file/mkdir-006.phpt",
    "ext/standard/tests/file/bug35740.phpt",
    "ext/standard/tests/file/dirname_error.phpt",
    "ext/standard/tests/file/bug30362.phpt",
    "ext/standard/tests/file/chmod_variation2.phpt",
    "ext/standard/tests/file/fflush_variation4.phpt",
    "ext/standard/tests/file/fgetcsv_variation4.phpt",
    "ext/standard/tests/file/copy_variation9.phpt",
    "ext/standard/tests/file/fgetcsv_variation13.phpt",
    "ext/standard/tests/file/is_dir_variation4.phpt",
    "ext/standard/tests/file/007_variation22.phpt",
    "ext/standard/tests/file/fread_variation3.phpt",
    "ext/standard/tests/file/fgets_variation3.phpt",
    "ext/standard/tests/file/rename_variation10.phpt",
    "ext/standard/tests/file/lstat_stat_variation7.phpt",
    "ext/standard/tests/file/fgetcsv_variation3.phpt",
    "ext/standard/tests/file/fgetcsv_variation22.phpt",
    "ext/standard/tests/file/fscanf_variation42.phpt",
    "ext/standard/tests/file/fscanf_variation49.phpt",
    "ext/standard/tests/file/pathinfo_basic2.phpt",
    "ext/standard/tests/file/fgets_variation6-win32.phpt",
    "ext/standard/tests/file/pathinfo_error.phpt",
    "ext/standard/tests/file/fputcsv_error.phpt",
    "ext/standard/tests/file/fflush_basic.phpt",
    "ext/standard/tests/file/is_dir_error.phpt",
    "ext/standard/tests/file/touch_variation6-win32.phpt",
    "ext/standard/tests/file/lstat_stat_basic.phpt",
    "ext/standard/tests/file/fread_basic.phpt",
    "ext/standard/tests/file/pclose_variation1.phpt",
    "ext/standard/tests/file/rename_variation3.phpt",
    "ext/standard/tests/file/fseek_dir_basic.phpt",
    "ext/standard/tests/file/bug41874_1.phpt",
    "ext/standard/tests/file/tempnam_variation1.phpt",
    "ext/standard/tests/file/fread_fwrite_basic.phpt",
    "ext/standard/tests/file/bug49047.phpt",
    "ext/standard/tests/file/fputcsv_variation7.phpt",
    "ext/standard/tests/file/fscanf_variation53.phpt",
    "ext/standard/tests/file/fileowner_variation3.phpt",
    "ext/standard/tests/file/chgrp.phpt",
    "ext/standard/tests/file/fscanf_variation46.phpt",
    "ext/standard/tests/file/is_dir_basic.phpt",
    "ext/standard/tests/file/fscanf_variation44.phpt",
    "ext/standard/tests/file/file_get_contents_variation1.phpt",
    "ext/standard/tests/file/basename_basic-win32.phpt",
    "ext/standard/tests/file/bug24313.phpt",
    "ext/standard/tests/file/ftruncate_variation2-win32.phpt",
    "ext/standard/tests/file/is_file_variation1.phpt",
    "ext/standard/tests/file/is_executable_variation3.phpt",
    "ext/standard/tests/file/symlink_link_linkinfo_is_link_variation2.phpt",
    "ext/standard/tests/file/mkdir-003.phpt",
    "ext/standard/tests/file/fnmatch_error.phpt",
    "ext/standard/tests/file/fscanf_variation20.phpt",
    "ext/standard/tests/file/tempnam_variation8-win32.phpt",
    "ext/standard/tests/file/bug40501.phpt",
    "ext/standard/tests/file/bug41874_2.phpt",
    "ext/standard/tests/file/fgetc_variation2.phpt",
    "ext/standard/tests/file/copy_error.phpt",
    "ext/standard/tests/file/fgets_error.phpt",
    "ext/standard/tests/file/rename_variation-win32.phpt",
    "ext/standard/tests/file/touch_error.phpt",
    "ext/standard/tests/file/fstat_error.phpt",
    "ext/standard/tests/file/fscanf_variation55.phpt",
    "ext/standard/tests/file/readfile_variation8-win32.phpt",
    "ext/standard/tests/file/fscanf_variation6.phpt",
    "ext/standard/tests/file/mkdir_variation1.phpt",
    "ext/standard/tests/file/is_file_error.phpt",
    "ext/standard/tests/file/stat_variation2-win32.phpt",
    "ext/standard/tests/file/lchown_error.phpt",
    "ext/standard/tests/file/fscanf_variation5.phpt",
    "ext/standard/tests/file/bug39863.phpt",
    "ext/standard/tests/file/copy_variation17.phpt",
    "ext/standard/tests/file/fstat_basic.phpt",
    "ext/standard/tests/file/fscanf_variation27.phpt",
    "ext/standard/tests/file/include_userstream_003.phpt",
    "ext/standard/tests/file/lstat_stat_variation14.phpt",
    "ext/standard/tests/file/006_variation1.phpt",
    "ext/standard/tests/file/lstat_stat_variation20.phpt",
    "ext/standard/tests/file/copy_variation2.phpt",
    "ext/standard/tests/file/fgetcsv_variation6.phpt",
    "ext/standard/tests/file/pathinfo_variation2.phpt",
    "ext/standard/tests/file/fgetcsv_variation1.phpt",
    "ext/standard/tests/file/007_variation17.phpt",
    "ext/standard/tests/file/bug43248.phpt",
    "ext/standard/tests/file/007_variation2.phpt",
    "ext/standard/tests/file/fscanf_variation28.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_variation4-win32.phpt",
    "ext/standard/tests/file/fread_variation1.phpt",
    "ext/standard/tests/file/is_executable_error.phpt",
    "ext/standard/tests/file/fscanf_variation12.phpt",
    "ext/standard/tests/file/realpath_basic4.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_variation5.phpt",
    "ext/standard/tests/file/fileowner_variation1.phpt",
    "ext/standard/tests/file/fpassthru_error.phpt",
    "ext/standard/tests/file/php_fd_wrapper_03.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_variation2.phpt",
    "ext/standard/tests/file/readfile_variation6.phpt",
    "ext/standard/tests/file/lstat_stat_variation4.phpt",
    "ext/standard/tests/file/copy_variation2-win32.phpt",
    "ext/standard/tests/file/pathinfo_variaton.phpt",
    "ext/standard/tests/file/fopen_variation10-win32.phpt",
    "ext/standard/tests/file/filesize_variation4.phpt",
    "ext/standard/tests/file/userstreams_002.phpt",
    "ext/standard/tests/file/rename_variation8.phpt",
    "ext/standard/tests/file/file_variation3.phpt",
    "ext/standard/tests/file/windows_links/bug48746_2.phpt",
    "ext/standard/tests/file/windows_links/bug48746_1.phpt",
    "ext/standard/tests/file/windows_links/bug48746.phpt",
    "ext/standard/tests/file/windows_links/bug48746_3.phpt",
    "ext/standard/tests/file/glob_variation.phpt",
    "ext/standard/tests/file/pathinfo_variation3.phpt",
    "ext/standard/tests/file/fileinode_variation.phpt",
    "ext/standard/tests/file/file_variation.phpt",
    "ext/standard/tests/file/fscanf_variation38.phpt",
    "ext/standard/tests/file/fpassthru_basic.phpt",
    "ext/standard/tests/file/realpath_cache.phpt",
    "ext/standard/tests/file/stat_variation5-win32.phpt",
    "ext/standard/tests/file/fopen_variation4.phpt",
    "ext/standard/tests/file/fopen_variation12.phpt",
    "ext/standard/tests/file/fscanf_variation39.phpt",
    "ext/standard/tests/file/fgetcsv_variation17.phpt",
    "ext/standard/tests/file/copy_variation12-win32.phpt",
    "ext/standard/tests/file/umask_variation2.phpt",
    "ext/standard/tests/file/rmdir_variation3-win32.phpt",
    "ext/standard/tests/file/pathinfo_basic.phpt",
    "ext/standard/tests/file/bug26615.phpt",
    "ext/standard/tests/file/touch_variation5.phpt",
    "ext/standard/tests/file/lstat_stat_variation5.phpt",
    "ext/standard/tests/file/007_variation10.phpt",
    "ext/standard/tests/file/fscanf_variation48.phpt",
    "ext/standard/tests/file/pathinfo_basic2-win32.phpt",
    "ext/standard/tests/file/fflush_error.phpt",
    "ext/standard/tests/file/fgets_variation6.phpt",
    "ext/standard/tests/file/filetype_error.phpt",
    "ext/standard/tests/file/fopen_variation17.phpt",
    "ext/standard/tests/file/is_writable_basic.phpt",
    "ext/standard/tests/file/is_executable_basic-win32.phpt",
    "ext/standard/tests/file/basename_error.phpt",
    "ext/standard/tests/file/unlink_error.phpt",
    "ext/standard/tests/file/filesize_variation3.phpt",
    "ext/standard/tests/file/is_readable_variation2.phpt",
    "ext/standard/tests/file/bug53241.phpt",
    "ext/standard/tests/file/lchgrp_basic.phpt",
    "ext/standard/tests/file/fpassthru_variation.phpt",
    "ext/standard/tests/file/lstat_stat_variation15.phpt",
    "ext/standard/tests/file/stream_supports_lock.phpt",
    "ext/standard/tests/file/symlink_link_linkinfo_is_link_variation1.phpt",
    "ext/standard/tests/file/bug43008.phpt",
    "ext/standard/tests/file/lstat_stat_variation16.phpt",
    "ext/standard/tests/file/is_writable_variation3.phpt",
    "ext/standard/tests/file/clearstatcache_error.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_variation7.phpt",
    "ext/standard/tests/file/bug41874_3.phpt",
    "ext/standard/tests/file/fgetcsv_variation28.phpt",
    "ext/standard/tests/file/file_get_contents_variation9.phpt",
    "ext/standard/tests/file/rename_variation4.phpt",
    "ext/standard/tests/file/realpath_error-win32.phpt",
    "ext/standard/tests/file/disk_total_space_basic.phpt",
    "ext/standard/tests/file/bug38450_3.phpt",
    "ext/standard/tests/file/fwrite_variation1-win32.phpt",
    "ext/standard/tests/file/userwrapper.phpt",
    "ext/standard/tests/file/unlink_variation4.phpt",
    "ext/standard/tests/file/ftruncate_variation2.phpt",
    "ext/standard/tests/file/rmdir_variation1.phpt",
    "ext/standard/tests/file/realpath_basic3.phpt",
    "ext/standard/tests/file/file_get_contents_basic.phpt",
    "ext/standard/tests/file/fgets_basic.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_error2.phpt",
    "ext/standard/tests/file/fstat_variation4.phpt",
    "ext/standard/tests/file/007_variation15.phpt",
    "ext/standard/tests/file/rename_variation6-win32.phpt",
    "ext/standard/tests/file/file_variation9.phpt",
    "ext/standard/tests/file/fgets_variation1.phpt",
    "ext/standard/tests/file/tempnam_variation6.phpt",
    "ext/standard/tests/file/readlink_variation1.phpt",
    "ext/standard/tests/file/007_variation9.phpt",
    "ext/standard/tests/file/popen_pclose_basic-win32.phpt",
    "ext/standard/tests/file/007_variation3.phpt",
    "ext/standard/tests/file/fputcsv_variation12.phpt",
    "ext/standard/tests/file/fnmatch_variation.phpt",
    "ext/standard/tests/file/is_dir_variation3.phpt",
    "ext/standard/tests/file/fileinode_variation3.phpt",
    "ext/standard/tests/file/readfile_error.phpt",
    "ext/standard/tests/file/userfilters.phpt",
    "ext/standard/tests/file/tempnam_variation2.phpt",
    "ext/standard/tests/file/fputcsv_variation13.phpt",
    "ext/standard/tests/file/fscanf_variation9.phpt",
    "ext/standard/tests/file/bug27508.phpt",
    "ext/standard/tests/file/flock.phpt",
    "ext/standard/tests/file/fseek_variation3.phpt",
    "ext/standard/tests/file/tempnam_variation6-win32.phpt",
    "ext/standard/tests/file/touch_variation4.phpt",
    "ext/standard/tests/file/file_get_contents_file_put_contents_error.phpt",
    "ext/standard/tests/file/stat_variation7-win32.phpt",
    "ext/standard/tests/file/007_variation19.phpt",
    "ext/standard/tests/file/fscanf_error.phpt",
    "ext/standard/tests/file/fputcsv_variation11.phpt",
    "ext/standard/tests/file/fscanf_variation40.phpt",
    "ext/standard/tests/file/rename_variation9.phpt",
    "ext/standard/tests/file/disk_total_space_variation.phpt",
    "ext/standard/tests/file/fscanf_variation52.phpt",
    "ext/standard/tests/file/filesize_error.phpt",
    "ext/standard/tests/file/stream_rfc2397_004.phpt",
    "ext/standard/tests/file/glob_variation-win32.phpt",
    "ext/standard/tests/file/is_dir_variation2.phpt",
    "ext/standard/tests/file/is_dir_variation1.phpt",
    "ext/standard/tests/file/bug44805.phpt",
    "ext/standard/tests/file/parse_ini_file.phpt",
    "ext/standard/tests/file/userstreams_003.phpt",
    "ext/standard/tests/file/symlink_link_linkinfo_is_link_variation6.phpt",
    "ext/standard/tests/file/tempnam_variation4.phpt",
    "ext/standard/tests/file/basename.phpt",
    "ext/standard/tests/file/userstreams.phpt",
    "ext/standard/tests/file/ftruncate_variation4.phpt",
    "ext/standard/tests/file/fstat_variation7.phpt",
    "ext/standard/tests/file/fwrite_variation4.phpt",
    "ext/standard/tests/file/basename_variation1.phpt",
    "ext/standard/tests/file/fgetcsv_variation24.phpt",
    "ext/standard/tests/file/parse_ini_file_variation3.phpt",
    "ext/standard/tests/file/basename_basic.phpt",
    "ext/standard/tests/file/fgetcsv_variation20.phpt",
    "ext/standard/tests/file/tempnam_variation7-win32.phpt",
    "ext/standard/tests/file/mkdir-005.phpt",
    "ext/standard/tests/file/fgets_variation4.phpt",
    "ext/standard/tests/file/007_variation12.phpt",
    "ext/standard/tests/file/bug26003.phpt",
    "ext/standard/tests/file/fseek_variation1.phpt",
    "ext/standard/tests/file/fgets_variation4-win32.phpt",
    "ext/standard/tests/file/005_variation.phpt",
    "ext/standard/tests/file/fgets_socket_variation1.phpt",
    "ext/standard/tests/file/chmod_variation4.phpt",
    "ext/standard/tests/file/file_exists_error.phpt",
    "ext/standard/tests/file/bug45985.phpt",
    "ext/standard/tests/file/mkdir_variation5.phpt",
    "ext/standard/tests/file/fseek_ftell_rewind_variation3.phpt",
    "ext/standard/tests/array/array_walk_recursive_object1.phpt",
    "ext/standard/tests/array/bug44181.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation1.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation5.phpt",
    "ext/standard/tests/array/array_fill_keys_variation3.phpt",
    "ext/standard/tests/array/asort_variation2.phpt",
    "ext/standard/tests/array/array_keys_variation_002.phpt",
    "ext/standard/tests/array/arsort_variation4.phpt",
    "ext/standard/tests/array/array_shift_variation4.phpt",
    "ext/standard/tests/array/array_product_variation3.phpt",
    "ext/standard/tests/array/usort_variation9.phpt",
    "ext/standard/tests/array/array_fill_keys.phpt",
    "ext/standard/tests/array/ksort_variation7.phpt",
    "ext/standard/tests/array/array_change_key_case_errors.phpt",
    "ext/standard/tests/array/bug36975.phpt",
    "ext/standard/tests/array/array_unique_variation7.phpt",
    "ext/standard/tests/array/array_pad_variation1.phpt",
    "ext/standard/tests/array/array_push_variation1.phpt",
    "ext/standard/tests/array/array_slice_variation6.phpt",
    "ext/standard/tests/array/natcasesort_variation2.phpt",
    "ext/standard/tests/array/asort_variation6.phpt",
    "ext/standard/tests/array/arsort_object2.phpt",
    "ext/standard/tests/array/natcasesort_error.phpt",
    "ext/standard/tests/array/array_diff_key_variation4.phpt",
    "ext/standard/tests/array/array_rand_variation3.phpt",
    "ext/standard/tests/array/array_key_exists_variation6.phpt",
    "ext/standard/tests/array/sort_variation11.phpt",
    "ext/standard/tests/array/array_pad_variation3.phpt",
    "ext/standard/tests/array/array_multisort_error.phpt",
    "ext/standard/tests/array/array_search_variation1.phpt",
    "ext/standard/tests/array/array_intersect_assoc_variation7.phpt",
    "ext/standard/tests/array/array_uintersect_variation4.phpt",
    "ext/standard/tests/array/array_key_exists_object1.phpt",
    "ext/standard/tests/array/array_unshift_variation5.phpt",
    "ext/standard/tests/array/array_shift_error.phpt",
    "ext/standard/tests/array/array_merge_recursive_variation7.phpt",
    "ext/standard/tests/array/array_uintersect_assoc_variation4.phpt",
    "ext/standard/tests/array/key_exists_variation1.phpt",
    "ext/standard/tests/array/array_combine_error2.phpt",
    "ext/standard/tests/array/reset_error.phpt",
    "ext/standard/tests/array/array_map_001.phpt",
    "ext/standard/tests/array/krsort_variation4.phpt",
    "ext/standard/tests/array/bug24897.phpt",
    "ext/standard/tests/array/asort_variation11.phpt",
    "ext/standard/tests/array/array_flip_variation3.phpt",
    "ext/standard/tests/array/max_error.phpt",
    "ext/standard/tests/array/array_filter_variation7.phpt",
    "ext/standard/tests/array/bug39576.phpt",
    "ext/standard/tests/array/bug32021.phpt",
    "ext/standard/tests/array/array_walk_recursive_object2.phpt",
    "ext/standard/tests/array/005.phpt",
    "ext/standard/tests/array/array_unique_variation4.phpt",
    "ext/standard/tests/array/asort_object1.phpt",
    "ext/standard/tests/array/array_chunk.phpt",
    "ext/standard/tests/array/array_unshift_variation8.phpt",
    "ext/standard/tests/array/array_diff_key_variation6.phpt",
    "ext/standard/tests/array/extract_variation9.phpt",
    "ext/standard/tests/array/array_filter_variation4.phpt",
    "ext/standard/tests/array/array_intersect_assoc_error.phpt",
    "ext/standard/tests/array/min_error.phpt",
    "ext/standard/tests/array/bug22463.phpt",
    "ext/standard/tests/array/array_chunk_variation17.phpt",
    "ext/standard/tests/array/array_splice_basic.phpt",
    "ext/standard/tests/array/array_multisort_variation8.phpt",
    "ext/standard/tests/array/key_variation1.phpt",
    "ext/standard/tests/array/array_map_variation19.phpt",
    "ext/standard/tests/array/bug33989.phpt",
    "ext/standard/tests/array/array_key_exists_variation8.phpt",
    "ext/standard/tests/array/ksort_variation11.phpt",
    "ext/standard/tests/array/bug41121.phpt",
    "ext/standard/tests/array/array_walk_recursive_variation8.phpt",
    "ext/standard/tests/array/uasort_variation10.phpt",
    "ext/standard/tests/array/array_intersect_assoc_variation10.phpt",
    "ext/standard/tests/array/ksort_variation4.phpt",
    "ext/standard/tests/array/array_chunk_variation23.phpt",
    "ext/standard/tests/array/array_diff_key_variation3.phpt",
    "ext/standard/tests/array/sizeof_variation5.phpt",
    "ext/standard/tests/array/array_walk_recursive_basic1.phpt",
    "ext/standard/tests/array/array_intersect_assoc_basic.phpt",
    "ext/standard/tests/array/array_chunk_variation22.phpt",
    "ext/standard/tests/array/ksort_variation8.phpt",
    "ext/standard/tests/array/bug25758.phpt",
    "ext/standard/tests/array/asort_basic.phpt",
    "ext/standard/tests/array/ksort_variation1.phpt",
    "ext/standard/tests/array/krsort_variation1.phpt",
    "ext/standard/tests/array/asort_error.phpt",
    "ext/standard/tests/array/array_fill_variation2.phpt",
    "ext/standard/tests/array/array_values_variation2.phpt",
    "ext/standard/tests/array/rsort_variation1.phpt",
    "ext/standard/tests/array/array_sum_variation4.phpt",
    "ext/standard/tests/array/rsort_basic.phpt",
    "ext/standard/tests/array/array_walk_recursive_error2.phpt",
    "ext/standard/tests/array/array_user_key_compare.phpt",
    "ext/standard/tests/array/natcasesort_variation11.phpt",
    "ext/standard/tests/array/prev_error2.phpt",
    "ext/standard/tests/array/array_values_variation7.phpt",
    "ext/standard/tests/array/array_splice_errors.phpt",
    "ext/standard/tests/array/var_export.phpt",
    "ext/standard/tests/array/extract_variation1.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation7.phpt",
    "ext/standard/tests/array/compact_variation2.phpt",
    "ext/standard/tests/array/array_sum_variation6.phpt",
    "ext/standard/tests/array/array_key_exists_variation1.phpt",
    "ext/standard/tests/array/asort_variation1.phpt",
    "ext/standard/tests/array/min_basiclong_64bit.phpt",
    "ext/standard/tests/array/array_intersect_key_variation3.phpt",
    "ext/standard/tests/array/current_variation3.phpt",
    "ext/standard/tests/array/array_push_basic.phpt",
    "ext/standard/tests/array/natcasesort_variation3.phpt",
    "ext/standard/tests/array/bug35022.phpt",
    "ext/standard/tests/array/array_map_variation3.phpt",
    "ext/standard/tests/array/shuffle_variation2.phpt",
    "ext/standard/tests/array/array_chunk_error.phpt",
    "ext/standard/tests/array/bug50006.phpt",
    "ext/standard/tests/array/array_pad_variation5.phpt",
    "ext/standard/tests/array/bug21182.phpt",
    "ext/standard/tests/array/array_walk_recursive_variation5.phpt",
    "ext/standard/tests/array/array_reduce_error.phpt",
    "ext/standard/tests/array/array_slice_variation4.phpt",
    "ext/standard/tests/array/end_error.phpt",
    "ext/standard/tests/array/array_udiff_uassoc_variation3.phpt",
    "ext/standard/tests/array/array_diff_assoc_variation8.phpt",
    "ext/standard/tests/array/key_basic.phpt",
    "ext/standard/tests/array/array_reverse_variation1.phpt",
    "ext/standard/tests/array/bug31158.phpt",
    "ext/standard/tests/array/array_intersect_assoc_variation4.phpt",
    "ext/standard/tests/array/array_map_object1.phpt",
    "ext/standard/tests/array/array_unique_variation5.phpt",
    "ext/standard/tests/array/key_variation4.phpt",
    "ext/standard/tests/array/array_diff_key.phpt",
    "ext/standard/tests/array/array_filter_variation9.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation8.phpt",
    "ext/standard/tests/array/shuffle_basic1.phpt",
    "ext/standard/tests/array/array_intersect_1.phpt",
    "ext/standard/tests/array/uasort_variation5.phpt",
    "ext/standard/tests/array/array_chunk_variation32.phpt",
    "ext/standard/tests/array/arsort_variation2.phpt",
    "ext/standard/tests/array/key_error.phpt",
    "ext/standard/tests/array/array_filter_object.phpt",
    "ext/standard/tests/array/array_keys_variation_004.phpt",
    "ext/standard/tests/array/array_pad_variation6.phpt",
    "ext/standard/tests/array/asort_object2.phpt",
    "ext/standard/tests/array/compact.phpt",
    "ext/standard/tests/array/bug42177.phpt",
    "ext/standard/tests/array/array_unshift_variation1.phpt",
    "ext/standard/tests/array/rsort_variation10.phpt",
    "ext/standard/tests/array/array_diff_key_variation8.phpt",
    "ext/standard/tests/array/array_values_variation1.phpt",
    "ext/standard/tests/array/array_walk.phpt",
    "ext/standard/tests/array/array_key_exists_variation7.phpt",
    "ext/standard/tests/array/array_splice_variation2.phpt",
    "ext/standard/tests/array/array_merge_variation1.phpt",
    "ext/standard/tests/array/sort_variation2.phpt",
    "ext/standard/tests/array/asort_variation5.phpt",
    "ext/standard/tests/array/sort_object1.phpt",
    "ext/standard/tests/array/bug34982.phpt",
    "ext/standard/tests/array/array_sum_variation7.phpt",
    "ext/standard/tests/array/array_udiff_uassoc_variation2.phpt",
    "ext/standard/tests/array/rsort_variation7.phpt",
    "ext/standard/tests/array/array_combine_variation2.phpt",
    "ext/standard/tests/array/array_intersect_assoc_variation2.phpt",
    "ext/standard/tests/array/array_intersect_key_error.phpt",
    "ext/standard/tests/array/array_map_variation2.phpt",
    "ext/standard/tests/array/array_chunk_variation6.phpt",
    "ext/standard/tests/array/array_udiff_uassoc_variation6.phpt",
    "ext/standard/tests/array/array_map_variation8.phpt",
    "ext/standard/tests/array/array_walk_rec_objects.phpt",
    "ext/standard/tests/array/array_slice_variation11.phpt",
    "ext/standard/tests/array/ksort_basic.phpt",
    "ext/standard/tests/array/usort_variation1.phpt",
    "ext/standard/tests/array/array_pad_variation7.phpt",
    "ext/standard/tests/array/array_key_exists_variation2.phpt",
    "ext/standard/tests/array/array_intersect_variation6.phpt",
    "ext/standard/tests/array/array_walk_recursive_variation9.phpt",
    "ext/standard/tests/array/array_chunk_variation25.phpt",
    "ext/standard/tests/array/array_keys_variation_001.phpt",
    "ext/standard/tests/array/bug24220.phpt",
    "ext/standard/tests/array/array_diff_variation5.phpt",
    "ext/standard/tests/array/array_values.phpt",
    "ext/standard/tests/array/array_udiff_uassoc_error.phpt",
    "ext/standard/tests/array/bug52534.phpt",
    "ext/standard/tests/array/array_merge_recursive_basic2.phpt",
    "ext/standard/tests/array/bug12776.phpt",
    "ext/standard/tests/array/array_chunk_variation26.phpt",
    "ext/standard/tests/array/array_udiff_assoc_variation2.phpt",
    "ext/standard/tests/array/arsort_variation5.phpt",
    "ext/standard/tests/array/array_values_variation5.phpt",
    "ext/standard/tests/array/asort_variation10.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation13.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation12.phpt",
    "ext/standard/tests/array/array_unique_variation1.phpt",
    "ext/standard/tests/array/array_values_errors.phpt",
    "ext/standard/tests/array/array_multisort_variation7.phpt",
    "ext/standard/tests/array/array_map_variation13.phpt",
    "ext/standard/tests/array/array_fill_variation3.phpt",
    "ext/standard/tests/array/array_rand_variation6.phpt",
    "ext/standard/tests/array/array_values_variation4.phpt",
    "ext/standard/tests/array/array_pop_errors.phpt",
    "ext/standard/tests/array/usort_object2.phpt",
    "ext/standard/tests/array/array_count_values.phpt",
    "ext/standard/tests/array/array_splice_variation1.phpt",
    "ext/standard/tests/array/array_flip_basic.phpt",
    "ext/standard/tests/array/array_uintersect_assoc_variation2.phpt",
    "ext/standard/tests/array/array_chunk_variation15.phpt",
    "ext/standard/tests/array/array_unshift_basic1.phpt",
    "ext/standard/tests/array/bug40709.phpt",
    "ext/standard/tests/array/array_chunk_variation18.phpt",
    "ext/standard/tests/array/array_intersect_ukey_variation1.phpt",
    "ext/standard/tests/array/bug44929.phpt",
    "ext/standard/tests/array/array_diff_assoc_variation10.phpt",
    "ext/standard/tests/array/001.phpt",
    "ext/standard/tests/array/bug33382.phpt",
    "ext/standard/tests/array/array_intersect_ukey_variation4.phpt",
    "ext/standard/tests/array/max_basiclong_64bit.phpt",
    "ext/standard/tests/array/array_diff_ukey_variation2.phpt",
    "ext/standard/tests/array/array_intersect_uassoc_variation4.phpt",
    "ext/standard/tests/array/array_diff_variation9.phpt",
    "ext/standard/tests/array/compact_basic.phpt",
    "ext/standard/tests/array/array_walk_variation1.phpt",
    "ext/standard/tests/array/krsort_variation5.phpt",
    "ext/standard/tests/array/reset_variation3.phpt",
    "ext/standard/tests/array/array_map_basic.phpt",
    "ext/standard/tests/array/array_product_variation4.phpt",
    "ext/standard/tests/array/sort_variation10.phpt",
    "ext/standard/tests/array/array_slice_variation9.phpt",
    "ext/standard/tests/array/array_count_values_variation.phpt",
    "ext/standard/tests/array/current_basic.phpt",
    "ext/standard/tests/array/extract_variation2.phpt",
    "ext/standard/tests/array/array_merge_basic.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation2.phpt",
    "ext/standard/tests/array/array_merge_recursive_variation3.phpt",
    "ext/standard/tests/array/array_udiff_uassoc_variation5.phpt",
    "ext/standard/tests/array/array_chunk_variation8.phpt",
    "ext/standard/tests/array/array_intersect_uassoc_variation2.phpt",
    "ext/standard/tests/array/array_merge_variation6.phpt",
    "ext/standard/tests/array/array_udiff_uassoc_basic.phpt",
    "ext/standard/tests/array/array_diff_ukey_variation3.phpt",
    "ext/standard/tests/array/array_merge_variation7.phpt",
    "ext/standard/tests/array/array_reduce_variation2.phpt",
    "ext/standard/tests/array/array_splice_variation3.phpt",
    "ext/standard/tests/array/array_walk_error2.phpt",
    "ext/standard/tests/array/004.phpt",
    "ext/standard/tests/array/bug34227.phpt",
    "ext/standard/tests/array/shuffle_basic2.phpt",
    "ext/standard/tests/array/array_diff_ukey_variation8.phpt",
    "ext/standard/tests/array/array_keys_variation_002_64bit.phpt",
    "ext/standard/tests/array/in_array_variation4.phpt",
    "ext/standard/tests/array/array_slice_basic.phpt",
    "ext/standard/tests/array/reset_basic.phpt",
    "ext/standard/tests/array/array_udiff_basic.phpt",
    "ext/standard/tests/array/asort_variation3.phpt",
    "ext/standard/tests/array/array_merge.phpt",
    "ext/standard/tests/array/array_fill_variation1_64bit.phpt",
    "ext/standard/tests/array/array_walk_object1.phpt",
    "ext/standard/tests/array/array_diff_variation1.phpt",
    "ext/standard/tests/array/array_diff_key_variation5.phpt",
    "ext/standard/tests/array/bug35014_64bit.phpt",
    "ext/standard/tests/array/array_merge_recursive_variation6.phpt",
    "ext/standard/tests/array/sizeof_variation4.phpt",
    "ext/standard/tests/array/array_uintersect_error.phpt",
    "ext/standard/tests/array/natcasesort_object1.phpt",
    "ext/standard/tests/array/natcasesort_variation8.phpt",
    "ext/standard/tests/array/array_walk_recursive1.phpt",
    "ext/standard/tests/array/array_walk_variation3.phpt",
    "ext/standard/tests/array/array_intersect_ukey_variation6.phpt",
    "ext/standard/tests/array/reset_variation2.phpt",
    "ext/standard/tests/array/array_reverse_variation4.phpt",
    "ext/standard/tests/array/array_replace.phpt",
    "ext/standard/tests/array/each_variation1.phpt",
    "ext/standard/tests/array/var_export3.phpt",
    "ext/standard/tests/array/array_values_variation6.phpt",
    "ext/standard/tests/array/end_variation3.phpt",
    "ext/standard/tests/array/array_reduce.phpt",
    "ext/standard/tests/array/extract_error.phpt",
    "ext/standard/tests/array/array_map_variation5.phpt",
    "ext/standard/tests/array/arsort_variation9.phpt",
    "ext/standard/tests/array/count_basic.phpt",
    "ext/standard/tests/array/array_intersect_ukey_variation2.phpt",
    "ext/standard/tests/array/array_merge_recursive_variation10.phpt",
    "ext/standard/tests/array/array_flip_variation4.phpt",
    "ext/standard/tests/array/array_intersect_variation2.phpt",
    "ext/standard/tests/array/array_intersect_basic.phpt",
    "ext/standard/tests/array/usort_variation2.phpt",
    "ext/standard/tests/array/array_map_variation12.phpt",
    "ext/standard/tests/array/bug35014.phpt",
    "ext/standard/tests/array/arsort_variation1.phpt",
    "ext/standard/tests/array/range_variation.phpt",
    "ext/standard/tests/array/each_variation2.phpt",
    "ext/standard/tests/array/array_rand_variation4.phpt",
    "ext/standard/tests/array/array_fill_variation1.phpt",
    "ext/standard/tests/array/bug20865.phpt",
    "ext/standard/tests/array/bug23581.phpt",
    "ext/standard/tests/array/ksort_variation3.phpt",
    "ext/standard/tests/array/array_udiff_uassoc_variation1.phpt",
    "ext/standard/tests/array/array_intersect_key_variation5.phpt",
    "ext/standard/tests/array/extract_safety.phpt",
    "ext/standard/tests/array/array_diff_variation4.phpt",
    "ext/standard/tests/array/shuffle_variation1.phpt",
    "ext/standard/tests/array/key_exists_basic.phpt",
    "ext/standard/tests/array/arsort_variation8.phpt",
    "ext/standard/tests/array/arsort_variation6.phpt",
    "ext/standard/tests/array/array_unshift_variation3.phpt",
    "ext/standard/tests/array/array_intersect_uassoc_error.phpt",
    "ext/standard/tests/array/rsort_variation6.phpt",
    "ext/standard/tests/array/array_udiff_variation2.phpt",
    "ext/standard/tests/array/array_chunk_variation1.phpt",
    "ext/standard/tests/array/ksort_variation6.phpt",
    "ext/standard/tests/array/array_unique_variation2.phpt",
    "ext/standard/tests/array/array_intersect_uassoc_basic.phpt",
    "ext/standard/tests/array/array_map_variation10.phpt",
    "ext/standard/tests/array/bug21918.phpt",
    "ext/standard/tests/array/krsort_variation9.phpt",
    "ext/standard/tests/array/array_unique_basic.phpt",
    "ext/standard/tests/array/extract_variation6.phpt",
    "ext/standard/tests/array/array_map_object2.phpt",
    "ext/standard/tests/array/array_rand_variation2.phpt",
    "ext/standard/tests/array/max_variation1.phpt",
    "ext/standard/tests/array/array_multisort_basic2.phpt",
    "ext/standard/tests/array/array_diff_basic.phpt",
    "ext/standard/tests/array/array_reverse_variation2.phpt",
    "ext/standard/tests/array/array_sum_variation3.phpt",
    "ext/standard/tests/array/extract_variation8.phpt",
    "ext/standard/tests/array/array_chunk_variation2.phpt",
    "ext/standard/tests/array/array_filter_basic.phpt",
    "ext/standard/tests/array/array_fill_keys_variation1.phpt",
    "ext/standard/tests/array/bug54459.phpt",
    "ext/standard/tests/array/var_export2.phpt",
    "ext/standard/tests/array/array_push.phpt",
    "ext/standard/tests/array/array_merge_recursive_variation9.phpt",
    "ext/standard/tests/array/array_shift_variation3.phpt",
    "ext/standard/tests/array/array_walk_object2.phpt",
    "ext/standard/tests/array/array_uintersect_assoc_error.phpt",
    "ext/standard/tests/array/003.phpt",
    "ext/standard/tests/array/extract_variation4.phpt",
    "ext/standard/tests/array/array_combine_variation3.phpt",
    "ext/standard/tests/array/array_map_variation18.phpt",
    "ext/standard/tests/array/array_udiff_error.phpt",
    "ext/standard/tests/array/array_diff_assoc_variation6.phpt",
    "ext/standard/tests/array/array_map_variation17.phpt",
    "ext/standard/tests/array/array_intersect_key_variation7.phpt",
    "ext/standard/tests/array/array_intersect_key_basic.phpt",
    "ext/standard/tests/array/array_flip_variation2.phpt",
    "ext/standard/tests/array/sizeof_variation3.phpt",
    "ext/standard/tests/array/array_intersect_ukey_variation7.phpt",
    "ext/standard/tests/array/array_shuffle_basic.phpt",
    "ext/standard/tests/array/bug22088.phpt",
    "ext/standard/tests/array/array_search1.phpt",
    "ext/standard/tests/array/array_uintersect_uassoc_variation2.phpt",
    "ext/standard/tests/array/rsort_error.phpt",
    "ext/standard/tests/array/bug24980.phpt",
    "ext/standard/tests/array/array_filter_error.phpt",
    "ext/standard/tests/array/sort_basic.phpt",
    "ext/standard/tests/array/array_merge_variation9.phpt",
    "ext/standard/tests/array/array_reduce_variation1.phpt",
    "ext/standard/tests/array/array_diff_key_variation2.phpt",
    "ext/standard/tests/array/array_fill_object.phpt",
    "ext/standard/tests/array/array_uintersect_uassoc_variation6.phpt",
    "ext/standard/tests/array/asort_variation7.phpt",
    "ext/standard/tests/array/krsort_object.phpt",
    "ext/standard/tests/array/array_uintersect_uassoc_variation4.phpt",
    "ext/standard/tests/array/array_merge_recursive_error.phpt",
    "ext/standard/tests/array/array_unshift.phpt",
    "ext/standard/tests/array/array_udiff_variation5.phpt",
    "ext/standard/tests/array/array_uintersect_assoc_variation5.phpt",
    "ext/standard/tests/array/uasort_variation1.phpt",
    "ext/standard/tests/array/array_filter_variation5.phpt",
    "ext/standard/tests/array/bug29493.phpt",
    "ext/standard/tests/array/in_array_variation3.phpt",
    "ext/standard/tests/array/array_slice_variation1.phpt",
    "ext/standard/tests/array/array_intersect_assoc_variation1.phpt",
    "ext/standard/tests/array/array_intersect_key_variation2.phpt",
    "ext/standard/tests/array/usort_error2.phpt",
    "ext/standard/tests/array/array_uintersect_assoc_basic.phpt",
    "ext/standard/tests/array/max_variation2.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation9.phpt",
    "ext/standard/tests/array/array_chunk_variation7.phpt",
    "ext/standard/tests/array/usort_variation8.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation11.phpt",
    "ext/standard/tests/array/array_intersect_uassoc_variation8.phpt",
    "ext/standard/tests/array/krsort_variation8.phpt",
    "ext/standard/tests/array/bug35821.phpt",
    "ext/standard/tests/array/array_reduce_variation3.phpt",
    "ext/standard/tests/array/arsort_variation11.phpt",
    "ext/standard/tests/array/array_pad_variation2.phpt",
    "ext/standard/tests/array/array_fill_error.phpt",
    "ext/standard/tests/array/array_splice_variation4.phpt",
    "ext/standard/tests/array/array_chunk_variation9.phpt",
    "ext/standard/tests/array/array_uintersect_basic.phpt",
    "ext/standard/tests/array/array_intersect_uassoc_variation1.phpt",
    "ext/standard/tests/array/array_walk_error1.phpt",
    "ext/standard/tests/array/end_basic.phpt",
    "ext/standard/tests/array/array_chunk_variation24.phpt",
    "ext/standard/tests/array/shuffle_variation5.phpt",
    "ext/standard/tests/array/bug50006_2.phpt",
    "ext/standard/tests/array/array_udiff_assoc_error.phpt",
    "ext/standard/tests/array/key_exists_variation2.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation3.phpt",
    "ext/standard/tests/array/count_error.phpt",
    "ext/standard/tests/array/006.phpt",
    "ext/standard/tests/array/array_slice_variation10.phpt",
    "ext/standard/tests/array/prev_variation1.phpt",
    "ext/standard/tests/array/array_walk_basic1.phpt",
    "ext/standard/tests/array/array_intersect_variation7.phpt",
    "ext/standard/tests/array/shuffle_variation4.phpt",
    "ext/standard/tests/array/array_merge_recursive_variation2.phpt",
    "ext/standard/tests/array/bug51552.phpt",
    "ext/standard/tests/array/array_intersect_uassoc_variation11.phpt",
    "ext/standard/tests/array/array_udiff_variation4.phpt",
    "ext/standard/tests/array/array_intersect_uassoc_variation9.phpt",
    "ext/standard/tests/array/array_diff_variation6.phpt",
    "ext/standard/tests/array/count_variation3.phpt",
    "ext/standard/tests/array/arsort_object1.phpt",
    "ext/standard/tests/array/array_walk_basic2.phpt",
    "ext/standard/tests/array/sizeof_error.phpt",
    "ext/standard/tests/array/array_fill.phpt",
    "ext/standard/tests/array/array_reverse_basic1.phpt",
    "ext/standard/tests/array/array_walk_objects.phpt",
    "ext/standard/tests/array/array_unique_variation6.phpt",
    "ext/standard/tests/array/krsort_variation3.phpt",
    "ext/standard/tests/array/ksort_variation9.phpt",
    "ext/standard/tests/array/array_slice_variation8.phpt",
    "ext/standard/tests/array/ksort_object.phpt",
    "ext/standard/tests/array/array_count_values_error.phpt",
    "ext/standard/tests/array/array_rand_variation5.phpt",
    "ext/standard/tests/array/krsort_variation2.phpt",
    "ext/standard/tests/array/in_array_variation2.phpt",
    "ext/standard/tests/array/array_chunk_variation28.phpt",
    "ext/standard/tests/array/arsort_basic.phpt",
    "ext/standard/tests/array/array_reverse_variation6.phpt",
    "ext/standard/tests/array/bug24766.phpt",
    "ext/standard/tests/array/array_push_variation3.phpt",
    "ext/standard/tests/array/bug14580.phpt",
    "ext/standard/tests/array/array_map_variation4.phpt",
    "ext/standard/tests/array/bug48484.phpt",
    "ext/standard/tests/array/current_variation4.phpt",
    "ext/standard/tests/array/array_pad_error.phpt",
    "ext/standard/tests/array/array_pop.phpt",
    "ext/standard/tests/array/array_push_variation6.phpt",
    "ext/standard/tests/array/extract_variation11.phpt",
    "ext/standard/tests/array/extract_variation5.phpt",
    "ext/standard/tests/array/bug43541.phpt",
    "ext/standard/tests/array/array_diff_assoc_basic.phpt",
    "ext/standard/tests/array/array_uintersect_variation2.phpt",
    "ext/standard/tests/array/array_filter_variation8.phpt",
    "ext/standard/tests/array/ksort_variation10.phpt",
    "ext/standard/tests/array/min_basic.phpt",
    "ext/standard/tests/array/array_unique_variation3.phpt",
    "ext/standard/tests/array/range_variation1.phpt",
    "ext/standard/tests/array/extract_variation7.phpt",
    "ext/standard/tests/array/sort_variation4.phpt",
    "ext/standard/tests/array/sort_variation1.phpt",
    "ext/standard/tests/array/end_64bit.phpt",
    "ext/standard/tests/array/extract_variation10.phpt",
    "ext/standard/tests/array/uasort_variation9.phpt",
    "ext/standard/tests/array/array_merge_recursive_variation1.phpt",
    "ext/standard/tests/array/array_udiff_uassoc_variation4.phpt",
    "ext/standard/tests/array/array_change_key_case_variation4.phpt",
    "ext/standard/tests/array/bug29253.phpt",
    "ext/standard/tests/array/array_slice_variation2.phpt",
    "ext/standard/tests/array/bug46873.phpt",
    "ext/standard/tests/array/array_intersect_variation8.phpt",
    "ext/standard/tests/array/array_walk_variation5.phpt",
    "ext/standard/tests/array/array_push_error2.phpt",
    "ext/standard/tests/array/bug50006_1.phpt",
    "ext/standard/tests/array/bug43495.phpt",
    "ext/standard/tests/array/end_variation2.phpt",
    "ext/standard/tests/array/sort_object2.phpt",
    "ext/standard/tests/array/arsort_error.phpt",
    "ext/standard/tests/array/bug26458.phpt",
    "ext/standard/tests/array/asort_variation8.phpt",
    "ext/standard/tests/array/next_variation1.phpt",
    "ext/standard/tests/array/array_unshift_variation4.phpt",
    "ext/standard/tests/array/array_uintersect_assoc_basic2.phpt",
    "ext/standard/tests/array/array_reverse_variation5.phpt",
    "ext/standard/tests/array/each_variation3.phpt",
    "ext/standard/tests/array/array_map_variation14.phpt",
    "ext/standard/tests/array/natcasesort_basic.phpt",
    "ext/standard/tests/array/array_sum_variation2.phpt",
    "ext/standard/tests/array/array_next_error1.phpt",
    "ext/standard/tests/array/array_udiff_assoc_variation4.phpt",
    "ext/standard/tests/array/locale_sort.phpt",
    "ext/standard/tests/array/sizeof_variation2.phpt",
    "ext/standard/tests/array/array_combine_variation6.phpt",
    "ext/standard/tests/array/array_diff_variation2.phpt",
    "ext/standard/tests/array/array_walk_recursive.phpt",
    "ext/standard/tests/array/array_diff_key_error.phpt",
    "ext/standard/tests/array/array_diff_assoc_variation2.phpt",
    "ext/standard/tests/array/bug45312.phpt",
    "ext/standard/tests/array/krsort_error.phpt",
    "ext/standard/tests/array/array_intersect_uassoc_variation3.phpt",
    "ext/standard/tests/array/array_combine_variation5.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation10.phpt",
    "ext/standard/tests/array/array_walk_recursive_variation1.phpt",
    "ext/standard/tests/array/usort_basic.phpt",
    "ext/standard/tests/array/array_intersect_variation10.phpt",
    "ext/standard/tests/array/array_multisort_variation1.phpt",
    "ext/standard/tests/array/natcasesort_variation10.phpt",
    "ext/standard/tests/array/007.phpt",
    "ext/standard/tests/array/bug42850.phpt",
    "ext/standard/tests/array/key_exists_error.phpt",
    "ext/standard/tests/array/array_diff_ukey_variation4.phpt",
    "ext/standard/tests/array/bug42233.phpt",
    "ext/standard/tests/array/array_uintersect_assoc_variation1.phpt",
    "ext/standard/tests/array/array_values_error.phpt",
    "ext/standard/tests/array/next_basic.phpt",
    "ext/standard/tests/array/array_intersect_ukey_variation8.phpt",
    "ext/standard/tests/array/bug38464.phpt",
    "ext/standard/tests/array/current_variation2.phpt",
    "ext/standard/tests/array/krsort_variation11.phpt",
    "ext/standard/tests/array/array_key_exists_variation3.phpt",
    "ext/standard/tests/array/bug28974.phpt",
    "ext/standard/tests/array/in_array_errors.phpt",
    "ext/standard/tests/array/in_array_variation1.phpt",
    "ext/standard/tests/array/krsort_variation10.phpt",
    "ext/standard/tests/array/uksort_error.phpt",
    "ext/standard/tests/array/array_push_variation2.phpt",
    "ext/standard/tests/array/sizeof_variation1.phpt",
    "ext/standard/tests/array/sizeof_basic1.phpt",
    "ext/standard/tests/array/array_intersect_uassoc_variation7.phpt",
    "ext/standard/tests/array/array_multisort_variation10.phpt",
    "ext/standard/tests/array/array_rand_basic2.phpt",
    "ext/standard/tests/array/array_diff_assoc_variation7.phpt",
    "ext/standard/tests/array/array_diff_error.phpt",
    "ext/standard/tests/array/array_combine.phpt",
    "ext/standard/tests/array/array_udiff_variation3.phpt",
    "ext/standard/tests/array/array_intersect_assoc_variation6.phpt",
    "ext/standard/tests/array/array_search_variation3.phpt",
    "ext/standard/tests/array/array_diff_uassoc_error.phpt",
    "ext/standard/tests/array/array_keys_variation_003.phpt",
    "ext/standard/tests/array/array_rand.phpt",
    "ext/standard/tests/array/array_unshift_variation6.phpt",
    "ext/standard/tests/array/array_map_variation6.phpt",
    "ext/standard/tests/array/array_multisort_variation5.phpt",
    "ext/standard/tests/array/array_slice_variation7.phpt",
    "ext/standard/tests/array/uasort_basic2.phpt",
    "ext/standard/tests/array/array_diff_ukey_variation11.phpt",
    "ext/standard/tests/array/array_intersect_key_variation6.phpt",
    "ext/standard/tests/array/array_key_exists_object2.phpt",
    "ext/standard/tests/array/array_chunk_variation11.phpt",
    "ext/standard/tests/array/current_variation1.phpt",
    "ext/standard/tests/array/array_search_variation4.phpt",
    "ext/standard/tests/array/uasort_variation8.phpt",
    "ext/standard/tests/array/array_udiff_variation1.phpt",
    "ext/standard/tests/array/array_merge_recursive_basic1.phpt",
    "ext/standard/tests/array/array_diff_ukey_variation5.phpt",
    "ext/standard/tests/array/array_diff_ukey_variation6.phpt",
    "ext/standard/tests/array/002.phpt",
    "ext/standard/tests/array/array_fill_keys_variation4.phpt",
    "ext/standard/tests/array/008.phpt",
    "ext/standard/tests/array/usort_variation6.phpt",
    "ext/standard/tests/array/array_intersect_ukey_variation10.phpt",
    "ext/standard/tests/array/array_values_variation.phpt",
    "ext/standard/tests/array/uksort_variation1.phpt",
    "ext/standard/tests/array/natcasesort_variation7.phpt",
    "ext/standard/tests/array/min_variation2.phpt",
    "ext/standard/tests/array/uksort_variation2.phpt",
    "ext/standard/tests/array/array_map_variation9.phpt",
    "ext/standard/tests/array/array_diff_assoc_variation1.phpt",
    "ext/standard/tests/array/prev_error1.phpt",
    "ext/standard/tests/array/array_diff_key_basic.phpt",
    "ext/standard/tests/array/array_fill_object_2_4.phpt",
    "ext/standard/tests/array/usort_error1.phpt",
    "ext/standard/tests/array/next_variation2.phpt",
    "ext/standard/tests/array/array_diff_ukey_basic.phpt",
    "ext/standard/tests/array/sort_variation3.phpt",
    "ext/standard/tests/array/array_walk_variation4.phpt",
    "ext/standard/tests/array/array_uintersect_variation5.phpt",
    "ext/standard/tests/array/natcasesort_object2.phpt",
    "ext/standard/tests/array/array_combine_variation4.phpt",
    "ext/standard/tests/array/array_sum.phpt",
    "ext/standard/tests/array/array_multisort_variation3.phpt",
    "ext/standard/tests/array/array_intersect_variation3.phpt",
    "ext/standard/tests/array/bug34066.phpt",
    "ext/standard/tests/array/sizeof_basic2.phpt",
    "ext/standard/tests/array/array_map_variation15.phpt",
    "ext/standard/tests/array/array_sum_variation5.phpt",
    "ext/standard/tests/array/sort_variation9.phpt",
    "ext/standard/tests/array/array_multisort_variation11.phpt",
    "ext/standard/tests/array/array_merge_variation8.phpt",
    "ext/standard/tests/array/array_values_variation3.phpt",
    "ext/standard/tests/array/array_merge_recursive_variation8.phpt",
    "ext/standard/tests/array/usort_variation5.phpt",
    "ext/standard/tests/array/array_diff_key2.phpt",
    "ext/standard/tests/array/array_chunk_variation4.phpt",
    "ext/standard/tests/array/array_reverse_error.phpt",
    "ext/standard/tests/array/array_merge_recursive_variation4.phpt",
    "ext/standard/tests/array/array_shift_variation1.phpt",
    "ext/standard/tests/array/array_change_key_case_variation.phpt",
    "ext/standard/tests/array/array_merge_recursive_variation5.phpt",
    "ext/standard/tests/array/uasort_object2.phpt",
    "ext/standard/tests/array/array_walk_recursive_error1.phpt",
    "ext/standard/tests/array/next_error.phpt",
    "ext/standard/tests/array/array_map_variation11.phpt",
    "ext/standard/tests/array/array_intersect_key_variation8.phpt",
    "ext/standard/tests/array/array_multisort_variation4.phpt",
    "ext/standard/tests/array/each_error.phpt",
    "ext/standard/tests/array/array_shift_variation2.phpt",
    "ext/standard/tests/array/each_variation5.phpt",
    "ext/standard/tests/array/ksort_error.phpt",
    "ext/standard/tests/array/array_keys_variation_005.phpt",
    "ext/standard/tests/array/array_push_error1.phpt",
    "ext/standard/tests/array/array_merge_variation10.phpt",
    "ext/standard/tests/array/array_sum_error.phpt",
    "ext/standard/tests/array/krsort_variation6.phpt",
    "ext/standard/tests/array/bug43505.phpt",
    "ext/standard/tests/array/array_map_variation16.phpt",
    "ext/standard/tests/array/array_chunk_variation13.phpt",
    "ext/standard/tests/array/array_change_key_case_variation7.phpt",
    "ext/standard/tests/array/array_intersect_variation9.phpt",
    "ext/standard/tests/array/rsort_variation4.phpt",
    "ext/standard/tests/array/array_intersect_ukey_variation5.phpt",
    "ext/standard/tests/array/sizeof_object2.phpt",
    "ext/standard/tests/array/array_combine_basic.phpt",
    "ext/standard/tests/array/array_chunk2.phpt",
    "ext/standard/tests/array/array_key_exists_variation4.phpt",
    "ext/standard/tests/array/reset_variation1.phpt",
    "ext/standard/tests/array/array_chunk_variation31.phpt",
    "ext/standard/tests/array/bug25359.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation6.phpt",
    "ext/standard/tests/array/bug30833.phpt",
    "ext/standard/tests/array/array_shift_variation5.phpt",
    "ext/standard/tests/array/array_uintersect_variation1.phpt",
    "ext/standard/tests/array/array_chunk_variation16.phpt",
    "ext/standard/tests/array/array_pad.phpt",
    "ext/standard/tests/array/shuffle_variation3.phpt",
    "ext/standard/tests/array/array_unique_error.phpt",
    "ext/standard/tests/array/array_chunk_variation12.phpt",
    "ext/standard/tests/array/array_map_object3.phpt",
    "ext/standard/tests/array/array_change_key_case_variation6.phpt",
    "ext/standard/tests/array/current_error.phpt",
    "ext/standard/tests/array/array_pop_variation.phpt",
    "ext/standard/tests/array/unexpected_array_mod_bug.phpt",
    "ext/standard/tests/array/bug31213.phpt",
    "ext/standard/tests/array/arsort_variation10.phpt",
    "ext/standard/tests/array/array_diff_ukey_variation1.phpt",
    "ext/standard/tests/array/array_flip_error.phpt",
    "ext/standard/tests/array/array_diff_ukey_variation9.phpt",
    "ext/standard/tests/array/bug42838.phpt",
    "ext/standard/tests/array/uasort_object1.phpt",
    "ext/standard/tests/array/array_chunk_variation27.phpt",
    "ext/standard/tests/array/array_walk_variation6.phpt",
    "ext/standard/tests/array/array_change_key_case.phpt",
    "ext/standard/tests/array/array_values_basic.phpt",
    "ext/standard/tests/array/array_change_key_case_variation5.phpt",
    "ext/standard/tests/array/usort_variation10.phpt",
    "ext/standard/tests/array/array_sum_basic.phpt",
    "ext/standard/tests/array/array_combine_variation1.phpt",
    "ext/standard/tests/array/array_walk_variation9.phpt",
    "ext/standard/tests/array/bug41686.phpt",
    "ext/standard/tests/array/array_reverse_basic2.phpt",
    "ext/standard/tests/array/rsort_variation2.phpt",
    "ext/standard/tests/array/array_chunk_variation14.phpt",
    "ext/standard/tests/array/array_chunk_variation21.phpt",
    "ext/standard/tests/array/array_intersect_variation5.phpt",
    "ext/standard/tests/array/array_intersect_assoc_variation9.phpt",
    "ext/standard/tests/array/array_intersect_uassoc_variation6.phpt",
    "ext/standard/tests/array/krsort_variation7.phpt",
    "ext/standard/tests/array/end_variation1.phpt",
    "ext/standard/tests/array/uasort_variation2.phpt",
    "ext/standard/tests/array/uasort_variation11.phpt",
    "ext/standard/tests/array/array_unshift_object.phpt",
    "ext/standard/tests/array/array_merge_variation2.phpt",
    "ext/standard/tests/array/array_chunk_variation5.phpt",
    "ext/standard/tests/array/arsort_variation7.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation4.phpt",
    "ext/standard/tests/array/array_chunk_basic1.phpt",
    "ext/standard/tests/array/sort_variation7.phpt",
    "ext/standard/tests/array/array_intersect_variation1.phpt",
    "ext/standard/tests/array/rsort_variation8.phpt",
    "ext/standard/tests/array/bug48224.phpt",
    "ext/standard/tests/array/array_fill_variation5.phpt",
    "ext/standard/tests/array/array_diff_uassoc_variation14.phpt",
    "ext/standard/tests/array/array_keys_error.phpt",
    "ext/standard/tests/array/arsort_variation3.phpt",
    "ext/standard/tests/array/array_unshift_basic2.phpt",
    "ext/standard/tests/array/array_product_variation5.phpt",
    "ext/standard/tests/array/min_variation1.phpt",
    "ext/standard/tests/array/array_combine_error1.phpt",
    "ext/standard/tests/array/prev_error3.phpt",
    "ext/standard/tests/array/array_product_error.phpt",
    "ext/standard/tests/array/array_diff_key_variation7.phpt",
    "ext/standard/tests/array/array_unshift_variation2.phpt",
    "ext/standard/tests/array/array_fill_keys_variation2.phpt",
    "ext/standard/tests/array/array_intersect_key_variation4.phpt",
    "ext/standard/tests/array/array_filter_variation3.phpt",
    "ext/standard/tests/array/array_diff_variation10.phpt",
    "ext/standard/tests/array/array_reverse_variation3.phpt",
    "ext/standard/tests/array/bug44182.phpt",
    "ext/standard/tests/array/bug20381.phpt",
    "ext/standard/tests/array/rsort_variation9.phpt",
    "ext/standard/tests/array/array_diff_1.phpt",
    "ext/standard/tests/array/each_variation4.phpt",
    "ext/standard/tests/array/array_filter.phpt",
    "ext/standard/tests/array/array_shift_variation7.phpt",
    "ext/standard/tests/array/uksort_basic.phpt",
    "ext/standard/tests/array/array_intersect_error.phpt",
    "ext/standard/tests/array/array_uintersect_uassoc_basic.phpt",
    "ext/standard/tests/array/usort_variation4.phpt",
    "ext/standard/tests/array/array_merge_variation4.phpt",
    "ext/standard/tests/array/array_change_key_case_variation8.phpt",
    "ext/standard/tests/array/array_rand_error.phpt",
    "ext/standard/tests/array/array_change_key_case_variation3.phpt",
    "ext/standard/tests/array/array_udiff_assoc_variation.phpt",
    "ext/standard/tests/array/key_variation3.phpt",
    "ext/standard/tests/array/usort_variation3.phpt",
    "ext/standard/tests/array/array_filter_variation1.phpt",
    "ext/standard/tests/array/array_udiff_assoc_basic.phpt",
    "ext/standard/tests/array/uasort_variation7.phpt",
    "ext/standard/tests/array/uasort_variation6.phpt",
    "ext/standard/tests/array/array_uintersect_uassoc_variation1.phpt",
    "ext/standard/tests/array/array_multisort_variation9.phpt",
    "ext/standard/tests/array/array_intersect_ukey_variation9.phpt",
    "ext/standard/tests/array/array_chunk_variation20.phpt",
    "ext/standard/tests/array/prev_variation2.phpt",
    "ext/standard/tests/array/array_uintersect_assoc_variation3.phpt",
    "ext/standard/tests/array/shuffle_error.phpt",
    "ext/standard/tests/array/array_intersect_assoc_variation5.phpt",
    "ext/standard/tests/array/array_diff_variation3.phpt",
    "ext/standard/tests/array/array_merge_variation5.phpt",
    "ext/standard/tests/array/array_unique_variation8.phpt",
    "ext/standard/tests/array/array_shift_basic.phpt",
    "ext/standard/tests/array/array_flip_variation1.phpt",
    "ext/standard/tests/array/asort_variation9.phpt",
    "ext/standard/tests/array/each.phpt",
    "ext/standard/tests/array/array_uintersect_uassoc_error.phpt",
    "ext/standard/tests/array/array_walk_recursive_variation7.phpt",
    "ext/standard/tests/array/count_recursive.phpt",
    "ext/standard/tests/array/compact_variation1.phpt",
    "ext/standard/tests/array/array_map_error.phpt",
    "ext/standard/tests/array/array_diff_variation8.phpt",
    "ext/standard/tests/array/bug30074.phpt",
    "ext/standard/tests/array/array_intersect_uassoc_variation10.phpt",
    "ext/standard/tests/array/count_variation1.phpt",
    "ext/standard/tests/array/bug34066_1.phpt",
    "ext/standard/tests/array/array_walk_variation8.phpt",
    "ext/standard/tests/array/array_pad_variation4.phpt",
    "ext/standard/tests/array/uasort_variation3.phpt",
    "ext/standard/tests/array/array_intersect_uassoc_variation5.phpt",
    "ext/standard/tests/array/bug48854.phpt",
    "ext/standard/tests/array/bug25708.phpt",
    "ext/standard/tests/array/array_intersect_assoc_variation3.phpt",
    "ext/standard/tests/array/end.phpt",
    "ext/standard/tests/array/array_map_variation1.phpt",
    "ext/standard/tests/array/array_flip.phpt",
    "ext/standard/tests/array/array_diff_assoc_variation3.phpt",
    "ext/standard/tests/array/array_search_variation2.phpt",
    "ext/standard/tests/array/array_shift_variation8.phpt",
    "ext/standard/tests/array/array_diff_assoc_variation4.phpt",
    "ext/standard/tests/array/array_values_variation_64bit.phpt",
    "ext/standard/tests/array/array_key_exists_basic.phpt",
    "ext/standard/tests/array/sizeof_object1.phpt",
    "ext/standard/tests/array/array_udiff_assoc_variation5.phpt",
    "ext/standard/tests/array/natcasesort_variation1.phpt",
    "ext/standard/tests/array/max_basic.phpt",
    "ext/standard/tests/array/natcasesort_variation6.phpt",
    "ext/standard/tests/array/array_keys_basic.phpt",
    "ext/standard/tests/array/array_uintersect_uassoc_variation3.phpt",
    "ext/standard/tests/array/each_basic.phpt",
    "ext/standard/tests/array/array_filter_variation6.phpt",
    "ext/standard/tests/array/natsort_basic.phpt",
    "ext/standard/tests/array/rsort_variation3.phpt",
    "ext/standard/tests/array/array_diff_assoc_variation5.phpt",
    "ext/standard/tests/array/uasort_error.phpt",
    "ext/standard/tests/array/array_next_error2.phpt",
    "ext/standard/tests/array/array_intersect_key.phpt",
    "ext/standard/tests/array/array_diff_assoc.phpt",
    "ext/standard/tests/array/array_fill_keys_error.phpt",
    "ext/standard/tests/array/array_intersect_ukey_variation3.phpt",
    "ext/standard/tests/array/compact_error.phpt",
    "ext/standard/tests/array/array_search.phpt",
    "ext/standard/tests/array/prev_basic.phpt",
    "ext/standard/tests/array/array_count_values2.phpt",
    "ext/standard/tests/array/array_walk_recursive_variation2.phpt",
    "ext/standard/tests/array/bug23788.phpt",
    "ext/standard/tests/array/array_multisort_basic1.phpt",
    "ext/standard/tests/array/sort_variation5.phpt",
    "ext/standard/tests/array/array_walk_recursive_basic2.phpt",
    "ext/standard/tests/array/array_walk_recursive_variation6.phpt",
    "ext/standard/tests/array/array_diff_ukey_variation10.phpt",
    "ext/standard/tests/array/max.phpt",
    "ext/standard/tests/array/ksort_variation2.phpt",
    "ext/standard/tests/array/array_change_key_case_variation2.phpt",
    "ext/standard/tests/array/array_sum_variation1.phpt",
    "ext/standard/tests/array/array_push_variation5.phpt",
    "ext/standard/tests/array/array_chunk_variation3.phpt",
    "ext/standard/tests/array/array_intersect_variation4.phpt",
    "ext/standard/tests/array/array_unshift_variation7.phpt",
    "ext/standard/tests/array/array_diff_variation7.phpt",
    "ext/standard/tests/array/array_udiff_assoc_variation1.phpt",
    "ext/standard/tests/array/array_product_variation1.phpt",
    "ext/standard/tests/array/count_variation2.phpt",
    "ext/standard/tests/array/array_walk_recursive_variation4.phpt",
    "ext/standard/tests/array/array_unshift_error.phpt",
    "ext/standard/tests/array/array_product_variation2.phpt",
    "ext/standard/tests/array/each_variation6.phpt",
    "ext/standard/tests/array/bug21998.phpt",
    "ext/standard/tests/array/krsort_basic.phpt",
    "ext/standard/tests/array/array_uintersect_variation3.phpt",
    "ext/standard/tests/array/array_chunk_basic2.phpt",
    "ext/standard/tests/array/array_flip_variation5.phpt",
    "ext/standard/tests/array/uasort_variation4.phpt",
    "ext/standard/tests/array/asort_variation4.phpt",
    "ext/standard/tests/array/sort_variation6.phpt",
    "ext/standard/tests/array/array_intersect_assoc_variation8.phpt",
    "ext/standard/tests/array/sort_error.phpt",
    "ext/standard/tests/array/array_walk_variation2.phpt",
    "ext/standard/tests/array/ksort_variation5.phpt",
    "ext/standard/tests/array/array_diff_uassoc_basic.phpt",
    "ext/standard/tests/array/rsort_variation11.phpt",
    "ext/standard/tests/array/array_intersect_key_variation1.phpt",
    "ext/standard/tests/array/array_intersect_ukey_error.phpt",
    "ext/standard/tests/array/array_fill_basic.phpt",
    "ext/standard/tests/array/sort_variation8.phpt",
    "ext/standard/tests/array/array_slice_error.phpt",
    "ext/standard/tests/array/array_slice_variation3.phpt",
    "ext/standard/tests/array/array_udiff_assoc_variation3.phpt",
    "ext/standard/tests/array/rsort_object2.phpt",
    "ext/standard/tests/array/range.phpt",
    "ext/standard/tests/array/rsort_object1.phpt",
    "ext/standard/tests/array/natcasesort_variation4.phpt",
    "ext/standard/tests/array/array_key_exists_variation5.phpt",
    "ext/standard/tests/array/array_rand_variation1.phpt",
    "ext/standard/tests/array/array_walk_variation7.phpt",
    "ext/standard/tests/array/array_multisort_variation2.phpt",
    "ext/standard/tests/array/range_variation1_64bit.phpt",
    "ext/standard/tests/array/array_merge_variation3.phpt",
    "ext/standard/tests/array/array_uintersect_uassoc_variation5.phpt",
    "ext/standard/tests/array/key_variation2.phpt",
    "ext/standard/tests/array/array_chunk_variation19.phpt",
    "ext/standard/tests/array/array_chunk_variation29.phpt",
    "ext/standard/tests/array/009.phpt",
    "ext/standard/tests/array/array_diff_ukey_variation7.phpt",
    "ext/standard/tests/array/rsort_variation5.phpt",
    "ext/standard/tests/array/array_chunk_variation10.phpt",
    "ext/standard/tests/array/array_intersect_ukey_basic.phpt",
    "ext/standard/tests/array/array_diff_key_variation1.phpt",
    "ext/standard/tests/array/natcasesort_variation9.phpt",
    "ext/standard/tests/array/usort_variation7.phpt",
    "ext/standard/tests/array/array_unshift_variation9.phpt",
    "ext/standard/tests/array/array_change_key_case_variation1.phpt",
    "ext/standard/tests/array/array_fill_variation4.phpt",
    "ext/standard/tests/array/array_rand_basic1.phpt",
    "ext/standard/tests/array/bug24198.phpt",
    "ext/standard/tests/array/array_diff_assoc_error.phpt",
    "ext/standard/tests/array/min.phpt",
    "ext/standard/tests/array/array_filter_variation2.phpt",
    "ext/standard/tests/array/array_merge_error.phpt",
    "ext/standard/tests/array/bug28739.phpt",
    "ext/standard/tests/array/extract_variation3.phpt",
    "ext/standard/tests/array/array_slice.phpt",
    "ext/standard/tests/array/array_walk_recursive_variation3.phpt",
    "ext/standard/tests/array/usort_object1.phpt",
    "ext/standard/tests/array/array_multisort_variation6.phpt",
    "ext/standard/tests/array/array_shift_variation6.phpt",
    "ext/standard/tests/array/array_map_variation7.phpt",
    "ext/standard/tests/array/bug30266.phpt",
    "ext/standard/tests/array/current_variation5.phpt",
    "ext/standard/tests/array/array_diff_ukey_error.phpt",
    "ext/standard/tests/array/array_chunk_variation30.phpt",
    "ext/standard/tests/array/array_search_errors.phpt",
    "ext/standard/tests/array/bug40191.phpt",
    "ext/standard/tests/array/natcasesort_variation5.phpt",
    "ext/standard/tests/array/array_diff_assoc_variation9.phpt",
    "ext/standard/tests/array/range_errors.phpt",
    "ext/standard/tests/array/array_key_exists_error.phpt",
    "ext/standard/tests/array/array_slice_variation5.phpt",
    "ext/standard/tests/array/uasort_basic1.phpt",
    "ext/standard/tests/array/array_key_exists.phpt",
    "ext/standard/tests/strings/sprintf_variation8.phpt",
    "ext/standard/tests/strings/vfprintf_basic3.phpt",
    "ext/standard/tests/strings/bug26973.phpt",
    "ext/standard/tests/strings/substr_replace.phpt",
    "ext/standard/tests/strings/ltrim_variation1.phpt",
    "ext/standard/tests/strings/strrpos_variation10.phpt",
    "ext/standard/tests/strings/trim_error.phpt",
    "ext/standard/tests/strings/strncasecmp_variation10.phpt",
    "ext/standard/tests/strings/strrchr_variation1.phpt",
    "ext/standard/tests/strings/strtr_variation6.phpt",
    "ext/standard/tests/strings/strspn_variation5.phpt",
    "ext/standard/tests/strings/strval_variation1.phpt",
    "ext/standard/tests/strings/printf_variation1.phpt",
    "ext/standard/tests/strings/hebrev_error.phpt",
    "ext/standard/tests/strings/sha1_file.phpt",
    "ext/standard/tests/strings/vsprintf_variation11_64bit.phpt",
    "ext/standard/tests/strings/bug45166.phpt",
    "ext/standard/tests/strings/vprintf_variation8.phpt",
    "ext/standard/tests/strings/fprintf_variation_005.phpt",
    "ext/standard/tests/strings/vfprintf_variation9.phpt",
    "ext/standard/tests/strings/sprintf_f.phpt",
    "ext/standard/tests/strings/strip_tags_variation4.phpt",
    "ext/standard/tests/strings/vprintf_variation1.phpt",
    "ext/standard/tests/strings/vfprintf_variation19.phpt",
    "ext/standard/tests/strings/convert_uudecode_basic.phpt",
    "ext/standard/tests/strings/htmlentities06.phpt",
    "ext/standard/tests/strings/show_source_basic.phpt",
    "ext/standard/tests/strings/str_split_variation6_64bit.phpt",
    "ext/standard/tests/strings/vsprintf_variation16_64bit.phpt",
    "ext/standard/tests/strings/money_format_basic1.phpt",
    "ext/standard/tests/strings/sprintf_variation52.phpt",
    "ext/standard/tests/strings/convert_uuencode_variation1.phpt",
    "ext/standard/tests/strings/ltrim_variation2.phpt",
    "ext/standard/tests/strings/strings001.phpt",
    "ext/standard/tests/strings/sprintf_variation20.phpt",
    "ext/standard/tests/strings/str_split_variation1.phpt",
    "ext/standard/tests/strings/fprintf_variation_002.phpt",
    "ext/standard/tests/strings/vsprintf_basic1.phpt",
    "ext/standard/tests/strings/setlocale_variation3.phpt",
    "ext/standard/tests/strings/strtr_variation8.phpt",
    "ext/standard/tests/strings/get_html_translation_table_error.phpt",
    "ext/standard/tests/strings/sprintf_f_3.phpt",
    "ext/standard/tests/strings/setlocale_variation5.phpt",
    "ext/standard/tests/strings/strripos_variation4.phpt",
    "ext/standard/tests/strings/explode_variation4.phpt",
    "ext/standard/tests/strings/sha1.phpt",
    "ext/standard/tests/strings/wordwrap_variation5.phpt",
    "ext/standard/tests/strings/vprintf_variation7.phpt",
    "ext/standard/tests/strings/005.phpt",
    "ext/standard/tests/strings/uuencode.phpt",
    "ext/standard/tests/strings/strtoupper1.phpt",
    "ext/standard/tests/strings/strrpos_variation15.phpt",
    "ext/standard/tests/strings/http_build_query.phpt",
    "ext/standard/tests/strings/nl2br_error.phpt",
    "ext/standard/tests/strings/trim1.phpt",
    "ext/standard/tests/strings/vsprintf_variation19.phpt",
    "ext/standard/tests/strings/chunk_split_variation11.phpt",
    "ext/standard/tests/strings/bug22187.phpt",
    "ext/standard/tests/strings/strtr_variation1.phpt",
    "ext/standard/tests/strings/str_replace_variation3.phpt",
    "ext/standard/tests/strings/crc32_variation1.phpt",
    "ext/standard/tests/strings/vprintf_variation9.phpt",
    "ext/standard/tests/strings/vfprintf_variation20.phpt",
    "ext/standard/tests/strings/addcslashes_002.phpt",
    "ext/standard/tests/strings/vsprintf_variation8.phpt",
    "ext/standard/tests/strings/str_pad.phpt",
    "ext/standard/tests/strings/stripos_variation12.phpt",
    "ext/standard/tests/strings/printf.phpt",
    "ext/standard/tests/strings/parse_str_basic3.phpt",
    "ext/standard/tests/strings/strrpos_variation11.phpt",
    "ext/standard/tests/strings/get_html_translation_table_basic4.phpt",
    "ext/standard/tests/strings/chunk_split.phpt",
    "ext/standard/tests/strings/stripos_variation6.phpt",
    "ext/standard/tests/strings/convert_uuencode_basic.phpt",
    "ext/standard/tests/strings/strlen.phpt",
    "ext/standard/tests/strings/html_entity_decode_koi8-r.phpt",
    "ext/standard/tests/strings/get_html_translation_table_basic9.phpt",
    "ext/standard/tests/strings/strspn_variation11.phpt",
    "ext/standard/tests/strings/soundex_error.phpt",
    "ext/standard/tests/strings/str_shuffle_variation1.phpt",
    "ext/standard/tests/strings/wordwrap.phpt",
    "ext/standard/tests/strings/nl_langinfo_basic.phpt",
    "ext/standard/tests/strings/vprintf_basic2.phpt",
    "ext/standard/tests/strings/get_html_translation_table_basic2.phpt",
    "ext/standard/tests/strings/sprintf_variation43.phpt",
    "ext/standard/tests/strings/crc32_variation2.phpt",
    "ext/standard/tests/strings/vsprintf_variation7.phpt",
    "ext/standard/tests/strings/vfprintf_variation8.phpt",
    "ext/standard/tests/strings/htmlentities-utf.phpt",
    "ext/standard/tests/strings/chop_basic.phpt",
    "ext/standard/tests/strings/hebrevc_variation2.phpt",
    "ext/standard/tests/strings/bug47481.phpt",
    "ext/standard/tests/strings/chop_variation1.phpt",
    "ext/standard/tests/strings/bug37244.phpt",
    "ext/standard/tests/strings/bug21338.phpt",
    "ext/standard/tests/strings/stripslashes_variation1.phpt",
    "ext/standard/tests/strings/vprintf_variation15_64bit.phpt",
    "ext/standard/tests/strings/strripos.phpt",
    "ext/standard/tests/strings/strpos_number.phpt",
    "ext/standard/tests/strings/vfprintf_variation14_64bit.phpt",
    "ext/standard/tests/strings/strspn_variation4.phpt",
    "ext/standard/tests/strings/strtr_variation5.phpt",
    "ext/standard/tests/strings/str_replace_error.phpt",
    "ext/standard/tests/strings/sprintf_variation34.phpt",
    "ext/standard/tests/strings/bug47322.phpt",
    "ext/standard/tests/strings/strspn_variation12.phpt",
    "ext/standard/tests/strings/sprintf_variation12.phpt",
    "ext/standard/tests/strings/strrev_variation3.phpt",
    "ext/standard/tests/strings/sprintf_variation49.phpt",
    "ext/standard/tests/strings/sprintf_basic7.phpt",
    "ext/standard/tests/strings/strip_tags_variation10.phpt",
    "ext/standard/tests/strings/vfprintf_error2.phpt",
    "ext/standard/tests/strings/strrpos_variation2.phpt",
    "ext/standard/tests/strings/convert_cyr_string_variation1.phpt",
    "ext/standard/tests/strings/str_word_count.phpt",
    "ext/standard/tests/strings/vfprintf_error3.phpt",
    "ext/standard/tests/strings/sprintf_variation51.phpt",
    "ext/standard/tests/strings/sprintf_basic5.phpt",
    "ext/standard/tests/strings/bug54055.phpt",
    "ext/standard/tests/strings/parse_str_error1.phpt",
    "ext/standard/tests/strings/bug42107.phpt",
    "ext/standard/tests/strings/strrpos_variation12.phpt",
    "ext/standard/tests/strings/strncasecmp_variation3.phpt",
    "ext/standard/tests/strings/vfprintf_variation21.phpt",
    "ext/standard/tests/strings/sprintf_basic2.phpt",
    "ext/standard/tests/strings/strlen_error.phpt",
    "ext/standard/tests/strings/strcspn_variation10.phpt",
    "ext/standard/tests/strings/bug27295.phpt",
    "ext/standard/tests/strings/vsprintf_variation15_64bit.phpt",
    "ext/standard/tests/strings/nl2br_variation5.phpt",
    "ext/standard/tests/strings/setlocale_variation1.phpt",
    "ext/standard/tests/strings/strtr_variation9.phpt",
    "ext/standard/tests/strings/sprintf_variation7.phpt",
    "ext/standard/tests/strings/rtrim_variation2.phpt",
    "ext/standard/tests/strings/strtok_variation7.phpt",
    "ext/standard/tests/strings/sprintf_variation47.phpt",
    "ext/standard/tests/strings/chunk_split_variation4.phpt",
    "ext/standard/tests/strings/html_entity_decode_macroman.phpt",
    "ext/standard/tests/strings/fprintf_variation_007_64bit.phpt",
    "ext/standard/tests/strings/vprintf_variation19_64bit.phpt",
    "ext/standard/tests/strings/sscanf_basic3.phpt",
    "ext/standard/tests/strings/bug22227.phpt",
    "ext/standard/tests/strings/substr_count_error.phpt",
    "ext/standard/tests/strings/addslashes_variation3.phpt",
    "ext/standard/tests/strings/strrpos_offset.phpt",
    "ext/standard/tests/strings/strtoupper1-win32.phpt",
    "ext/standard/tests/strings/bug38770.phpt",
    "ext/standard/tests/strings/strrpos.phpt",
    "ext/standard/tests/strings/vsprintf_variation4_64bit.phpt",
    "ext/standard/tests/strings/chr_ord.phpt",
    "ext/standard/tests/strings/crypt_blowfish_variation1.phpt",
    "ext/standard/tests/strings/nl_langinfo_variation1.phpt",
    "ext/standard/tests/strings/htmlspecialchars_decode_variation5.phpt",
    "ext/standard/tests/strings/vsprintf_basic8.phpt",
    "ext/standard/tests/strings/bug23894.phpt",
    "ext/standard/tests/strings/sprintf_variation10.phpt",
    "ext/standard/tests/strings/sprintf_variation9.phpt",
    "ext/standard/tests/strings/sprintf_variation40_64bit.phpt",
    "ext/standard/tests/strings/count_chars_variation1.phpt",
    "ext/standard/tests/strings/printf_basic2.phpt",
    "ext/standard/tests/strings/bug28386.phpt",
    "ext/standard/tests/strings/vprintf_basic1.phpt",
    "ext/standard/tests/strings/vprintf_variation11_64bit.phpt",
    "ext/standard/tests/strings/strtr_variation4.phpt",
    "ext/standard/tests/strings/printf_basic7.phpt",
    "ext/standard/tests/strings/htmlentities18.phpt",
    "ext/standard/tests/strings/bug22904.phpt",
    "ext/standard/tests/strings/htmlspecialchars_decode_error.phpt",
    "ext/standard/tests/strings/str_rot13_error.phpt",
    "ext/standard/tests/strings/strip_tags_variation3.phpt",
    "ext/standard/tests/strings/stripos_variation9.phpt",
    "ext/standard/tests/strings/vprintf_variation12_64bit.phpt",
    "ext/standard/tests/strings/strncmp_variation6.phpt",
    "ext/standard/tests/strings/bug28633.phpt",
    "ext/standard/tests/strings/setlocale_variation4.phpt",
    "ext/standard/tests/strings/strncasecmp_variation2.phpt",
    "ext/standard/tests/strings/chr_error.phpt",
    "ext/standard/tests/strings/get_html_translation_table_variation1.phpt",
    "ext/standard/tests/strings/strcspn.phpt",
    "ext/standard/tests/strings/substr.phpt",
    "ext/standard/tests/strings/strrpos_variation1.phpt",
    "ext/standard/tests/strings/stristr_variation1.phpt",
    "ext/standard/tests/strings/vfprintf_variation16_64bit.phpt",
    "ext/standard/tests/strings/ucwords_variation3.phpt",
    "ext/standard/tests/strings/crypt_blowfish_variation2.phpt",
    "ext/standard/tests/strings/vprintf_variation19.phpt",
    "ext/standard/tests/strings/strrchr_variation7.phpt",
    "ext/standard/tests/strings/setlocale_error.phpt",
    "ext/standard/tests/strings/count_chars.phpt",
    "ext/standard/tests/strings/crypt.phpt",
    "ext/standard/tests/strings/bug38322.phpt",
    "ext/standard/tests/strings/wordwrap_basic.phpt",
    "ext/standard/tests/strings/sscanf_variation1.phpt",
    "ext/standard/tests/strings/vfprintf_variation1.phpt",
    "ext/standard/tests/strings/stripos_variation15.phpt",
    "ext/standard/tests/strings/sprintf_variation3.phpt",
    "ext/standard/tests/strings/quotemeta_basic.phpt",
    "ext/standard/tests/strings/sprintf_variation42.phpt",
    "ext/standard/tests/strings/str_shuffle_error.phpt",
    "ext/standard/tests/strings/printf_64bit.phpt",
    "ext/standard/tests/strings/parse_str_basic4.phpt",
    "ext/standard/tests/strings/ucwords_variation4.phpt",
    "ext/standard/tests/strings/trim_variation2.phpt",
    "ext/standard/tests/strings/vfprintf_basic2.phpt",
    "ext/standard/tests/strings/bug40432.phpt",
    "ext/standard/tests/strings/nl2br.phpt",
    "ext/standard/tests/strings/md5_basic2.phpt",
    "ext/standard/tests/strings/vsprintf_variation2.phpt",
    "ext/standard/tests/strings/strspn_error.phpt",
    "ext/standard/tests/strings/str_replace_basic.phpt",
    "ext/standard/tests/strings/strncmp_variation4.phpt",
    "ext/standard/tests/strings/bug20108.phpt",
    "ext/standard/tests/strings/addslashes_variation2.phpt",
    "ext/standard/tests/strings/bug36306.phpt",
    "ext/standard/tests/strings/vprintf_variation14.phpt",
    "ext/standard/tests/strings/htmlentities07.phpt",
    "ext/standard/tests/strings/strncasecmp_basic.phpt",
    "ext/standard/tests/strings/str_replace_variation2.phpt",
    "ext/standard/tests/strings/strspn_basic.phpt",
    "ext/standard/tests/strings/pathinfo.phpt",
    "ext/standard/tests/strings/htmlentities17.phpt",
    "ext/standard/tests/strings/chunk_split_variation10.phpt",
    "ext/standard/tests/strings/bug27675.phpt",
    "ext/standard/tests/strings/sscanf_basic2.phpt",
    "ext/standard/tests/strings/strstr2.phpt",
    "ext/standard/tests/strings/004.phpt",
    "ext/standard/tests/strings/bug22224.phpt",
    "ext/standard/tests/strings/url_t.phpt",
    "ext/standard/tests/strings/vprintf_variation15.phpt",
    "ext/standard/tests/strings/vprintf_basic3.phpt",
    "ext/standard/tests/strings/chop_variation4.phpt",
    "ext/standard/tests/strings/strcspn_error.phpt",
    "ext/standard/tests/strings/strrchr_variation5.phpt",
    "ext/standard/tests/strings/strip_tags_variation2.phpt",
    "ext/standard/tests/strings/sprintf_variation32.phpt",
    "ext/standard/tests/strings/similar_text_error.phpt",
    "ext/standard/tests/strings/bug21744.phpt",
    "ext/standard/tests/strings/vsprintf_variation14_64bit.phpt",
    "ext/standard/tests/strings/vfprintf_basic4.phpt",
    "ext/standard/tests/strings/chop_error.phpt",
    "ext/standard/tests/strings/strip_tags_variation7.phpt",
    "ext/standard/tests/strings/strspn_variation9.phpt",
    "ext/standard/tests/strings/sprintf_variation18.phpt",
    "ext/standard/tests/strings/sscanf_basic4.phpt",
    "ext/standard/tests/strings/sscanf_variation2.phpt",
    "ext/standard/tests/strings/strcspn_variation3.phpt",
    "ext/standard/tests/strings/htmlentities12.phpt",
    "ext/standard/tests/strings/stripos_variation1.phpt",
    "ext/standard/tests/strings/join_variation2.phpt",
    "ext/standard/tests/strings/sprintf_variation29.phpt",
    "ext/standard/tests/strings/vsprintf_basic4.phpt",
    "ext/standard/tests/strings/crypt_sha512.phpt",
    "ext/standard/tests/strings/strrchr_variation10.phpt",
    "ext/standard/tests/strings/vprintf_variation17.phpt",
    "ext/standard/tests/strings/printf_basic5.phpt",
    "ext/standard/tests/strings/strcspn_variation11.phpt",
    "ext/standard/tests/strings/strcspn_variation4.phpt",
    "ext/standard/tests/strings/strspn_variation7.phpt",
    "ext/standard/tests/strings/addcslashes_004.phpt",
    "ext/standard/tests/strings/vprintf_error.phpt",
    "ext/standard/tests/strings/strncmp_variation8.phpt",
    "ext/standard/tests/strings/stripcslashes_error.phpt",
    "ext/standard/tests/strings/strtok_variation3.phpt",
    "ext/standard/tests/strings/vsprintf_variation3.phpt",
    "ext/standard/tests/strings/convert_cyr_string_basic.phpt",
    "ext/standard/tests/strings/bug22207.phpt",
    "ext/standard/tests/strings/str_split_variation4.phpt",
    "ext/standard/tests/strings/basename_variation.phpt",
    "ext/standard/tests/strings/strcspn_variation7.phpt",
    "ext/standard/tests/strings/convert_cyr_string_error.phpt",
    "ext/standard/tests/strings/vprintf_basic7.phpt",
    "ext/standard/tests/strings/vprintf_variation6.phpt",
    "ext/standard/tests/strings/htmlentities02.phpt",
    "ext/standard/tests/strings/stripos_error.phpt",
    "ext/standard/tests/strings/bug26819.phpt",
    "ext/standard/tests/strings/stripos_variation2.phpt",
    "ext/standard/tests/strings/vprintf_variation16.phpt",
    "ext/standard/tests/strings/chunk_split_variation3.phpt",
    "ext/standard/tests/strings/count_chars_variation2.phpt",
    "ext/standard/tests/strings/sprintf_basic1.phpt",
    "ext/standard/tests/strings/bug50052.phpt",
    "ext/standard/tests/strings/nl2br_variation4.phpt",
    "ext/standard/tests/strings/strripos_basic1.phpt",
    "ext/standard/tests/strings/vprintf_variation16_64bit.phpt",
    "ext/standard/tests/strings/vsprintf_basic7_64bit.phpt",
    "ext/standard/tests/strings/stripslashes_basic.phpt",
    "ext/standard/tests/strings/sscanf_error.phpt",
    "ext/standard/tests/strings/strtok_variation6.phpt",
    "ext/standard/tests/strings/bug54238.phpt",
    "ext/standard/tests/strings/strtr_variation2.phpt",
    "ext/standard/tests/strings/htmlentities23.phpt",
    "ext/standard/tests/strings/strip_tags_variation5.phpt",
    "ext/standard/tests/strings/printf_basic3.phpt",
    "ext/standard/tests/strings/vfprintf_variation11_64bit.phpt",
    "ext/standard/tests/strings/htmlentities01.phpt",
    "ext/standard/tests/strings/quoted_printable_encode_001.phpt",
    "ext/standard/tests/strings/vprintf_variation5.phpt",
    "ext/standard/tests/strings/strtr_variation3.phpt",
    "ext/standard/tests/strings/strncasecmp_variation4.phpt",
    "ext/standard/tests/strings/strcspn_variation8.phpt",
    "ext/standard/tests/strings/htmlentities22.phpt",
    "ext/standard/tests/strings/sprintf_variation44.phpt",
    "ext/standard/tests/strings/bug21453.phpt",
    "ext/standard/tests/strings/str_pad_variation2.phpt",
    "ext/standard/tests/strings/dirname_basic.phpt",
    "ext/standard/tests/strings/crc32_error.phpt",
    "ext/standard/tests/strings/bug45485.phpt",
    "ext/standard/tests/strings/chop_variation2.phpt",
    "ext/standard/tests/strings/bug51059.phpt",
    "ext/standard/tests/strings/get_html_translation_table_basic3.phpt",
    "ext/standard/tests/strings/bug53319.phpt",
    "ext/standard/tests/strings/bug54454.phpt",
    "ext/standard/tests/strings/strtoupper.phpt",
    "ext/standard/tests/strings/vfprintf_error1.phpt",
    "ext/standard/tests/strings/sprintf_variation13.phpt",
    "ext/standard/tests/strings/sha1_basic.phpt",
    "ext/standard/tests/strings/explode_variation3.phpt",
    "ext/standard/tests/strings/vfprintf_basic1.phpt",
    "ext/standard/tests/strings/sprintf_variation23.phpt",
    "ext/standard/tests/strings/vprintf_basic6.phpt",
    "ext/standard/tests/strings/hebrevc_basic.phpt",
    "ext/standard/tests/strings/quoted_printable_decode_variation1.phpt",
    "ext/standard/tests/strings/vfprintf_basic7.phpt",
    "ext/standard/tests/strings/chunk_split_error.phpt",
    "ext/standard/tests/strings/bug44703.phpt",
    "ext/standard/tests/strings/strrpos_variation5.phpt",
    "ext/standard/tests/strings/bug20169.phpt",
    "ext/standard/tests/strings/fprintf_variation_009.phpt",
    "ext/standard/tests/strings/bug46578.phpt",
    "ext/standard/tests/strings/sprintf_variation16.phpt",
    "ext/standard/tests/strings/sprintf_error.phpt",
    "ext/standard/tests/strings/ucwords_variation2.phpt",
    "ext/standard/tests/strings/stripos_basic2.phpt",
    "ext/standard/tests/strings/strtr_variation7.phpt",
    "ext/standard/tests/strings/stripos_variation4.phpt",
    "ext/standard/tests/strings/strip_tags_error.phpt",
    "ext/standard/tests/strings/strrchr.phpt",
    "ext/standard/tests/strings/bug37262.phpt",
    "ext/standard/tests/strings/fprintf_variation_007.phpt",
    "ext/standard/tests/strings/strlen_variation1.phpt",
    "ext/standard/tests/strings/fprintf_error.phpt",
    "ext/standard/tests/strings/quoted_printable_encode_002.phpt",
    "ext/standard/tests/strings/fprintf_variation_004.phpt",
    "ext/standard/tests/strings/sprintf_variation35_64bit.phpt",
    "ext/standard/tests/strings/bug36392.phpt",
    "ext/standard/tests/strings/vsprintf_variation5.phpt",
    "ext/standard/tests/strings/number_format_error.phpt",
    "ext/standard/tests/strings/vsprintf_variation18.phpt",
    "ext/standard/tests/strings/dirname_variation.phpt",
    "ext/standard/tests/strings/sprintf_variation41.phpt",
    "ext/standard/tests/strings/strrchr_variation3.phpt",
    "ext/standard/tests/strings/ord_variation1.phpt",
    "ext/standard/tests/strings/strtr.phpt",
    "ext/standard/tests/strings/chr_variation1.phpt",
    "ext/standard/tests/strings/get_html_translation_table_basic7.phpt",
    "ext/standard/tests/strings/strrpos_error.phpt",
    "ext/standard/tests/strings/strncmp_variation1.phpt",
    "ext/standard/tests/strings/hebrev_variation1.phpt",
    "ext/standard/tests/strings/strstr.phpt",
    "ext/standard/tests/strings/sprintf_variation11.phpt",
    "ext/standard/tests/strings/vprintf_basic5.phpt",
    "ext/standard/tests/strings/htmlspecialchars_decode_variation4.phpt",
    "ext/standard/tests/strings/sprintf_variation24.phpt",
    "ext/standard/tests/strings/printf_variation2.phpt",
    "ext/standard/tests/strings/strripos_variation5.phpt",
    "ext/standard/tests/strings/strspn_variation1.phpt",
    "ext/standard/tests/strings/sprintf_basic7_64bit.phpt",
    "ext/standard/tests/strings/strspn.phpt",
    "ext/standard/tests/strings/trim_variation1.phpt",
    "ext/standard/tests/strings/str_pad_variation1.phpt",
    "ext/standard/tests/strings/stripos_variation5.phpt",
    "ext/standard/tests/strings/strrpos_basic2.phpt",
    "ext/standard/tests/strings/bug39621.phpt",
    "ext/standard/tests/strings/fprintf_variation_006_64bit.phpt",
    "ext/standard/tests/strings/quoted_printable_decode_basic.phpt",
    "ext/standard/tests/strings/chunk_split_variation6.phpt",
    "ext/standard/tests/strings/strtok_variation2.phpt",
    "ext/standard/tests/strings/strtok_error.phpt",
    "ext/standard/tests/strings/html_entity_decode_win1252.phpt",
    "ext/standard/tests/strings/strripos_offset.phpt",
    "ext/standard/tests/strings/006.phpt",
    "ext/standard/tests/strings/str_replace.phpt",
    "ext/standard/tests/strings/html_entity_decode2.phpt",
    "ext/standard/tests/strings/strncasecmp_variation11.phpt",
    "ext/standard/tests/strings/vprintf_variation14_64bit.phpt",
    "ext/standard/tests/strings/setlocale_basic1.phpt",
    "ext/standard/tests/strings/ltrim_error.phpt",
    "ext/standard/tests/strings/stripos.phpt",
    "ext/standard/tests/strings/strripos_basic2.phpt",
    "ext/standard/tests/strings/sprintf_variation28.phpt",
    "ext/standard/tests/strings/strncasecmp_variation7.phpt",
    "ext/standard/tests/strings/convert_cyr_string.phpt",
    "ext/standard/tests/strings/implode.phpt",
    "ext/standard/tests/strings/unpack_error.phpt",
    "ext/standard/tests/strings/sprintf_variation22.phpt",
    "ext/standard/tests/strings/sprintf_variation27.phpt",
    "ext/standard/tests/strings/ord_basic.phpt",
    "ext/standard/tests/strings/join_variation5.phpt",
    "ext/standard/tests/strings/setlocale_variation2.phpt",
    "ext/standard/tests/strings/sprintf_variation46.phpt",
    "ext/standard/tests/strings/bug27276.phpt",
    "ext/standard/tests/strings/vsprintf_basic2.phpt",
    "ext/standard/tests/strings/strcspn_basic.phpt",
    "ext/standard/tests/strings/sprintf_variation30.phpt",
    "ext/standard/tests/strings/bug40915.phpt",
    "ext/standard/tests/strings/sprintf_variation39.phpt",
    "ext/standard/tests/strings/stripcslashes_basic.phpt",
    "ext/standard/tests/strings/substr_replace_error.phpt",
    "ext/standard/tests/strings/explode_variation2.phpt",
    "ext/standard/tests/strings/htmlentities13.phpt",
    "ext/standard/tests/strings/explode_variation6.phpt",
    "ext/standard/tests/strings/strrev_variation4.phpt",
    "ext/standard/tests/strings/sprintf_variation5.phpt",
    "ext/standard/tests/strings/strncmp_variation2.phpt",
    "ext/standard/tests/strings/vsprintf_variation14.phpt",
    "ext/standard/tests/strings/vsprintf_basic3.phpt",
    "ext/standard/tests/strings/strrchr_error.phpt",
    "ext/standard/tests/strings/vprintf_variation10.phpt",
    "ext/standard/tests/strings/strncasecmp_variation8.phpt",
    "ext/standard/tests/strings/vfprintf_variation15.phpt",
    "ext/standard/tests/strings/convert_uudecode_variation1.phpt",
    "ext/standard/tests/strings/stripos_basic1.phpt",
    "ext/standard/tests/strings/hebrev_variation2.phpt",
    "ext/standard/tests/strings/str_split_basic.phpt",
    "ext/standard/tests/strings/vsprintf_variation1.phpt",
    "ext/standard/tests/strings/vsprintf_variation9.phpt",
    "ext/standard/tests/strings/htmlentities16.phpt",
    "ext/standard/tests/strings/sprintf_basic8_64bit.phpt",
    "ext/standard/tests/strings/vsprintf_variation6.phpt",
    "ext/standard/tests/strings/sprintf_variation35.phpt",
    "ext/standard/tests/strings/sscanf_basic1.phpt",
    "ext/standard/tests/strings/stripslashes_variation5.phpt",
    "ext/standard/tests/strings/bug24312.phpt",
    "ext/standard/tests/strings/htmlentities03.phpt",
    "ext/standard/tests/strings/chop_variation3.phpt",
    "ext/standard/tests/strings/bug35817.phpt",
    "ext/standard/tests/strings/vprintf_variation4_64bit.phpt",
    "ext/standard/tests/strings/soundex.phpt",
    "ext/standard/tests/strings/strrpos_variation8.phpt",
    "ext/standard/tests/strings/parse_str_basic1.phpt",
    "ext/standard/tests/strings/vfprintf_variation13_64bit.phpt",
    "ext/standard/tests/strings/str_split_variation2.phpt",
    "ext/standard/tests/strings/sha1_variation2.phpt",
    "ext/standard/tests/strings/vsprintf_variation12_64bit.phpt",
    "ext/standard/tests/strings/strncmp_variation3.phpt",
    "ext/standard/tests/strings/sprintf_variation41_64bit.phpt",
    "ext/standard/tests/strings/strspn_variation2.phpt",
    "ext/standard/tests/strings/bin2hex_error.phpt",
    "ext/standard/tests/strings/strpbrk_error.phpt",
    "ext/standard/tests/strings/ucwords_error.phpt",
    "ext/standard/tests/strings/strncmp_variation5.phpt",
    "ext/standard/tests/strings/strrchr_variation6.phpt",
    "ext/standard/tests/strings/explode1.phpt",
    "ext/standard/tests/strings/vprintf_variation4.phpt",
    "ext/standard/tests/strings/strncasecmp_variation6.phpt",
    "ext/standard/tests/strings/chunk_split_variation2.phpt",
    "ext/standard/tests/strings/vsprintf_basic7.phpt",
    "ext/standard/tests/strings/explode_variation1.phpt",
    "ext/standard/tests/strings/strrpos_variation14.phpt",
    "ext/standard/tests/strings/sprintf_basic6.phpt",
    "ext/standard/tests/strings/money_format_error.phpt",
    "ext/standard/tests/strings/rtrim_error.phpt",
    "ext/standard/tests/strings/fprintf_variation_003_64bit.phpt",
    "ext/standard/tests/strings/htmlentities_html5.phpt",
    "ext/standard/tests/strings/explode_variation5.phpt",
    "ext/standard/tests/strings/vsprintf_variation10.phpt",
    "ext/standard/tests/strings/get_html_translation_table_variation2.phpt",
    "ext/standard/tests/strings/crc32_variation3.phpt",
    "ext/standard/tests/strings/addslashes_basic.phpt",
    "ext/standard/tests/strings/strrev_variation2.phpt",
    "ext/standard/tests/strings/setlocale_basic2.phpt",
    "ext/standard/tests/strings/md5_error.phpt",
    "ext/standard/tests/strings/sprintf_variation53.phpt",
    "ext/standard/tests/strings/printf_basic9.phpt",
    "ext/standard/tests/strings/sscanf_basic5.phpt",
    "ext/standard/tests/strings/007.phpt",
    "ext/standard/tests/strings/sprintf_basic3.phpt",
    "ext/standard/tests/strings/sprintf_variation15.phpt",
    "ext/standard/tests/strings/sprintf_variation26.phpt",
    "ext/standard/tests/strings/stristr_error.phpt",
    "ext/standard/tests/strings/strncasecmp_variation5.phpt",
    "ext/standard/tests/strings/get_meta_tags.phpt",
    "ext/standard/tests/strings/strval_basic.phpt",
    "ext/standard/tests/strings/html_entity_decode3.phpt",
    "ext/standard/tests/strings/strcoll_error.phpt",
    "ext/standard/tests/strings/convert_uuencode_error.phpt",
    "ext/standard/tests/strings/nl2br_variation2.phpt",
    "ext/standard/tests/strings/bug27278.phpt",
    "ext/standard/tests/strings/strcasecmp.phpt",
    "ext/standard/tests/strings/bin2hex_basic.phpt",
    "ext/standard/tests/strings/ltrim_basic.phpt",
    "ext/standard/tests/strings/vsprintf_variation11.phpt",
    "ext/standard/tests/strings/sprintf_basic9.phpt",
    "ext/standard/tests/strings/vfprintf_variation16.phpt",
    "ext/standard/tests/strings/crc32_variation4.phpt",
    "ext/standard/tests/strings/vfprintf_basic6.phpt",
    "ext/standard/tests/strings/strrchr_variation9.phpt",
    "ext/standard/tests/strings/bug40637.phpt",
    "ext/standard/tests/strings/pack.phpt",
    "ext/standard/tests/strings/strip_tags_variation11.phpt",
    "ext/standard/tests/strings/vprintf_variation11.phpt",
    "ext/standard/tests/strings/strval_variation2.phpt",
    "ext/standard/tests/strings/chunk_split_variation1.phpt",
    "ext/standard/tests/strings/strrpos_variation9.phpt",
    "ext/standard/tests/strings/strspn_variation3.phpt",
    "ext/standard/tests/strings/strrev_basic.phpt",
    "ext/standard/tests/strings/str_split_variation6.phpt",
    "ext/standard/tests/strings/strrpos_basic1.phpt",
    "ext/standard/tests/strings/htmlspecialchars_decode_variation1.phpt",
    "ext/standard/tests/strings/htmlentities.phpt",
    "ext/standard/tests/strings/md5_basic1.phpt",
    "ext/standard/tests/strings/htmlentities20.phpt",
    "ext/standard/tests/strings/strspn_variation6.phpt",
    "ext/standard/tests/strings/stripos_variation11.phpt",
    "ext/standard/tests/strings/get_html_translation_table_basic8.phpt",
    "ext/standard/tests/strings/strip_tags_variation6.phpt",
    "ext/standard/tests/strings/print_basic.phpt",
    "ext/standard/tests/strings/str_split_variation5.phpt",
    "ext/standard/tests/strings/sprintf_variation50.phpt",
    "ext/standard/tests/strings/bug50847.phpt",
    "ext/standard/tests/strings/str_rot13_basic.phpt",
    "ext/standard/tests/strings/lcfirst.phpt",
    "ext/standard/tests/strings/vsprintf_variation19_64bit.phpt",
    "ext/standard/tests/strings/vfprintf_variation14.phpt",
    "ext/standard/tests/strings/strip_tags_basic2.phpt",
    "ext/standard/tests/strings/get_html_translation_table_basic6.phpt",
    "ext/standard/tests/strings/vfprintf_basic5.phpt",
    "ext/standard/tests/strings/fprintf_variation_006.phpt",
    "ext/standard/tests/strings/bug48709.phpt",
    "ext/standard/tests/strings/htmlentities19.phpt",
    "ext/standard/tests/strings/bug24281.phpt",
    "ext/standard/tests/strings/sscanf_basic7.phpt",
    "ext/standard/tests/strings/str_pad_variation5.phpt",
    "ext/standard/tests/strings/strtr_error.phpt",
    "ext/standard/tests/strings/sprintf_variation33.phpt",
    "ext/standard/tests/strings/strncasecmp_error.phpt",
    "ext/standard/tests/strings/bug33605.phpt",
    "ext/standard/tests/strings/vfprintf_variation3.phpt",
    "ext/standard/tests/strings/vsprintf_variation13.phpt",
    "ext/standard/tests/strings/fprintf_variation_003.phpt",
    "ext/standard/tests/strings/bug53021.phpt",
    "ext/standard/tests/strings/vsprintf_variation15.phpt",
    "ext/standard/tests/strings/vfprintf_variation4.phpt",
    "ext/standard/tests/strings/ucwords_basic.phpt",
    "ext/standard/tests/strings/setlocale_basic3.phpt",
    "ext/standard/tests/strings/substr_count_basic.phpt",
    "ext/standard/tests/strings/ucwords_variation1.phpt",
    "ext/standard/tests/strings/str_pad_variation3.phpt",
    "ext/standard/tests/strings/rtrim_basic.phpt",
    "ext/standard/tests/strings/strtolower-win32.phpt",
    "ext/standard/tests/strings/addcslashes_003.phpt",
    "ext/standard/tests/strings/similar_text_basic.phpt",
    "ext/standard/tests/strings/sprintf_variation4.phpt",
    "ext/standard/tests/strings/show_source_variation1.phpt",
    "ext/standard/tests/strings/addslashes_error.phpt",
    "ext/standard/tests/strings/str_shuffle.phpt",
    "ext/standard/tests/strings/vfprintf_variation10.phpt",
    "ext/standard/tests/strings/sprintf_variation17.phpt",
    "ext/standard/tests/strings/sscanf_basic8.phpt",
    "ext/standard/tests/strings/strrpos_variation4.phpt",
    "ext/standard/tests/strings/ucfirst.phpt",
    "ext/standard/tests/strings/strripos_variation3.phpt",
    "ext/standard/tests/strings/stristr.phpt",
    "ext/standard/tests/strings/sprintf_variation19.phpt",
    "ext/standard/tests/strings/strnatcmp_error.phpt",
    "ext/standard/tests/strings/vfprintf_error4.phpt",
    "ext/standard/tests/strings/vfprintf_variation15_64bit.phpt",
    "ext/standard/tests/strings/htmlspecialchars_basic.phpt",
    "ext/standard/tests/strings/vprintf_variation13.phpt",
    "ext/standard/tests/strings/add-and-stripcslashes.phpt",
    "ext/standard/tests/strings/stripslashes_variation2.phpt",
    "ext/standard/tests/strings/stripslashes_variation3.phpt",
    "ext/standard/tests/strings/explode.phpt",
    "ext/standard/tests/strings/vfprintf_variation6.phpt",
    "ext/standard/tests/strings/bug27457.phpt",
    "ext/standard/tests/strings/strtok_variation5.phpt",
    "ext/standard/tests/strings/strtok_variation4.phpt",
    "ext/standard/tests/strings/nl2br_variation3.phpt",
    "ext/standard/tests/strings/vsprintf_basic9.phpt",
    "ext/standard/tests/strings/sprintf_variation37.phpt",
    "ext/standard/tests/strings/str_split_variation2_64bit.phpt",
    "ext/standard/tests/strings/chunk_split_variation5.phpt",
    "ext/standard/tests/strings/htmlentities10.phpt",
    "ext/standard/tests/strings/show_source_variation2.phpt",
    "ext/standard/tests/strings/str_split_variation7_64bit.phpt",
    "ext/standard/tests/strings/vprintf_variation13_64bit.phpt",
    "ext/standard/tests/strings/chunk_split_variation7.phpt",
    "ext/standard/tests/strings/bug36148.phpt",
    "ext/standard/tests/strings/strlen_basic.phpt",
    "ext/standard/tests/strings/strnatcasecmp_basic.phpt",
    "ext/standard/tests/strings/unpack.phpt",
    "ext/standard/tests/strings/printf_error.phpt",
    "ext/standard/tests/strings/strrchr_variation2.phpt",
    "ext/standard/tests/strings/sprintf_variation36.phpt",
    "ext/standard/tests/strings/strcspn_variation1.phpt",
    "ext/standard/tests/strings/bug23650.phpt",
    "ext/standard/tests/strings/strrchr_basic.phpt",
    "ext/standard/tests/strings/chunk_split_variation13.phpt",
    "ext/standard/tests/strings/bug40754.phpt",
    "ext/standard/tests/strings/strrpos_variation3.phpt",
    "ext/standard/tests/strings/007-win32.phpt",
    "ext/standard/tests/strings/html_entity_decode1.phpt",
    "ext/standard/tests/strings/dirname_error.phpt",
    "ext/standard/tests/strings/sprintf_variation40.phpt",
    "ext/standard/tests/strings/trim_basic.phpt",
    "ext/standard/tests/strings/bug20934.phpt",
    "ext/standard/tests/strings/strncmp_variation7.phpt",
    "ext/standard/tests/strings/join_variation3.phpt",
    "ext/standard/tests/strings/levenshtein.phpt",
    "ext/standard/tests/strings/htmlspecialchars_decode_variation7.phpt",
    "ext/standard/tests/strings/money_format_variation1.phpt",
    "ext/standard/tests/strings/sprintf_variation1.phpt",
    "ext/standard/tests/strings/sha1_error.phpt",
    "ext/standard/tests/strings/number_format_basic.phpt",
    "ext/standard/tests/strings/wordwrap_variation3.phpt",
    "ext/standard/tests/strings/htmlentities15.phpt",
    "ext/standard/tests/strings/stripos_variation8.phpt",
    "ext/standard/tests/strings/006-win32.phpt",
    "ext/standard/tests/strings/bug47168.phpt",
    "ext/standard/tests/strings/fprintf_variation_001.phpt",
    "ext/standard/tests/strings/md5.phpt",
    "ext/standard/tests/strings/strrchr_variation12.phpt",
    "ext/standard/tests/strings/parse_str_basic2.phpt",
    "ext/standard/tests/strings/strnatcasecmp_error.phpt",
    "ext/standard/tests/strings/stripos_variation3.phpt",
    "ext/standard/tests/strings/sprintf_variation45.phpt",
    "ext/standard/tests/strings/print_variation1.phpt",
    "ext/standard/tests/strings/strrpos_variation7.phpt",
    "ext/standard/tests/strings/strip_tags_variation8.phpt",
    "ext/standard/tests/strings/vsprintf_variation13_64bit.phpt",
    "ext/standard/tests/strings/vfprintf_variation5.phpt",
    "ext/standard/tests/strings/strcspn_variation2.phpt",
    "ext/standard/tests/strings/strripos_variation1.phpt",
    "ext/standard/tests/strings/bug39873.phpt",
    "ext/standard/tests/strings/bug25671.phpt",
    "ext/standard/tests/strings/sprintf_variation31.phpt",
    "ext/standard/tests/strings/bug21730.phpt",
    "ext/standard/tests/strings/vsprintf_basic5.phpt",
    "ext/standard/tests/strings/vfprintf_variation12_64bit.phpt",
    "ext/standard/tests/strings/str_getcsv_001.phpt",
    "ext/standard/tests/strings/sprintf_variation38.phpt",
    "ext/standard/tests/strings/htmlentities_html4.phpt",
    "ext/standard/tests/strings/stristr2.phpt",
    "ext/standard/tests/strings/htmlentities05.phpt",
    "ext/standard/tests/strings/bug44242.phpt",
    "ext/standard/tests/strings/wordwrap_error.phpt",
    "ext/standard/tests/strings/strspn_variation10.phpt",
    "ext/standard/tests/strings/strncasecmp_variation1.phpt",
    "ext/standard/tests/strings/strcoll.phpt",
    "ext/standard/tests/strings/vsprintf_variation17.phpt",
    "ext/standard/tests/strings/str_shuffle_basic.phpt",
    "ext/standard/tests/strings/htmlentities21.phpt",
    "ext/standard/tests/strings/chunk_split_variation9.phpt",
    "ext/standard/tests/strings/strval.phpt",
    "ext/standard/tests/strings/strrchr_variation4.phpt",
    "ext/standard/tests/strings/stristr_variation2.phpt",
    "ext/standard/tests/strings/vsprintf_variation16.phpt",
    "ext/standard/tests/strings/strnatcasecmp_variation1.phpt",
    "ext/standard/tests/strings/str_replace_variation1.phpt",
    "ext/standard/tests/strings/htmlspecialchars_decode_variation2.phpt",
    "ext/standard/tests/strings/strtok_basic.phpt",
    "ext/standard/tests/strings/htmlentities08.phpt",
    "ext/standard/tests/strings/strrev_variation1.phpt",
    "ext/standard/tests/strings/htmlspecialchars_decode_variation6.phpt",
    "ext/standard/tests/strings/bug40704.phpt",
    "ext/standard/tests/strings/join_variation6.phpt",
    "ext/standard/tests/strings/sha1raw.phpt",
    "ext/standard/tests/strings/count_chars_basic.phpt",
    "ext/standard/tests/strings/htmlentities11.phpt",
    "ext/standard/tests/strings/bug24208.phpt",
    "ext/standard/tests/strings/strtolower.phpt",
    "ext/standard/tests/strings/printf_basic6.phpt",
    "ext/standard/tests/strings/highlight_file.phpt",
    "ext/standard/tests/strings/quotemeta_error.phpt",
    "ext/standard/tests/strings/chop_variation5.phpt",
    "ext/standard/tests/strings/add-and-stripslashes.phpt",
    "ext/standard/tests/strings/crc32.phpt",
    "ext/standard/tests/strings/strncasecmp_variation9.phpt",
    "ext/standard/tests/strings/html_entity_decode_html5.phpt",
    "ext/standard/tests/strings/quoted_printable_decode_error.phpt",
    "ext/standard/tests/strings/strcspn_variation12.phpt",
    "ext/standard/tests/strings/strrchr_variation11.phpt",
    "ext/standard/tests/strings/str_split_error.phpt",
    "ext/standard/tests/strings/strip_tags_variation1.phpt",
    "ext/standard/tests/strings/sha1_variation1.phpt",
    "ext/standard/tests/strings/addcslashes_001.phpt",
    "ext/standard/tests/strings/html_entity_decode_iso8859-15.phpt",
    "ext/standard/tests/strings/strncmp_error.phpt",
    "ext/standard/tests/strings/moneyformat.phpt",
    "ext/standard/tests/strings/stripos_variation13.phpt",
    "ext/standard/tests/strings/crc32_basic.phpt",
    "ext/standard/tests/strings/strrpos_variation13.phpt",
    "ext/standard/tests/strings/bug33076.phpt",
    "ext/standard/tests/strings/wordwrap_variation4.phpt",
    "ext/standard/tests/strings/vfprintf_basic7_64bit.phpt",
    "ext/standard/tests/strings/vsprintf_variation12.phpt",
    "ext/standard/tests/strings/sprintf_basic4.phpt",
    "ext/standard/tests/strings/str_pad_variation4.phpt",
    "ext/standard/tests/strings/bug29538.phpt",
    "ext/standard/tests/strings/printf_basic4.phpt",
    "ext/standard/tests/strings/vprintf_basic8.phpt",
    "ext/standard/tests/strings/html_entity_decode_html4.phpt",
    "ext/standard/tests/strings/stripos_variation7.phpt",
    "ext/standard/tests/strings/fprintf_variation_008.phpt",
    "ext/standard/tests/strings/bin2hex_variation1.phpt",
    "ext/standard/tests/strings/htmlspecialchars.phpt",
    "ext/standard/tests/strings/count_chars_error.phpt",
    "ext/standard/tests/strings/crypt_sha256.phpt",
    "ext/standard/tests/strings/bug47443.phpt",
    "ext/standard/tests/strings/hebrevc_variation1.phpt",
    "ext/standard/tests/strings/hebrev_basic.phpt",
    "ext/standard/tests/strings/htmlspecialchars_decode_basic.phpt",
    "ext/standard/tests/strings/bug51899.phpt",
    "ext/standard/tests/strings/ord_error.phpt",
    "ext/standard/tests/strings/bug20927.phpt",
    "ext/standard/tests/strings/sprintf_variation29_64bit.phpt",
    "ext/standard/tests/strings/join_basic.phpt",
    "ext/standard/tests/strings/strncmp_variation9.phpt",
    "ext/standard/tests/strings/strcspn_variation5.phpt",
    "ext/standard/tests/strings/htmlspecialchars_decode_variation3.phpt",
    "ext/standard/tests/strings/strpbrk_basic.phpt",
    "ext/standard/tests/strings/stristr_basic.phpt",
    "ext/standard/tests/strings/ltrim.phpt",
    "ext/standard/tests/strings/bug54322.phpt",
    "ext/standard/tests/strings/bug43927.phpt",
    "ext/standard/tests/strings/chr_basic.phpt",
    "ext/standard/tests/strings/vfprintf_basic9.phpt",
    "ext/standard/tests/strings/sprintf_variation21.phpt",
    "ext/standard/tests/strings/strtr_basic.phpt",
    "ext/standard/tests/strings/strspn_variation8.phpt",
    "ext/standard/tests/strings/strcmp.phpt",
    "ext/standard/tests/strings/bug20261.phpt",
    "ext/standard/tests/strings/implode1.phpt",
    "ext/standard/tests/strings/sscanf_basic6.phpt",
    "ext/standard/tests/strings/strip_tags_basic1.phpt",
    "ext/standard/tests/strings/vfprintf_variation18.phpt",
    "ext/standard/tests/strings/basename_error.phpt",
    "ext/standard/tests/strings/explode_bug.phpt",
    "ext/standard/tests/strings/get_html_translation_table_basic5.phpt",
    "ext/standard/tests/strings/substr_count_variation_002.phpt",
    "ext/standard/tests/strings/htmlentities04.phpt",
    "ext/standard/tests/strings/str_repeat.phpt",
    "ext/standard/tests/strings/sprintf_variation2.phpt",
    "ext/standard/tests/strings/vprintf_variation12.phpt",
    "ext/standard/tests/strings/htmlentities09.phpt",
    "ext/standard/tests/strings/vprintf_variation18.phpt",
    "ext/standard/tests/strings/sprintf_variation48.phpt",
    "ext/standard/tests/strings/bug25707.phpt",
    "ext/standard/tests/strings/strrchr_variation8.phpt",
    "ext/standard/tests/strings/strcspn_variation9.phpt",
    "ext/standard/tests/strings/bug24098.phpt",
    "ext/standard/tests/strings/explode_error.phpt",
    "ext/standard/tests/strings/soundex_basic.phpt",
    "ext/standard/tests/strings/strpos.phpt",
    "ext/standard/tests/strings/html_entity_decode_cp866.phpt",
    "ext/standard/tests/strings/strrev_error.phpt",
    "ext/standard/tests/strings/str_split_variation7.phpt",
    "ext/standard/tests/strings/str_word_count1.phpt",
    "ext/standard/tests/strings/join_variation4.phpt",
    "ext/standard/tests/strings/strrev.phpt",
    "ext/standard/tests/strings/bug47842.phpt",
    "ext/standard/tests/strings/strnatcmp_basic.phpt",
    "ext/standard/tests/strings/stripcslashes_variation1.phpt",
    "ext/standard/tests/strings/vprintf_variation3.phpt",
    "ext/standard/tests/strings/convert_uudecode_error.phpt",
    "ext/standard/tests/strings/md5_file.phpt",
    "ext/standard/tests/strings/vfprintf_variation4_64bit.phpt",
    "ext/standard/tests/strings/bug29075.phpt",
    "ext/standard/tests/strings/htmlentities-utf-3.phpt",
    "ext/standard/tests/strings/sprintf_f_2.phpt",
    "ext/standard/tests/strings/strripos_variation2.phpt",
    "ext/standard/tests/strings/strncmp_basic.phpt",
    "ext/standard/tests/strings/strripos_error.phpt",
    "ext/standard/tests/strings/str_split_variation3.phpt",
    "ext/standard/tests/strings/vprintf_basic4.phpt",
    "ext/standard/tests/strings/sprintf_variation28_64bit.phpt",
    "ext/standard/tests/strings/rtrim.phpt",
    "ext/standard/tests/strings/html_entity_decode_iso8859-5.phpt",
    "ext/standard/tests/strings/stripos_variation10.phpt",
    "ext/standard/tests/strings/join_error.phpt",
    "ext/standard/tests/strings/vprintf_variation2.phpt",
    "ext/standard/tests/strings/chunk_split_basic.phpt",
    "ext/standard/tests/strings/sprintf_variation34_64bit.phpt",
    "ext/standard/tests/strings/chunk_split_variation8.phpt",
    "ext/standard/tests/strings/vfprintf_variation7.phpt",
    "ext/standard/tests/strings/printf_basic8.phpt",
    "ext/standard/tests/strings/strip_tags_variation9.phpt",
    "ext/standard/tests/strings/sprintf_variation14.phpt",
    "ext/standard/tests/strings/htmlentities-utf-2.phpt",
    "ext/standard/tests/strings/sprintf_variation44_64bit.phpt",
    "ext/standard/tests/strings/printf_basic1.phpt",
    "ext/standard/tests/strings/htmlentities24.phpt",
    "ext/standard/tests/strings/vfprintf_variation11.phpt",
    "ext/standard/tests/strings/addslashes_variation1.phpt",
    "ext/standard/tests/strings/vfprintf_basic8.phpt",
    "ext/standard/tests/strings/bug36944.phpt",
    "ext/standard/tests/strings/bin2hex.phpt",
    "ext/standard/tests/strings/bug29119.phpt",
    "ext/standard/tests/strings/sprintf_variation6.phpt",
    "ext/standard/tests/strings/vsprintf_error.phpt",
    "ext/standard/tests/strings/join_variation1.phpt",
    "ext/standard/tests/strings/vsprintf_basic6.phpt",
    "ext/standard/tests/strings/get_html_translation_table_basic1.phpt",
    "ext/standard/tests/strings/fprintf_variation_008_64bit.phpt",
    "ext/standard/tests/strings/vfprintf_basic.phpt",
    "ext/standard/tests/strings/bug26817.phpt",
    "ext/standard/tests/strings/htmlentities14.phpt",
    "ext/standard/tests/strings/php_strip_whitespace.phpt",
    "ext/standard/tests/strings/vprintf_basic7_64bit.phpt",
    "ext/standard/tests/strings/md5raw.phpt",
    "ext/standard/tests/strings/chunk_split_variation12.phpt",
    "ext/standard/tests/strings/vfprintf_variation19_64bit.phpt",
    "ext/standard/tests/strings/wordwrap_variation2.phpt",
    "ext/standard/tests/strings/rtrim_variation1.phpt",
    "ext/standard/tests/strings/strrpos_variation6.phpt",
    "ext/standard/tests/strings/str_ireplace.phpt",
    "ext/standard/tests/strings/bug49785.phpt",
    "ext/standard/tests/strings/vprintf_basic9.phpt",
    "ext/standard/tests/strings/strcspn_variation6.phpt",
    "ext/standard/tests/strings/substr_compare.phpt",
    "ext/standard/tests/strings/vfprintf_variation17.phpt",
    "ext/standard/tests/strings/bug34214.phpt",
    "ext/standard/tests/strings/sprintf_variation25.phpt",
    "ext/standard/tests/strings/bug42208.phpt",
    "ext/standard/tests/strings/basename.phpt",
    "ext/standard/tests/strings/bug39032.phpt",
    "ext/standard/tests/strings/nl2br_variation1.phpt",
    "ext/standard/tests/strings/money_format_variation2.phpt",
    "ext/standard/tests/strings/strip_tags.phpt",
    "ext/standard/tests/strings/trim.phpt",
    "ext/standard/tests/strings/strtok_variation1.phpt",
    "ext/standard/tests/strings/bug26878.phpt",
    "ext/standard/tests/strings/html_entity_decode_win1251.phpt",
    "ext/standard/tests/strings/substr_count_variation_001.phpt",
    "ext/standard/tests/strings/metaphone.phpt",
    "ext/standard/tests/strings/basename_basic.phpt",
    "ext/standard/tests/strings/wordwrap_variation1.phpt",
    "ext/standard/tests/strings/stripslashes_error.phpt",
    "ext/standard/tests/strings/vsprintf_variation4.phpt",
    "ext/standard/tests/strings/sprintf_variation4_64bit.phpt",
    "ext/standard/tests/strings/sprintf_basic8.phpt",
    "ext/standard/tests/strings/hebrevc_error.phpt",
    "ext/standard/tests/strings/bug22008.phpt",
    "ext/standard/tests/strings/nl_langinfo_error1.phpt",
    "ext/standard/tests/strings/bug39350.phpt",
    "ext/standard/tests/strings/vfprintf_variation12.phpt",
    "ext/standard/tests/strings/vfprintf_variation13.phpt",
    "ext/standard/tests/strings/stripos_variation14.phpt",
    "ext/standard/tests/strings/strval_error.phpt",
    "ext/standard/tests/strings/stripslashes_variation4.phpt",
    "ext/standard/tests/general_functions/parse_ini_string_001.phpt",
    "ext/standard/tests/general_functions/escapeshellarg_error.phpt",
    "ext/standard/tests/general_functions/bug44295.phpt",
    "ext/standard/tests/general_functions/proc_nice_error.phpt",
    "ext/standard/tests/general_functions/bug44667.phpt",
    "ext/standard/tests/general_functions/getrusage_error.phpt",
    "ext/standard/tests/general_functions/php_uname_error.phpt",
    "ext/standard/tests/general_functions/callbacks_002.phpt",
    "ext/standard/tests/general_functions/is_callable_basic1.phpt",
    "ext/standard/tests/general_functions/bug29038.phpt",
    "ext/standard/tests/general_functions/gettype_settype_variation4.phpt",
    "ext/standard/tests/general_functions/var_export_basic1.phpt",
    "ext/standard/tests/general_functions/getopt_002.phpt",
    "ext/standard/tests/general_functions/005.phpt",
    "ext/standard/tests/general_functions/var_export_basic6.phpt",
    "ext/standard/tests/general_functions/gettype_settype_basic.phpt",
    "ext/standard/tests/general_functions/bug41445.phpt",
    "ext/standard/tests/general_functions/ob_get_flush_basic.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation4.phpt",
    "ext/standard/tests/general_functions/sys_getloadavg.phpt",
    "ext/standard/tests/general_functions/var_export_basic5.phpt",
    "ext/standard/tests/general_functions/bug44394_2.phpt",
    "ext/standard/tests/general_functions/get_defined_constants_error.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation9.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation11.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation6.phpt",
    "ext/standard/tests/general_functions/ob_start_closures.phpt",
    "ext/standard/tests/general_functions/phpcredits2.phpt",
    "ext/standard/tests/general_functions/import_request3.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation1.phpt",
    "ext/standard/tests/general_functions/intval_variation1.phpt",
    "ext/standard/tests/general_functions/gettype_settype_error.phpt",
    "ext/standard/tests/general_functions/get_cfg_var_error.phpt",
    "ext/standard/tests/general_functions/get_cfg_var_variation5.phpt",
    "ext/standard/tests/general_functions/bug49692.phpt",
    "ext/standard/tests/general_functions/usleep_basic.phpt",
    "ext/standard/tests/general_functions/getservbyname_basic.phpt",
    "ext/standard/tests/general_functions/getopt_004.phpt",
    "ext/standard/tests/general_functions/gettype_settype_variation6.phpt",
    "ext/standard/tests/general_functions/proc_nice_variation7.phpt",
    "ext/standard/tests/general_functions/is_callable_variation2.phpt",
    "ext/standard/tests/general_functions/bug43293_1.phpt",
    "ext/standard/tests/general_functions/var_export_basic3.phpt",
    "ext/standard/tests/general_functions/php_uname_basic.phpt",
    "ext/standard/tests/general_functions/sleep_basic.phpt",
    "ext/standard/tests/general_functions/get_extension_funcs_basic.phpt",
    "ext/standard/tests/general_functions/getopt_005.phpt",
    "ext/standard/tests/general_functions/bug41445_1.phpt",
    "ext/standard/tests/general_functions/is_null.phpt",
    "ext/standard/tests/general_functions/getservbyport_variation1.phpt",
    "ext/standard/tests/general_functions/phpinfo.phpt",
    "ext/standard/tests/general_functions/import_request1.phpt",
    "ext/standard/tests/general_functions/parse_ini_file-win32.phpt",
    "ext/standard/tests/general_functions/floatval_variation1.phpt",
    "ext/standard/tests/general_functions/getrusage_basic.phpt",
    "ext/standard/tests/general_functions/sleep_error.phpt",
    "ext/standard/tests/general_functions/001.phpt",
    "ext/standard/tests/general_functions/is_resource.phpt",
    "ext/standard/tests/general_functions/escapeshellarg_basic.phpt",
    "ext/standard/tests/general_functions/getopt_003.phpt",
    "ext/standard/tests/general_functions/get_cfg_var_variation4.phpt",
    "ext/standard/tests/general_functions/get_resource_type_error.phpt",
    "ext/standard/tests/general_functions/intval_variation2.phpt",
    "ext/standard/tests/general_functions/set_magic_quotes_runtime_variation1.phpt",
    "ext/standard/tests/general_functions/bug41518.phpt",
    "ext/standard/tests/general_functions/bug25038.phpt",
    "ext/standard/tests/general_functions/004.phpt",
    "ext/standard/tests/general_functions/bug27678.phpt",
    "ext/standard/tests/general_functions/parse_ini_string_002.phpt",
    "ext/standard/tests/general_functions/get_defined_vars_basic.phpt",
    "ext/standard/tests/general_functions/is_bool.phpt",
    "ext/standard/tests/general_functions/get_cfg_var_variation7.phpt",
    "ext/standard/tests/general_functions/var_export_error1.phpt",
    "ext/standard/tests/general_functions/is_scalar.phpt",
    "ext/standard/tests/general_functions/get_cfg_var_variation3.phpt",
    "ext/standard/tests/general_functions/proc_open02.phpt",
    "ext/standard/tests/general_functions/bug35229.phpt",
    "ext/standard/tests/general_functions/get_cfg_var_variation9.phpt",
    "ext/standard/tests/general_functions/call_user_func_return.phpt",
    "ext/standard/tests/general_functions/import_request.phpt",
    "ext/standard/tests/general_functions/rand.phpt",
    "ext/standard/tests/general_functions/usleep_error.phpt",
    "ext/standard/tests/general_functions/get_extension_funcs_error.phpt",
    "ext/standard/tests/general_functions/phpinfo2.phpt",
    "ext/standard/tests/general_functions/is_numeric.phpt",
    "ext/standard/tests/general_functions/bug31190.phpt",
    "ext/standard/tests/general_functions/highlight_heredoc.phpt",
    "ext/standard/tests/general_functions/is_array.phpt",
    "ext/standard/tests/general_functions/is_resource_error.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation10.phpt",
    "ext/standard/tests/general_functions/isset_basic2.phpt",
    "ext/standard/tests/general_functions/bug40752.phpt",
    "ext/standard/tests/general_functions/get_include_path_basic.phpt",
    "ext/standard/tests/general_functions/ob_get_flush_error.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation7.phpt",
    "ext/standard/tests/general_functions/003.phpt",
    "ext/standard/tests/general_functions/bug48660.phpt",
    "ext/standard/tests/general_functions/proc_nice_variation6.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation3.phpt",
    "ext/standard/tests/general_functions/get_resource_type_basic.phpt",
    "ext/standard/tests/general_functions/get_loaded_extensions_error.phpt",
    "ext/standard/tests/general_functions/bug48768.phpt",
    "ext/standard/tests/general_functions/bug47857.phpt",
    "ext/standard/tests/general_functions/get_resource_type_variation1.phpt",
    "ext/standard/tests/general_functions/proc_nice_basic.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation8.phpt",
    "ext/standard/tests/general_functions/getservbyport_error.phpt",
    "ext/standard/tests/general_functions/bug49056.phpt",
    "ext/standard/tests/general_functions/getmypid_basic.phpt",
    "ext/standard/tests/general_functions/debug_zval_dump_v.phpt",
    "ext/standard/tests/general_functions/get_cfg_var_variation8.phpt",
    "ext/standard/tests/general_functions/debug_zval_dump_o.phpt",
    "ext/standard/tests/general_functions/gettype_settype_variation1.phpt",
    "ext/standard/tests/general_functions/006.phpt",
    "ext/standard/tests/general_functions/getrusage.phpt",
    "ext/standard/tests/general_functions/floatval_basic.phpt",
    "ext/standard/tests/general_functions/debug_zval_dump_b.phpt",
    "ext/standard/tests/general_functions/escapeshellarg_variation1-win32.phpt",
    "ext/standard/tests/general_functions/get_included_files.phpt",
    "ext/standard/tests/general_functions/bug47027.phpt",
    "ext/standard/tests/general_functions/get_defined_constants_basic.phpt",
    "ext/standard/tests/general_functions/error_get_last.phpt",
    "ext/standard/tests/general_functions/bug43293_3.phpt",
    "ext/standard/tests/general_functions/is_callable_basic2.phpt",
    "ext/standard/tests/general_functions/getrusage_variation1.phpt",
    "ext/standard/tests/general_functions/callbacks_001.phpt",
    "ext/standard/tests/general_functions/is_float_64bit.phpt",
    "ext/standard/tests/general_functions/floatval.phpt",
    "ext/standard/tests/general_functions/print_r.phpt",
    "ext/standard/tests/general_functions/is_string.phpt",
    "ext/standard/tests/general_functions/is_object.phpt",
    "ext/standard/tests/general_functions/bug44394.phpt",
    "ext/standard/tests/general_functions/007.phpt",
    "ext/standard/tests/general_functions/var_dump.phpt",
    "ext/standard/tests/general_functions/is_int_64bit.phpt",
    "ext/standard/tests/general_functions/import_request2.phpt",
    "ext/standard/tests/general_functions/debug_zval_dump_e.phpt",
    "ext/standard/tests/general_functions/escapeshellarg_variation1.phpt",
    "ext/standard/tests/general_functions/call_user_func_array_variation_003.phpt",
    "ext/standard/tests/general_functions/gettype_settype_variation3.phpt",
    "ext/standard/tests/general_functions/intval.phpt",
    "ext/standard/tests/general_functions/002.phpt",
    "ext/standard/tests/general_functions/008.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation12.phpt",
    "ext/standard/tests/general_functions/get_cfg_var_variation2.phpt",
    "ext/standard/tests/general_functions/gettype_settype_variation7.phpt",
    "ext/standard/tests/general_functions/include_path.phpt",
    "ext/standard/tests/general_functions/get_loaded_extensions_basic.phpt",
    "ext/standard/tests/general_functions/putenv.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation14.phpt",
    "ext/standard/tests/general_functions/is_int.phpt",
    "ext/standard/tests/general_functions/get_magic_quotes_runtime.phpt",
    "ext/standard/tests/general_functions/phpcredits.phpt",
    "ext/standard/tests/general_functions/escapeshellarg_basic-win32.phpt",
    "ext/standard/tests/general_functions/getopt.phpt",
    "ext/standard/tests/general_functions/var_export_error3.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation13.phpt",
    "ext/standard/tests/general_functions/floatval_error.phpt",
    "ext/standard/tests/general_functions/set_magic_quotes_runtime_error.phpt",
    "ext/standard/tests/general_functions/var_export_basic8.phpt",
    "ext/standard/tests/general_functions/bug50732.phpt",
    "ext/standard/tests/general_functions/type.phpt",
    "ext/standard/tests/general_functions/php_uname_variation1.phpt",
    "ext/standard/tests/general_functions/uniqid_error.phpt",
    "ext/standard/tests/general_functions/bug46587.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation2.phpt",
    "ext/standard/tests/general_functions/head.phpt",
    "ext/standard/tests/general_functions/var_export_basic4.phpt",
    "ext/standard/tests/general_functions/strval.phpt",
    "ext/standard/tests/general_functions/print_r_64bit.phpt",
    "ext/standard/tests/general_functions/gettype_settype_variation5.phpt",
    "ext/standard/tests/general_functions/bug43293_2.phpt",
    "ext/standard/tests/general_functions/bug34794.phpt",
    "ext/standard/tests/general_functions/var_dump_64bit.phpt",
    "ext/standard/tests/general_functions/var_export-locale.phpt",
    "ext/standard/tests/general_functions/get_extension_funcs_variation.phpt",
    "ext/standard/tests/general_functions/isset_basic1.phpt",
    "ext/standard/tests/general_functions/proc_nice_variation5.phpt",
    "ext/standard/tests/general_functions/getservbyname_error.phpt",
    "ext/standard/tests/general_functions/proc_nice_variation2.phpt",
    "ext/standard/tests/general_functions/bug36011.phpt",
    "ext/standard/tests/general_functions/var_export_basic7.phpt",
    "ext/standard/tests/general_functions/bug41037.phpt",
    "ext/standard/tests/general_functions/set_magic_quotes_runtime_basic.phpt",
    "ext/standard/tests/general_functions/call_user_method.phpt",
    "ext/standard/tests/general_functions/var_export_basic2.phpt",
    "ext/standard/tests/general_functions/dl-cve-2007-4887.phpt",
    "ext/standard/tests/general_functions/bug47859.phpt",
    "ext/standard/tests/general_functions/proc_open.phpt",
    "ext/standard/tests/general_functions/bug42272.phpt",
    "ext/standard/tests/general_functions/is_resource_basic.phpt",
    "ext/standard/tests/general_functions/parse_ini_booleans.phpt",
    "ext/standard/tests/general_functions/gettype_settype_variation8.phpt",
    "ext/standard/tests/general_functions/ini_get_all.phpt",
    "ext/standard/tests/general_functions/get_magic_quotes_gpc.phpt",
    "ext/standard/tests/general_functions/is_callable_variation1.phpt",
    "ext/standard/tests/general_functions/bug50690.phpt",
    "ext/standard/tests/general_functions/closures_002.phpt",
    "ext/standard/tests/general_functions/getservbyname_variation5.phpt",
    "ext/standard/tests/general_functions/call_user_func_array_variation_001.phpt",
    "ext/standard/tests/general_functions/is_float.phpt",
    "ext/standard/tests/general_functions/get_cfg_var_basic.phpt",
    "ext/standard/tests/general_functions/proc_nice_variation3.phpt",
    "ext/standard/tests/general_functions/var_export_error2.phpt",
    "ext/standard/tests/general_functions/escapeshellcmd-win32.phpt",
    "ext/standard/tests/general_functions/parse_ini_basic.phpt",
    "ext/standard/tests/general_functions/bug52138.phpt",
    "ext/standard/tests/general_functions/get_cfg_var_variation6.phpt",
    "ext/standard/tests/general_functions/009.phpt",
    "ext/standard/tests/general_functions/parse_ini_file.phpt",
    "ext/standard/tests/general_functions/proc_nice_variation1.phpt",
    "ext/standard/tests/general_functions/bug40398.phpt",
    "ext/standard/tests/general_functions/is_callable_error.phpt",
    "ext/standard/tests/general_functions/gettype_settype_variation2.phpt",
    "ext/standard/tests/general_functions/ob_get_length_basic.phpt",
    "ext/standard/tests/general_functions/call_user_func_array_variation_002.phpt",
    "ext/standard/tests/general_functions/get_cfg_var_variation1.phpt",
    "ext/standard/tests/general_functions/bug49847.phpt",
    "ext/standard/tests/general_functions/debug_zval_dump_b_64bit.phpt",
    "ext/standard/tests/general_functions/bug44461.phpt",
    "ext/standard/tests/general_functions/getservbyport_basic.phpt",
    "ext/standard/tests/general_functions/bug41970.phpt",
    "ext/standard/tests/general_functions/bug44487.phpt",
    "ext/standard/tests/image/image_type_to_mime_type_variation3.phpt",
    "ext/standard/tests/image/image_type_to_extension.phpt",
    "ext/standard/tests/image/image_type_to_mime_type.phpt",
    "ext/standard/tests/image/getimagesize_variation3.phpt",
    "ext/standard/tests/image/getimagesize_error1.phpt",
    "ext/standard/tests/image/getimagesize_swc.phpt",
    "ext/standard/tests/image/getimagesize_variation_005.phpt",
    "ext/standard/tests/image/getimagesize_246x247.phpt",
    "ext/standard/tests/image/getimagesize_tif_mm.phpt",
    "ext/standard/tests/image/iptcembed_001.phpt",
    "ext/standard/tests/image/getimagesize_384x385.phpt",
    "ext/standard/tests/image/getimagesize_variation_003.phpt",
    "ext/standard/tests/image/bug13213.phpt",
    "ext/standard/tests/image/getimagesize_variation1.phpt",
    "ext/standard/tests/image/getimagesizefromstring1.phpt",
    "ext/standard/tests/image/getimagesize_variation4.phpt",
    "ext/standard/tests/image/getimagesize_xbm.phpt",
    "ext/standard/tests/image/getimagesize_wbmp.phpt",
    "ext/standard/tests/image/getimagesize_jpgapp.phpt",
    "ext/standard/tests/image/getimagesize_basic.phpt",
    "ext/standard/tests/image/getimagesize.phpt",
    "ext/standard/tests/image/getimagesize_variation2.phpt",
    "ext/standard/tests/class_object/get_parent_class_error_001.phpt",
    "ext/standard/tests/class_object/get_class_vars_error.phpt",
    "ext/standard/tests/class_object/is_a_error_001.phpt",
    "ext/standard/tests/class_object/get_class_methods_basic_002.phpt",
    "ext/standard/tests/class_object/trait_exists_basic_001.phpt",
    "ext/standard/tests/class_object/class_exists_variation_002.phpt",
    "ext/standard/tests/class_object/is_a_variation_002.phpt",
    "ext/standard/tests/class_object/get_declared_traits_error_001.phpt",
    "ext/standard/tests/class_object/get_declared_classes_variation1.phpt",
    "ext/standard/tests/class_object/is_subclass_of_variation_004.phpt",
    "ext/standard/tests/class_object/interface_exists_error.phpt",
    "ext/standard/tests/class_object/get_class_methods_basic_001.phpt",
    "ext/standard/tests/class_object/is_subclass_of_variation_002.phpt",
    "ext/standard/tests/class_object/get_class_variation_002.phpt",
    "ext/standard/tests/class_object/class_exists_variation_003.phpt",
    "ext/standard/tests/class_object/is_subclass_of_variation_001.phpt",
    "ext/standard/tests/class_object/get_declared_classes_error_001.phpt",
    "ext/standard/tests/class_object/get_class_error_001.phpt",
    "ext/standard/tests/class_object/forward_static_call_002.phpt",
    "ext/standard/tests/class_object/get_parent_class_variation_002.phpt",
    "ext/standard/tests/class_object/interface_exists_variation1.phpt",
    "ext/standard/tests/class_object/method_exists_error_001.phpt",
    "ext/standard/tests/class_object/trait_class_exists_variation_003.phpt",
    "ext/standard/tests/class_object/trait_exists_variation_002.phpt",
    "ext/standard/tests/class_object/get_declared_classes_basic_001.phpt",
    "ext/standard/tests/class_object/get_object_vars_variation_003.phpt",
    "ext/standard/tests/class_object/get_class_methods_variation_002.phpt",
    "ext/standard/tests/class_object/get_class_vars_variation1.phpt",
    "ext/standard/tests/class_object/method_exists_basic_003.phpt",
    "ext/standard/tests/class_object/class_exists_error_001.phpt",
    "ext/standard/tests/class_object/interface_exists_variation2.phpt",
    "ext/standard/tests/class_object/trait_exists_variation_001.phpt",
    "ext/standard/tests/class_object/forward_static_call_001.phpt",
    "ext/standard/tests/class_object/get_parent_class_variation_001.phpt",
    "ext/standard/tests/class_object/get_declared_traits_variation1.phpt",
    "ext/standard/tests/class_object/is_subclass_of_variation_003.phpt",
    "ext/standard/tests/class_object/get_object_vars_basic_001.phpt",
    "ext/standard/tests/class_object/get_object_vars_basic_002.phpt",
    "ext/standard/tests/class_object/is_a_variation_001.phpt",
    "ext/standard/tests/class_object/class_exists_variation_001.phpt",
    "ext/standard/tests/class_object/get_object_vars_variation_001.phpt",
    "ext/standard/tests/class_object/get_declared_interfaces_basic_001.phpt",
    "ext/standard/tests/class_object/get_declared_interfaces_error_001.phpt",
    "ext/standard/tests/class_object/class_exists_basic_001.phpt",
    "ext/standard/tests/class_object/trait_exists_error_001.phpt",
    "ext/standard/tests/class_object/interface_exists_variation3.phpt",
    "ext/standard/tests/class_object/method_exists_basic_002.phpt",
    "ext/standard/tests/class_object/method_exists_basic_001.phpt",
    "ext/standard/tests/class_object/property_exists_variation1.phpt",
    "ext/standard/tests/class_object/is_subclass_of_error_001.phpt",
    "ext/standard/tests/class_object/method_exists_variation_003.phpt",
    "ext/standard/tests/class_object/forward_static_call_003.phpt",
    "ext/standard/tests/class_object/get_declared_interfaces_variation1.phpt",
    "ext/standard/tests/class_object/get_class_methods_error_001.phpt",
    "ext/standard/tests/class_object/get_declared_traits_basic_001.phpt",
    "ext/standard/tests/class_object/interface_exists_variation4.phpt",
    "ext/standard/tests/class_object/is_a_variation_003.phpt",
    "ext/standard/tests/class_object/get_class_vars_variation2.phpt",
    "ext/standard/tests/class_object/method_exists_variation_001.phpt",
    "ext/standard/tests/class_object/get_class_methods_variation_001.phpt",
    "ext/standard/tests/class_object/get_class_variation_001.phpt",
    "ext/standard/tests/class_object/get_object_vars_variation_002.phpt",
    "ext/standard/tests/class_object/get_object_vars_error_001.phpt",
    "ext/standard/tests/class_object/method_exists_variation_002.phpt",
    "ext/standard/tests/class_object/property_exists_error.phpt",
    "ext/standard/tests/class_object/get_class_methods_basic_003.phpt",
    "ext/standard/tests/assert/assert_closures.phpt",
    "ext/standard/tests/assert/assert_basic1.phpt",
    "ext/standard/tests/assert/assert_error1.phpt",
    "ext/standard/tests/assert/assert_basic4.phpt",
    "ext/standard/tests/assert/assert03.phpt",
    "ext/standard/tests/assert/assert_variation.phpt",
    "ext/standard/tests/assert/assert_basic2.phpt",
    "ext/standard/tests/assert/assert_basic3.phpt",
    "ext/standard/tests/assert/assert02.phpt",
    "ext/standard/tests/assert/assert_basic.phpt",
    "ext/standard/tests/assert/assert_error.phpt",
    "ext/standard/tests/assert/assert04.phpt",
    "ext/standard/tests/assert/assert.phpt",
    "ext/standard/tests/assert/assert_basic5.phpt",
    "ext/standard/tests/streams/bug48309.phpt",
    "ext/standard/tests/streams/bug49936_win32.phpt",
    "ext/standard/tests/streams/stream_get_meta_data_process_basic.phpt",
    "ext/standard/tests/streams/stream_get_meta_data_file_variation1.phpt",
    "ext/standard/tests/streams/stream_socket_pair.phpt",
    "ext/standard/tests/streams/stream_get_meta_data_file_variation5.phpt",
    "ext/standard/tests/streams/stream_copy_to_stream_socket.phpt",
    "ext/standard/tests/streams/stream_get_contents_002.phpt",
    "ext/standard/tests/streams/bug46024.phpt",
    "ext/standard/tests/streams/stream_get_meta_data_socket_variation2.phpt",
    "ext/standard/tests/streams/bug46426.phpt",
    "ext/standard/tests/streams/bug49936.phpt",
    "ext/standard/tests/streams/stream_get_meta_data_socket_variation1.phpt",
    "ext/standard/tests/streams/stream_context_get_params_001.phpt",
    "ext/standard/tests/streams/bug47997.phpt",
    "ext/standard/tests/streams/bug44818.phpt",
    "ext/standard/tests/streams/stream_set_chunk_size.phpt",
    "ext/standard/tests/streams/stream_is_local.phpt",
    "ext/standard/tests/streams/stream_get_meta_data_file_error.phpt",
    "ext/standard/tests/streams/stream_get_line_nb.phpt",
    "ext/standard/tests/streams/stream_resolve_include_path.phpt",
    "ext/standard/tests/streams/bug53427.phpt",
    "ext/standard/tests/streams/stream_get_meta_data_socket_basic.phpt",
    "ext/standard/tests/streams/bug53903.phpt",
    "ext/standard/tests/streams/stream_get_meta_data_socket_variation3.phpt",
    "ext/standard/tests/streams/bug44712.phpt",
    "ext/standard/tests/streams/stream_get_contents_001.phpt",
    "ext/standard/tests/streams/bug54623.phpt",
    "ext/standard/tests/streams/stream_set_timeout_error.phpt",
    "ext/standard/tests/streams/stream_get_meta_data_socket_variation4.phpt",
    "ext/standard/tests/streams/stream_get_meta_data_file_variation2.phpt",
    "ext/standard/tests/streams/stream_get_meta_data_file_basic.phpt",
    "ext/standard/tests/streams/stream_get_meta_data_dir_basic.phpt",
    "ext/standard/tests/streams/stream_get_meta_data_file_variation4.phpt",
    "ext/standard/tests/math/abs_basiclong_64bit.phpt",
    "ext/standard/tests/math/deg2rad_error.phpt",
    "ext/standard/tests/math/mt_rand_error.phpt",
    "ext/standard/tests/math/bindec_error.phpt",
    "ext/standard/tests/math/floor_basiclong_64bit.phpt",
    "ext/standard/tests/math/hexdec_variation1_64bit.phpt",
    "ext/standard/tests/math/round_large_exp.phpt",
    "ext/standard/tests/math/dechex_basiclong_64bit.phpt",
    "ext/standard/tests/math/bug30695.phpt",
    "ext/standard/tests/math/bindec_variation1_64bit.phpt",
    "ext/standard/tests/math/cosh_error.phpt",
    "ext/standard/tests/math/log1p_basiclong_64bit.phpt",
    "ext/standard/tests/math/tanh_variation.phpt",
    "ext/standard/tests/math/decoct_basic.phpt",
    "ext/standard/tests/math/round_basiclong_64bit.phpt",
    "ext/standard/tests/math/fmod_variation2.phpt",
    "ext/standard/tests/math/acos_variation.phpt",
    "ext/standard/tests/math/decbin_error.phpt",
    "ext/standard/tests/math/bug21523.phpt",
    "ext/standard/tests/math/octdec_error.phpt",
    "ext/standard/tests/math/cosh_basic.phpt",
    "ext/standard/tests/math/tan_basic.phpt",
    "ext/standard/tests/math/cos_basiclong_64bit.phpt",
    "ext/standard/tests/math/acos_error.phpt",
    "ext/standard/tests/math/hypot_error.phpt",
    "ext/standard/tests/math/octdec_basic_64bit.phpt",
    "ext/standard/tests/math/acosh_basic.phpt",
    "ext/standard/tests/math/mt_srand_error.phpt",
    "ext/standard/tests/math/sqrt_variation.phpt",
    "ext/standard/tests/math/bug25694.phpt",
    "ext/standard/tests/math/round_prerounding.phpt",
    "ext/standard/tests/math/abs_error.phpt",
    "ext/standard/tests/math/ceil_basic.phpt",
    "ext/standard/tests/math/hexdec_error.phpt",
    "ext/standard/tests/math/srand_variation1.phpt",
    "ext/standard/tests/math/tanh_basic.phpt",
    "ext/standard/tests/math/pow_variation1.phpt",
    "ext/standard/tests/math/decbin_basiclong_64bit.phpt",
    "ext/standard/tests/math/decbin_variation1.phpt",
    "ext/standard/tests/math/log_variation2.phpt",
    "ext/standard/tests/math/dechex_variation1.phpt",
    "ext/standard/tests/math/round.phpt",
    "ext/standard/tests/math/fmod_variation1.phpt",
    "ext/standard/tests/math/rad2deg_basiclong_64bit.phpt",
    "ext/standard/tests/math/dechex_error.phpt",
    "ext/standard/tests/math/hypot_variation1.phpt",
    "ext/standard/tests/math/floor_variation1.phpt",
    "ext/standard/tests/math/log10_basiclong_64bit.phpt",
    "ext/standard/tests/math/pow_variation2.phpt",
    "ext/standard/tests/math/asinh_basiclong_64bit.phpt",
    "ext/standard/tests/math/log.phpt",
    "ext/standard/tests/math/abs.phpt",
    "ext/standard/tests/math/hexdec_variation1.phpt",
    "ext/standard/tests/math/is_nan_variation1.phpt",
    "ext/standard/tests/math/atan_basiclong_64bit.phpt",
    "ext/standard/tests/math/pow_error.phpt",
    "ext/standard/tests/math/getrandmax_error.phpt",
    "ext/standard/tests/math/sqrt_basic.phpt",
    "ext/standard/tests/math/acos_basiclong_64bit.phpt",
    "ext/standard/tests/math/atan2_variation1.phpt",
    "ext/standard/tests/math/sin_variation.phpt",
    "ext/standard/tests/math/atan2_error.phpt",
    "ext/standard/tests/math/acosh_variation.phpt",
    "ext/standard/tests/math/dechex_variation1_64bit.phpt",
    "ext/standard/tests/math/round_variation2.phpt",
    "ext/standard/tests/math/sqrt_error.phpt",
    "ext/standard/tests/math/round_variation1.phpt",
    "ext/standard/tests/math/cos_basic.phpt",
    "ext/standard/tests/math/srand_basic.phpt",
    "ext/standard/tests/math/is_infinite_basic.phpt",
    "ext/standard/tests/math/atanh_variation.phpt",
    "ext/standard/tests/math/atan_basic.phpt",
    "ext/standard/tests/math/acosh_basiclong_64bit.phpt",
    "ext/standard/tests/math/log10_variation.phpt",
    "ext/standard/tests/math/rand_basic.phpt",
    "ext/standard/tests/math/is_infinite_variation1.phpt",
    "ext/standard/tests/math/pow_basic2.phpt",
    "ext/standard/tests/math/is_finite_error.phpt",
    "ext/standard/tests/math/asinh_basic.phpt",
    "ext/standard/tests/math/is_finite_basic.phpt",
    "ext/standard/tests/math/round_basic.phpt",
    "ext/standard/tests/math/acos_basic.phpt",
    "ext/standard/tests/math/fmod_error.phpt",
    "ext/standard/tests/math/fmod_basiclong_64bit.phpt",
    "ext/standard/tests/math/constants.phpt",
    "ext/standard/tests/math/decoct_error.phpt",
    "ext/standard/tests/math/ceil_variation1.phpt",
    "ext/standard/tests/math/hexdec.phpt",
    "ext/standard/tests/math/tanh_error.phpt",
    "ext/standard/tests/math/asin_variation.phpt",
    "ext/standard/tests/math/hypot_basiclong_64bit.phpt",
    "ext/standard/tests/math/floor_basic.phpt",
    "ext/standard/tests/math/atan2_variation2.phpt",
    "ext/standard/tests/math/log_error.phpt",
    "ext/standard/tests/math/hypot_basic.phpt",
    "ext/standard/tests/math/is_nan_basic.phpt",
    "ext/standard/tests/math/exp_basic.phpt",
    "ext/standard/tests/math/atanh_error.phpt",
    "ext/standard/tests/math/pow_basic.phpt",
    "ext/standard/tests/math/base_convert_error.phpt",
    "ext/standard/tests/math/srand_error.phpt",
    "ext/standard/tests/math/floor_error.phpt",
    "ext/standard/tests/math/log_basiclong_64bit.phpt",
    "ext/standard/tests/math/tan_variation.phpt",
    "ext/standard/tests/math/rad2deg_variation.phpt",
    "ext/standard/tests/math/exp_variation1.phpt",
    "ext/standard/tests/math/mt_srand_variation1.phpt",
    "ext/standard/tests/math/lcg_value_basic.phpt",
    "ext/standard/tests/math/atan_error.phpt",
    "ext/standard/tests/math/constants_basic.phpt",
    "ext/standard/tests/math/is_nan_error.phpt",
    "ext/standard/tests/math/decbin_variation1_64bit.phpt",
    "ext/standard/tests/math/atanh_basic.phpt",
    "ext/standard/tests/math/hexdec_basic.phpt",
    "ext/standard/tests/math/pow_variation1_64bit.phpt",
    "ext/standard/tests/math/bug25665.phpt",
    "ext/standard/tests/math/exp_error.phpt",
    "ext/standard/tests/math/is_infinite_error.phpt",
    "ext/standard/tests/math/log1p_variation1.phpt",
    "ext/standard/tests/math/hexdec_basiclong_64bit.phpt",
    "ext/standard/tests/math/bug45712.phpt",
    "ext/standard/tests/math/log10_basic.phpt",
    "ext/standard/tests/math/log1p_basic.phpt",
    "ext/standard/tests/math/is_nan_basiclong_64bit.phpt",
    "ext/standard/tests/math/deg2rad_variation.phpt",
    "ext/standard/tests/math/mt_getrandmax_basic.phpt",
    "ext/standard/tests/math/mt_srand_basic.phpt",
    "ext/standard/tests/math/atanh_basiclong_64bit.phpt",
    "ext/standard/tests/math/sin_basic.phpt",
    "ext/standard/tests/math/is_finite_basiclong_64bit.phpt",
    "ext/standard/tests/math/asin_basic.phpt",
    "ext/standard/tests/math/fmod_basic.phpt",
    "ext/standard/tests/math/asinh_error.phpt",
    "ext/standard/tests/math/sin_basiclong_64bit.phpt",
    "ext/standard/tests/math/log_basic.phpt",
    "ext/standard/tests/math/getrandmax_basic.phpt",
    "ext/standard/tests/math/asin_basiclong_64bit.phpt",
    "ext/standard/tests/math/base_convert_variation3.phpt",
    "ext/standard/tests/math/expm1_basiclong_64bit.phpt",
    "ext/standard/tests/math/rand_variation1.phpt",
    "ext/standard/tests/math/cosh_variation.phpt",
    "ext/standard/tests/math/floorceil.phpt",
    "ext/standard/tests/math/cosh_basiclong_64bit.phpt",
    "ext/standard/tests/math/deg2rad_basiclong_64bit.phpt",
    "ext/standard/tests/math/hexdec_basic_64bit.phpt",
    "ext/standard/tests/math/sinh_basic.phpt",
    "ext/standard/tests/math/hypot_variation2.phpt",
    "ext/standard/tests/math/sqrt_basiclong_64bit.phpt",
    "ext/standard/tests/math/rand_error.phpt",
    "ext/standard/tests/math/pow.phpt",
    "ext/standard/tests/math/number_format_basic.phpt",
    "ext/standard/tests/math/ceil_error.phpt",
    "ext/standard/tests/math/octdec_basic.phpt",
    "ext/standard/tests/math/tan_error.phpt",
    "ext/standard/tests/math/ceil_basiclong_64bit.phpt",
    "ext/standard/tests/math/log1p_error.phpt",
    "ext/standard/tests/math/mt_getrandmax_error.phpt",
    "ext/standard/tests/math/atan_variation.phpt",
    "ext/standard/tests/math/is_finite_variation1.phpt",
    "ext/standard/tests/math/sinh_error.phpt",
    "ext/standard/tests/math/number_format_multichar.phpt",
    "ext/standard/tests/math/cos_error.phpt",
    "ext/standard/tests/math/dechex_basic.phpt",
    "ext/standard/tests/math/log_variation1.phpt",
    "ext/standard/tests/math/tanh_basiclong_64bit.phpt",
    "ext/standard/tests/math/octdec_variation1.phpt",
    "ext/standard/tests/math/decoct_variation1_64bit.phpt",
    "ext/standard/tests/math/round_error.phpt",
    "ext/standard/tests/math/abs_variation.phpt",
    "ext/standard/tests/math/mt_rand_basic.phpt",
    "ext/standard/tests/math/rand_variation2.phpt",
    "ext/standard/tests/math/bug27646.phpt",
    "ext/standard/tests/math/decbin_basic.phpt",
    "ext/standard/tests/math/log10_error.phpt",
    "ext/standard/tests/math/pow_basic_64bit.phpt",
    "ext/standard/tests/math/bindec_basic.phpt",
    "ext/standard/tests/math/atan2_basic.phpt",
    "ext/standard/tests/math/expm1_variation1.phpt",
    "ext/standard/tests/math/bug28228.phpt",
    "ext/standard/tests/math/deg2rad_basic.phpt",
    "ext/standard/tests/math/base_convert_variation2.phpt",
    "ext/standard/tests/math/octdec_basiclong_64bit.phpt",
    "ext/standard/tests/math/cos_variation.phpt",
    "ext/standard/tests/math/base_convert_basic.phpt",
    "ext/standard/tests/math/expm1_basic.phpt",
    "ext/standard/tests/math/sinh_variation.phpt",
    "ext/standard/tests/math/pi_basic.phpt",
    "ext/standard/tests/math/exp_basiclong_64bit.phpt",
    "ext/standard/tests/math/expm1_error.phpt",
    "ext/standard/tests/math/sinh_basiclong_64bit.phpt",
    "ext/standard/tests/math/atan2_basiclong_64bit.phpt",
    "ext/standard/tests/math/asinh_variation.phpt",
    "ext/standard/tests/math/sin_error.phpt",
    "ext/standard/tests/math/decoct_variation1.phpt",
    "ext/standard/tests/math/bindec_basic_64bit.phpt",
    "ext/standard/tests/math/round_modes.phpt",
    "ext/standard/tests/math/bindec_variation1.phpt",
    "ext/standard/tests/math/pow_basiclong_64bit.phpt",
    "ext/standard/tests/math/acosh_error.phpt",
    "ext/standard/tests/math/is_infinite_basiclong_64bit.phpt",
    "ext/standard/tests/math/tan_basiclong_64bit.phpt",
    "ext/standard/tests/math/rad2deg_error.phpt",
    "ext/standard/tests/math/bug30069.phpt",
    "ext/standard/tests/math/base_convert_variation1.phpt",
    "ext/standard/tests/math/decoct_basiclong_64bit.phpt",
    "ext/standard/tests/math/bug24142.phpt",
    "ext/standard/tests/math/rad2deg_basic.phpt",
    "ext/standard/tests/math/mt_rand_variation1.phpt",
    "ext/standard/tests/math/bindec_basiclong_64bit.phpt",
    "ext/standard/tests/math/mt_rand_variation2.phpt",
    "ext/standard/tests/math/abs_basic.phpt",
    "ext/standard/tests/math/asin_error.phpt",
    "ext/standard/tests/versioning/version_compare.phpt",
    "ext/standard/tests/versioning/php_sapi_name_variation001.phpt",
    "ext/standard/tests/versioning/php_sapi_name.phpt",
    "ext/standard/tests/time/001.phpt",
    "ext/standard/tests/time/bug38524.phpt",
    "ext/standard/tests/time/strptime_error.phpt",
    "ext/standard/tests/time/idate.phpt",
    "ext/standard/tests/time/strptime_basic.phpt",
    "ext/standard/tests/php_logo_guid.phpt",
    "ext/standard/tests/dir/closedir_variation2.phpt",
    "ext/standard/tests/dir/readdir_variation1.phpt",
    "ext/standard/tests/dir/dir_error.phpt",
    "ext/standard/tests/dir/scandir_variation4.phpt",
    "ext/standard/tests/dir/dir_variation7.phpt",
    "ext/standard/tests/dir/dir_variation4.phpt",
    "ext/standard/tests/dir/closedir_variation3.phpt",
    "ext/standard/tests/dir/scandir_basic.phpt",
    "ext/standard/tests/dir/scandir_variation6.phpt",
    "ext/standard/tests/dir/opendir_variation4.phpt",
    "ext/standard/tests/dir/scandir_variation1.phpt",
    "ext/standard/tests/dir/chdir_error2.phpt",
    "ext/standard/tests/dir/chdir_error1.phpt",
    "ext/standard/tests/dir/closedir_variation1.phpt",
    "ext/standard/tests/dir/readdir_variation6.phpt",
    "ext/standard/tests/dir/rewinddir_basic.phpt",
    "ext/standard/tests/dir/scandir_error1.phpt",
    "ext/standard/tests/dir/readdir_variation7.phpt",
    "ext/standard/tests/dir/readdir_variation5.phpt",
    "ext/standard/tests/dir/readdir_variation3.phpt",
    "ext/standard/tests/dir/scandir_variation5.phpt",
    "ext/standard/tests/dir/opendir_error2.phpt",
    "ext/standard/tests/dir/opendir_variation7.phpt",
    "ext/standard/tests/dir/scandir_variation2.phpt",
    "ext/standard/tests/dir/getcwd_basic.phpt",
    "ext/standard/tests/dir/opendir_variation6.phpt",
    "ext/standard/tests/dir/opendir_variation3.phpt",
    "ext/standard/tests/dir/dir_variation8.phpt",
    "ext/standard/tests/dir/dir_variation1.phpt",
    "ext/standard/tests/dir/readdir_basic.phpt",
    "ext/standard/tests/dir/dir_variation9.phpt",
    "ext/standard/tests/dir/chdir_variation2.phpt",
    "ext/standard/tests/dir/opendir_error1.phpt",
    "ext/standard/tests/dir/scandir_variation3.phpt",
    "ext/standard/tests/dir/readdir_error.phpt",
    "ext/standard/tests/dir/opendir_basic.phpt",
    "ext/standard/tests/dir/scandir_variation10.phpt",
    "ext/standard/tests/dir/dir_basic.phpt",
    "ext/standard/tests/dir/closedir_error.phpt",
    "ext/standard/tests/dir/dir_variation3.phpt",
    "ext/standard/tests/dir/opendir_variation2.phpt",
    "ext/standard/tests/dir/readdir_variation2.phpt",
    "ext/standard/tests/dir/scandir_error2.phpt",
    "ext/standard/tests/dir/opendir_variation5.phpt",
    "ext/standard/tests/dir/closedir_basic.phpt",
    "ext/standard/tests/dir/chdir_variation1.phpt",
    "ext/standard/tests/dir/rewinddir_error.phpt",
    "ext/standard/tests/dir/rewinddir_variation3.phpt",
    "ext/standard/tests/dir/chdir_basic.phpt",
    "ext/standard/tests/dir/readdir_variation4.phpt",
    "ext/standard/tests/dir/scandir_variation8.phpt",
    "ext/standard/tests/dir/dir_variation2.phpt",
    "ext/standard/tests/dir/opendir_variation1.phpt",
    "ext/standard/tests/dir/getcwd_error.phpt",
    "ext/standard/tests/dir/scandir_variation9.phpt",
    "ext/standard/tests/dir/scandir_variation7.phpt",
    "ext/standard/tests/dir/rewinddir_variation1.phpt",
    "ext/standard/tests/dir/rewinddir_variation2.phpt",
    "ext/standard/tests/dir/dir_variation5.phpt",
    "ext/standard/tests/dir/dir_variation6.phpt",
    "ext/standard/tests/zend_logo_guid.phpt",
    "ext/standard/tests/filters/bug22538.phpt",
    "ext/standard/tests/filters/php_user_filter_03.phpt",
    "ext/standard/tests/filters/001.phpt",
    "ext/standard/tests/filters/bug35916.phpt",
    "ext/standard/tests/filters/php_user_filter_02.phpt",
    "ext/standard/tests/filters/chunked_001.phpt",
    "ext/standard/tests/filters/stream_filter_remove_error.phpt",
    "ext/standard/tests/filters/basic.phpt",
    "ext/standard/tests/filters/filter_errors_user.phpt",
    "ext/standard/tests/filters/bug46164-2.phpt",
    "ext/standard/tests/filters/php_user_filter_01.phpt",
    "ext/standard/tests/filters/stream_filter_remove_basic.phpt",
    "ext/standard/tests/filters/bug46164-1.phpt",
    "ext/standard/tests/filters/filter_errors_convert_base64_decode.phpt",
    "ext/standard/tests/filters/read.phpt",
    "ext/standard/tests/filters/filter_errors_zlib_inflate.phpt",
    "ext/standard/tests/serialize/serialization_miscTypes_001.phpt",
    "ext/standard/tests/serialize/serialization_objects_014.phpt",
    "ext/standard/tests/serialize/005.phpt",
    "ext/standard/tests/serialize/serialization_objects_009.phpt",
    "ext/standard/tests/serialize/unserializeS.phpt",
    "ext/standard/tests/serialize/serialization_objects_015.phpt",
    "ext/standard/tests/serialize/serialization_objects_004.phpt",
    "ext/standard/tests/serialize/serialization_objects_005.phpt",
    "ext/standard/tests/serialize/bug30234.phpt",
    "ext/standard/tests/serialize/precision.phpt",
    "ext/standard/tests/serialize/serialization_arrays_001.phpt",
    "ext/standard/tests/serialize/001.phpt",
    "ext/standard/tests/serialize/serialization_objects_011.phpt",
    "ext/standard/tests/serialize/bug36424.phpt",
    "ext/standard/tests/serialize/serialization_arrays_002.phpt",
    "ext/standard/tests/serialize/004.phpt",
    "ext/standard/tests/serialize/bug14293.phpt",
    "ext/standard/tests/serialize/bug25378.phpt",
    "ext/standard/tests/serialize/bug31442.phpt",
    "ext/standard/tests/serialize/bug21957.phpt",
    "ext/standard/tests/serialize/serialization_resources_001.phpt",
    "ext/standard/tests/serialize/serialization_objects_012.phpt",
    "ext/standard/tests/serialize/serialization_objects_010.phpt",
    "ext/standard/tests/serialize/003.phpt",
    "ext/standard/tests/serialize/bug23298.phpt",
    "ext/standard/tests/serialize/006.phpt",
    "ext/standard/tests/serialize/bug43614.phpt",
    "ext/standard/tests/serialize/serialization_objects_008.phpt",
    "ext/standard/tests/serialize/serialization_error_001.phpt",
    "ext/standard/tests/serialize/bug42919.phpt",
    "ext/standard/tests/serialize/serialization_precision_002.phpt",
    "ext/standard/tests/serialize/serialization_objects_013.phpt",
    "ext/standard/tests/serialize/serialization_objects_006.phpt",
    "ext/standard/tests/serialize/bug26762.phpt",
    "ext/standard/tests/serialize/bug37947.phpt",
    "ext/standard/tests/serialize/serialization_arrays_003.phpt",
    "ext/standard/tests/serialize/bug31402.phpt",
    "ext/standard/tests/serialize/serialization_objects_007.phpt",
    "ext/standard/tests/serialize/002.phpt",
    "ext/standard/tests/serialize/serialization_objects_003.phpt",
    "ext/standard/tests/serialize/serialization_objects_002.phpt",
    "ext/standard/tests/serialize/bug24063.phpt",
    "ext/standard/tests/serialize/bug28325.phpt",
    "ext/standard/tests/serialize/bug27469.phpt",
    "ext/standard/tests/serialize/serialization_precision_001.phpt",
    "ext/standard/tests/serialize/bug46882.phpt",
    "ext/standard/tests/serialize/bug45706.phpt",
    "ext/standard/tests/serialize/serialization_objects_001.phpt",
    "ext/standard/tests/serialize/serialization_arrays_004.phpt",
    "ext/standard/tests/serialize/incomplete_class.phpt",
    "ext/standard/tests/serialize/serialization_arrays_005.phpt",
    "ext/standard/tests/url/rawurlencode_error_001.phpt",
    "ext/standard/tests/url/parse_url_variation_002_32bit.phpt",
    "ext/standard/tests/url/parse_url_basic_001.phpt",
    "ext/standard/tests/url/parse_url_error_001.phpt",
    "ext/standard/tests/url/base64_decode_error_001.phpt",
    "ext/standard/tests/url/parse_url_basic_008.phpt",
    "ext/standard/tests/url/parse_url_variation_001.phpt",
    "ext/standard/tests/url/bug53248.phpt",
    "ext/standard/tests/url/parse_url_basic_005.phpt",
    "ext/standard/tests/url/urlencode_error_001.phpt",
    "ext/standard/tests/url/get_headers_error_001.phpt",
    "ext/standard/tests/url/get_headers_error_002.phpt",
    "ext/standard/tests/url/rawurldecode_variation_001.phpt",
    "ext/standard/tests/url/base64_encode_basic_002.phpt",
    "ext/standard/tests/url/base64_decode_basic_001.phpt",
    "ext/standard/tests/url/parse_url_basic_009.phpt",
    "ext/standard/tests/url/rawurldecode_error_001.phpt",
    "ext/standard/tests/url/base64_decode_basic_002.phpt",
    "ext/standard/tests/url/rawurlencode_variation_001.phpt",
    "ext/standard/tests/url/parse_url_basic_006.phpt",
    "ext/standard/tests/url/urlencode_variation_001.phpt",
    "ext/standard/tests/url/urldecode_error_001.phpt",
    "ext/standard/tests/url/urldecode_variation_001.phpt",
    "ext/standard/tests/url/base64_decode_variation_001.phpt",
    "ext/standard/tests/url/parse_url_basic_004.phpt",
    "ext/standard/tests/url/base64_encode_basic_001.phpt",
    "ext/standard/tests/url/bug47174.phpt",
    "ext/standard/tests/url/base64_encode_error_001.phpt",
    "ext/standard/tests/url/base64_decode_variation_002.phpt",
    "ext/standard/tests/url/parse_url_error_002.phpt",
    "ext/standard/tests/url/bug52327.phpt",
    "ext/standard/tests/url/parse_url_basic_003.phpt",
    "ext/standard/tests/url/parse_url_basic_010.phpt",
    "ext/standard/tests/url/parse_url_basic_007.phpt",
    "ext/standard/tests/url/parse_url_variation_002_64bit.phpt",
    "ext/standard/tests/url/base64_encode_variation_001.phpt",
    "ext/standard/tests/url/bug54180.phpt",
    "ext/standard/tests/url/parse_url_basic_002.phpt",
    "ext/standard/tests/network/inet_ipv6.phpt",
    "ext/standard/tests/network/udgloop.phpt",
    "ext/standard/tests/network/long2ip_error.phpt",
    "ext/standard/tests/network/udp4loop.phpt",
    "ext/standard/tests/network/gethostbyname_error003.phpt",
    "ext/standard/tests/network/gethostbyname_error005.phpt",
    "ext/standard/tests/network/fsockopen_error.phpt",
    "ext/standard/tests/network/gethostbyaddr_error.phpt",
    "ext/standard/tests/network/tcp6loop.phpt",
    "ext/standard/tests/network/ip2long_variation1.phpt",
    "ext/standard/tests/network/closelog_error.phpt",
    "ext/standard/tests/network/gethostbyname_basic003.phpt",
    "ext/standard/tests/network/shutdown.phpt",
    "ext/standard/tests/network/gethostbyname_basic002.phpt",
    "ext/standard/tests/network/syslog_error.phpt",
    "ext/standard/tests/network/gethostbyname_error006.phpt",
    "ext/standard/tests/network/unixloop.phpt",
    "ext/standard/tests/network/gethostbynamel_error.phpt",
    "ext/standard/tests/network/gethostbyname_error001.phpt",
    "ext/standard/tests/network/closelog_basic.phpt",
    "ext/standard/tests/network/ip.phpt",
    "ext/standard/tests/network/fsockopen_basic.phpt",
    "ext/standard/tests/network/gethostbynamel_basic1.phpt",
    "ext/standard/tests/network/gethostbyname_error004.phpt",
    "ext/standard/tests/network/fsockopen_variation2.phpt",
    "ext/standard/tests/network/bug41347.phpt",
    "ext/standard/tests/network/tcp4loop.phpt",
    "ext/standard/tests/network/socket_get_status_basic.phpt",
    "ext/standard/tests/network/gethostbyaddr_basic1.phpt",
    "ext/standard/tests/network/gethostbyname_error002.phpt",
    "ext/standard/tests/network/syslog_basic-win32.phpt",
    "ext/standard/tests/network/udp6loop.phpt",
    "ext/standard/tests/network/gethostbyname_basic001.phpt",
    "ext/standard/tests/network/ip2long_error.phpt",
    "ext/standard/tests/network/bug20134.phpt",
    "ext/standard/tests/network/http-stream.phpt",
    "ext/standard/tests/network/long2ip_variation1.phpt",
    "ext/standard/tests/network/inet.phpt",
    "ext/standard/tests/network/ip_x86_64.phpt",
    "ext/standard/tests/network/fsockopen_variation1.phpt",
    "ext/standard/tests/network/getmxrr.phpt",
    "ext/standard/tests/misc/get_browser_basic.phpt",
    "ext/standard/tests/misc/time_nanosleep_error3.phpt",
    "ext/standard/tests/misc/time_nanosleep_error5.phpt",
    "ext/standard/tests/misc/time_nanosleep_error2.phpt",
    "ext/standard/tests/misc/time_nanosleep_basic.phpt",
    "ext/standard/tests/misc/time_nanosleep_error1.phpt",
    "ext/standard/tests/misc/syslog_basic.phpt",
    "ext/standard/tests/misc/get_browser_error.phpt",
    "ext/standard/tests/misc/get_browser_variation1.phpt",
    "ext/standard/tests/misc/time_nanosleep_error4.phpt",
    "ext/standard/tests/misc/syslog_parameters.phpt",
    "ext/odbc/tests/odbc_free_result_001.phpt",
    "ext/odbc/tests/odbc_data_source_001.phpt",
    "ext/odbc/tests/odbc_columns_001.phpt",
    "ext/odbc/tests/odbc_columnprivileges_001.phpt",
    "ext/odbc/tests/odbc_exec_002.phpt",
    "ext/odbc/tests/odbc_tables_001.phpt",
    "ext/odbc/tests/odbc_exec_001.phpt",
    "ext/xmlreader/tests/010.phpt",
    "ext/xmlreader/tests/005.phpt",
    "ext/xmlreader/tests/expand.phpt",
    "ext/xmlreader/tests/011.phpt",
    "ext/xmlreader/tests/001.phpt",
    "ext/xmlreader/tests/004.phpt",
    "ext/xmlreader/tests/bug36743.phpt",
    "ext/xmlreader/tests/003.phpt",
    "ext/xmlreader/tests/006.phpt",
    "ext/xmlreader/tests/007.phpt",
    "ext/xmlreader/tests/002.phpt",
    "ext/xmlreader/tests/008.phpt",
    "ext/xmlreader/tests/013.phpt",
    "ext/xmlreader/tests/012.phpt",
    "ext/xmlreader/tests/bug42139.phpt",
    "ext/xmlreader/tests/009.phpt",
    "ext/pcntl/tests/pcntl_alarm.phpt",
    "ext/pcntl/tests/pcntl_get_last_error.phpt",
    "ext/pcntl/tests/pcntl_wait.phpt",
    "ext/pcntl/tests/001.phpt",
    "ext/pcntl/tests/pcntl_fork_basic.phpt",
    "ext/pcntl/tests/pcntl_exec_2.phpt",
    "ext/pcntl/tests/003.phpt",
    "ext/pcntl/tests/pcntl_exec_3.phpt",
    "ext/pcntl/tests/bug47566.phpt",
    "ext/pcntl/tests/002.phpt",
    "ext/pcntl/tests/signal_closure_handler.phpt",
    "ext/pcntl/tests/pcntl_signal_dispatch.phpt",
    "ext/pcntl/tests/pcntl_exec.phpt",
    "ext/pcntl/tests/pcntl_fork_variation.phpt",
    "ext/pcntl/tests/pcntl_signal.phpt",
    "ext/gmp/tests/027.phpt",
    "ext/gmp/tests/033.phpt",
    "ext/gmp/tests/010.phpt",
    "ext/gmp/tests/bug32773.phpt",
    "ext/gmp/tests/005.phpt",
    "ext/gmp/tests/014.phpt",
    "ext/gmp/tests/029.phpt",
    "ext/gmp/tests/011.phpt",
    "ext/gmp/tests/017.phpt",
    "ext/gmp/tests/036.phpt",
    "ext/gmp/tests/028.phpt",
    "ext/gmp/tests/001.phpt",
    "ext/gmp/tests/015.phpt",
    "ext/gmp/tests/026.phpt",
    "ext/gmp/tests/004.phpt",
    "ext/gmp/tests/bug50283.phpt",
    "ext/gmp/tests/003.phpt",
    "ext/gmp/tests/020.phpt",
    "ext/gmp/tests/022.phpt",
    "ext/gmp/tests/006.phpt",
    "ext/gmp/tests/031.phpt",
    "ext/gmp/tests/023.phpt",
    "ext/gmp/tests/007.phpt",
    "ext/gmp/tests/gmp_nextprime.phpt",
    "ext/gmp/tests/002.phpt",
    "ext/gmp/tests/008.phpt",
    "ext/gmp/tests/018.phpt",
    "ext/gmp/tests/040.phpt",
    "ext/gmp/tests/034.phpt",
    "ext/gmp/tests/039.phpt",
    "ext/gmp/tests/035.phpt",
    "ext/gmp/tests/013.phpt",
    "ext/gmp/tests/037.phpt",
    "ext/gmp/tests/016.phpt",
    "ext/gmp/tests/025.phpt",
    "ext/gmp/tests/012.phpt",
    "ext/gmp/tests/038.phpt",
    "ext/gmp/tests/021.phpt",
    "ext/gmp/tests/030.phpt",
    "ext/gmp/tests/032.phpt",
    "ext/gmp/tests/019.phpt",
    "ext/gmp/tests/024.phpt",
    "ext/gmp/tests/009.phpt",
    "ext/gmp/tests/bug52906.phpt",
    "ext/ereg/tests/eregi_variation_003.phpt",
    "ext/ereg/tests/split_basic_001.phpt",
    "ext/ereg/tests/010.phpt",
    "ext/ereg/tests/spliti_variation_001.phpt",
    "ext/ereg/tests/005.phpt",
    "ext/ereg/tests/eregi_error_002.phpt",
    "ext/ereg/tests/spliti_variation_003.phpt",
    "ext/ereg/tests/ereg_replace_variation_001.phpt",
    "ext/ereg/tests/014.phpt",
    "ext/ereg/tests/split_variation_003.phpt",
    "ext/ereg/tests/split_error_002.phpt",
    "ext/ereg/tests/spliti_basic_002.phpt",
    "ext/ereg/tests/011.phpt",
    "ext/ereg/tests/eregi_variation_001.phpt",
    "ext/ereg/tests/001.phpt",
    "ext/ereg/tests/ereg_replace_variation_003.phpt",
    "ext/ereg/tests/015.phpt",
    "ext/ereg/tests/ereg_error_001.phpt",
    "ext/ereg/tests/004.phpt",
    "ext/ereg/tests/spliti_error_002.phpt",
    "ext/ereg/tests/split_variation_002.phpt",
    "ext/ereg/tests/spliti_variation_004.phpt",
    "ext/ereg/tests/eregi_variation_002.phpt",
    "ext/ereg/tests/spliti_basic_003.phpt",
    "ext/ereg/tests/eregi_error_001.phpt",
    "ext/ereg/tests/spliti_basic_001.phpt",
    "ext/ereg/tests/ereg_replace_error_001.phpt",
    "ext/ereg/tests/eregi_replace_variation_002.phpt",
    "ext/ereg/tests/ereg_basic_001.phpt",
    "ext/ereg/tests/003.phpt",
    "ext/ereg/tests/eregi_variation_004.phpt",
    "ext/ereg/tests/eregi_basic_003.phpt",
    "ext/ereg/tests/ereg_variation_001.phpt",
    "ext/ereg/tests/006.phpt",
    "ext/ereg/tests/eregi_basic.phpt",
    "ext/ereg/tests/split_error_001.phpt",
    "ext/ereg/tests/spliti_basic_004.phpt",
    "ext/ereg/tests/ereg_variation_003.phpt",
    "ext/ereg/tests/spliti_error_001.phpt",
    "ext/ereg/tests/sql_regcase_basic_001.phpt",
    "ext/ereg/tests/sql_regcase_error_001.phpt",
    "ext/ereg/tests/007.phpt",
    "ext/ereg/tests/split_variation_004.phpt",
    "ext/ereg/tests/002.phpt",
    "ext/ereg/tests/008.phpt",
    "ext/ereg/tests/split_basic_002.phpt",
    "ext/ereg/tests/eregi_replace_basic.phpt",
    "ext/ereg/tests/eregi_replace_variation_001.phpt",
    "ext/ereg/tests/spliti_variation_002.phpt",
    "ext/ereg/tests/013.phpt",
    "ext/ereg/tests/eregi_replace_basic_001.phpt",
    "ext/ereg/tests/016.phpt",
    "ext/ereg/tests/ereg_replace_basic_001.phpt",
    "ext/ereg/tests/eregi_basic_001.phpt",
    "ext/ereg/tests/eregi_replace_variation_003.phpt",
    "ext/ereg/tests/012.phpt",
    "ext/ereg/tests/split_basic_003.phpt",
    "ext/ereg/tests/sql_regcase_variation_001.phpt",
    "ext/ereg/tests/eregi_replace_basic_002.phpt",
    "ext/ereg/tests/eregi_basic_002.phpt",
    "ext/ereg/tests/ereg_error_002.phpt",
    "ext/ereg/tests/eregi_replace_error_001.phpt",
    "ext/ereg/tests/ereg_replace_error_002.phpt",
    "ext/ereg/tests/ereg_basic_004.phpt",
    "ext/ereg/tests/eregi_replace_error_002.phpt",
    "ext/ereg/tests/ereg_basic_003.phpt",
    "ext/ereg/tests/ereg_replace_basic_002.phpt",
    "ext/ereg/tests/ereg_replace_variation_002.phpt",
    "ext/ereg/tests/eregi_basic_004.phpt",
    "ext/ereg/tests/ereg_basic_002.phpt",
    "ext/ereg/tests/009.phpt",
    "ext/ereg/tests/ereg_variation_004.phpt",
    "ext/ereg/tests/split_variation_001.phpt",
    "ext/ereg/tests/ereg_variation_002.phpt",
    "ext/ctype/tests/ctype_upper_variation2.phpt",
    "ext/ctype/tests/ctype_alnum_error.phpt",
    "ext/ctype/tests/ctype_cntrl_variation1.phpt",
    "ext/ctype/tests/ctype_graph_basic.phpt",
    "ext/ctype/tests/ctype_punct_variation1.phpt",
    "ext/ctype/tests/ctype_alnum_variation4.phpt",
    "ext/ctype/tests/ctype_alpha_variation4.phpt",
    "ext/ctype/tests/ctype_digit_variation2.phpt",
    "ext/ctype/tests/ctype_graph_variation3.phpt",
    "ext/ctype/tests/ctype_xdigit_basic.phpt",
    "ext/ctype/tests/ctype_digit_variation3.phpt",
    "ext/ctype/tests/bug25745.phpt",
    "ext/ctype/tests/ctype_upper_error.phpt",
    "ext/ctype/tests/ctype_space_variation4.phpt",
    "ext/ctype/tests/ctype_graph_variation2.phpt",
    "ext/ctype/tests/001.phpt",
    "ext/ctype/tests/ctype_lower_basic.phpt",
    "ext/ctype/tests/ctype_digit_error.phpt",
    "ext/ctype/tests/ctype_space_variation3.phpt",
    "ext/ctype/tests/ctype_lower_variation4.phpt",
    "ext/ctype/tests/ctype_upper_variation4.phpt",
    "ext/ctype/tests/ctype_space_variation2.phpt",
    "ext/ctype/tests/ctype_cntrl_variation4.phpt",
    "ext/ctype/tests/ctype_cntrl_basic.phpt",
    "ext/ctype/tests/ctype_print_variation2.phpt",
    "ext/ctype/tests/ctype_digit_variation1.phpt",
    "ext/ctype/tests/ctype_graph_variation4.phpt",
    "ext/ctype/tests/ctype_xdigit_variation2.phpt",
    "ext/ctype/tests/ctype_xdigit_variation4.phpt",
    "ext/ctype/tests/ctype_alpha_variation1.phpt",
    "ext/ctype/tests/ctype_digit_basic.phpt",
    "ext/ctype/tests/ctype_digit_variation4.phpt",
    "ext/ctype/tests/ctype_cntrl_error.phpt",
    "ext/ctype/tests/ctype_alpha_variation2.phpt",
    "ext/ctype/tests/ctype_cntrl_variation3.phpt",
    "ext/ctype/tests/ctype_graph_variation1.phpt",
    "ext/ctype/tests/ctype_punct_variation3.phpt",
    "ext/ctype/tests/ctype_upper_basic.phpt",
    "ext/ctype/tests/ctype_space_variation1.phpt",
    "ext/ctype/tests/ctype_lower_variation1.phpt",
    "ext/ctype/tests/ctype_alpha_basic.phpt",
    "ext/ctype/tests/ctype_alnum_variation2.phpt",
    "ext/ctype/tests/ctype_xdigit_variation3.phpt",
    "ext/ctype/tests/ctype_alnum_variation3.phpt",
    "ext/ctype/tests/002.phpt",
    "ext/ctype/tests/ctype_alnum_basic.phpt",
    "ext/ctype/tests/ctype_graph_error.phpt",
    "ext/ctype/tests/ctype_alpha_variation3.phpt",
    "ext/ctype/tests/ctype_punct_error.phpt",
    "ext/ctype/tests/ctype_lower_error.phpt",
    "ext/ctype/tests/ctype_alpha_error.phpt",
    "ext/ctype/tests/ctype_xdigit_variation1.phpt",
    "ext/ctype/tests/ctype_lower_variation2.phpt",
    "ext/ctype/tests/ctype_space_error.phpt",
    "ext/ctype/tests/ctype_punct_variation4.phpt",
    "ext/ctype/tests/ctype_alnum_variation1.phpt",
    "ext/ctype/tests/ctype_upper_variation3.phpt",
    "ext/ctype/tests/ctype_print_error.phpt",
    "ext/ctype/tests/bug34645.phpt",
    "ext/ctype/tests/ctype_punct_variation2.phpt",
    "ext/ctype/tests/ctype_cntrl_variation2.phpt",
    "ext/ctype/tests/ctype_print_variation4.phpt",
    "ext/ctype/tests/ctype_xdigit_error.phpt",
    "ext/ctype/tests/ctype_space_basic.phpt",
    "ext/ctype/tests/ctype_punct_basic.phpt",
    "ext/ctype/tests/ctype_print_basic.phpt",
    "ext/ctype/tests/ctype_lower_variation3.phpt",
    "ext/ctype/tests/ctype_print_variation3.phpt",
    "ext/ctype/tests/ctype_upper_variation1.phpt",
    "ext/ctype/tests/ctype_print_variation1.phpt",
    "ext/fileinfo/tests/finfo_close_basic.phpt",
    "ext/fileinfo/tests/finfo_open_error.phpt",
    "ext/fileinfo/tests/finfo_phpinfo_basic.phpt",
    "ext/fileinfo/tests/mime_content_type_001.phpt",
    "ext/fileinfo/tests/finfo_open_basic.phpt",
    "ext/fileinfo/tests/finfo_close_error.phpt",
    "ext/fileinfo/tests/mime_content_type_002.phpt",
    "ext/fileinfo/tests/finfo_set_flags_basic.phpt",
    "ext/fileinfo/tests/finfo_file_stream_001.phpt",
    "ext/fileinfo/tests/finfo_file_basic.phpt",
    "ext/fileinfo/tests/finfo_file_001.phpt",
    "ext/fileinfo/tests/finfo_buffer_error.phpt",
    "ext/fileinfo/tests/finfo_open_001.phpt",
    "ext/fileinfo/tests/finfo_open_variation1.phpt",
    "ext/fileinfo/tests/finfo_buffer_variation1.phpt",
    "ext/fileinfo/tests/finfo_buffer_basic.phpt",
    "ext/fileinfo/tests/finfo_file_002.phpt",
    "ext/reflection/tests/ReflectionObject_getConstructor_error.phpt",
    "ext/reflection/tests/bug41061.phpt",
    "ext/reflection/tests/bug30146.phpt",
    "ext/reflection/tests/ReflectionMethod_006.phpt",
    "ext/reflection/tests/bug37816.phpt",
    "ext/reflection/tests/ReflectionFunction_getFileName.001.phpt",
    "ext/reflection/tests/ReflectionExtension_constructor_error.phpt",
    "ext/reflection/tests/bug49719.phpt",
    "ext/reflection/tests/bug45139.phpt",
    "ext/reflection/tests/010.phpt",
    "ext/reflection/tests/ReflectionMethod_getDocComment_basic.phpt",
    "ext/reflection/tests/ReflectionClass_getName_error1.phpt",
    "ext/reflection/tests/ReflectionClass_getMethods_002.phpt",
    "ext/reflection/tests/ReflectionObject_getName_error.phpt",
    "ext/reflection/tests/ReflectionClass_isAbstract_basic.phpt",
    "ext/reflection/tests/bug36308.phpt",
    "ext/reflection/tests/ReflectionClass_export_basic1.phpt",
    "ext/reflection/tests/ReflectionParameter_export_basic.phpt",
    "ext/reflection/tests/005.phpt",
    "ext/reflection/tests/ReflectionParameter_invalidMethodInConstructor.phpt",
    "ext/reflection/tests/closures_003.phpt",
    "ext/reflection/tests/ReflectionClass_getMethods_003.phpt",
    "ext/reflection/tests/014.phpt",
    "ext/reflection/tests/ReflectionClass_getDocComment_002.phpt",
    "ext/reflection/tests/ReflectionProperty_export_basic.phpt",
    "ext/reflection/tests/ReflectionObject_isInstantiable_variation.phpt",
    "ext/reflection/tests/ReflectionObject_isInstance_basic.phpt",
    "ext/reflection/tests/ReflectionClass_hasConstant_001.phpt",
    "ext/reflection/tests/ReflectionClass_isInternal_basic.phpt",
    "ext/reflection/tests/bug52854.phpt",
    "ext/reflection/tests/ReflectionClass_getMethods_001.phpt",
    "ext/reflection/tests/ReflectionExtension_getClassNames_variation1.phpt",
    "ext/reflection/tests/bug51911.phpt",
    "ext/reflection/tests/bug39884.phpt",
    "ext/reflection/tests/bug26695.phpt",
    "ext/reflection/tests/ReflectionClass_getNamespaceName.phpt",
    "ext/reflection/tests/ReflectionClass_constructor_001.phpt",
    "ext/reflection/tests/011.phpt",
    "ext/reflection/tests/ReflectionExtension_info_basic.phpt",
    "ext/reflection/tests/ReflectionClass_FileInfo_basic.phpt",
    "ext/reflection/tests/ReflectionMethod_invokeArgs_basic.phpt",
    "ext/reflection/tests/bug30148.phpt",
    "ext/reflection/tests/bug38942.phpt",
    "ext/reflection/tests/ReflectionClass_getConstants_error.phpt",
    "ext/reflection/tests/property_exists.phpt",
    "ext/reflection/tests/ReflectionClass_modifiers_002.phpt",
    "ext/reflection/tests/ReflectionClass_isSubclassOf_error1.phpt",
    "ext/reflection/tests/bug45765.phpt",
    "ext/reflection/tests/ReflectionObject_isInternal_basic.phpt",
    "ext/reflection/tests/017.phpt",
    "ext/reflection/tests/bug53915.phpt",
    "ext/reflection/tests/ReflectionExtension_isTemporary.phpt",
    "ext/reflection/tests/ReflectionClass_newInstance_001.phpt",
    "ext/reflection/tests/ReflectionClass_getProperty_003.phpt",
    "ext/reflection/tests/ReflectionClass_toString_001.phpt",
    "ext/reflection/tests/ReflectionClass_setStaticPropertyValue_001_2_4.phpt",
    "ext/reflection/tests/ReflectionClass_isInternal_error.phpt",
    "ext/reflection/tests/ReflectionFunction_getExtensionName.phpt",
    "ext/reflection/tests/ReflectionClass_getConstructor_error.phpt",
    "ext/reflection/tests/bug52057.phpt",
    "ext/reflection/tests/ReflectionClass_getConstant_error.phpt",
    "ext/reflection/tests/ReflectionClass_toString_002.phpt",
    "ext/reflection/tests/traits001.phpt",
    "ext/reflection/tests/bug38465.phpt",
    "ext/reflection/tests/ReflectionMethod_setAccessible.phpt",
    "ext/reflection/tests/ReflectionMethod_getModifiers_basic.phpt",
    "ext/reflection/tests/ReflectionProperty_getDocComment_error.phpt",
    "ext/reflection/tests/bug37964.phpt",
    "ext/reflection/tests/ReflectionParameter_export_error2.phpt",
    "ext/reflection/tests/001.phpt",
    "ext/reflection/tests/ReflectionParameter_export_error3.phpt",
    "ext/reflection/tests/bug39001.phpt",
    "ext/reflection/tests/ReflectionClass_getConstructor_basic.phpt",
    "ext/reflection/tests/ReflectionClass_hasMethod_basic.phpt",
    "ext/reflection/tests/ReflectionClass_getProperties_002.phpt",
    "ext/reflection/tests/015.phpt",
    "ext/reflection/tests/ReflectionClass_isInterface_basic.phpt",
    "ext/reflection/tests/traits004.phpt",
    "ext/reflection/tests/026.phpt",
    "ext/reflection/tests/004.phpt",
    "ext/reflection/tests/ReflectionObject_isUserDefined_error.phpt",
    "ext/reflection/tests/ReflectionClass_getConstants_basic.phpt",
    "ext/reflection/tests/bug46064_2.phpt",
    "ext/reflection/tests/ReflectionClass_getInterfaceNames_basic.phpt",
    "ext/reflection/tests/ReflectionParameter_canBePassedByValue.phpt",
    "ext/reflection/tests/ReflectionClass_getInterfaces_004.phpt",
    "ext/reflection/tests/bug49074.phpt",
    "ext/reflection/tests/closures_001.phpt",
    "ext/reflection/tests/ReflectionClass_getInterfaces_002.phpt",
    "ext/reflection/tests/bug48336.phpt",
    "ext/reflection/tests/ReflectionMethod_returnsReference_basic.phpt",
    "ext/reflection/tests/ReflectionClass_getStaticPropertyValue_001_2_4.phpt",
    "ext/reflection/tests/ReflectionClass_getProperty_004.phpt",
    "ext/reflection/tests/bug40794.phpt",
    "ext/reflection/tests/ReflectionMethod_basic1.phpt",
    "ext/reflection/tests/ReflectionObject_getName_basic.phpt",
    "ext/reflection/tests/ReflectionFunction_construct.001.phpt",
    "ext/reflection/tests/ReflectionClass_hasProperty_001.phpt",
    "ext/reflection/tests/ReflectionClass_newInstanceArgs_002.phpt",
    "ext/reflection/tests/bug33312.phpt",
    "ext/reflection/tests/ReflectionClass_getModifiers_basic.phpt",
    "ext/reflection/tests/ReflectionExtension_getDependencies_variation2.phpt",
    "ext/reflection/tests/ReflectionObject_isSubclassOf.002.phpt",
    "ext/reflection/tests/bug39067.phpt",
    "ext/reflection/tests/ReflectionClass_isInstantiable_basic.phpt",
    "ext/reflection/tests/ReflectionProperty_isDefault_basic.phpt",
    "ext/reflection/tests/parameters_002.phpt",
    "ext/reflection/tests/bug33389.phpt",
    "ext/reflection/tests/ReflectionClass_hasConstant_basic.phpt",
    "ext/reflection/tests/ReflectionParameter_getPosition_basic.phpt",
    "ext/reflection/tests/ReflectionClass_isIterateable_001.phpt",
    "ext/reflection/tests/ReflectionClass_getDefaultProperties_001.phpt",
    "ext/reflection/tests/ReflectionMethod_constructor_basic.phpt",
    "ext/reflection/tests/ReflectionClass_isUserDefined_error.phpt",
    "ext/reflection/tests/ReflectionClass_getProperty_001.phpt",
    "ext/reflection/tests/bug38217.phpt",
    "ext/reflection/tests/bug49092.phpt",
    "ext/reflection/tests/bug30209.phpt",
    "ext/reflection/tests/bug38132.phpt",
    "ext/reflection/tests/003.phpt",
    "ext/reflection/tests/ReflectionClass_getExtension_basic.phpt",
    "ext/reflection/tests/ReflectionParameter_001.phpt",
    "ext/reflection/tests/ReflectionObject_export_basic1.phpt",
    "ext/reflection/tests/bug31651.phpt",
    "ext/reflection/tests/ReflectionObject_isSubclassOf_error.phpt",
    "ext/reflection/tests/ReflectionFunction_getClosureThis.phpt",
    "ext/reflection/tests/bug29986.phpt",
    "ext/reflection/tests/ReflectionFunction_isDeprecated_basic.phpt",
    "ext/reflection/tests/ReflectionFunction_getFileName.002.phpt",
    "ext/reflection/tests/ReflectionMethod_getClosure_error.phpt",
    "ext/reflection/tests/ReflectionFunction_isDisabled_basic.phpt",
    "ext/reflection/tests/bug38653.phpt",
    "ext/reflection/tests/ReflectionClass_getInterfaces_001.phpt",
    "ext/reflection/tests/ReflectionMethod_getDocComment_error.phpt",
    "ext/reflection/tests/ReflectionProperty_error.phpt",
    "ext/reflection/tests/bug48757.phpt",
    "ext/reflection/tests/020.phpt",
    "ext/reflection/tests/022.phpt",
    "ext/reflection/tests/parameters_001.phpt",
    "ext/reflection/tests/ReflectionProperty_constructor_error.phpt",
    "ext/reflection/tests/bug40431.phpt",
    "ext/reflection/tests/bug43926.phpt",
    "ext/reflection/tests/006.phpt",
    "ext/reflection/tests/ReflectionMethod_invoke_error1.phpt",
    "ext/reflection/tests/ReflectionObject_isInstantiable_basic.phpt",
    "ext/reflection/tests/ReflectionClass_getModifierNames_basic.phpt",
    "ext/reflection/tests/ReflectionClass_getName_basic.phpt",
    "ext/reflection/tests/bug32981.phpt",
    "ext/reflection/tests/ReflectionClass_newInstanceArgs_001.phpt",
    "ext/reflection/tests/023.phpt",
    "ext/reflection/tests/bug36337.phpt",
    "ext/reflection/tests/ReflectionObject_isSubclassOf_basic.phpt",
    "ext/reflection/tests/ReflectionProperty_constructor_variation1.phpt",
    "ext/reflection/tests/ReflectionClass_getStaticPropertyValue_002.phpt",
    "ext/reflection/tests/ReflectionObject_constructor_error.phpt",
    "ext/reflection/tests/ReflectionMethod_basic4.phpt",
    "ext/reflection/tests/bug38194.phpt",
    "ext/reflection/tests/ReflectionClass_isCloneable_001.phpt",
    "ext/reflection/tests/ReflectionMethod_getDeclaringClass_basic.phpt",
    "ext/reflection/tests/ReflectionProperty_export_error.phpt",
    "ext/reflection/tests/ReflectionParameter_003.phpt",
    "ext/reflection/tests/bug45571.phpt",
    "ext/reflection/tests/ReflectionMethod_invoke_basic.phpt",
    "ext/reflection/tests/ReflectionMethod_basic2.phpt",
    "ext/reflection/tests/bug26640.phpt",
    "ext/reflection/tests/ReflectionObject_getConstants_basic.phpt",
    "ext/reflection/tests/ReflectionObject_constructor_basic.phpt",
    "ext/reflection/tests/ReflectionProperty_getDeclaringClass_variation1.phpt",
    "ext/reflection/tests/ReflectionProperty_getValue_error.phpt",
    "ext/reflection/tests/ReflectionObject_isInstantiable_error.phpt",
    "ext/reflection/tests/ReflectionMethod_basic3.phpt",
    "ext/reflection/tests/ReflectionClass_getParentClass_001.phpt",
    "ext/reflection/tests/ReflectionClass_constructor_002.phpt",
    "ext/reflection/tests/ReflectionObject_isInstance_error.phpt",
    "ext/reflection/tests/ReflectionFunction_001.phpt",
    "ext/reflection/tests/ReflectionClass_isInstance_basic.phpt",
    "ext/reflection/tests/ReflectionClass_getDocComment_001.phpt",
    "ext/reflection/tests/ReflectionExtension_export_basic.phpt",
    "ext/reflection/tests/ReflectionObject_getConstant_error.phpt",
    "ext/reflection/tests/007.phpt",
    "ext/reflection/tests/ReflectionMethod_invoke_error2.phpt",
    "ext/reflection/tests/ReflectionClass_getProperties_001.phpt",
    "ext/reflection/tests/ReflectionObject___toString_basic1.phpt",
    "ext/reflection/tests/ReflectionClass_isSubclassOf_error.phpt",
    "ext/reflection/tests/ReflectionClass_setStaticPropertyValue_001.phpt",
    "ext/reflection/tests/ReflectionObject_getConstructor_basic.phpt",
    "ext/reflection/tests/ReflectionClass_hasMethod_001.phpt",
    "ext/reflection/tests/bug51905.phpt",
    "ext/reflection/tests/ReflectionFunction_isClosure_basic.phpt",
    "ext/reflection/tests/ReflectionObject_getName_error1.phpt",
    "ext/reflection/tests/ReflectionProperty_getModifiers_basic.phpt",
    "ext/reflection/tests/ReflectionExtension_getClassNames_basic.phpt",
    "ext/reflection/tests/ReflectionClass_hasConstant_002.phpt",
    "ext/reflection/tests/002.phpt",
    "ext/reflection/tests/static_properties_002.phpt",
    "ext/reflection/tests/008.phpt",
    "ext/reflection/tests/018.phpt",
    "ext/reflection/tests/ReflectionMethod_constructor_error1.phpt",
    "ext/reflection/tests/ReflectionFunction_getClosure_basic.phpt",
    "ext/reflection/tests/ReflectionClass_export_basic2.phpt",
    "ext/reflection/tests/ReflectionClass_getDefaultProperties_002.phpt",
    "ext/reflection/tests/ReflectionMethod_getStaticVariables_basic.phpt",
    "ext/reflection/tests/ReflectionClass_isFinal_basic.phpt",
    "ext/reflection/tests/ReflectionExtension_getVersion_basic.phpt",
    "ext/reflection/tests/ReflectionClass_getName_error.phpt",
    "ext/reflection/tests/ReflectionFunction_getNamespaceName.phpt",
    "ext/reflection/tests/ReflectionMethod_invokeArgs_error1.phpt",
    "ext/reflection/tests/ReflectionClass_getProperties_003.phpt",
    "ext/reflection/tests/ReflectionClass_isInstantiable_variation.phpt",
    "ext/reflection/tests/ReflectionClass_getMethod_002.phpt",
    "ext/reflection/tests/013.phpt",
    "ext/reflection/tests/ReflectionProperty_getModifiers.001.phpt",
    "ext/reflection/tests/016.phpt",
    "ext/reflection/tests/ReflectionClass_modifiers_001.phpt",
    "ext/reflection/tests/ReflectionMethod_invokeArgs_error3.phpt",
    "ext/reflection/tests/ReflectionObject_getConstants_error.phpt",
    "ext/reflection/tests/ReflectionClass_getParentClass.phpt",
    "ext/reflection/tests/ReflectionClass_getProperty_002.phpt",
    "ext/reflection/tests/ReflectionClass_implementsInterface_001.phpt",
    "ext/reflection/tests/ReflectionMethod_invokeArgs_error2.phpt",
    "ext/reflection/tests/closures_003_v1.phpt",
    "ext/reflection/tests/bug30856.phpt",
    "ext/reflection/tests/025.phpt",
    "ext/reflection/tests/bug36434.phpt",
    "ext/reflection/tests/ReflectionExtension_constructor_basic.phpt",
    "ext/reflection/tests/ReflectionClass_isIterateable_variation1.phpt",
    "ext/reflection/tests/ReflectionClass_getStaticPropertyValue_001.phpt",
    "ext/reflection/tests/012.phpt",
    "ext/reflection/tests/ReflectionObject_export_basic2.phpt",
    "ext/reflection/tests/021.phpt",
    "ext/reflection/tests/ReflectionClass_getExtension_variation.phpt",
    "ext/reflection/tests/ReflectionClass_FileInfo_error.phpt",
    "ext/reflection/tests/ReflectionClass_getInterfaces_003.phpt",
    "ext/reflection/tests/ReflectionMethod_getClosure_basic.phpt",
    "ext/reflection/tests/ReflectionClass_hasProperty_basic.phpt",
    "ext/reflection/tests/bug29268.phpt",
    "ext/reflection/tests/ReflectionClass_isSubclassOf_002.phpt",
    "ext/reflection/tests/traits002.phpt",
    "ext/reflection/tests/ReflectionObject_isInternal_error.phpt",
    "ext/reflection/tests/ReflectionMethod_getClosureThis.phpt",
    "ext/reflection/tests/ReflectionClass_isCloneable_002.phpt",
    "ext/reflection/tests/traits005.phpt",
    "ext/reflection/tests/ReflectionFunction_getDocComment.001.phpt",
    "ext/reflection/tests/ReflectionParameter_export_error.phpt",
    "ext/reflection/tests/ReflectionParameter_002.phpt",
    "ext/reflection/tests/ReflectionFunction_getExtension.phpt",
    "ext/reflection/tests/ReflectionObject_export_basic3.phpt",
    "ext/reflection/tests/ReflectionClass_isInstantiable_error.phpt",
    "ext/reflection/tests/bug42976.phpt",
    "ext/reflection/tests/closures_004.phpt",
    "ext/reflection/tests/ReflectionClass_isIterateable_basic.phpt",
    "ext/reflection/tests/ReflectionClass_CannotClone_basic.phpt",
    "ext/reflection/tests/ReflectionFunction_getClosure_error.phpt",
    "ext/reflection/tests/ReflectionExtension_getClasses_basic.phpt",
    "ext/reflection/tests/bug46205.phpt",
    "ext/reflection/tests/ReflectionClass_setStaticPropertyValue_002.phpt",
    "ext/reflection/tests/ReflectionProperty_setValue_error.phpt",
    "ext/reflection/tests/ReflectionProperty_basic1.phpt",
    "ext/reflection/tests/ReflectionClass_toString_003.phpt",
    "ext/reflection/tests/ReflectionProperty_basic2.phpt",
    "ext/reflection/tests/ReflectionParameter_getDeclaringFunction_basic.phpt",
    "ext/reflection/tests/closures_002.phpt",
    "ext/reflection/tests/ReflectionMethod_constructor_error2.phpt",
    "ext/reflection/tests/ReflectionObject_getConstant_basic.phpt",
    "ext/reflection/tests/019.phpt",
    "ext/reflection/tests/ReflectionExtension_isPersistant.phpt",
    "ext/reflection/tests/ReflectionProperty_setAccessible.phpt",
    "ext/reflection/tests/ReflectionClass_getConstant_basic.phpt",
    "ext/reflection/tests/ReflectionObject___toString_basic2.phpt",
    "ext/reflection/tests/ReflectionExtension_getName_basic.phpt",
    "ext/reflection/tests/ReflectionProperty_getDocComment_basic.phpt",
    "ext/reflection/tests/bug29523.phpt",
    "ext/reflection/tests/ReflectionObject_FileInfo_error.phpt",
    "ext/reflection/tests/024.phpt",
    "ext/reflection/tests/ReflectionClass_isSubclassOf_basic.phpt",
    "ext/reflection/tests/ReflectionClass_isUserDefined_basic.phpt",
    "ext/reflection/tests/ReflectionClass_getMethod_001.phpt",
    "ext/reflection/tests/ReflectionParameter_isDefault.phpt",
    "ext/reflection/tests/traits003.phpt",
    "ext/reflection/tests/ReflectionParameter_toString_basic.phpt",
    "ext/reflection/tests/009.phpt",
    "ext/reflection/tests/ReflectionClass_getExtensionName_variation.phpt",
    "ext/reflection/tests/ReflectionClass_hasProperty_002.phpt",
    "ext/reflection/tests/bug46064.phpt",
    "ext/reflection/tests/ReflectionClass_getExtensionName_basic.phpt",
    "ext/reflection/tests/ReflectionExtension_getDependencies_basic.phpt",
    "ext/reflection/tests/bug29828.phpt",
    "ext/reflection/tests/bug47254.phpt",
    "ext/reflection/tests/ReflectionClass_hasMethod_002.phpt",
    "ext/reflection/tests/ReflectionObject_FileInfo_basic.phpt",
    "ext/reflection/tests/ReflectionClass_isInstance_error.phpt",
    "ext/reflection/tests/bug41884.phpt",
    "ext/reflection/tests/ReflectionObject_isUserDefined_basic.phpt",
    "ext/reflection/tests/bug53366.phpt",
    "ext/reflection/tests/bug30961.phpt",
    "ext/xsl/tests/xslt012.phpt",
    "ext/xsl/tests/xsltprocessor_registerPHPFunctions-null.phpt",
    "ext/xsl/tests/xslt006.phpt",
    "ext/xsl/tests/xsltprocessor_removeParameter.phpt",
    "ext/xsl/tests/xslt011.phpt",
    "ext/xsl/tests/xsltprocessor_registerPHPFunctions-string-notallowed.phpt",
    "ext/xsl/tests/xsltprocessor_registerPHPFunctions-string.phpt",
    "ext/xsl/tests/xsl-phpinfo.phpt",
    "ext/xsl/tests/xsltprocessor_registerPHPFunctions-funcundef.phpt",
    "ext/xsl/tests/xslt004.phpt",
    "ext/xsl/tests/xslt003.phpt",
    "ext/xsl/tests/xslt001.phpt",
    "ext/xsl/tests/xsltprocessor_setparameter-nostring.phpt",
    "ext/xsl/tests/xslt010.phpt",
    "ext/xsl/tests/xsltprocessor_removeParameter-wrongparams.phpt",
    "ext/xsl/tests/xsltprocessor_registerPHPFunctions-array.phpt",
    "ext/xsl/tests/xsltprocessor_getParameter-wrongparam.phpt",
    "ext/xsl/tests/xslt008.phpt",
    "ext/xsl/tests/bug26384.phpt",
    "ext/xsl/tests/xslt007.phpt",
    "ext/xsl/tests/xsltprocessor_registerPHPFunctions-array-multiple.phpt",
    "ext/xsl/tests/bug33853.phpt",
    "ext/xsl/tests/xsltprocessor_registerPHPFunctions-funcnostring.phpt",
    "ext/xsl/tests/xsltprocessor_registerPHPFunctions-allfuncs.phpt",
    "ext/xsl/tests/xslt009.phpt",
    "ext/xsl/tests/xslt002.phpt",
    "ext/xsl/tests/bug48221.phpt",
    "ext/xsl/tests/xslt005.phpt",
    "ext/xsl/tests/xsltprocessor_registerPHPFunctions-string-multiple.phpt",
    "ext/xsl/tests/xsltprocessor_registerPHPFunctions-array-notallowed.phpt",
    "ext/xsl/tests/xsltprocessor_setparameter-errorquote.phpt",
    "ext/xsl/tests/xsltprocessor_getParameter.phpt",
    "ext/xsl/tests/xsltprocessor_getParameter-invalidparam.phpt",
    "ext/xsl/tests/xsltprocessor_removeParameter-invalidparam.phpt",
    "ext/sysvshm/tests/005.phpt",
    "ext/sysvshm/tests/001.phpt",
    "ext/sysvshm/tests/004.phpt",
    "ext/sysvshm/tests/003.phpt",
    "ext/sysvshm/tests/006.phpt",
    "ext/sysvshm/tests/007.phpt",
    "ext/sysvshm/tests/002.phpt",
    "ext/filter/tests/050.phpt",
    "ext/filter/tests/027.phpt",
    "ext/filter/tests/bug48762.phpt",
    "ext/filter/tests/033.phpt",
    "ext/filter/tests/bug39846.phpt",
    "ext/filter/tests/filter_data.phpt",
    "ext/filter/tests/010.phpt",
    "ext/filter/tests/045.phpt",
    "ext/filter/tests/005.phpt",
    "ext/filter/tests/bug53150.phpt",
    "ext/filter/tests/054.phpt",
    "ext/filter/tests/callback_closure.phpt",
    "ext/filter/tests/014.phpt",
    "ext/filter/tests/044.phpt",
    "ext/filter/tests/029.phpt",
    "ext/filter/tests/bug53037.phpt",
    "ext/filter/tests/011.phpt",
    "ext/filter/tests/017.phpt",
    "ext/filter/tests/036.phpt",
    "ext/filter/tests/bug39763.phpt",
    "ext/filter/tests/bug42718-2.phpt",
    "ext/filter/tests/043.phpt",
    "ext/filter/tests/028.phpt",
    "ext/filter/tests/001.phpt",
    "ext/filter/tests/041.phpt",
    "ext/filter/tests/bug49274.phpt",
    "ext/filter/tests/015.phpt",
    "ext/filter/tests/bug7715.phpt",
    "ext/filter/tests/026.phpt",
    "ext/filter/tests/004.phpt",
    "ext/filter/tests/bug7733.phpt",
    "ext/filter/tests/bug44445.phpt",
    "ext/filter/tests/048.phpt",
    "ext/filter/tests/bug8315.phpt",
    "ext/filter/tests/003.phpt",
    "ext/filter/tests/049.phpt",
    "ext/filter/tests/020.phpt",
    "ext/filter/tests/022.phpt",
    "ext/filter/tests/006.phpt",
    "ext/filter/tests/031.phpt",
    "ext/filter/tests/023.phpt",
    "ext/filter/tests/046.phpt",
    "ext/filter/tests/callback_non_modified_var.phpt",
    "ext/filter/tests/bug46973.phpt",
    "ext/filter/tests/bug52929.phpt",
    "ext/filter/tests/007.phpt",
    "ext/filter/tests/bug44779.phpt",
    "ext/filter/tests/bug42718.phpt",
    "ext/filter/tests/PMOPB45.phpt",
    "ext/filter/tests/002.phpt",
    "ext/filter/tests/008.phpt",
    "ext/filter/tests/018.phpt",
    "ext/filter/tests/040.phpt",
    "ext/filter/tests/034.phpt",
    "ext/filter/tests/039.phpt",
    "ext/filter/tests/035.phpt",
    "ext/filter/tests/013.phpt",
    "ext/filter/tests/037.phpt",
    "ext/filter/tests/016.phpt",
    "ext/filter/tests/bug52209.phpt",
    "ext/filter/tests/047.phpt",
    "ext/filter/tests/025.phpt",
    "ext/filter/tests/012.phpt",
    "ext/filter/tests/038.phpt",
    "ext/filter/tests/bug50632.phpt",
    "ext/filter/tests/bug50158.phpt",
    "ext/filter/tests/021.phpt",
    "ext/filter/tests/052.phpt",
    "ext/filter/tests/bug7586.phpt",
    "ext/filter/tests/030.phpt",
    "ext/filter/tests/032.phpt",
    "ext/filter/tests/019.phpt",
    "ext/filter/tests/053.phpt",
    "ext/filter/tests/051.phpt",
    "ext/filter/tests/024.phpt",
    "ext/filter/tests/009.phpt",
    "ext/filter/tests/bug47745.phpt",
    "ext/filter/tests/042.phpt",
    "ext/filter/tests/bug51192.phpt",
    "ext/filter/tests/bug47435.phpt",
    "ext/simplexml/tests/027.phpt",
    "ext/simplexml/tests/sxe_003.phpt",
    "ext/simplexml/tests/profile06.phpt",
    "ext/simplexml/tests/033.phpt",
    "ext/simplexml/tests/010.phpt",
    "ext/simplexml/tests/005.phpt",
    "ext/simplexml/tests/profile09.phpt",
    "ext/simplexml/tests/bug41582.phpt",
    "ext/simplexml/tests/014.phpt",
    "ext/simplexml/tests/029.phpt",
    "ext/simplexml/tests/profile04.phpt",
    "ext/simplexml/tests/simplexml_load_file.phpt",
    "ext/simplexml/tests/011.phpt",
    "ext/simplexml/tests/profile03.phpt",
    "ext/simplexml/tests/017.phpt",
    "ext/simplexml/tests/bug38354.phpt",
    "ext/simplexml/tests/036.phpt",
    "ext/simplexml/tests/SimpleXMLElement_addAttribute_basic.phpt",
    "ext/simplexml/tests/bug42369.phpt",
    "ext/simplexml/tests/bug39760.phpt",
    "ext/simplexml/tests/profile10.phpt",
    "ext/simplexml/tests/028.phpt",
    "ext/simplexml/tests/profile13.phpt",
    "ext/simplexml/tests/001.phpt",
    "ext/simplexml/tests/profile02.phpt",
    "ext/simplexml/tests/009b.phpt",
    "ext/simplexml/tests/015.phpt",
    "ext/simplexml/tests/bug41867.phpt",
    "ext/simplexml/tests/026.phpt",
    "ext/simplexml/tests/004.phpt",
    "ext/simplexml/tests/bug44478.phpt",
    "ext/simplexml/tests/016a.phpt",
    "ext/simplexml/tests/bug51615.phpt",
    "ext/simplexml/tests/000.phpt",
    "ext/simplexml/tests/profile05.phpt",
    "ext/simplexml/tests/bug24392.phpt",
    "ext/simplexml/tests/bug45553.phpt",
    "ext/simplexml/tests/003.phpt",
    "ext/simplexml/tests/profile11.phpt",
    "ext/simplexml/tests/020.phpt",
    "ext/simplexml/tests/022.phpt",
    "ext/simplexml/tests/006.phpt",
    "ext/simplexml/tests/031.phpt",
    "ext/simplexml/tests/bug41175.phpt",
    "ext/simplexml/tests/bug41947.phpt",
    "ext/simplexml/tests/023.phpt",
    "ext/simplexml/tests/simplexml_import_dom.phpt",
    "ext/simplexml/tests/sxe_002.phpt",
    "ext/simplexml/tests/bug37076_1.phpt",
    "ext/simplexml/tests/bug38406.phpt",
    "ext/simplexml/tests/bug48601.phpt",
    "ext/simplexml/tests/bug38424.phpt",
    "ext/simplexml/tests/007.phpt",
    "ext/simplexml/tests/profile12.phpt",
    "ext/simplexml/tests/bug37076.phpt",
    "ext/simplexml/tests/002.phpt",
    "ext/simplexml/tests/SimpleXMLElement_addAttribute_required_attribute_name.phpt",
    "ext/simplexml/tests/bug40451.phpt",
    "ext/simplexml/tests/008.phpt",
    "ext/simplexml/tests/profile07.phpt",
    "ext/simplexml/tests/018.phpt",
    "ext/simplexml/tests/034.phpt",
    "ext/simplexml/tests/035.phpt",
    "ext/simplexml/tests/profile01.phpt",
    "ext/simplexml/tests/013.phpt",
    "ext/simplexml/tests/016.phpt",
    "ext/simplexml/tests/bug26976.phpt",
    "ext/simplexml/tests/bug36611.phpt",
    "ext/simplexml/tests/bug46003.phpt",
    "ext/simplexml/tests/bug27010.phpt",
    "ext/simplexml/tests/025.phpt",
    "ext/simplexml/tests/012.phpt",
    "ext/simplexml/tests/bug41861.phpt",
    "ext/simplexml/tests/021.phpt",
    "ext/simplexml/tests/014a.phpt",
    "ext/simplexml/tests/sxe_005.phpt",
    "ext/simplexml/tests/bug37565.phpt",
    "ext/simplexml/tests/bug37386.phpt",
    "ext/simplexml/tests/bug46048.phpt",
    "ext/simplexml/tests/030.phpt",
    "ext/simplexml/tests/bug46047.phpt",
    "ext/simplexml/tests/profile08.phpt",
    "ext/simplexml/tests/bug38347.phpt",
    "ext/simplexml/tests/sxe_004.phpt",
    "ext/simplexml/tests/032.phpt",
    "ext/simplexml/tests/bug43221.phpt",
    "ext/simplexml/tests/014b.phpt",
    "ext/simplexml/tests/019.phpt",
    "ext/simplexml/tests/024.phpt",
    "ext/simplexml/tests/009.phpt",
    "ext/simplexml/tests/bug35785.phpt",
    "ext/simplexml/tests/sxe_001.phpt",
    "ext/simplexml/tests/bug42259.phpt",
    "ext/simplexml/tests/bug39662.phpt",
    "ext/bcmath/tests/bcpowmod_error1.phpt",
    "ext/bcmath/tests/bcadd_error1.phpt",
    "ext/bcmath/tests/bcmod_error2.phpt",
    "ext/bcmath/tests/bcdiv_error1.phpt",
    "ext/bcmath/tests/bcsqrt.phpt",
    "ext/bcmath/tests/bcscale_variation001.phpt",
    "ext/bcmath/tests/bcpow.phpt",
    "ext/bcmath/tests/bcpowmod_error3.phpt",
    "ext/bcmath/tests/bcpowmod_error2.phpt",
    "ext/bcmath/tests/bcmod.phpt",
    "ext/bcmath/tests/bcscale_variation002.phpt",
    "ext/bcmath/tests/bccomp_variation002.phpt",
    "ext/bcmath/tests/bcpow_variation001.phpt",
    "ext/bcmath/tests/bcsqrt_error1.phpt",
    "ext/bcmath/tests/bcmul.phpt",
    "ext/bcmath/tests/bccomp_variation001.phpt",
    "ext/bcmath/tests/bcpowmod.phpt",
    "ext/bcmath/tests/bcdiv_error2.phpt",
    "ext/bcmath/tests/bcdiv.phpt",
    "ext/bcmath/tests/bcpow_error3.phpt",
    "ext/bcmath/tests/bcadd_variation001.phpt",
    "ext/bcmath/tests/bcsub.phpt",
    "ext/bcmath/tests/bcscale.phpt",
    "ext/bcmath/tests/bcmod_error1.phpt",
    "ext/bcmath/tests/bcsub_error1.phpt",
    "ext/bcmath/tests/bccomp.phpt",
    "ext/bcmath/tests/bcmul_error1.phpt",
    "ext/bcmath/tests/bcsqrt_variation001.phpt",
    "ext/bcmath/tests/bcadd.phpt",
    "ext/bcmath/tests/bcsqrt_error2.phpt",
    "ext/bz2/tests/with_files.phpt",
    "ext/bz2/tests/005.phpt",
    "ext/bz2/tests/bz2_filter_compress.phpt",
    "ext/bz2/tests/001.phpt",
    "ext/bz2/tests/004.phpt",
    "ext/bz2/tests/bz2_filter_decompress.phpt",
    "ext/bz2/tests/003.phpt",
    "ext/bz2/tests/002.phpt",
    "ext/bz2/tests/with_strings.phpt",
    "ext/libxml/tests/001.phpt",
    "ext/libxml/tests/bug42112.phpt",
    "ext/libxml/tests/004.phpt",
    "ext/libxml/tests/003.phpt",
    "ext/libxml/tests/002.phpt",
    "ext/libxml/tests/bug54440.phpt",
    "ext/phar/tests/027.phpt",
    "ext/phar/tests/rename_dir.phpt",
    "ext/phar/tests/open_for_write_newfile_b.phpt",
    "ext/phar/tests/pharfileinfo_construct.phpt",
    "ext/phar/tests/frontcontroller16.phpt",
    "ext/phar/tests/bug46060.phpt",
    "ext/phar/tests/033.phpt",
    "ext/phar/tests/phar_get_supportedcomp4.phpt",
    "ext/phar/tests/bug48377.phpt",
    "ext/phar/tests/mkdir.phpt",
    "ext/phar/tests/010.phpt",
    "ext/phar/tests/phar_buildfromdirectory4.phpt",
    "ext/phar/tests/frontcontroller32.phpt",
    "ext/phar/tests/opendir.phpt",
    "ext/phar/tests/phar_unlinkarchive.phpt",
    "ext/phar/tests/phar_offset_check.phpt",
    "ext/phar/tests/ini_set.phpt",
    "ext/phar/tests/005.phpt",
    "ext/phar/tests/bug46178.phpt",
    "ext/phar/tests/phar_buildfromiterator8.phpt",
    "ext/phar/tests/nophar_web.phpt",
    "ext/phar/tests/fopen_edgecases.phpt",
    "ext/phar/tests/security.phpt",
    "ext/phar/tests/create_new_phar_c.phpt",
    "ext/phar/tests/phar_setdefaultstub.phpt",
    "ext/phar/tests/phar_oo_003.phpt",
    "ext/phar/tests/phar_construct_invalidurl.phpt",
    "ext/phar/tests/frontcontroller1.phpt",
    "ext/phar/tests/bug48377.2.phpt",
    "ext/phar/tests/014.phpt",
    "ext/phar/tests/phar_oo_getcontentsgz.phpt",
    "ext/phar/tests/frontcontroller2.phpt",
    "ext/phar/tests/029.phpt",
    "ext/phar/tests/phar_stub_write.phpt",
    "ext/phar/tests/phar_gobyebye.phpt",
    "ext/phar/tests/phar_buildfromiterator5.phpt",
    "ext/phar/tests/phar_convert_tar.phpt",
    "ext/phar/tests/phar_oo_getcontents.phpt",
    "ext/phar/tests/frontcontroller25.phpt",
    "ext/phar/tests/phar_oo_005_5.2.phpt",
    "ext/phar/tests/phar_begin_setstub_commitU.phpt",
    "ext/phar/tests/phpinfo_004.phpt",
    "ext/phar/tests/phar_oo_iswriteable.phpt",
    "ext/phar/tests/011.phpt",
    "ext/phar/tests/phar_ctx_001.phpt",
    "ext/phar/tests/phar_convert_tar3.phpt",
    "ext/phar/tests/019bU.phpt",
    "ext/phar/tests/phar_create_in_cwd.phpt",
    "ext/phar/tests/017.phpt",
    "ext/phar/tests/phar_extract3.phpt",
    "ext/phar/tests/phar_oo_002U.phpt",
    "ext/phar/tests/open_for_write_existing_c.phpt",
    "ext/phar/tests/open_for_write_newfile_b_5_2.phpt",
    "ext/phar/tests/metadata_write.phpt",
    "ext/phar/tests/bug13727.phpt",
    "ext/phar/tests/cached_manifest_1U.phpt",
    "ext/phar/tests/phar_copy.phpt",
    "ext/phar/tests/pharfileinfo_compression.phpt",
    "ext/phar/tests/frontcontroller18.phpt",
    "ext/phar/tests/metadata_writeU.phpt",
    "ext/phar/tests/phar_gzip.phpt",
    "ext/phar/tests/opendir_edgecases.phpt",
    "ext/phar/tests/open_for_write_newfile_c.phpt",
    "ext/phar/tests/phar_oo_002.phpt",
    "ext/phar/tests/phar_oo_007.phpt",
    "ext/phar/tests/readfile_edgecases.phpt",
    "ext/phar/tests/phar_buildfromiterator1.phpt",
    "ext/phar/tests/frontcontroller5.phpt",
    "ext/phar/tests/pharfileinfo_setmetadata.phpt",
    "ext/phar/tests/028.phpt",
    "ext/phar/tests/bug45218_SLOWTESTU.phpt",
    "ext/phar/tests/mounteddirU.phpt",
    "ext/phar/tests/phar_metadata_read.phpt",
    "ext/phar/tests/open_for_write_existing_b.phpt",
    "ext/phar/tests/file_get_contents.phpt",
    "ext/phar/tests/001.phpt",
    "ext/phar/tests/readfile.phpt",
    "ext/phar/tests/pharfileinfo_getcrc32.phpt",
    "ext/phar/tests/bug47085.phpt",
    "ext/phar/tests/015.phpt",
    "ext/phar/tests/frontcontroller12.phpt",
    "ext/phar/tests/026.phpt",
    "ext/phar/tests/phar_oo_005.phpt",
    "ext/phar/tests/withphar.phpt",
    "ext/phar/tests/frontcontroller33.phpt",
    "ext/phar/tests/phar_setalias.phpt",
    "ext/phar/tests/004.phpt",
    "ext/phar/tests/phar_oo_012_confirm.phpt",
    "ext/phar/tests/badparameters.phpt",
    "ext/phar/tests/delete_in_phar_b.phpt",
    "ext/phar/tests/stat.phpt",
    "ext/phar/tests/phar_oo_010.phpt",
    "ext/phar/tests/phar_begin_setstub_commit.phpt",
    "ext/phar/tests/frontcontroller29.phpt",
    "ext/phar/tests/frontcontroller19.phpt",
    "ext/phar/tests/phar_oo_005U.phpt",
    "ext/phar/tests/open_for_write_newfile.phpt",
    "ext/phar/tests/phar_buildfromiterator2.phpt",
    "ext/phar/tests/webphar_compilefail.phpt",
    "ext/phar/tests/phar_buildfromdirectory5.phpt",
    "ext/phar/tests/frontcontroller34.phpt",
    "ext/phar/tests/metadata_read.phpt",
    "ext/phar/tests/frontcontroller17.phpt",
    "ext/phar/tests/phar_oo_uncompressall.phpt",
    "ext/phar/tests/phar_get_supportedcomp2.phpt",
    "ext/phar/tests/test_signaturealgos.phpt",
    "ext/phar/tests/003.phpt",
    "ext/phar/tests/frontcontroller20.phpt",
    "ext/phar/tests/phar_setsignaturealgo1.phpt",
    "ext/phar/tests/phar_extract2.phpt",
    "ext/phar/tests/stat2.phpt",
    "ext/phar/tests/phar_oo_compressallgz.phpt",
    "ext/phar/tests/phar_dotted_path.phpt",
    "ext/phar/tests/phar_decompress.phpt",
    "ext/phar/tests/open_for_write_existing_c_5_2.phpt",
    "ext/phar/tests/phar_gzipU.phpt",
    "ext/phar/tests/frontcontroller11.phpt",
    "ext/phar/tests/phar_extract.phpt",
    "ext/phar/tests/rmdir.phpt",
    "ext/phar/tests/frontcontroller14.phpt",
    "ext/phar/tests/015b.phpt",
    "ext/phar/tests/frontcontroller9.phpt",
    "ext/phar/tests/027U.phpt",
    "ext/phar/tests/fopen_edgecases2.phpt",
    "ext/phar/tests/020.phpt",
    "ext/phar/tests/022.phpt",
    "ext/phar/tests/phar_convert_repeated.phpt",
    "ext/phar/tests/006.phpt",
    "ext/phar/tests/031.phpt",
    "ext/phar/tests/phar_oo_011.phpt",
    "ext/phar/tests/frontcontroller22.phpt",
    "ext/phar/tests/test_alias_unset.phpt",
    "ext/phar/tests/phar_buildfromdirectory6.phpt",
    "ext/phar/tests/phar_convert_repeated_b.phpt",
    "ext/phar/tests/003a.phpt",
    "ext/phar/tests/phar_buildfromiterator4.phpt",
    "ext/phar/tests/invalid_setstubalias.phpt",
    "ext/phar/tests/023.phpt",
    "ext/phar/tests/phar_setalias2.phpt",
    "ext/phar/tests/019c.phpt",
    "ext/phar/tests/phar_oo_012.phpt",
    "ext/phar/tests/phar_get_supported_signatures_001a.phpt",
    "ext/phar/tests/ini_set_off.phpt",
    "ext/phar/tests/phar_stub.phpt",
    "ext/phar/tests/fgc_edgecases.phpt",
    "ext/phar/tests/phar_buildfromiterator7.phpt",
    "ext/phar/tests/phar_dir_iterate.phpt",
    "ext/phar/tests/stat2_5.3.phpt",
    "ext/phar/tests/refcount1_5_2.phpt",
    "ext/phar/tests/frontcontroller6.phpt",
    "ext/phar/tests/metadata_write_commit.phpt",
    "ext/phar/tests/phar_createdefaultstub.phpt",
    "ext/phar/tests/phar_stub_write_file.phpt",
    "ext/phar/tests/bug52013.phpt",
    "ext/phar/tests/frontcontroller7.phpt",
    "ext/phar/tests/phar_oo_001U.phpt",
    "ext/phar/tests/phpinfo_002.phpt",
    "ext/phar/tests/cached_manifest_1.phpt",
    "ext/phar/tests/016b.phpt",
    "ext/phar/tests/phar_oo_compressed_002.phpt",
    "ext/phar/tests/phar_buildfromdirectory1.phpt",
    "ext/phar/tests/007.phpt",
    "ext/phar/tests/test_unset.phpt",
    "ext/phar/tests/phar_oo_getmodified.phpt",
    "ext/phar/tests/phar_oo_compressed_001b.phpt",
    "ext/phar/tests/readfile_edgecasesU.phpt",
    "ext/phar/tests/create_new_phar.phpt",
    "ext/phar/tests/019b.phpt",
    "ext/phar/tests/delete_in_phar.phpt",
    "ext/phar/tests/bug54395.phpt",
    "ext/phar/tests/phar_get_supportedcomp1.phpt",
    "ext/phar/tests/tar/rename_dir.phpt",
    "ext/phar/tests/tar/open_for_write_newfile_b.phpt",
    "ext/phar/tests/tar/033.phpt",
    "ext/phar/tests/tar/tar_openssl_hash.phpt",
    "ext/phar/tests/tar/tar_nostub.phpt",
    "ext/phar/tests/tar/frontcontroller19.phar.phpt",
    "ext/phar/tests/tar/frontcontroller17.phar.phpt",
    "ext/phar/tests/tar/frontcontroller18.phar.phpt",
    "ext/phar/tests/tar/phar_buildfromiterator8.phpt",
    "ext/phar/tests/tar/phar_setdefaultstub.phpt",
    "ext/phar/tests/tar/frontcontroller13.phar.phpt",
    "ext/phar/tests/tar/phar_buildfromiterator5.phpt",
    "ext/phar/tests/tar/frontcontroller5.phar.phpt",
    "ext/phar/tests/tar/phar_begin_setstub_commitU.phpt",
    "ext/phar/tests/tar/links6.phpt",
    "ext/phar/tests/tar/tar_makegz.phpt",
    "ext/phar/tests/tar/tar_004U.phpt",
    "ext/phar/tests/tar/frontcontroller20.phar.phpt",
    "ext/phar/tests/tar/frontcontroller11.phar.phpt",
    "ext/phar/tests/tar/open_for_write_existing_c.phpt",
    "ext/phar/tests/tar/open_for_write_newfile_b_5_2.phpt",
    "ext/phar/tests/tar/frontcontroller21.phar.phpt",
    "ext/phar/tests/tar/phar_copy.phpt",
    "ext/phar/tests/tar/tar_bz2.phpt",
    "ext/phar/tests/tar/open_for_write_newfile_c.phpt",
    "ext/phar/tests/tar/links3.phpt",
    "ext/phar/tests/tar/open_for_write_existing_b.phpt",
    "ext/phar/tests/tar/exists_as_phar.phpt",
    "ext/phar/tests/tar/frontcontroller2.phar.phpt",
    "ext/phar/tests/tar/links5.phpt",
    "ext/phar/tests/tar/phar_setalias.phpt",
    "ext/phar/tests/tar/delete_in_phar_b.phpt",
    "ext/phar/tests/tar/frontcontroller9.phar.phpt",
    "ext/phar/tests/tar/phar_begin_setstub_commit.phpt",
    "ext/phar/tests/tar/bignames_overflow.phpt",
    "ext/phar/tests/tar/bug49910.phpt",
    "ext/phar/tests/tar/open_for_write_newfile.phpt",
    "ext/phar/tests/tar/allU.phpt",
    "ext/phar/tests/tar/frontcontroller3.phar.phpt",
    "ext/phar/tests/tar/require_hash.phpt",
    "ext/phar/tests/tar/phar_convert_phar3.phpt",
    "ext/phar/tests/tar/open_for_write_existing_c_5_2.phpt",
    "ext/phar/tests/tar/tar_002.phpt",
    "ext/phar/tests/tar/rmdir.phpt",
    "ext/phar/tests/tar/phar_convert_phar.phpt",
    "ext/phar/tests/tar/truncated.phpt",
    "ext/phar/tests/tar/phar_buildfromiterator4.phpt",
    "ext/phar/tests/tar/bignames.phpt",
    "ext/phar/tests/tar/phar_setalias2.phpt",
    "ext/phar/tests/tar/tar_makebz2.phpt",
    "ext/phar/tests/tar/phar_stub.phpt",
    "ext/phar/tests/tar/phar_buildfromiterator7.phpt",
    "ext/phar/tests/tar/refcount1_5_2.phpt",
    "ext/phar/tests/tar/frontcontroller8.phar.phpt",
    "ext/phar/tests/tar/links2.phpt",
    "ext/phar/tests/tar/frontcontroller16.phar.phpt",
    "ext/phar/tests/tar/tar_001.phpt",
    "ext/phar/tests/tar/tar_gzip.phpt",
    "ext/phar/tests/tar/delete_in_phar.phpt",
    "ext/phar/tests/tar/delete.phpt",
    "ext/phar/tests/tar/phar_magic.phpt",
    "ext/phar/tests/tar/open_for_write_newfile_c_5_2.phpt",
    "ext/phar/tests/tar/frontcontroller12.phar.phpt",
    "ext/phar/tests/tar/frontcontroller10.phar.phpt",
    "ext/phar/tests/tar/tar_004.phpt",
    "ext/phar/tests/tar/phar_setsignaturealgo2.phpt",
    "ext/phar/tests/tar/create_new_and_modify.phpt",
    "ext/phar/tests/tar/dir.phpt",
    "ext/phar/tests/tar/create_new_phar_b.phpt",
    "ext/phar/tests/tar/badalias.phpt",
    "ext/phar/tests/tar/phar_convert_phar2.phpt",
    "ext/phar/tests/tar/rename.phpt",
    "ext/phar/tests/tar/refcount1.phpt",
    "ext/phar/tests/tar/tar_003.phpt",
    "ext/phar/tests/tar/frontcontroller7.phar.phpt",
    "ext/phar/tests/tar/frontcontroller4.phar.phpt",
    "ext/phar/tests/tar/links4.phpt",
    "ext/phar/tests/tar/phar_convert_phar4.phpt",
    "ext/phar/tests/tar/tar_nohash.phpt",
    "ext/phar/tests/tar/alias_acrobatics.phpt",
    "ext/phar/tests/tar/033a.phpt",
    "ext/phar/tests/tar/open_for_write_existing_b_5_2.phpt",
    "ext/phar/tests/tar/frontcontroller15.phar.phpt",
    "ext/phar/tests/tar/phar_buildfromiterator9.phpt",
    "ext/phar/tests/tar/phar_buildfromiterator6.phpt",
    "ext/phar/tests/tar/badchecksum.phpt",
    "ext/phar/tests/tar/links.phpt",
    "ext/phar/tests/tar/frontcontroller6.phar.phpt",
    "ext/phar/tests/tar/frontcontroller1.phar.phpt",
    "ext/phar/tests/tar/phar_commitwrite.phpt",
    "ext/phar/tests/tar/phar_stub_error.phpt",
    "ext/phar/tests/tar/tar_bz2U.phpt",
    "ext/phar/tests/tar/delete_in_phar_confirm.phpt",
    "ext/phar/tests/tar/tar_gzipU.phpt",
    "ext/phar/tests/tar/all.phpt",
    "ext/phar/tests/tar/frontcontroller14.phar.phpt",
    "ext/phar/tests/tar/open_for_write_existing.phpt",
    "ext/phar/tests/frontcontroller23.phpt",
    "ext/phar/tests/delete.phpt",
    "ext/phar/tests/002.phpt",
    "ext/phar/tests/frontcontroller15.phpt",
    "ext/phar/tests/phar_magic.phpt",
    "ext/phar/tests/008.phpt",
    "ext/phar/tests/open_for_write_newfile_c_5_2.phpt",
    "ext/phar/tests/phar_oo_004U.phpt",
    "ext/phar/tests/018.phpt",
    "ext/phar/tests/phar_convert_zip.phpt",
    "ext/phar/tests/phar_offset_get_error.phpt",
    "ext/phar/tests/metadata_write_commitU.phpt",
    "ext/phar/tests/phar_buildfromdirectory2.phpt",
    "ext/phar/tests/phpinfo_003.phpt",
    "ext/phar/tests/phar_oo_nosig.phpt",
    "ext/phar/tests/rename_dir_and_mount.phpt",
    "ext/phar/tests/include_path.phpt",
    "ext/phar/tests/phar_oo_004.phpt",
    "ext/phar/tests/013.phpt",
    "ext/phar/tests/phar_setsignaturealgo2.phpt",
    "ext/phar/tests/016.phpt",
    "ext/phar/tests/bug45218_SLOWTEST.phpt",
    "ext/phar/tests/phar_mount.phpt",
    "ext/phar/tests/phar_get_supported_signatures_002a.phpt",
    "ext/phar/tests/phpinfo_001.phpt",
    "ext/phar/tests/create_new_and_modify.phpt",
    "ext/phar/tests/fopen5.2.phpt",
    "ext/phar/tests/fopen_edgecases2U.phpt",
    "ext/phar/tests/phar_get_supported_signatures_002.phpt",
    "ext/phar/tests/018U.phpt",
    "ext/phar/tests/withphar_web.phpt",
    "ext/phar/tests/dir.phpt",
    "ext/phar/tests/phar_convert_again.phpt",
    "ext/phar/tests/phar_buildfromiterator3.phpt",
    "ext/phar/tests/create_new_phar_b.phpt",
    "ext/phar/tests/025.phpt",
    "ext/phar/tests/phar_buildfromdirectory3.phpt",
    "ext/phar/tests/012.phpt",
    "ext/phar/tests/phar_metadata_readU.phpt",
    "ext/phar/tests/addfuncs.phpt",
    "ext/phar/tests/phar_oo_009.phpt",
    "ext/phar/tests/019cU.phpt",
    "ext/phar/tests/fopen.phpt",
    "ext/phar/tests/phar_oo_001.phpt",
    "ext/phar/tests/rename.phpt",
    "ext/phar/tests/nophar.phpt",
    "ext/phar/tests/phar_metadata_writeU.phpt",
    "ext/phar/tests/phar_oo_006.phpt",
    "ext/phar/tests/021.phpt",
    "ext/phar/tests/refcount1.phpt",
    "ext/phar/tests/frontcontroller13.phpt",
    "ext/phar/tests/frontcontroller27.phpt",
    "ext/phar/tests/mounteddir.phpt",
    "ext/phar/tests/phar_isvalidpharfilename.phpt",
    "ext/phar/tests/phar_oo_012b.phpt",
    "ext/phar/tests/frontcontroller30.phpt",
    "ext/phar/tests/phar_buildfromiterator10.phpt",
    "ext/phar/tests/alias_acrobatics.phpt",
    "ext/phar/tests/033a.phpt",
    "ext/phar/tests/frontcontroller4.phpt",
    "ext/phar/tests/open_for_write_existing_b_5_2.phpt",
    "ext/phar/tests/fatal_error_webphar.phpt",
    "ext/phar/tests/phar_metadata_write.phpt",
    "ext/phar/tests/phar_buildfromiterator9.phpt",
    "ext/phar/tests/pharfileinfo_destruct.phpt",
    "ext/phar/tests/phar_buildfromiterator6.phpt",
    "ext/phar/tests/030.phpt",
    "ext/phar/tests/phar_oo_compressed_001.phpt",
    "ext/phar/tests/ini_set_offU.phpt",
    "ext/phar/tests/frontcontroller8.phpt",
    "ext/phar/tests/metadata_readU.phpt",
    "ext/phar/tests/bug13786.phpt",
    "ext/phar/tests/phar_oo_011b.phpt",
    "ext/phar/tests/create_path_error.phpt",
    "ext/phar/tests/frontcontroller21.phpt",
    "ext/phar/tests/032.phpt",
    "ext/phar/tests/phar_commitwrite.phpt",
    "ext/phar/tests/phar_stub_error.phpt",
    "ext/phar/tests/phar_oo_008.phpt",
    "ext/phar/tests/019.phpt",
    "ext/phar/tests/cache_list/copyonwrite25.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller16.phpt",
    "ext/phar/tests/cache_list/copyonwrite3.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller32.phpt",
    "ext/phar/tests/cache_list/copyonwrite22.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller1.phpt",
    "ext/phar/tests/cache_list/frontcontroller2.phpt",
    "ext/phar/tests/cache_list/frontcontroller25.phpt",
    "ext/phar/tests/cache_list/frontcontroller18.phpt",
    "ext/phar/tests/cache_list/frontcontroller5.phpt",
    "ext/phar/tests/cache_list/copyonwrite15.phar.phpt",
    "ext/phar/tests/cache_list/copyonwrite18.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller12.phpt",
    "ext/phar/tests/cache_list/frontcontroller33.phpt",
    "ext/phar/tests/cache_list/frontcontroller29.phpt",
    "ext/phar/tests/cache_list/frontcontroller19.phpt",
    "ext/phar/tests/cache_list/frontcontroller34.phpt",
    "ext/phar/tests/cache_list/frontcontroller17.phpt",
    "ext/phar/tests/cache_list/frontcontroller20.phpt",
    "ext/phar/tests/cache_list/copyonwrite20.phar.phpt",
    "ext/phar/tests/cache_list/copyonwrite7.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller11.phpt",
    "ext/phar/tests/cache_list/frontcontroller14.phpt",
    "ext/phar/tests/cache_list/frontcontroller9.phpt",
    "ext/phar/tests/cache_list/frontcontroller22.phpt",
    "ext/phar/tests/cache_list/copyonwrite14.phar.phpt",
    "ext/phar/tests/cache_list/copyonwrite13.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller6.phpt",
    "ext/phar/tests/cache_list/frontcontroller7.phpt",
    "ext/phar/tests/cache_list/copyonwrite2.phar.phpt",
    "ext/phar/tests/cache_list/copyonwrite5.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller23.phpt",
    "ext/phar/tests/cache_list/copyonwrite19.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller15.phpt",
    "ext/phar/tests/cache_list/copyonwrite10.phar.phpt",
    "ext/phar/tests/cache_list/copyonwrite21.phar.phpt",
    "ext/phar/tests/cache_list/copyonwrite4a.phpt",
    "ext/phar/tests/cache_list/copyonwrite16.phar.phpt",
    "ext/phar/tests/cache_list/copyonwrite4.phar.phpt",
    "ext/phar/tests/cache_list/copyonwrite23.phar.phpt",
    "ext/phar/tests/cache_list/copyonwrite1.phar.phpt",
    "ext/phar/tests/cache_list/copyonwrite17.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller13.phpt",
    "ext/phar/tests/cache_list/frontcontroller27.phpt",
    "ext/phar/tests/cache_list/frontcontroller30.phpt",
    "ext/phar/tests/cache_list/copyonwrite9.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller4.phpt",
    "ext/phar/tests/cache_list/copyonwrite11.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller8.phpt",
    "ext/phar/tests/cache_list/frontcontroller21.phpt",
    "ext/phar/tests/cache_list/copyonwrite12.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller26.phpt",
    "ext/phar/tests/cache_list/frontcontroller31.phpt",
    "ext/phar/tests/cache_list/frontcontroller10.phpt",
    "ext/phar/tests/cache_list/copyonwrite6.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller3.phpt",
    "ext/phar/tests/cache_list/frontcontroller28.phpt",
    "ext/phar/tests/cache_list/copyonwrite8.phar.phpt",
    "ext/phar/tests/cache_list/frontcontroller24.phpt",
    "ext/phar/tests/cache_list/copyonwrite24.phar.phpt",
    "ext/phar/tests/frontcontroller26.phpt",
    "ext/phar/tests/phar_oo_compressallbz2.phpt",
    "ext/phar/tests/phar_running.phpt",
    "ext/phar/tests/zf_test.phpt",
    "ext/phar/tests/delete_in_phar_confirm.phpt",
    "ext/phar/tests/frontcontroller31.phpt",
    "ext/phar/tests/phar_get_supported_signatures_001.phpt",
    "ext/phar/tests/frontcontroller10.phpt",
    "ext/phar/tests/frontcontroller3.phpt",
    "ext/phar/tests/frontcontroller28.phpt",
    "ext/phar/tests/024.phpt",
    "ext/phar/tests/front.phar.phpt",
    "ext/phar/tests/phar_oo_compressed_002b.phpt",
    "ext/phar/tests/phar_convert_tar2.phpt",
    "ext/phar/tests/bug46032.phpt",
    "ext/phar/tests/009.phpt",
    "ext/phar/tests/phar_bz2.phpt",
    "ext/phar/tests/invalid_alias.phpt",
    "ext/phar/tests/include_path_advanced.phpt",
    "ext/phar/tests/017U.phpt",
    "ext/phar/tests/zip/rename_dir.phpt",
    "ext/phar/tests/zip/open_for_write_newfile_b.phpt",
    "ext/phar/tests/zip/033.phpt",
    "ext/phar/tests/zip/corrupt_004.phpt",
    "ext/phar/tests/zip/frontcontroller19.phar.phpt",
    "ext/phar/tests/zip/frontcontroller17.phar.phpt",
    "ext/phar/tests/zip/frontcontroller18.phar.phpt",
    "ext/phar/tests/zip/phar_buildfromiterator8.phpt",
    "ext/phar/tests/zip/phar_setdefaultstub.phpt",
    "ext/phar/tests/zip/frontcontroller13.phar.phpt",
    "ext/phar/tests/zip/phar_buildfromiterator5.phpt",
    "ext/phar/tests/zip/frontcontroller5.phar.phpt",
    "ext/phar/tests/zip/phar_begin_setstub_commitU.phpt",
    "ext/phar/tests/zip/phar_magicU.phpt",
    "ext/phar/tests/zip/frontcontroller20.phar.phpt",
    "ext/phar/tests/zip/frontcontroller11.phar.phpt",
    "ext/phar/tests/zip/open_for_write_existing_c.phpt",
    "ext/phar/tests/zip/notphar.phpt",
    "ext/phar/tests/zip/open_for_write_newfile_b_5_2.phpt",
    "ext/phar/tests/zip/frontcontroller21.phar.phpt",
    "ext/phar/tests/zip/phar_copy.phpt",
    "ext/phar/tests/zip/getalias.phpt",
    "ext/phar/tests/zip/open_for_write_newfile_c.phpt",
    "ext/phar/tests/zip/open_for_write_existing_b.phpt",
    "ext/phar/tests/zip/exists_as_phar.phpt",
    "ext/phar/tests/zip/frontcontroller2.phar.phpt",
    "ext/phar/tests/zip/phar_setalias.phpt",
    "ext/phar/tests/zip/delete_in_phar_b.phpt",
    "ext/phar/tests/zip/frontcontroller9.phar.phpt",
    "ext/phar/tests/zip/phar_begin_setstub_commit.phpt",
    "ext/phar/tests/zip/odt.phpt",
    "ext/phar/tests/zip/open_for_write_newfile.phpt",
    "ext/phar/tests/zip/allU.phpt",
    "ext/phar/tests/zip/frontcontroller3.phar.phpt",
    "ext/phar/tests/zip/corrupt_005.phpt",
    "ext/phar/tests/zip/phar_oo_compressallgz.phpt",
    "ext/phar/tests/zip/corrupt_003.phpt",
    "ext/phar/tests/zip/open_for_write_existing_c_5_2.phpt",
    "ext/phar/tests/zip/rmdir.phpt",
    "ext/phar/tests/zip/phar_convert_phar.phpt",
    "ext/phar/tests/zip/bzip2.phpt",
    "ext/phar/tests/zip/corrupt_010.phpt",
    "ext/phar/tests/zip/phar_buildfromiterator4.phpt",
    "ext/phar/tests/zip/corrupt_001.phpt",
    "ext/phar/tests/zip/corrupt_002.phpt",
    "ext/phar/tests/zip/phar_setalias2.phpt",
    "ext/phar/tests/zip/corrupt_009.phpt",
    "ext/phar/tests/zip/corrupt_008.phpt",
    "ext/phar/tests/zip/phar_stub.phpt",
    "ext/phar/tests/zip/phar_buildfromiterator7.phpt",
    "ext/phar/tests/zip/refcount1_5_2.phpt",
    "ext/phar/tests/zip/metadata_write_commit.phpt",
    "ext/phar/tests/zip/zlib.phpt",
    "ext/phar/tests/zip/frontcontroller8.phar.phpt",
    "ext/phar/tests/zip/frontcontroller16.phar.phpt",
    "ext/phar/tests/zip/corrupt_007.phpt",
    "ext/phar/tests/zip/delete_in_phar.phpt",
    "ext/phar/tests/zip/largezip.phpt",
    "ext/phar/tests/zip/bug48791.phpt",
    "ext/phar/tests/zip/delete.phpt",
    "ext/phar/tests/zip/phar_magic.phpt",
    "ext/phar/tests/zip/open_for_write_newfile_c_5_2.phpt",
    "ext/phar/tests/zip/frontcontroller12.phar.phpt",
    "ext/phar/tests/zip/metadata_write_commitU.phpt",
    "ext/phar/tests/zip/frontcontroller10.phar.phpt",
    "ext/phar/tests/zip/phar_setsignaturealgo2.phpt",
    "ext/phar/tests/zip/create_new_and_modify.phpt",
    "ext/phar/tests/zip/dir.phpt",
    "ext/phar/tests/zip/create_new_phar_b.phpt",
    "ext/phar/tests/zip/badalias.phpt",
    "ext/phar/tests/zip/rename.phpt",
    "ext/phar/tests/zip/unixzip.phpt",
    "ext/phar/tests/zip/refcount1.phpt",
    "ext/phar/tests/zip/frontcontroller7.phar.phpt",
    "ext/phar/tests/zip/frontcontroller4.phar.phpt",
    "ext/phar/tests/zip/corrupt_006.phpt",
    "ext/phar/tests/zip/alias_acrobatics.phpt",
    "ext/phar/tests/zip/033a.phpt",
    "ext/phar/tests/zip/open_for_write_existing_b_5_2.phpt",
    "ext/phar/tests/zip/frontcontroller15.phar.phpt",
    "ext/phar/tests/zip/phar_buildfromiterator9.phpt",
    "ext/phar/tests/zip/phar_buildfromiterator6.phpt",
    "ext/phar/tests/zip/frontcontroller6.phar.phpt",
    "ext/phar/tests/zip/frontcontroller1.phar.phpt",
    "ext/phar/tests/zip/phar_commitwrite.phpt",
    "ext/phar/tests/zip/phar_stub_error.phpt",
    "ext/phar/tests/zip/phar_oo_compressallbz2.phpt",
    "ext/phar/tests/zip/delete_in_phar_confirm.phpt",
    "ext/phar/tests/zip/all.phpt",
    "ext/phar/tests/zip/frontcontroller14.phar.phpt",
    "ext/phar/tests/zip/open_for_write_existing.phpt",
    "ext/phar/tests/phar_get_supportedcomp3.phpt",
    "ext/phar/tests/frontcontroller24.phpt",
    "ext/phar/tests/open_for_write_existing.phpt",
    "ext/phar/tests/pharfileinfo_chmod.phpt",
    "ext/sysvsem/tests/sysv.phpt",
    "ext/dom/tests/DOMDocument_createProcessingInstruction_error.phpt",
    "ext/dom/tests/DOMDocumentType_internalSubset_error_001.phpt",
    "ext/dom/tests/DOMCharacterData_appendData_error_001.phpt",
    "ext/dom/tests/dom002.phpt",
    "ext/dom/tests/DOMDocument_saveHTMLFile_invalid_filename.phpt",
    "ext/dom/tests/bug36756.phpt",
    "ext/dom/tests/bug42082.phpt",
    "ext/dom/tests/DOMDocument_schemaValidateSource_error4.phpt",
    "ext/dom/tests/DOMCharacterData_data_error_002.phpt",
    "ext/dom/tests/DOMComment_appendData_basic.phpt",
    "ext/dom/tests/dom006.phpt",
    "ext/dom/tests/bug38474.phpt",
    "ext/dom/tests/bug34276.phpt",
    "ext/dom/tests/DOMDocument_createAttribute_basic.phpt",
    "ext/dom/tests/DOMDocumentType_notations_error_001.phpt",
    "ext/dom/tests/DOMDocument_preserveWhiteSpace_basic.phpt",
    "ext/dom/tests/DOMComment_construct_basic_001.phpt",
    "ext/dom/tests/DOMComment_construct_error_001.phpt",
    "ext/dom/tests/bug50661.phpt",
    "ext/dom/tests/dom001.phpt",
    "ext/dom/tests/bug35342.phpt",
    "ext/dom/tests/DOMComment_replaceData_error2.phpt",
    "ext/dom/tests/DOMNode_hasChildNodes_basic.phpt",
    "ext/dom/tests/domxpath.phpt",
    "ext/dom/tests/DOMDocumentType_publicId_error_001.phpt",
    "ext/dom/tests/DOMDocument_save_basic.phpt",
    "ext/dom/tests/dom_set_attr_node.phpt",
    "ext/dom/tests/domdocument_createentityreference_002.phpt",
    "ext/dom/tests/DOMDocumentFragment_construct_basic_001.phpt",
    "ext/dom/tests/bug47430.phpt",
    "ext/dom/tests/DOMDocument_saveHTML_error2.phpt",
    "ext/dom/tests/domdocument_createentityreference_001.phpt",
    "ext/dom/tests/DOMNode_normalize_basic.phpt",
    "ext/dom/tests/bug38949.phpt",
    "ext/dom/tests/dom007.phpt",
    "ext/dom/tests/DOMDocument_schemaValidate_error1.phpt",
    "ext/dom/tests/domdocument_createcomment_error_001.phpt",
    "ext/dom/tests/DOMDocument_createEntityReference_basic.phpt",
    "ext/dom/tests/DOMDocumentType_basic_001.phpt",
    "ext/dom/tests/DOMDocument_implementationRead_basic.phpt",
    "ext/dom/tests/DOMDocument_loadHTML_basic.phpt",
    "ext/dom/tests/bug41374.phpt",
    "ext/dom/tests/DOMDocument_createAttribute_variation.phpt",
    "ext/dom/tests/dom_comment_variation.phpt",
    "ext/dom/tests/bug44648.phpt",
    "ext/dom/tests/DOMComment_insertData_error2.phpt",
    "ext/dom/tests/DOMDocument_encoding_basic.phpt",
    "ext/dom/tests/DOMDocument_resolveExternals_basic.phpt",
    "ext/dom/tests/dom004.phpt",
    "ext/dom/tests/DOMCharacterData_substringData_basic_001.phpt",
    "ext/dom/tests/regsiter_node_class.phpt",
    "ext/dom/tests/DOMCharacterData_appendData_basic.phpt",
    "ext/dom/tests/DOMDocumentType_publicId_basic_001.phpt",
    "ext/dom/tests/DOMCharacterData_replaceData_error_001.phpt",
    "ext/dom/tests/DOMDocument_schemaValidate_error4.phpt",
    "ext/dom/tests/DOMAttr_ownerElement_error_001.phpt",
    "ext/dom/tests/dom003.phpt",
    "ext/dom/tests/bug52656.phpt",
    "ext/dom/tests/DOMDocumentFragment_appendXML_basic_001.phpt",
    "ext/dom/tests/dom_comment_basic.phpt",
    "ext/dom/tests/dom_xinclude.phpt",
    "ext/dom/tests/domchardata.phpt",
    "ext/dom/tests/dom005.phpt",
    "ext/dom/tests/DOMDocument_preserveWhiteSpace_variations.phpt",
    "ext/dom/tests/DOMDocument_validate_error1.phpt",
    "ext/dom/tests/dom_import_simplexml.phpt",
    "ext/dom/tests/DOMDocument_schemaValidate_basic.phpt",
    "ext/dom/tests/bug37456.phpt",
    "ext/dom/tests/DOMDocument_validate_on_parse_basic.phpt",
    "ext/dom/tests/DOMAttr_value_basic_002.phpt",
    "ext/dom/tests/DOMDocument_savexml_basic.phpt",
    "ext/dom/tests/DOMComment_replaceData_error1.phpt",
    "ext/dom/tests/DOMDocumentFragment_construct_basic_002.phpt",
    "ext/dom/tests/bug49463.phpt",
    "ext/dom/tests/DOMDocument_saveHTML_variant1.phpt",
    "ext/dom/tests/DOMNode_issamenode_basic.phpt",
    "ext/dom/tests/DOMDocument_saveHTMLFile_formatOutput.phpt",
    "ext/dom/tests/DOMCharacterData_deleteData_basic_001.phpt",
    "ext/dom/tests/DOMDocumentFragment_appendXML_error_002.phpt",
    "ext/dom/tests/bug38438.phpt",
    "ext/dom/tests/bug41257.phpt",
    "ext/dom/tests/canonicalization.phpt",
    "ext/dom/tests/DOMDocumentFragment_construct_error_001.phpt",
    "ext/dom/tests/DOMDocument_standalone_basic.phpt",
    "ext/dom/tests/DOMComment_appendData_basic_Sullivan.phpt",
    "ext/dom/tests/DOMDocumentType_systemId_error_001.phpt",
    "ext/dom/tests/DOMDocumentFragment_appendXML_hasChildNodes_basic.phpt",
    "ext/dom/tests/DOMDocument_saveHTMLFile_error2.phpt",
    "ext/dom/tests/DOMComment_insertData_basic.phpt",
    "ext/dom/tests/DOMCharacterData_length_error_001.phpt",
    "ext/dom/tests/domattributes.phpt",
    "ext/dom/tests/bug43364.phpt",
    "ext/dom/tests/DOMNode_hasChildNodes.phpt",
    "ext/dom/tests/DOMDocument_documentURI_basic.phpt",
    "ext/dom/tests/DOMComment_replaceData_basic.phpt",
    "ext/dom/tests/DOMDocument_validate_external_dtd.phpt",
    "ext/dom/tests/bug32615.phpt",
    "ext/dom/tests/bug49490.phpt",
    "ext/dom/tests/DOMDocumentFragment_appendXML_error_001.phpt",
    "ext/dom/tests/DOMCharacterData_insertData_error_001.phpt",
    "ext/dom/tests/DOMDocument_saveHTMLFile_error1.phpt",
    "ext/dom/tests/DOMDocument_schemaValidate_error5.phpt",
    "ext/dom/tests/bug46185.phpt",
    "ext/dom/tests/DOMAttr_value_basic_001.phpt",
    "ext/dom/tests/DOMCharacterData_deleteData_error_002.phpt",
    "ext/dom/tests/DOMDocument_schemaValidateSource_error3.phpt",
    "ext/dom/tests/bug46849.phpt",
    "ext/dom/tests/bug37277.phpt",
    "ext/dom/tests/DOMDocument_createAttribute_error1.phpt",
    "ext/dom/tests/bug47848.phpt",
    "ext/dom/tests/bug35673.phpt",
    "ext/dom/tests/DOMDocument_schemaValidate_error2.phpt",
    "ext/dom/tests/DOMComment_insertData_error1.phpt",
    "ext/dom/tests/bug47849.phpt",
    "ext/dom/tests/DOMNode_insertBefore.phpt",
    "ext/dom/tests/DOMAttr_name_basic_001.phpt",
    "ext/dom/tests/bug45251.phpt",
    "ext/dom/tests/DOMDocument_schemaValidate_error3.phpt",
    "ext/dom/tests/bug40836.phpt",
    "ext/dom/tests/DOMDocumentType_entities_error_001.phpt",
    "ext/dom/tests/DOMDocument_strictErrorChecking_basic.phpt",
    "ext/dom/tests/DOMText_appendData_basic.phpt",
    "ext/dom/tests/DOMDocument_schemaValidateSource_error1.phpt",
    "ext/dom/tests/DOMNode_cloneNode_basic.phpt",
    "ext/dom/tests/dom_create_element.phpt",
    "ext/dom/tests/DOMDocument_saveHTML_variant2.phpt",
    "ext/dom/tests/DOMDocument_schemaValidateSource_basic.phpt",
    "ext/dom/tests/DOMCharacterData_data_basic_002.phpt",
    "ext/dom/tests/bug28721.phpt",
    "ext/dom/tests/domelement.phpt",
    "ext/dom/tests/DOMAttr_construct_error_001.phpt",
    "ext/dom/tests/DOMDocumentType_name_error_001.phpt",
    "ext/dom/tests/DOMDocument_schemaValidateSource_error2.phpt",
    "ext/dom/tests/DOMDocument_createAttribute_error.phpt",
    "ext/dom/tests/DOMNode_replaceChild_basic.phpt",
    "ext/dom/tests/DOMElement_hasAttributes_basic.phpt",
    "ext/dom/tests/DOMDocument_createProcessingInstruction_basic.phpt",
    "ext/dom/tests/bug38850.phpt",
    "ext/dom/tests/DOMCDATASection_construct_error_001.phpt",
    "ext/dom/tests/DOMNode_removeChild_basic.phpt",
    "ext/dom/tests/DOMDocument_strictErrorChecking_variation.phpt",
    "ext/dom/tests/DOMDocument_validate_error2.phpt",
    "ext/dom/tests/DOMDocument_validate_basic.phpt",
    "ext/dom/tests/DOMCharacterData_deleteData_error_001.phpt",
    "ext/dom/tests/DOMDocument_saveHTML_basic.phpt",
    "ext/dom/tests/DOMDocument_validate_on_parse_variation.phpt",
    "ext/dom/tests/DOMDocument_saveHTMLFile_basic.phpt",
    "ext/dom/tests/DOMDocumentType_systemId_basic_001.phpt",
    "ext/dom/tests/bug46335.phpt",
    "ext/dom/tests/bug28817.phpt",
    "ext/dom/tests/DOMDocumentFragment_appendXML_error_003.phpt",
    "ext/dom/tests/DOMDocument_config_basic.phpt",
    "ext/pcre/tests/preg_quote_basic.phpt",
    "ext/pcre/tests/preg_grep_basic.phpt",
    "ext/pcre/tests/preg_match_error1.phpt",
    "ext/pcre/tests/preg_match_variation1.phpt",
    "ext/pcre/tests/bug21732.phpt",
    "ext/pcre/tests/bug44925.phpt",
    "ext/pcre/tests/005.phpt",
    "ext/pcre/tests/locales.phpt",
    "ext/pcre/tests/split2.phpt",
    "ext/pcre/tests/preg_match_all_error3.phpt",
    "ext/pcre/tests/bug38600.phpt",
    "ext/pcre/tests/preg_replace_callback2.phpt",
    "ext/pcre/tests/bug40909.phpt",
    "ext/pcre/tests/bug44214_2.phpt",
    "ext/pcre/tests/preg_match_basic_edit.phpt",
    "ext/pcre/tests/pcre_extended.phpt",
    "ext/pcre/tests/pcre_anchored.phpt",
    "ext/pcre/tests/bug44214.phpt",
    "ext/pcre/tests/preg_match_all_error1.phpt",
    "ext/pcre/tests/preg_match_all_error.phpt",
    "ext/pcre/tests/preg_match_all_edit_basic.phpt",
    "ext/pcre/tests/bug20528.phpt",
    "ext/pcre/tests/preg_quote_error.phpt",
    "ext/pcre/tests/preg_split_error2.phpt",
    "ext/pcre/tests/bug33200.phpt",
    "ext/pcre/tests/ungreedy.phpt",
    "ext/pcre/tests/cache_limit.phpt",
    "ext/pcre/tests/001.phpt",
    "ext/pcre/tests/study.phpt",
    "ext/pcre/tests/multiline.phpt",
    "ext/pcre/tests/bug47229.phpt",
    "ext/pcre/tests/004.phpt",
    "ext/pcre/tests/preg_replace.phpt",
    "ext/pcre/tests/bug47662.phpt",
    "ext/pcre/tests/preg_match_all_basic.phpt",
    "ext/pcre/tests/bug21758.phpt",
    "ext/pcre/tests/recursion_limit.phpt",
    "ext/pcre/tests/dollar_endonly.phpt",
    "ext/pcre/tests/preg_split_basic.phpt",
    "ext/pcre/tests/preg_last_error_error.phpt",
    "ext/pcre/tests/split.phpt",
    "ext/pcre/tests/invalid_utf8.phpt",
    "ext/pcre/tests/003.phpt",
    "ext/pcre/tests/bug26927.phpt",
    "ext/pcre/tests/preg_replace_callback.phpt",
    "ext/pcre/tests/preg_split_error.phpt",
    "ext/pcre/tests/pcre_extra.phpt",
    "ext/pcre/tests/preg_replace_callback_error.phpt",
    "ext/pcre/tests/bug41148.phpt",
    "ext/pcre/tests/006.phpt",
    "ext/pcre/tests/bug44191.phpt",
    "ext/pcre/tests/backtrack_limit.phpt",
    "ext/pcre/tests/preg_quote_error1.phpt",
    "ext/pcre/tests/preg_grep_error2.phpt",
    "ext/pcre/tests/preg_match_basic.phpt",
    "ext/pcre/tests/preg_replace_edit_basic.phpt",
    "ext/pcre/tests/delimiters.phpt",
    "ext/pcre/tests/bug52971.phpt",
    "ext/pcre/tests/preg_replace_basic.phpt",
    "ext/pcre/tests/007.phpt",
    "ext/pcre/tests/preg_replace_error1.phpt",
    "ext/pcre/tests/bug37800.phpt",
    "ext/pcre/tests/invalid_utf8_offset.phpt",
    "ext/pcre/tests/002.phpt",
    "ext/pcre/tests/match_flags3.phpt",
    "ext/pcre/tests/preg_grep_error.phpt",
    "ext/pcre/tests/preg_split_error1.phpt",
    "ext/pcre/tests/bug37911.phpt",
    "ext/pcre/tests/pcre.constants.phpt",
    "ext/pcre/tests/match_flags.phpt",
    "ext/pcre/tests/bug40195.phpt",
    "ext/pcre/tests/preg_filter.phpt",
    "ext/pcre/tests/bug34790.phpt",
    "ext/pcre/tests/preg_replace_callback_basic.phpt",
    "ext/pcre/tests/match_flags2.phpt",
    "ext/pcre/tests/preg_grep_error1.phpt",
    "ext/pcre/tests/preg_match_error2.phpt",
    "ext/pcre/tests/bug42945.phpt",
    "ext/pcre/tests/preg_match_basic_002.phpt",
    "ext/pcre/tests/preg_replace2.phpt",
    "ext/pcre/tests/bug41638.phpt",
    "ext/pcre/tests/preg_replace_callback3.phpt",
    "ext/pcre/tests/grep2.phpt",
    "ext/pcre/tests/preg_match_all_error2.phpt",
    "ext/pcre/tests/bug41050.phpt",
    "ext/pcre/tests/preg_replace_callback_error1.phpt",
    "ext/pcre/tests/bug42298.phpt",
    "ext/pcre/tests/pcre_count.phpt",
    "ext/pcre/tests/preg_match_error.phpt",
    "ext/pcre/tests/grep.phpt",
    "ext/pcre/tests/preg_replace_error.phpt",
    "ext/pcre/tests/bug27103.phpt",
    "ext/pcre/tests/bug52732.phpt",
    "ext/pcre/tests/preg_replace_variation1.phpt",
    "ext/pcre/tests/bug42737.phpt",
    "ext/pcre/tests/bug27011.phpt",
    "ext/pcre/tests/preg_replace_error2.phpt",
    "ext/oci8/tests/coll_003.phpt",
    "ext/oci8/tests/bug40415.phpt",
    "ext/oci8/tests/bug36096.phpt",
    "ext/oci8/tests/coll_007.phpt",
    "ext/oci8/tests/array_bind_002.phpt",
    "ext/oci8/tests/function_aliases.phpt",
    "ext/oci8/tests/bug27303_3.phpt",
    "ext/oci8/tests/connect_1_old.phpt",
    "ext/oci8/tests/bug38173.phpt",
    "ext/oci8/tests/bug42134.phpt",
    "ext/oci8/tests/array_bind_str1.phpt",
    "ext/oci8/tests/drcp_connection_class.phpt",
    "ext/oci8/tests/error.phpt",
    "ext/oci8/tests/bug37220.phpt",
    "ext/oci8/tests/field_funcs_old.phpt",
    "ext/oci8/tests/coll_005.phpt",
    "ext/oci8/tests/lob_014.phpt",
    "ext/oci8/tests/array_bind_003.phpt",
    "ext/oci8/tests/coll_004.phpt",
    "ext/oci8/tests/define_old.phpt",
    "ext/oci8/tests/cursor_bind.phpt",
    "ext/oci8/tests/bind_char_1_11gR1.phpt",
    "ext/oci8/tests/array_bind_010.phpt",
    "ext/oci8/tests/fetch_all3.phpt",
    "ext/oci8/tests/lob_036.phpt",
    "ext/oci8/tests/fetch_object.phpt",
    "ext/oci8/tests/driver_name.phpt",
    "ext/oci8/tests/lob_039.phpt",
    "ext/oci8/tests/coll_014_func.phpt",
    "ext/oci8/tests/lob_010.phpt",
    "ext/oci8/tests/define3.phpt",
    "ext/oci8/tests/drcp_conn_close2.phpt",
    "ext/oci8/tests/bug51291.phpt",
    "ext/oci8/tests/coll_002.phpt",
    "ext/oci8/tests/lob_null.phpt",
    "ext/oci8/tests/pecl_bug10194_blob.phpt",
    "ext/oci8/tests/drcp_pconn_close1.phpt",
    "ext/oci8/tests/coll_009.phpt",
    "ext/oci8/tests/drcp_scope5.phpt",
    "ext/oci8/tests/bind_rowid.phpt",
    "ext/oci8/tests/field_funcs.phpt",
    "ext/oci8/tests/lob_034.phpt",
    "ext/oci8/tests/num.phpt",
    "ext/oci8/tests/connect_with_charset_001.phpt",
    "ext/oci8/tests/bind_long_raw.phpt",
    "ext/oci8/tests/drcp_conn_close1.phpt",
    "ext/oci8/tests/cursor_bind_err.phpt",
    "ext/oci8/tests/lob_002.phpt",
    "ext/oci8/tests/connect_scope_try3.phpt",
    "ext/oci8/tests/minfo.phpt",
    "ext/oci8/tests/coll_013_func.phpt",
    "ext/oci8/tests/lob_007.phpt",
    "ext/oci8/tests/fetch_all4.phpt",
    "ext/oci8/tests/password_2.phpt",
    "ext/oci8/tests/bind_char_1.phpt",
    "ext/oci8/tests/coll_010_func.phpt",
    "ext/oci8/tests/coll_012_func.phpt",
    "ext/oci8/tests/bug36403.phpt",
    "ext/oci8/tests/array_bind_float1.phpt",
    "ext/oci8/tests/bug38161.phpt",
    "ext/oci8/tests/drcp_scope4.phpt",
    "ext/oci8/tests/array_bind_014.phpt",
    "ext/oci8/tests/bug36010.phpt",
    "ext/oci8/tests/bug27303_4.phpt",
    "ext/oci8/tests/bug35973.phpt",
    "ext/oci8/tests/bind_long.phpt",
    "ext/oci8/tests/uncommitted.phpt",
    "ext/oci8/tests/coll_015.phpt",
    "ext/oci8/tests/refcur_prefetch_3.phpt",
    "ext/oci8/tests/error_old.phpt",
    "ext/oci8/tests/descriptors.phpt",
    "ext/oci8/tests/define.phpt",
    "ext/oci8/tests/array_bind_001.phpt",
    "ext/oci8/tests/lob_011.phpt",
    "ext/oci8/tests/pecl_bug10194_blob_64.phpt",
    "ext/oci8/tests/connect_without_oracle_home_old.phpt",
    "ext/oci8/tests/password.phpt",
    "ext/oci8/tests/drcp_newconnect.phpt",
    "ext/oci8/tests/oci_execute_segfault.phpt",
    "ext/oci8/tests/cursors_old.phpt",
    "ext/oci8/tests/bind_char_4_11gR1.phpt",
    "ext/oci8/tests/statement_type_old.phpt",
    "ext/oci8/tests/drcp_privileged.phpt",
    "ext/oci8/tests/array_bind_006.phpt",
    "ext/oci8/tests/coll_017.phpt",
    "ext/oci8/tests/conn_attr_3.phpt",
    "ext/oci8/tests/privileged_connect1.phpt",
    "ext/oci8/tests/error_parse.phpt",
    "ext/oci8/tests/lob_temp1.phpt",
    "ext/oci8/tests/cursors.phpt",
    "ext/oci8/tests/array_bind_date1.phpt",
    "ext/oci8/tests/bug37581.phpt",
    "ext/oci8/tests/serverversion.phpt",
    "ext/oci8/tests/pecl_bug10194.phpt",
    "ext/oci8/tests/coll_002_func.phpt",
    "ext/oci8/tests/fetch_assoc.phpt",
    "ext/oci8/tests/bug43492_2.phpt",
    "ext/oci8/tests/fetch_row.phpt",
    "ext/oci8/tests/bug42496_2.phpt",
    "ext/oci8/tests/bug51253.phpt",
    "ext/oci8/tests/lob_020.phpt",
    "ext/oci8/tests/coll_003_func.phpt",
    "ext/oci8/tests/bug44113.phpt",
    "ext/oci8/tests/define2.phpt",
    "ext/oci8/tests/coll_016.phpt",
    "ext/oci8/tests/testping.phpt",
    "ext/oci8/tests/bug44206.phpt",
    "ext/oci8/tests/coll_012.phpt",
    "ext/oci8/tests/commit_old.phpt",
    "ext/oci8/tests/lob_018.phpt",
    "ext/oci8/tests/bind_empty.phpt",
    "ext/oci8/tests/coll_006_func.phpt",
    "ext/oci8/tests/fetch_all5.phpt",
    "ext/oci8/tests/coll_011_func.phpt",
    "ext/oci8/tests/coll_017_func.phpt",
    "ext/oci8/tests/lob_031.phpt",
    "ext/oci8/tests/lob_030.phpt",
    "ext/oci8/tests/array_bind_008.phpt",
    "ext/oci8/tests/lob_028.phpt",
    "ext/oci8/tests/field_funcs1.phpt",
    "ext/oci8/tests/lob_009.phpt",
    "ext/oci8/tests/password_new.phpt",
    "ext/oci8/tests/lob_042.phpt",
    "ext/oci8/tests/fetch_array.phpt",
    "ext/oci8/tests/fetch_all2.phpt",
    "ext/oci8/tests/conn_attr_1.phpt",
    "ext/oci8/tests/array_bind_str.phpt",
    "ext/oci8/tests/connect_1.phpt",
    "ext/oci8/tests/coll_006.phpt",
    "ext/oci8/tests/b47243_2.phpt",
    "ext/oci8/tests/statement_type.phpt",
    "ext/oci8/tests/lob_038.phpt",
    "ext/oci8/tests/error1.phpt",
    "ext/oci8/tests/fetch.phpt",
    "ext/oci8/tests/error2.phpt",
    "ext/oci8/tests/old_oci_close.phpt",
    "ext/oci8/tests/bug27303_4_11gR1.phpt",
    "ext/oci8/tests/coll_010.phpt",
    "ext/oci8/tests/lob_032.phpt",
    "ext/oci8/tests/connect_scope_try5.phpt",
    "ext/oci8/tests/fetch_into.phpt",
    "ext/oci8/tests/lob_023.phpt",
    "ext/oci8/tests/array_bind_float.phpt",
    "ext/oci8/tests/pecl_bug8816.phpt",
    "ext/oci8/tests/bug45458.phpt",
    "ext/oci8/tests/prefetch.phpt",
    "ext/oci8/tests/bug47189.phpt",
    "ext/oci8/tests/lob_017.phpt",
    "ext/oci8/tests/bug42173.phpt",
    "ext/oci8/tests/coll_008.phpt",
    "ext/oci8/tests/lob_005.phpt",
    "ext/oci8/tests/define5.phpt",
    "ext/oci8/tests/connect.phpt",
    "ext/oci8/tests/coll_016_func.phpt",
    "ext/oci8/tests/extauth_01.phpt",
    "ext/oci8/tests/statement_cache.phpt",
    "ext/oci8/tests/lob_037.phpt",
    "ext/oci8/tests/define1.phpt",
    "ext/oci8/tests/connect_scope_try1.phpt",
    "ext/oci8/tests/select_null.phpt",
    "ext/oci8/tests/conn_attr_4.phpt",
    "ext/oci8/tests/lob_016.phpt",
    "ext/oci8/tests/lob_026.phpt",
    "ext/oci8/tests/bug42496_1.phpt",
    "ext/oci8/tests/pecl_bug6109.phpt",
    "ext/oci8/tests/default_prefetch2.phpt",
    "ext/oci8/tests/bug40078.phpt",
    "ext/oci8/tests/lob_041.phpt",
    "ext/oci8/tests/close.phpt",
    "ext/oci8/tests/fetch_into1.phpt",
    "ext/oci8/tests/lob_033.phpt",
    "ext/oci8/tests/coll_009_func.phpt",
    "ext/oci8/tests/commit_001.phpt",
    "ext/oci8/tests/connect_old.phpt",
    "ext/oci8/tests/bug27303_1.phpt",
    "ext/oci8/tests/bug46994.phpt",
    "ext/oci8/tests/lob_008.phpt",
    "ext/oci8/tests/coll_018.phpt",
    "ext/oci8/tests/conn_attr_5.phpt",
    "ext/oci8/tests/drcp_pconn_close2.phpt",
    "ext/oci8/tests/array_bind_009.phpt",
    "ext/oci8/tests/edition_2.phpt",
    "ext/oci8/tests/refcur_prefetch_1.phpt",
    "ext/oci8/tests/bug26133.phpt",
    "ext/oci8/tests/null_byte_2.phpt",
    "ext/oci8/tests/bug43497_92.phpt",
    "ext/oci8/tests/lob_012.phpt",
    "ext/oci8/tests/bug32325.phpt",
    "ext/oci8/tests/array_bind_011.phpt",
    "ext/oci8/tests/coll_013.phpt",
    "ext/oci8/tests/connect_without_oracle_home_old_11.phpt",
    "ext/oci8/tests/drcp_characterset.phpt",
    "ext/oci8/tests/drcp_cclass1.phpt",
    "ext/oci8/tests/lob_001.phpt",
    "ext/oci8/tests/array_bind_012.phpt",
    "ext/oci8/tests/array_bind_int.phpt",
    "ext/oci8/tests/prefetch_old.phpt",
    "ext/oci8/tests/b47243_1.phpt",
    "ext/oci8/tests/lob_aliases.phpt",
    "ext/oci8/tests/password_old.phpt",
    "ext/oci8/tests/connect_scope1.phpt",
    "ext/oci8/tests/commit_002.phpt",
    "ext/oci8/tests/lob_027.phpt",
    "ext/oci8/tests/field_funcs2.phpt",
    "ext/oci8/tests/bug27303_2.phpt",
    "ext/oci8/tests/lob_035.phpt",
    "ext/oci8/tests/xmltype_02.phpt",
    "ext/oci8/tests/coll_011.phpt",
    "ext/oci8/tests/bind_char_4.phpt",
    "ext/oci8/tests/lob_temp.phpt",
    "ext/oci8/tests/b47243_3.phpt",
    "ext/oci8/tests/drcp_connect1.phpt",
    "ext/oci8/tests/coll_015_func.phpt",
    "ext/oci8/tests/drcp_scope1.phpt",
    "ext/oci8/tests/bug47281.phpt",
    "ext/oci8/tests/conn_attr_2.phpt",
    "ext/oci8/tests/bug41069.phpt",
    "ext/oci8/tests/coll_004_func.phpt",
    "ext/oci8/tests/coll_001.phpt",
    "ext/oci8/tests/drcp_scope2.phpt",
    "ext/oci8/tests/lob_024.phpt",
    "ext/oci8/tests/connect_scope2.phpt",
    "ext/oci8/tests/fetch_object_2.phpt",
    "ext/oci8/tests/array_bind_int1.phpt",
    "ext/oci8/tests/lob_043.phpt",
    "ext/oci8/tests/array_bind_004.phpt",
    "ext/oci8/tests/lob_015.phpt",
    "ext/oci8/tests/error_bind.phpt",
    "ext/oci8/tests/bug27303_1_11gR1.phpt",
    "ext/oci8/tests/lob_006.phpt",
    "ext/oci8/tests/default_prefetch1.phpt",
    "ext/oci8/tests/edition_1.phpt",
    "ext/oci8/tests/bind_char_2_11gR1.phpt",
    "ext/oci8/tests/bind_char_3_11gR1.phpt",
    "ext/oci8/tests/connect_scope_try6.phpt",
    "ext/oci8/tests/exec_fetch.phpt",
    "ext/oci8/tests/bug43497.phpt",
    "ext/oci8/tests/reflection2.phpt",
    "ext/oci8/tests/fetch_into2.phpt",
    "ext/oci8/tests/array_bind_013.phpt",
    "ext/oci8/tests/lob_019.phpt",
    "ext/oci8/tests/connect_scope_try4.phpt",
    "ext/oci8/tests/default_prefetch.phpt",
    "ext/oci8/tests/bug44008.phpt",
    "ext/oci8/tests/bug42841.phpt",
    "ext/oci8/tests/coll_014.phpt",
    "ext/oci8/tests/bind_char_3.phpt",
    "ext/oci8/tests/xmltype_01.phpt",
    "ext/oci8/tests/array_bind_date.phpt",
    "ext/oci8/tests/null_byte_1.phpt",
    "ext/oci8/tests/lob_025.phpt",
    "ext/oci8/tests/extauth_04.phpt",
    "ext/oci8/tests/connect_scope_try2.phpt",
    "ext/oci8/tests/extauth_02.phpt",
    "ext/oci8/tests/define4.phpt",
    "ext/oci8/tests/lob_029.phpt",
    "ext/oci8/tests/lob_004.phpt",
    "ext/oci8/tests/array_bind_005.phpt",
    "ext/oci8/tests/connect_without_oracle_home_11.phpt",
    "ext/oci8/tests/pecl_bug16842.phpt",
    "ext/oci8/tests/execute_mode.phpt",
    "ext/oci8/tests/lob_021.phpt",
    "ext/oci8/tests/reflection1.phpt",
    "ext/oci8/tests/bind_char_2.phpt",
    "ext/oci8/tests/lob_022.phpt",
    "ext/oci8/tests/bug27303_2_11gR1.phpt",
    "ext/oci8/tests/bind_raw.phpt",
    "ext/oci8/tests/lob_003.phpt",
    "ext/oci8/tests/connect_without_oracle_home.phpt",
    "ext/oci8/tests/debug.phpt",
    "ext/oci8/tests/extauth_03.phpt",
    "ext/oci8/tests/bug43492.phpt",
    "ext/oci8/tests/persistent.phpt",
    "ext/oci8/tests/lob_040.phpt",
    "ext/oci8/tests/fetch_all.phpt",
    "ext/oci8/tests/pecl_bug16035.phpt",
    "ext/oci8/tests/coll_019.phpt",
    "ext/oci8/tests/refcur_prefetch_2.phpt",
    "ext/oci8/tests/array_bind_007.phpt",
    "ext/oci8/tests/drcp_scope3.phpt",
    "ext/oci8/tests/old_oci_close1.phpt",
    "ext/oci8/tests/privileged_connect.phpt",
    "ext/oci8/tests/lob_013.phpt",
    "ext/exif/tests/bug34704.phpt",
    "ext/exif/tests/exif015.phpt",
    "ext/exif/tests/exif004.phpt",
    "ext/exif/tests/exif027.phpt",
    "ext/exif/tests/exif_imagetype_basic.phpt",
    "ext/exif/tests/exif026.phpt",
    "ext/exif/tests/exif008.phpt",
    "ext/exif/tests/exif013.phpt",
    "ext/exif/tests/exif024.phpt",
    "ext/exif/tests/exif021.phpt",
    "ext/exif/tests/exif014.phpt",
    "ext/exif/tests/exif017.phpt",
    "ext/exif/tests/exif025.phpt",
    "ext/exif/tests/exif000.phpt",
    "ext/exif/tests/exif012.phpt",
    "ext/exif/tests/exif002.phpt",
    "ext/exif/tests/exif009.phpt",
    "ext/exif/tests/exif022.phpt",
    "ext/exif/tests/exif_imagetype_error.phpt",
    "ext/exif/tests/exif003.phpt",
    "ext/exif/tests/exif016.phpt",
    "ext/exif/tests/exif011.phpt",
    "ext/exif/tests/exif006.phpt",
    "ext/exif/tests/exif_read_exif_data_basic.phpt",
    "ext/exif/tests/exif018.phpt",
    "ext/exif/tests/exif_imagetype_variation1.phpt",
    "ext/exif/tests/exif_tagname_basic.phpt",
    "ext/exif/tests/exif001.phpt",
    "ext/exif/tests/bug54002.phpt",
    "ext/exif/tests/exif010.phpt",
    "ext/exif/tests/exif005.phpt",
    "ext/exif/tests/bug48378.phpt",
    "ext/exif/tests/exif007.phpt",
    "ext/exif/tests/exif020.phpt",
    "ext/exif/tests/exif_tagname_error.phpt",
    "ext/exif/tests/exif_tagname_variation1.phpt",
    "ext/exif/tests/exif023.phpt",
    "ext/json/tests/bug42785.phpt",
    "ext/json/tests/005.phpt",
    "ext/json/tests/bug41504.phpt",
    "ext/json/tests/bug54058.phpt",
    "ext/json/tests/bug42090.phpt",
    "ext/json/tests/json_decode_basic.phpt",
    "ext/json/tests/pass003.phpt",
    "ext/json/tests/serialize.phpt",
    "ext/json/tests/001.phpt",
    "ext/json/tests/004.phpt",
    "ext/json/tests/pass001.1_64bit.phpt",
    "ext/json/tests/bug43941.phpt",
    "ext/json/tests/bug41067.phpt",
    "ext/json/tests/json_encode_numeric.phpt",
    "ext/json/tests/json_encode_pretty_print.phpt",
    "ext/json/tests/003.phpt",
    "ext/json/tests/json_encode_basic.phpt",
    "ext/json/tests/bug41567.phpt",
    "ext/json/tests/bug47644.phpt",
    "ext/json/tests/006.phpt",
    "ext/json/tests/json_encode_error.phpt",
    "ext/json/tests/json_encode_unescaped_slashes.phpt",
    "ext/json/tests/bug46944.phpt",
    "ext/json/tests/bug40503.phpt",
    "ext/json/tests/007.phpt",
    "ext/json/tests/pass001.1.phpt",
    "ext/json/tests/bug45791.phpt",
    "ext/json/tests/002.phpt",
    "ext/json/tests/008.phpt",
    "ext/json/tests/json_encode_basic_utf8.phpt",
    "ext/json/tests/bug41403.phpt",
    "ext/json/tests/json_decode_error.phpt",
    "ext/json/tests/bug46215.phpt",
    "ext/json/tests/fail001.phpt",
    "ext/json/tests/bug41034.phpt",
    "ext/json/tests/pass001.phpt",
    "ext/json/tests/pass002.phpt",
    "ext/pspell/tests/005.phpt",
    "ext/pspell/tests/001.phpt",
    "ext/pspell/tests/004.phpt",
    "ext/pspell/tests/003.phpt",
    "ext/pspell/tests/002.phpt",
    "ext/intl/tests/resourcebundle_internal.phpt",
    "ext/intl/tests/formatter_get_locale.phpt",
    "ext/intl/tests/dateformat_get_locale.phpt",
    "ext/intl/tests/dateformat_format_parse.phpt",
    "ext/intl/tests/formatter_get_set_attribute.phpt",
    "ext/intl/tests/formatter_parse.phpt",
    "ext/intl/tests/resourcebundle_create.phpt",
    "ext/intl/tests/transliterator_get_error_message_error.phpt",
    "ext/intl/tests/collator_compare.phpt",
    "ext/intl/tests/transliterator_get_error_code_basic.phpt",
    "ext/intl/tests/locale_get_display_variant.phpt",
    "ext/intl/tests/transliterator_create_inverse_error.phpt",
    "ext/intl/tests/locale_get_script.phpt",
    "ext/intl/tests/dateformat_get_set_calendar.phpt",
    "ext/intl/tests/transliterator_property_id.phpt",
    "ext/intl/tests/transliterator_create_basic.phpt",
    "ext/intl/tests/normalizer_normalize.phpt",
    "ext/intl/tests/formatter_get_set_pattern.phpt",
    "ext/intl/tests/msgfmt_get_error.phpt",
    "ext/intl/tests/msgfmt_parse.phpt",
    "ext/intl/tests/regression_sort_eq.phpt",
    "ext/intl/tests/idn.phpt",
    "ext/intl/tests/dateformat_clone.phpt",
    "ext/intl/tests/dateformat_get_datetype.phpt",
    "ext/intl/tests/collator_sort.phpt",
    "ext/intl/tests/bug53512.phpt",
    "ext/intl/tests/collator_get_error_code.phpt",
    "ext/intl/tests/dateformat_set_timezone_id.phpt",
    "ext/intl/tests/bug12887.phpt",
    "ext/intl/tests/locale_get_display_name.phpt",
    "ext/intl/tests/transliterator_transliterate_error.phpt",
    "ext/intl/tests/msgfmt_fail.phpt",
    "ext/intl/tests/locale_get_display_script.phpt",
    "ext/intl/tests/transliterator_get_error_message_basic.phpt",
    "ext/intl/tests/locale_accept.phpt",
    "ext/intl/tests/formatter_fail.phpt",
    "ext/intl/tests/dateformat_parse.phpt",
    "ext/intl/tests/formatter_format_currency.phpt",
    "ext/intl/tests/dateformat_get_timetype.phpt",
    "ext/intl/tests/regression_sortwsk_eq.phpt",
    "ext/intl/tests/collator_asort.phpt",
    "ext/intl/tests/spoofchecker_001.phpt",
    "ext/intl/tests/msgfmt_clone.phpt",
    "ext/intl/tests/locale_get_display_region.phpt",
    "ext/intl/tests/spoofchecker_003.phpt",
    "ext/intl/tests/transliterator_get_error_code_error.phpt",
    "ext/intl/tests/transliterator_transliterate_basic.phpt",
    "ext/intl/tests/formatter_format.phpt",
    "ext/intl/tests/collator_get_set_strength.phpt",
    "ext/intl/tests/resourcebundle_iterator.phpt",
    "ext/intl/tests/locale_get_all_variants.phpt",
    "ext/intl/tests/resourcebundle_locales.phpt",
    "ext/intl/tests/collation_customization.phpt",
    "ext/intl/tests/bug14562.phpt",
    "ext/intl/tests/dateformat_get_timezone_id.phpt",
    "ext/intl/tests/locale_lookup.phpt",
    "ext/intl/tests/bug48227.phpt",
    "ext/intl/tests/intl_is_failure.phpt",
    "ext/intl/tests/formatter_get_set_symbol.phpt",
    "ext/intl/tests/collator_create.phpt",
    "ext/intl/tests/dateformat_parse_localtime_parsepos.phpt",
    "ext/intl/tests/collator_get_set_attribute.phpt",
    "ext/intl/tests/locale_get_keywords.phpt",
    "ext/intl/tests/locale_parse_locale.phpt",
    "ext/intl/tests/collator_get_locale.phpt",
    "ext/intl/tests/intl_error_name.phpt",
    "ext/intl/tests/intl_get_error_code.phpt",
    "ext/intl/tests/badargs.phpt",
    "ext/intl/tests/formatter_parse_currency.phpt",
    "ext/intl/tests/transliterator_clone.phpt",
    "ext/intl/tests/locale_compose_locale.phpt",
    "ext/intl/tests/locale_get_primary_language.phpt",
    "ext/intl/tests/locale_set_default.phpt",
    "ext/intl/tests/transliterator_transliterate_variant1.phpt",
    "ext/intl/tests/msgfmt_get_locale.phpt",
    "ext/intl/tests/transliterator_create_inverse_basic.phpt",
    "ext/intl/tests/collator_sort_with_sort_keys.phpt",
    "ext/intl/tests/dateformat_localtime.phpt",
    "ext/intl/tests/formatter_format_conv.phpt",
    "ext/intl/tests/regression_sort_and_cow.phpt",
    "ext/intl/tests/locale_get_display_language.phpt",
    "ext/intl/tests/transliterator_list_ids_basic.phpt",
    "ext/intl/tests/bug50590.phpt",
    "ext/intl/tests/dateformat_parse_timestamp_parsepos.phpt",
    "ext/intl/tests/collator_get_sort_key.phpt",
    "ext/intl/tests/transliterator_create_from_rule_basic.phpt",
    "ext/intl/tests/grapheme.phpt",
    "ext/intl/tests/locale_get_region.phpt",
    "ext/intl/tests/spoofchecker_004.phpt",
    "ext/intl/tests/regression_sortwsk_and_cow.phpt",
    "ext/intl/tests/intl_get_error_message.phpt",
    "ext/intl/tests/transliterator_create_error.phpt",
    "ext/intl/tests/formatter_get_error.phpt",
    "ext/intl/tests/dateformat_is_set_lenient.phpt",
    "ext/intl/tests/locale_filter_matches.phpt",
    "ext/intl/tests/spoofchecker_002.phpt",
    "ext/intl/tests/transliterator_list_ids_error.phpt",
    "ext/intl/tests/resourcebundle_individual.phpt",
    "ext/intl/tests/transliterator_create_from_rule_error.phpt",
    "ext/intl/tests/collator_get_error_message.phpt",
    "ext/intl/tests/formatter_get_set_text_attribute.phpt",
    "ext/intl/tests/formatter_clone.phpt",
    "ext/intl/tests/dateformat_format.phpt",
    "ext/intl/tests/msgfmt_get_set_pattern.phpt",
    "ext/intl/tests/locale_get_default.phpt",
    "ext/intl/tests/msgfmt_format.phpt",
    "ext/intl/tests/dateformat_get_set_pattern.phpt",
    "ext/intl/tests/resourcebundle_arrayaccess.phpt",
    "ext/soap/tests/server009.phpt",
    "ext/soap/tests/classmap004.phpt",
    "ext/soap/tests/server021.phpt",
    "ext/soap/tests/classmap003.phpt",
    "ext/soap/tests/server006.phpt",
    "ext/soap/tests/typemap004.phpt",
    "ext/soap/tests/server030.phpt",
    "ext/soap/tests/typemap008.phpt",
    "ext/soap/tests/server022.phpt",
    "ext/soap/tests/bugs/bug37013.phpt",
    "ext/soap/tests/bugs/bug31755.phpt",
    "ext/soap/tests/bugs/bug29109.phpt",
    "ext/soap/tests/bugs/bug35273.phpt",
    "ext/soap/tests/bugs/bug41004.phpt",
    "ext/soap/tests/bugs/bug46419.phpt",
    "ext/soap/tests/bugs/bug32776.phpt",
    "ext/soap/tests/bugs/bug36614.phpt",
    "ext/soap/tests/bugs/bug42692.phpt",
    "ext/soap/tests/bugs/bug37278.phpt",
    "ext/soap/tests/bugs/bug42326.phpt",
    "ext/soap/tests/bugs/bug30994.phpt",
    "ext/soap/tests/bugs/bug50698_4.phpt",
    "ext/soap/tests/bugs/bug42488.phpt",
    "ext/soap/tests/bugs/bug36999.phpt",
    "ext/soap/tests/bugs/bug34643.phpt",
    "ext/soap/tests/bugs/bug42183.phpt",
    "ext/soap/tests/bugs/multiport.phpt",
    "ext/soap/tests/bugs/bug50698_1.phpt",
    "ext/soap/tests/bugs/bug47273.phpt",
    "ext/soap/tests/bugs/bug41337_2.phpt",
    "ext/soap/tests/bugs/bug27742.phpt",
    "ext/soap/tests/bugs/bug41477.phpt",
    "ext/soap/tests/bugs/bug37083.phpt",
    "ext/soap/tests/bugs/bug38536.phpt",
    "ext/soap/tests/bugs/bug38067.phpt",
    "ext/soap/tests/bugs/bug41097.phpt",
    "ext/soap/tests/bugs/bug31422.phpt",
    "ext/soap/tests/bugs/bug31695.phpt",
    "ext/soap/tests/bugs/bug40609.phpt",
    "ext/soap/tests/bugs/bug34449.phpt",
    "ext/soap/tests/bugs/bug36226-2.phpt",
    "ext/soap/tests/bugs/bug28969.phpt",
    "ext/soap/tests/bugs/bug36575.phpt",
    "ext/soap/tests/bugs/bug29830.phpt",
    "ext/soap/tests/bugs/bug42359.phpt",
    "ext/soap/tests/bugs/bug36629.phpt",
    "ext/soap/tests/bugs/bug46427.phpt",
    "ext/soap/tests/bugs/bug42086.phpt",
    "ext/soap/tests/bugs/bug30928.phpt",
    "ext/soap/tests/bugs/bug44811.phpt",
    "ext/soap/tests/bugs/bug36226.phpt",
    "ext/soap/tests/bugs/bug36908.phpt",
    "ext/soap/tests/bugs/bug42151.phpt",
    "ext/soap/tests/bugs/bug34453.phpt",
    "ext/soap/tests/bugs/bug41337.phpt",
    "ext/soap/tests/bugs/bug44882.phpt",
    "ext/soap/tests/bugs/bug50698_3.phpt",
    "ext/soap/tests/bugs/bug29061.phpt",
    "ext/soap/tests/bugs/bug39832.phpt",
    "ext/soap/tests/bugs/bug35142.phpt",
    "ext/soap/tests/bugs/bug39121.phpt",
    "ext/soap/tests/bugs/bug47049.phpt",
    "ext/soap/tests/bugs/bug39815.phpt",
    "ext/soap/tests/bugs/bug50762.phpt",
    "ext/soap/tests/bugs/bug30106.phpt",
    "ext/soap/tests/bugs/bug28985.phpt",
    "ext/soap/tests/bugs/bug29839.phpt",
    "ext/soap/tests/bugs/bug32941.phpt",
    "ext/soap/tests/bugs/bug30799.phpt",
    "ext/soap/tests/bugs/bug38055.phpt",
    "ext/soap/tests/bugs/bug42214.phpt",
    "ext/soap/tests/bugs/bug30045.phpt",
    "ext/soap/tests/bugs/bug38005.phpt",
    "ext/soap/tests/bugs/bug29795.phpt",
    "ext/soap/tests/bugs/bug38004.phpt",
    "ext/soap/tests/bugs/bug41566.phpt",
    "ext/soap/tests/bugs/bug29236.phpt",
    "ext/soap/tests/bugs/bug50698_2.phpt",
    "ext/soap/tests/bugs/bug30175.phpt",
    "ext/soap/tests/bugs/bug28751.phpt",
    "ext/soap/tests/bugs/bug27722.phpt",
    "ext/soap/tests/bugs/bug29844.phpt",
    "ext/soap/tests/bugs/bug43045.phpt",
    "ext/soap/tests/bugs/bug34657.phpt",
    "ext/soap/tests/server015.phpt",
    "ext/soap/tests/server014.phpt",
    "ext/soap/tests/server016.phpt",
    "ext/soap/tests/bug46760.phpt",
    "ext/soap/tests/interop/Round3/GroupE/r3_groupE_list_002w.phpt",
    "ext/soap/tests/interop/Round3/GroupE/r3_groupE_list_006w.phpt",
    "ext/soap/tests/interop/Round3/GroupE/r3_groupE_list_005w.phpt",
    "ext/soap/tests/interop/Round3/GroupE/r3_groupE_list_001w.phpt",
    "ext/soap/tests/interop/Round3/GroupE/r3_groupE_list_003w.phpt",
    "ext/soap/tests/interop/Round3/GroupE/r3_groupE_list_004w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_rpcenc_002w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_rpcenc_001w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_doclitparams_001w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_compound1_002w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_import2_001w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_import3_001w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_doclit_002w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_compound1_001w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_import1_001w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_rpcenc_003w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_doclit_003w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_doclitparams_004w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_import3_002w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_compound2_001w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_doclit_001w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_compound1_003w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_doclitparams_002w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_doclitparams_003w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_rpcenc_004w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_doclit_004w.phpt",
    "ext/soap/tests/interop/Round3/GroupD/r3_groupD_emptysa_001w.phpt",
    "ext/soap/tests/interop/Round3/GroupF/r3_groupF_headers_001w.phpt",
    "ext/soap/tests/interop/Round3/GroupF/r3_groupF_headers_002w.phpt",
    "ext/soap/tests/interop/Round3/GroupF/r3_groupF_headers_004w.phpt",
    "ext/soap/tests/interop/Round3/GroupF/r3_groupF_headers_003w.phpt",
    "ext/soap/tests/interop/Round3/GroupF/r3_groupF_ext_001w.phpt",
    "ext/soap/tests/interop/Round3/GroupF/r3_groupF_extreq_001w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_029w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_022w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_008w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_027w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_013w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_015w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_025w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_010w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_009w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_012w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_018w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_020w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_017w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_005w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_030w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_014w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_023w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_028w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_016w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_007w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_024w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_004w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_031w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_026w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_011w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_033w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_034w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_002w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_006w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_021w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_001w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_003w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_019w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_035w.phpt",
    "ext/soap/tests/interop/Round4/GroupI/r4_groupI_xsd_032w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_rpcenc_005w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_soapfault_001w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_012w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_017w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_rpcenc_009w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_015w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_014w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_005w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_rpcenc_007w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_doclit_002w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_doclit_009w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_001w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_soapfault_003w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_001w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_009w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_012w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_rpcenc_010w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_rpcenc_008w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_004w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_doclit_010w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_002w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_doclit_005w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_doclit_008w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_003w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_015w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_017w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_005w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_rpcenc_006w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_013w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_doclit_006w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_016w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_003w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_rpcenc_001w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_soapfault_002w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_007w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_010w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_009w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_002w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_010w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_008w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_soapfault_004w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_011w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_011w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_007w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_008w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_016w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_rpcenc_003w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_doclit_004w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_rpcenc_002w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_rpcenc_004w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_014w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_doclit_007w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_006w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_013w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_rpcenc_004w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_doclit_003w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_complex_doclit_001w.phpt",
    "ext/soap/tests/interop/Round4/GroupH/r4_groupH_simple_doclit_006w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_011w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_011s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_002w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_016p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_020w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_001p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_015p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_017w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_003s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_015s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_007w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_007s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_017p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_005w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_001w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_011p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_021w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_024p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_024s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_021s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_004s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_012w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_004p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_022w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_013w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_023w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_013s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_023s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_002s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_013p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_003p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_010s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_019w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_009s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_005s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_021p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_003w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_008w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_019p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_023p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_005p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_012p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_008p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_010w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_020p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_018w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_002p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_018p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_019s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_007p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_022p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_006p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_014s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_015w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_009p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_010p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_012s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_006w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_024w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_014p.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_001s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_008s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_020s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_006s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_016s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_018s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_014w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_009w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_022s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_017s.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_004w.phpt",
    "ext/soap/tests/interop/Round2/Base/r2_base_016w.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_005p.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_002w.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_002p.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_001p.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_004s.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_004w.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_003w.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_001s.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_005w.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_001w.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_005s.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_003s.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_002s.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_004p.phpt",
    "ext/soap/tests/interop/Round2/GroupB/r2_groupB_003p.phpt",
    "ext/soap/tests/server023.phpt",
    "ext/soap/tests/setheaders.phpt",
    "ext/soap/tests/typemap007.phpt",
    "ext/soap/tests/server025.phpt",
    "ext/soap/tests/classmap002.phpt",
    "ext/soap/tests/typemap001.phpt",
    "ext/soap/tests/server010.phpt",
    "ext/soap/tests/server004.phpt",
    "ext/soap/tests/server020.phpt",
    "ext/soap/tests/server005.phpt",
    "ext/soap/tests/server001.phpt",
    "ext/soap/tests/server012.phpt",
    "ext/soap/tests/server007.phpt",
    "ext/soap/tests/any.phpt",
    "ext/soap/tests/typemap013.phpt",
    "ext/soap/tests/server028.phpt",
    "ext/soap/tests/server026.phpt",
    "ext/soap/tests/server018.phpt",
    "ext/soap/tests/transport001.phpt",
    "ext/soap/tests/server029.phpt",
    "ext/soap/tests/bug48557.phpt",
    "ext/soap/tests/typemap012.phpt",
    "ext/soap/tests/server017.phpt",
    "ext/soap/tests/typemap010.phpt",
    "ext/soap/tests/typemap011.phpt",
    "ext/soap/tests/server008.phpt",
    "ext/soap/tests/soap12/T10.phpt",
    "ext/soap/tests/soap12/T80.phpt",
    "ext/soap/tests/soap12/T28.phpt",
    "ext/soap/tests/soap12/T45.phpt",
    "ext/soap/tests/soap12/T14.phpt",
    "ext/soap/tests/soap12/T41.phpt",
    "ext/soap/tests/soap12/T36.phpt",
    "ext/soap/tests/soap12/T12.phpt",
    "ext/soap/tests/soap12/T30.phpt",
    "ext/soap/tests/soap12/T40.phpt",
    "ext/soap/tests/soap12/T66.phpt",
    "ext/soap/tests/soap12/T77_1.phpt",
    "ext/soap/tests/soap12/T72.phpt",
    "ext/soap/tests/soap12/T55.phpt",
    "ext/soap/tests/soap12/T74.phpt",
    "ext/soap/tests/soap12/T15.phpt",
    "ext/soap/tests/soap12/T13.phpt",
    "ext/soap/tests/soap12/T48.phpt",
    "ext/soap/tests/soap12/T02.phpt",
    "ext/soap/tests/soap12/T65.phpt",
    "ext/soap/tests/soap12/T03.phpt",
    "ext/soap/tests/soap12/T56.phpt",
    "ext/soap/tests/soap12/T04.phpt",
    "ext/soap/tests/soap12/T78.phpt",
    "ext/soap/tests/soap12/T34.phpt",
    "ext/soap/tests/soap12/T35.phpt",
    "ext/soap/tests/soap12/T76_1.phpt",
    "ext/soap/tests/soap12/T75.phpt",
    "ext/soap/tests/soap12/T73.phpt",
    "ext/soap/tests/soap12/T58.phpt",
    "ext/soap/tests/soap12/T26.phpt",
    "ext/soap/tests/soap12/T24.phpt",
    "ext/soap/tests/soap12/T47.phpt",
    "ext/soap/tests/soap12/T38_1.phpt",
    "ext/soap/tests/soap12/T46.phpt",
    "ext/soap/tests/soap12/T69.phpt",
    "ext/soap/tests/soap12/T64.phpt",
    "ext/soap/tests/soap12/T76_2.phpt",
    "ext/soap/tests/soap12/T61.phpt",
    "ext/soap/tests/soap12/T77_3.phpt",
    "ext/soap/tests/soap12/T33.phpt",
    "ext/soap/tests/soap12/T23.phpt",
    "ext/soap/tests/soap12/T25.phpt",
    "ext/soap/tests/soap12/T54.phpt",
    "ext/soap/tests/soap12/T49.phpt",
    "ext/soap/tests/soap12/T50.phpt",
    "ext/soap/tests/soap12/T05.phpt",
    "ext/soap/tests/soap12/T59.phpt",
    "ext/soap/tests/soap12/T22.phpt",
    "ext/soap/tests/soap12/T42.phpt",
    "ext/soap/tests/soap12/T29.phpt",
    "ext/soap/tests/soap12/T52.phpt",
    "ext/soap/tests/soap12/T70.phpt",
    "ext/soap/tests/soap12/T63.phpt",
    "ext/soap/tests/soap12/T60.phpt",
    "ext/soap/tests/soap12/T71.phpt",
    "ext/soap/tests/soap12/T19.phpt",
    "ext/soap/tests/soap12/T67.phpt",
    "ext/soap/tests/soap12/T51.phpt",
    "ext/soap/tests/soap12/T31.phpt",
    "ext/soap/tests/soap12/T39.phpt",
    "ext/soap/tests/soap12/T77_2.phpt",
    "ext/soap/tests/soap12/T37.phpt",
    "ext/soap/tests/soap12/T11.phpt",
    "ext/soap/tests/soap12/T32.phpt",
    "ext/soap/tests/soap12/T27.phpt",
    "ext/soap/tests/soap12/T57.phpt",
    "ext/soap/tests/soap12/T68.phpt",
    "ext/soap/tests/soap12/T38_2.phpt",
    "ext/soap/tests/soap12/T53.phpt",
    "ext/soap/tests/soap12/T44.phpt",
    "ext/soap/tests/soap12/T01.phpt",
    "ext/soap/tests/soap12/T43.phpt",
    "ext/soap/tests/typemap009.phpt",
    "ext/soap/tests/typemap006.phpt",
    "ext/soap/tests/typemap002.phpt",
    "ext/soap/tests/typemap005.phpt",
    "ext/soap/tests/server024.phpt",
    "ext/soap/tests/server019.phpt",
    "ext/soap/tests/typemap003.phpt",
    "ext/soap/tests/server027.phpt",
    "ext/soap/tests/schema/schema039.phpt",
    "ext/soap/tests/schema/schema051.phpt",
    "ext/soap/tests/schema/schema047.phpt",
    "ext/soap/tests/schema/schema019.phpt",
    "ext/soap/tests/schema/schema043.phpt",
    "ext/soap/tests/schema/schema052.phpt",
    "ext/soap/tests/schema/schema042.phpt",
    "ext/soap/tests/schema/schema054.phpt",
    "ext/soap/tests/schema/schema006.phpt",
    "ext/soap/tests/schema/schema009.phpt",
    "ext/soap/tests/schema/schema071.phpt",
    "ext/soap/tests/schema/schema059.phpt",
    "ext/soap/tests/schema/schema074.phpt",
    "ext/soap/tests/schema/schema016.phpt",
    "ext/soap/tests/schema/schema004.phpt",
    "ext/soap/tests/schema/schema023.phpt",
    "ext/soap/tests/schema/schema002.phpt",
    "ext/soap/tests/schema/schema044.phpt",
    "ext/soap/tests/schema/schema063.phpt",
    "ext/soap/tests/schema/schema028.phpt",
    "ext/soap/tests/schema/schema005.phpt",
    "ext/soap/tests/schema/schema049.phpt",
    "ext/soap/tests/schema/schema025.phpt",
    "ext/soap/tests/schema/schema073.phpt",
    "ext/soap/tests/schema/schema079.phpt",
    "ext/soap/tests/schema/schema008.phpt",
    "ext/soap/tests/schema/schema072.phpt",
    "ext/soap/tests/schema/schema027.phpt",
    "ext/soap/tests/schema/schema018.phpt",
    "ext/soap/tests/schema/schema045.phpt",
    "ext/soap/tests/schema/schema003.phpt",
    "ext/soap/tests/schema/schema015.phpt",
    "ext/soap/tests/schema/schema011.phpt",
    "ext/soap/tests/schema/schema032.phpt",
    "ext/soap/tests/schema/schema055.phpt",
    "ext/soap/tests/schema/schema058.phpt",
    "ext/soap/tests/schema/schema010.phpt",
    "ext/soap/tests/schema/schema048.phpt",
    "ext/soap/tests/schema/schema029.phpt",
    "ext/soap/tests/schema/schema060.phpt",
    "ext/soap/tests/schema/schema017.phpt",
    "ext/soap/tests/schema/schema040.phpt",
    "ext/soap/tests/schema/schema068.phpt",
    "ext/soap/tests/schema/schema014.phpt",
    "ext/soap/tests/schema/schema007.phpt",
    "ext/soap/tests/schema/schema066.phpt",
    "ext/soap/tests/schema/schema070.phpt",
    "ext/soap/tests/schema/schema065.phpt",
    "ext/soap/tests/schema/schema035.phpt",
    "ext/soap/tests/schema/schema022.phpt",
    "ext/soap/tests/schema/schema082.phpt",
    "ext/soap/tests/schema/schema081.phpt",
    "ext/soap/tests/schema/schema085.phpt",
    "ext/soap/tests/schema/schema026.phpt",
    "ext/soap/tests/schema/schema080.phpt",
    "ext/soap/tests/schema/schema069.phpt",
    "ext/soap/tests/schema/schema030.phpt",
    "ext/soap/tests/schema/schema057.phpt",
    "ext/soap/tests/schema/schema083.phpt",
    "ext/soap/tests/schema/schema034.phpt",
    "ext/soap/tests/schema/schema076.phpt",
    "ext/soap/tests/schema/schema031.phpt",
    "ext/soap/tests/schema/schema024.phpt",
    "ext/soap/tests/schema/schema075.phpt",
    "ext/soap/tests/schema/schema053.phpt",
    "ext/soap/tests/schema/schema033.phpt",
    "ext/soap/tests/schema/schema062.phpt",
    "ext/soap/tests/schema/schema036.phpt",
    "ext/soap/tests/schema/schema020.phpt",
    "ext/soap/tests/schema/schema050.phpt",
    "ext/soap/tests/schema/schema021.phpt",
    "ext/soap/tests/schema/schema046.phpt",
    "ext/soap/tests/schema/schema077.phpt",
    "ext/soap/tests/schema/schema037.phpt",
    "ext/soap/tests/schema/schema056.phpt",
    "ext/soap/tests/schema/schema013.phpt",
    "ext/soap/tests/schema/schema012.phpt",
    "ext/soap/tests/schema/schema061.phpt",
    "ext/soap/tests/schema/schema084.phpt",
    "ext/soap/tests/schema/schema038.phpt",
    "ext/soap/tests/schema/schema064.phpt",
    "ext/soap/tests/schema/schema041.phpt",
    "ext/soap/tests/schema/schema001.phpt",
    "ext/soap/tests/schema/schema067.phpt",
    "ext/soap/tests/schema/schema078.phpt",
    "ext/soap/tests/server031.phpt",
    "ext/soap/tests/server013.phpt",
    "ext/soap/tests/server002.phpt",
    "ext/soap/tests/server003.phpt",
    "ext/soap/tests/classmap001.phpt",
    "ext/soap/tests/server011.phpt",
    "ext/shmop/tests/001.phpt",
    "ext/tokenizer/tests/token_get_all_variation10.phpt",
    "ext/tokenizer/tests/token_get_all_variation5.phpt",
    "ext/tokenizer/tests/token_get_all_variation11.phpt",
    "ext/tokenizer/tests/token_get_all_variation1.phpt",
    "ext/tokenizer/tests/token_get_all_variation14.phpt",
    "ext/tokenizer/tests/token_get_all_variation7.phpt",
    "ext/tokenizer/tests/token_get_all_variation8.phpt",
    "ext/tokenizer/tests/token_get_all_variation3.phpt",
    "ext/tokenizer/tests/token_get_all_variation13.phpt",
    "ext/tokenizer/tests/001.phpt",
    "ext/tokenizer/tests/token_get_all_variation4.phpt",
    "ext/tokenizer/tests/token_get_all_variation12.phpt",
    "ext/tokenizer/tests/token_get_all_variation2.phpt",
    "ext/tokenizer/tests/token_get_all_basic.phpt",
    "ext/tokenizer/tests/token_get_all_error.phpt",
    "ext/tokenizer/tests/token_get_all_variation6.phpt",
    "ext/tokenizer/tests/003.phpt",
    "ext/tokenizer/tests/token_get_all_variation9.phpt",
    "ext/tokenizer/tests/token_get_all_variation17.phpt",
    "ext/tokenizer/tests/bug54089.phpt",
    "ext/tokenizer/tests/token_get_all_variation16.phpt",
    "ext/tokenizer/tests/002.phpt",
    "ext/tokenizer/tests/token_get_all_variation19.phpt",
    "ext/tokenizer/tests/bug26463.phpt",
    "ext/tokenizer/tests/token_get_all_variation15.phpt",
    "ext/tokenizer/tests/token_get_all_variation18.phpt",
    "ext/snmp/tests/snmp_set_enum_print.phpt",
    "ext/snmp/tests/snmpset-nomib.phpt",
    "ext/snmp/tests/snmpgetnext.phpt",
    "ext/snmp/tests/snmp3-error.phpt",
    "ext/snmp/tests/snmp3.phpt",
    "ext/snmp/tests/snmp2_getnext.phpt",
    "ext/snmp/tests/snmp-object-error.phpt",
    "ext/snmp/tests/snmp-object-setSecurity_error.phpt",
    "ext/snmp/tests/snmp2_set.phpt",
    "ext/snmp/tests/wrong_hostname.phpt",
    "ext/snmp/tests/snmp2_real_walk.phpt",
    "ext/snmp/tests/snmprealwalk.phpt",
    "ext/snmp/tests/snmp-object-errno-errstr.phpt",
    "ext/snmp/tests/generic_timeout_error.phpt",
    "ext/snmp/tests/snmpwalk.phpt",
    "ext/snmp/tests/snmp_get_quick_print.phpt",
    "ext/snmp/tests/snmp2_walk.phpt",
    "ext/snmp/tests/snmp2_get.phpt",
    "ext/snmp/tests/snmp-object.phpt",
    "ext/snmp/tests/snmpset.phpt",
    "ext/snmp/tests/snmp_set_oid_output_format.phpt",
    "ext/snmp/tests/snmp_get_valueretrieval.phpt",
    "ext/snmp/tests/snmp-object-properties.phpt",
    "ext/snmp/tests/snmpget.phpt",
    "ext/snmp/tests/snmp_getvalue.phpt",
    "ext/snmp/tests/snmp_read_mib.phpt",
    "ext/snmp/tests/snmp2_set-nomib.phpt",
    "ext/xml/tests/xml_set_character_data_handler_variation1.phpt",
    "ext/xml/tests/xml_parser_get_option_variation2.phpt",
    "ext/xml/tests/xml_set_end_namespace_decl_handler_error.phpt",
    "ext/xml/tests/bug26614_libxml.phpt",
    "ext/xml/tests/xml_set_notation_decl_handler_variation1.phpt",
    "ext/xml/tests/xml_set_object_variation2.phpt",
    "ext/xml/tests/xml_parser_create_error.phpt",
    "ext/xml/tests/bug27908.phpt",
    "ext/xml/tests/xml_set_external_entity_ref_handler_error.phpt",
    "ext/xml/tests/utf8_decode_error.phpt",
    "ext/xml/tests/bug35447.phpt",
    "ext/xml/tests/xml_get_error_code_error.phpt",
    "ext/xml/tests/bug25666.phpt",
    "ext/xml/tests/xml_set_start_namespace_decl_handler_error.phpt",
    "ext/xml/tests/xml_set_element_handler_variation1.phpt",
    "ext/xml/tests/xml011.phpt",
    "ext/xml/tests/xml_parser_create_ns_variation1.phpt",
    "ext/xml/tests/xml_get_current_column_number_error.phpt",
    "ext/xml/tests/xml_set_start_namespace_decl_handler_variation1.phpt",
    "ext/xml/tests/xml_set_character_data_handler_error.phpt",
    "ext/xml/tests/xml003.phpt",
    "ext/xml/tests/xml001.phpt",
    "ext/xml/tests/xml_parser_create_variation1.phpt",
    "ext/xml/tests/xml_set_unparsed_entity_decl_handler_variation1.phpt",
    "ext/xml/tests/xml_parser_set_option_basic.phpt",
    "ext/xml/tests/xml_set_default_handler_variation1.phpt",
    "ext/xml/tests/xml_set_processing_instruction_handler_basic.phpt",
    "ext/xml/tests/xml_get_current_byte_index_error.phpt",
    "ext/xml/tests/xml_set_end_namespace_decl_handler_variation1.phpt",
    "ext/xml/tests/xml_parser_free_error.phpt",
    "ext/xml/tests/xml_parser_set_option_variation1.phpt",
    "ext/xml/tests/bug26528.phpt",
    "ext/xml/tests/xml_set_unparsed_entity_decl_handler_error.phpt",
    "ext/xml/tests/bug32001b.phpt",
    "ext/xml/tests/xml_parser_set_option_variation2.phpt",
    "ext/xml/tests/xml_parse_into_struct_variation1.phpt",
    "ext/xml/tests/xml_parser_get_option_error.phpt",
    "ext/xml/tests/xml_parser_set_option_variation3.phpt",
    "ext/xml/tests/xml002.phpt",
    "ext/xml/tests/xml_get_error_code_variation1.phpt",
    "ext/xml/tests/xml_parser_create_ns_error.phpt",
    "ext/xml/tests/bug26614.phpt",
    "ext/xml/tests/xml_get_current_column_number_variation1.phpt",
    "ext/xml/tests/xml010.phpt",
    "ext/xml/tests/bug43957.phpt",
    "ext/xml/tests/xml_parse_into_struct_variation.phpt",
    "ext/xml/tests/xml_set_processing_instruction_handler_variation1.phpt",
    "ext/xml/tests/xml_closures_001.phpt",
    "ext/xml/tests/xml_parse_into_struct_error.phpt",
    "ext/xml/tests/xml007.phpt",
    "ext/xml/tests/xml_set_start_namespace_decl_handler_basic.phpt",
    "ext/xml/tests/xml_parser_set_option_error.phpt",
    "ext/xml/tests/xml_error_string_variation1.phpt",
    "ext/xml/tests/utf8_decode_variation1.phpt",
    "ext/xml/tests/xml_parser_get_option_variation1.phpt",
    "ext/xml/tests/xml006.phpt",
    "ext/xml/tests/xml_parse_variation1.phpt",
    "ext/xml/tests/utf8_encode_variation1.phpt",
    "ext/xml/tests/xml_set_element_handler_error.phpt",
    "ext/xml/tests/xml_set_notation_decl_handler_error.phpt",
    "ext/xml/tests/utf8_encode_error.phpt",
    "ext/xml/tests/bug49687.phpt",
    "ext/xml/tests/xml009.phpt",
    "ext/xml/tests/bug32001.phpt",
    "ext/xml/tests/xml_set_object_variation1.phpt",
    "ext/xml/tests/xml_parser_free_variation1.phpt",
    "ext/xml/tests/xml_set_processing_instruction_handler_error.phpt",
    "ext/xml/tests/xml_set_object_error.phpt",
    "ext/xml/tests/xml_parse_error.phpt",
    "ext/xml/tests/xml_set_external_entity_ref_handler_variation1.phpt",
    "ext/xml/tests/bug46699.phpt",
    "ext/xml/tests/xml_error_string_error.phpt",
    "ext/xml/tests/xml_set_default_handler_error.phpt",
    "ext/xml/tests/xml_set_notation_decl_handler_basic.phpt",
    "ext/xml/tests/xml004.phpt",
    "ext/xml/tests/xml_get_current_byte_index_variation1.phpt",
    "ext/xml/tests/xml_get_current_line_number_error.phpt",
    "ext/xml/tests/bug50576.phpt",
    "ext/xml/tests/xml_get_current_line_number_variation1.phpt",
    "ext/xml/tests/bug30266.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_nextrowset.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_phpinfo.phpt",
    "ext/pdo_mysql/tests/bug_42499.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_fetch_serialize_simple.phpt",
    "ext/pdo_mysql/tests/pdo_mysql___construct_options_libmysql.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_fetch_non_select.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_pconnect.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_case.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_blobs.phpt",
    "ext/pdo_mysql/tests/show_tables.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_types.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_connection_status.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_prepare_match_against.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_bindvalue.phpt",
    "ext/pdo_mysql/tests/bug_41698.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_prepare_native_placeholder_everywhere.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_fetchobject.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_prepare_native.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_variable_columncount.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_fetch_table_names.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_exec_load_data.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_exec_select.phpt",
    "ext/pdo_mysql/tests/bug_51670.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_class_constants.phpt",
    "ext/pdo_mysql/tests/bug_41997.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_columncount.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_autocommit.phpt",
    "ext/pdo_mysql/tests/bug_39858.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_bindparam.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_exec.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_prepare_native_named_placeholder.phpt",
    "ext/pdo_mysql/tests/bug53782.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_prepare_load_data.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_errorinfo.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_prepare_native_dup_named_placeholder.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_driver_name.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_prepare_emulated_anonymous.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_begintransaction.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_statement_class.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_prefetch.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_prepare_native_column.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_errorcode.phpt",
    "ext/pdo_mysql/tests/bug_33689.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_client_version.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_prepare_emulated.phpt",
    "ext/pdo_mysql/tests/bug_39483.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_prepare_emulated_placeholder_everywhere.phpt",
    "ext/pdo_mysql/tests/pdo_mysql___construct_options.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_prepare_native_mixed_style.phpt",
    "ext/pdo_mysql/tests/pdo_mysql___construct_uri.phpt",
    "ext/pdo_mysql/tests/common.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_bindparam_types.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_rowcount.phpt",
    "ext/pdo_mysql/tests/bug_43371.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_subclass.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_prepare_native_clear_error.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_blobfromsteam.phpt",
    "ext/pdo_mysql/tests/bug_37445.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_interface.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_errorcode.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_getcolumnmeta.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_connect_charset.phpt",
    "ext/pdo_mysql/tests/pecl_bug_5802.phpt",
    "ext/pdo_mysql/tests/bug_pecl_7976.phpt",
    "ext/pdo_mysql/tests/bug_44454.phpt",
    "ext/pdo_mysql/tests/pecl_bug_5780.phpt",
    "ext/pdo_mysql/tests/bug44327.phpt",
    "ext/pdo_mysql/tests/bug41125.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_max_buffer_size.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_oracle_nulls.phpt",
    "ext/pdo_mysql/tests/bug_44707.phpt",
    "ext/pdo_mysql/tests/pecl_bug_5200.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_bit.phpt",
    "ext/pdo_mysql/tests/bug53551.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_unbuffered_2050.phpt",
    "ext/pdo_mysql/tests/bug_pecl_12925.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_bindcolumn.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_closecursor.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_commit.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_get_attribute.phpt",
    "ext/pdo_mysql/tests/pdo_mysql___construct_ini.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_rollback.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_server_version.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_init_command.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_fetch_serialize.phpt",
    "ext/pdo_mysql/tests/last_insert_id.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_types_zerofill.phpt",
    "ext/pdo_mysql/tests/bug_50323.phpt",
    "ext/pdo_mysql/tests/bug_41125.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_closecursor_empty.phpt",
    "ext/pdo_mysql/tests/bug46292.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_server_info.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_multiquery.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_stmt_errorinfo.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_fetch_both.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_exec_ddl.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_last_insert_id.phpt",
    "ext/pdo_mysql/tests/bug_45120.phpt",
    "ext/pdo_mysql/tests/pdo_mysql_attr_errmode.phpt",
    "ext/pdo_mysql/tests/pdo_mysql___construct.phpt",
    "ext/interbase/tests/ibase_rollback_001.phpt",
    "ext/interbase/tests/005.phpt",
    "ext/interbase/tests/ibase_num_fields_001.phpt",
    "ext/interbase/tests/ibase_free_query_001.phpt",
    "ext/interbase/tests/bug45575.phpt",
    "ext/interbase/tests/004.phpt",
    "ext/interbase/tests/ibase_affected_rows_001.phpt",
    "ext/interbase/tests/ibase_param_info_001.phpt",
    "ext/interbase/tests/ibase_close_001.phpt",
    "ext/interbase/tests/003.phpt",
    "ext/interbase/tests/ibase_trans_001.phpt",
    "ext/interbase/tests/ibase_num_params_001.phpt",
    "ext/interbase/tests/006.phpt",
    "ext/interbase/tests/007.phpt",
    "ext/interbase/tests/002.phpt",
    "ext/interbase/tests/008.phpt",
    "ext/interbase/tests/ibase_errmsg_001.phpt",
    "ext/interbase/tests/bug46247.phpt",
    "ext/interbase/tests/bug45373.phpt",
    "ext/interbase/tests/ibase_drop_db_001.phpt",
    "ext/interbase/tests/bug46543.phpt",
    "ext/interbase/tests/ibase_trans_002.phpt",
    "ext/tidy/tests/027.phpt",
    "ext/tidy/tests/033.phpt",
    "ext/tidy/tests/010.phpt",
    "ext/tidy/tests/005.phpt",
    "ext/tidy/tests/014.phpt",
    "ext/tidy/tests/029.phpt",
    "ext/tidy/tests/011.phpt",
    "ext/tidy/tests/tidy_error1.phpt",
    "ext/tidy/tests/bug_50558.phpt",
    "ext/tidy/tests/017.phpt",
    "ext/tidy/tests/tidy_error.phpt",
    "ext/tidy/tests/028.phpt",
    "ext/tidy/tests/001.phpt",
    "ext/tidy/tests/015.phpt",
    "ext/tidy/tests/026.phpt",
    "ext/tidy/tests/004.phpt",
    "ext/tidy/tests/003.phpt",
    "ext/tidy/tests/020.phpt",
    "ext/tidy/tests/022.phpt",
    "ext/tidy/tests/006.phpt",
    "ext/tidy/tests/031.phpt",
    "ext/tidy/tests/023.phpt",
    "ext/tidy/tests/007.phpt",
    "ext/tidy/tests/002.phpt",
    "ext/tidy/tests/008.phpt",
    "ext/tidy/tests/018.phpt",
    "ext/tidy/tests/034.phpt",
    "ext/tidy/tests/035.phpt",
    "ext/tidy/tests/013.phpt",
    "ext/tidy/tests/016.phpt",
    "ext/tidy/tests/025.phpt",
    "ext/tidy/tests/012.phpt",
    "ext/tidy/tests/021.phpt",
    "ext/tidy/tests/030.phpt",
    "ext/tidy/tests/032.phpt",
    "ext/tidy/tests/019.phpt",
    "ext/tidy/tests/024.phpt",
    "ext/tidy/tests/009.phpt",
    "ext/iconv/tests/iconv_strpos_variation2.phpt",
    "ext/iconv/tests/iconv_strrpos-charset-length-cve-2007-4840.phpt",
    "ext/iconv/tests/iconv_substr_basic.phpt",
    "ext/iconv/tests/iconv_substr_error2.phpt",
    "ext/iconv/tests/eucjp2sjis.phpt",
    "ext/iconv/tests/iconv_mime_decode_headers_variation2.phpt",
    "ext/iconv/tests/iconv_strpos.phpt",
    "ext/iconv/tests/iconv_mime_decode_variation3.phpt",
    "ext/iconv/tests/iconv_strrpos_error2.phpt",
    "ext/iconv/tests/iconv_strlen_variation1.phpt",
    "ext/iconv/tests/iconv_mime_decode_headers_variation3.phpt",
    "ext/iconv/tests/iconv_strrpos_variation3.phpt",
    "ext/iconv/tests/translit-failure.phpt",
    "ext/iconv/tests/eucjp2utf8.phpt",
    "ext/iconv/tests/iconv004.phpt",
    "ext/iconv/tests/iconv_mime_decode_variation1.phpt",
    "ext/iconv/tests/iconv_strpos_error2.phpt",
    "ext/iconv/tests/ob_iconv_handler-charset-length-cve-2007-4840.phpt",
    "ext/iconv/tests/iconv_set_encoding-charset-length-cve-2007-4840.phpt",
    "ext/iconv/tests/iconv_strlen_basic.phpt",
    "ext/iconv/tests/translit-utf8.phpt",
    "ext/iconv/tests/bug48289.phpt",
    "ext/iconv/tests/iconv_mime_decode_headers.phpt",
    "ext/iconv/tests/bug37773.phpt",
    "ext/iconv/tests/iconv_mime_decode.phpt",
    "ext/iconv/tests/iconv_strpos_variation5.phpt",
    "ext/iconv/tests/bug53304.phpt",
    "ext/iconv/tests/iconv_mime_decode_headers-charset-length-cve-2007-4840.phpt",
    "ext/iconv/tests/iconv_strlen_error2.phpt",
    "ext/iconv/tests/iconv_strlen_error1.phpt",
    "ext/iconv/tests/iconv_strpos_variation3_64bit.phpt",
    "ext/iconv/tests/iconv_get_encoding_basic.phpt",
    "ext/iconv/tests/iconv_strpos_basic.phpt",
    "ext/iconv/tests/iconv003.phpt",
    "ext/iconv/tests/iconv_strlen_variation2.phpt",
    "ext/iconv/tests/ob_iconv_handler.phpt",
    "ext/iconv/tests/iconv_strpos_error1.phpt",
    "ext/iconv/tests/iconv_substr_error1.phpt",
    "ext/iconv/tests/iconv_strlen-charset-length-cve-2007-4840.phpt",
    "ext/iconv/tests/iconv_strrpos.phpt",
    "ext/iconv/tests/iconv-charset-length-cve-2007-4840.phpt",
    "ext/iconv/tests/iconv_basic.phpt",
    "ext/iconv/tests/iconv001.phpt",
    "ext/iconv/tests/iconv002.phpt",
    "ext/iconv/tests/iconv_get_encoding_error.phpt",
    "ext/iconv/tests/iconv_stream_filter_delimiter.phpt",
    "ext/iconv/tests/iconv_mime_decode-charset-length-cve-2007-4840.phpt",
    "ext/iconv/tests/iconv_strrpos_error1.phpt",
    "ext/iconv/tests/iconv_mime_encode.phpt",
    "ext/iconv/tests/iconv_strpos-charset-length-cve-2007-4840.phpt",
    "ext/iconv/tests/bug52941.phpt",
    "ext/iconv/tests/iconv_strpos_variation4.phpt",
    "ext/iconv/tests/iconv_strpos_variation3.phpt",
    "ext/iconv/tests/iconv_set_encoding_error.phpt",
    "ext/iconv/tests/iconv_substr.phpt",
    "ext/iconv/tests/bug52211.phpt",
    "ext/iconv/tests/bug51250.phpt",
    "ext/iconv/tests/bug37176.phpt",
    "ext/iconv/tests/iconv_strrpos_variation2.phpt",
    "ext/iconv/tests/iconv_strrpos_variation1.phpt",
    "ext/iconv/tests/iconv_encoding_basic.phpt",
    "ext/iconv/tests/iconv_mime_decode_headers_variation1.phpt",
    "ext/iconv/tests/iconv_set_encoding_variation.phpt",
    "ext/iconv/tests/iconv_strlen.phpt",
    "ext/iconv/tests/iconv_strpos_variation1.phpt",
    "ext/iconv/tests/iconv_strrpos_basic.phpt",
    "ext/iconv/tests/iconv_substr-charset-length-cve-2007-4783.phpt",
    "ext/iconv/tests/iconv_mime_decode_variation2.phpt",
    "ext/iconv/tests/eucjp2iso2022jp.phpt",
    "ext/iconv/tests/iconv_stream_filter.phpt",
    "ext/zip/tests/zip_open.phpt",
    "ext/zip/tests/zip_entry_open.phpt",
    "ext/zip/tests/oo_getnameindex.phpt",
    "ext/zip/tests/zip_entry_name.phpt",
    "ext/zip/tests/oo_addfile.phpt",
    "ext/zip/tests/zip_entry_compressedsize.phpt",
    "ext/zip/tests/001.phpt",
    "ext/zip/tests/bug38943_2.phpt",
    "ext/zip/tests/bug38944.phpt",
    "ext/zip/tests/oo_getstatusstring.phpt",
    "ext/zip/tests/oo_open.phpt",
    "ext/zip/tests/zip_entry_compressionmethod.phpt",
    "ext/zip/tests/oo_delete.phpt",
    "ext/zip/tests/zip_read.phpt",
    "ext/zip/tests/bug53579.phpt",
    "ext/zip/tests/zip_entry_filesize.phpt",
    "ext/zip/tests/bug53885.phpt",
    "ext/zip/tests/oo_rename.phpt",
    "ext/zip/tests/bug53603.phpt",
    "ext/zip/tests/zip_close.phpt",
    "ext/zip/tests/zip_open_error.phpt",
    "ext/zip/tests/bug38943.phpt",
    "ext/zip/tests/oo_extract.phpt",
    "ext/zip/tests/bug40228.phpt",
    "ext/zip/tests/oo_properties.phpt",
    "ext/zip/tests/pecl12414.phpt",
    "ext/zip/tests/bug7658.phpt",
    "ext/zip/tests/bug7214.phpt",
    "ext/zip/tests/bug8009.phpt",
    "ext/zip/tests/bug47667.phpt",
    "ext/zip/tests/bug14962.phpt",
    "ext/zip/tests/oo_setcomment.phpt",
    "ext/zip/tests/oo_getcomment.phpt",
    "ext/zip/tests/oo_namelocate.phpt",
    "ext/zip/tests/bug49072.phpt",
    "ext/zip/tests/stream_meta_data.phpt",
    "ext/zip/tests/oo_ext_zip.phpt",
    "ext/zip/tests/zip_entry_read.phpt",
    "ext/zip/tests/bug8700.phpt",
    "ext/zip/tests/oo_close.phpt",
    "ext/zip/tests/bug53854.phpt",
    "ext/zip/tests/oo_addemptydir.phpt",
    "ext/zip/tests/bug11216.phpt",
    "ext/zip/tests/oo_stream.phpt",
    "ext/xmlwriter/tests/010.phpt",
    "ext/xmlwriter/tests/005.phpt",
    "ext/xmlwriter/tests/OO_004.phpt",
    "ext/xmlwriter/tests/OO_007.phpt",
    "ext/xmlwriter/tests/xmlwriter_write_dtd_basic_001.phpt",
    "ext/xmlwriter/tests/OO_003.phpt",
    "ext/xmlwriter/tests/001.phpt",
    "ext/xmlwriter/tests/bug41326.phpt",
    "ext/xmlwriter/tests/004.phpt",
    "ext/xmlwriter/tests/OO_006.phpt",
    "ext/xmlwriter/tests/OO_008.phpt",
    "ext/xmlwriter/tests/bug41287.phpt",
    "ext/xmlwriter/tests/003.phpt",
    "ext/xmlwriter/tests/OO_002.phpt",
    "ext/xmlwriter/tests/006.phpt",
    "ext/xmlwriter/tests/bug48204.phpt",
    "ext/xmlwriter/tests/xmlwriter_open_uri_error_003.phpt",
    "ext/xmlwriter/tests/xmlwriter_set_indent_string_basic_001.phpt",
    "ext/xmlwriter/tests/bug39504.phpt",
    "ext/xmlwriter/tests/007.phpt",
    "ext/xmlwriter/tests/OO_005.phpt",
    "ext/xmlwriter/tests/xmlwriter_open_uri_error_005.phpt",
    "ext/xmlwriter/tests/002.phpt",
    "ext/xmlwriter/tests/008.phpt",
    "ext/xmlwriter/tests/xmlwriter_write_dtd_error_001.phpt",
    "ext/xmlwriter/tests/xmlwriter_open_uri_error_002.phpt",
    "ext/xmlwriter/tests/xmlwriter_open_uri_error_004.phpt",
    "ext/xmlwriter/tests/xmlwriter_open_uri_error_001.phpt",
    "ext/xmlwriter/tests/OO_001.phpt",
    "ext/xmlwriter/tests/xmlwriter_write_attribute_ns_basic_001.phpt",
    "ext/xmlwriter/tests/xmlwriter_set_indent_string_error_001.phpt",
    "ext/xmlwriter/tests/OO_009.phpt",
    "ext/xmlwriter/tests/xmlwriter_write_attribute_ns_error_001.phpt",
    "ext/xmlwriter/tests/009.phpt",
    "ext/calendar/tests/jdtojulian.phpt",
    "ext/calendar/tests/jdtofrench.phpt",
    "ext/calendar/tests/juliantojd.phpt",
    "ext/calendar/tests/jdtomonthname.phpt",
    "ext/calendar/tests/jddayofweek.phpt",
    "ext/calendar/tests/easter_days.phpt",
    "ext/calendar/tests/gregoriantojd.phpt",
    "ext/calendar/tests/frenchtojd.phpt",
    "ext/calendar/tests/cal_info.phpt",
    "ext/calendar/tests/bug52744.phpt",
    "ext/calendar/tests/easter_date.phpt",
    "ext/calendar/tests/jdmonthname.phpt",
    "ext/calendar/tests/jewishtojd.phpt",
    "ext/calendar/tests/cal_from_jd.phpt",
    "ext/calendar/tests/jdtojewish.phpt",
    "ext/calendar/tests/cal_to_jd.phpt",
    "ext/calendar/tests/unixtojd.phpt",
    "ext/calendar/tests/bug53574.phpt",
    "ext/calendar/tests/cal_days_in_month.phpt",
    "ext/calendar/tests/jdtogregorian.phpt",
    "ext/calendar/tests/jdtounix.phpt",
    "tests/func/010.phpt",
    "tests/func/001.phpt",
    "tests/func/004.phpt",
    "tests/func/003.phpt",
    "tests/func/006.phpt",
    "tests/func/007.phpt",
    "tests/func/002.phpt",
    "tests/func/008.phpt",
    "tests/func/ini_alter.phpt",
    "tests/func/009.phpt",
    "tests/security/open_basedir_fileinode.phpt",
    "tests/security/open_basedir_error_log.phpt",
    "tests/security/open_basedir_dir.phpt",
    "tests/security/open_basedir_glob-win32.phpt",
    "tests/security/open_basedir_filegroup.phpt",
    "tests/security/open_basedir_tempnam.phpt",
    "tests/security/open_basedir_is_file.phpt",
    "tests/security/open_basedir_fileperms.phpt",
    "tests/security/open_basedir_symlink.phpt",
    "tests/security/open_basedir_fileowner.phpt",
    "tests/security/open_basedir_unlink.phpt",
    "tests/security/open_basedir_filectime.phpt",
    "tests/security/open_basedir_lstat.phpt",
    "tests/security/open_basedir_filemtime.phpt",
    "tests/security/open_basedir_file_put_contents.phpt",
    "tests/security/open_basedir_linkinfo.phpt",
    "tests/security/bug53226.phpt",
    "tests/security/open_basedir_filesize.phpt",
    "tests/security/open_basedir_fopen.phpt",
    "tests/security/open_basedir_file_get_contents.phpt",
    "tests/security/open_basedir_touch.phpt",
    "tests/security/open_basedir_stat.phpt",
    "tests/security/open_basedir_parse_ini_file.phpt",
    "tests/security/open_basedir_chmod.phpt",
    "tests/security/open_basedir_file.phpt",
    "tests/security/open_basedir_filetype.phpt",
    "tests/security/magic_quotes_gpc.phpt",
    "tests/security/open_basedir_copy_variation1.phpt",
    "tests/security/open_basedir_is_executable.phpt",
    "tests/security/open_basedir_chdir.phpt",
    "tests/security/open_basedir_opendir.phpt",
    "tests/security/open_basedir_is_readable.phpt",
    "tests/security/open_basedir_disk_free_space.phpt",
    "tests/security/open_basedir_rename.phpt",
    "tests/security/open_basedir_mkdir.phpt",
    "tests/security/open_basedir_rmdir.phpt",
    "tests/security/open_basedir_error_log_variation.phpt",
    "tests/security/open_basedir_glob.phpt",
    "tests/security/open_basedir_readlink.phpt",
    "tests/security/open_basedir_fileatime.phpt",
    "tests/security/open_basedir_file_exists.phpt",
    "tests/security/open_basedir_glob_variation.phpt",
    "tests/security/open_basedir_is_dir.phpt",
    "tests/security/open_basedir_is_link.phpt",
    "tests/security/open_basedir_scandir.phpt",
    "tests/security/open_basedir_link.phpt",
    "tests/security/open_basedir_is_writable.phpt",
    "tests/security/open_basedir_realpath.phpt",
    "tests/security/open_basedir_copy.phpt",
    "tests/basic/027.phpt",
    "tests/basic/010.phpt",
    "tests/basic/bug51709_1.phpt",
    "tests/basic/005.phpt",
    "tests/basic/bug45986.phpt",
    "tests/basic/014.phpt",
    "tests/basic/029.phpt",
    "tests/basic/php_real_logo_guid.phpt",
    "tests/basic/011.phpt",
    "tests/basic/017.phpt",
    "tests/basic/028.phpt",
    "tests/basic/001.phpt",
    "tests/basic/bug53180.phpt",
    "tests/basic/015.phpt",
    "tests/basic/026.phpt",
    "tests/basic/004.phpt",
    "tests/basic/rfc1867_anonymous_upload.phpt",
    "tests/basic/rfc1867_garbled_mime_headers.phpt",
    "tests/basic/rfc1867_file_upload_disabled.phpt",
    "tests/basic/bug46313.phpt",
    "tests/basic/php_egg_logo_guid.phpt",
    "tests/basic/003.phpt",
    "tests/basic/enable_post_data_reading_01.phpt",
    "tests/basic/rfc1867_max_file_uploads_empty_files.phpt",
    "tests/basic/rfc1867_array_upload.phpt",
    "tests/basic/rfc1867_post_max_size.phpt",
    "tests/basic/020.phpt",
    "tests/basic/022.phpt",
    "tests/basic/rfc1867_empty_upload.phpt",
    "tests/basic/006.phpt",
    "tests/basic/rfc1867_missing_boundary_2.phpt",
    "tests/basic/023.phpt",
    "tests/basic/bug46759.phpt",
    "tests/basic/rfc1867_missing_boundary.phpt",
    "tests/basic/rfc1867_max_file_size.phpt",
    "tests/basic/rfc1867_boundary_1.phpt",
    "tests/basic/007.phpt",
    "tests/basic/enable_post_data_reading_04.phpt",
    "tests/basic/bug51709_2.phpt",
    "tests/basic/rfc1867_invalid_boundary.phpt",
    "tests/basic/002.phpt",
    "tests/basic/008.phpt",
    "tests/basic/018.phpt",
    "tests/basic/bug29971.phpt",
    "tests/basic/rfc1867_malicious_input.phpt",
    "tests/basic/013.phpt",
    "tests/basic/rfc1867_post_max_filesize.phpt",
    "tests/basic/016.phpt",
    "tests/basic/025.phpt",
    "tests/basic/012.phpt",
    "tests/basic/021.phpt",
    "tests/basic/enable_post_data_reading_02.phpt",
    "tests/basic/rfc1867_max_file_uploads_empty_files_debug.phpt",
    "tests/basic/php_logo_guid.phpt",
    "tests/basic/bug46313-win.phpt",
    "tests/basic/req44164.phpt",
    "tests/basic/rfc1867_boundary_2.phpt",
    "tests/basic/019.phpt",
    "tests/basic/zend_logo_guid.phpt",
    "tests/basic/bug20539.phpt",
    "tests/basic/024.phpt",
    "tests/basic/009.phpt",
    "tests/basic/enable_post_data_reading_03.phpt",
    "tests/output/bug46897.phpt",
    "tests/output/ob_008.phpt",
    "tests/output/ob_get_length_error_001.phpt",
    "tests/output/ob_end_flush_basic_001.phpt",
    "tests/output/ob_implicit_flush_variation_001.phpt",
    "tests/output/ob_012.phpt",
    "tests/output/ob_013.phpt",
    "tests/output/ob_007.phpt",
    "tests/output/ob_009.phpt",
    "tests/output/ob_get_contents_error_001.phpt",
    "tests/output/ob_006.phpt",
    "tests/output/ob_end_flush_error_001.phpt",
    "tests/output/ob_017.phpt",
    "tests/output/ob_011.phpt",
    "tests/output/ob_start_callbacks.phpt",
    "tests/output/flush_error_001.phpt",
    "tests/output/ob_implicit_flush_error_001.phpt",
    "tests/output/ob_end_clean_basic_001.phpt",
    "tests/output/ob_start_basic_unerasable_003.phpt",
    "tests/output/ob_get_level_basic_001.phpt",
    "tests/output/ob_get_contents_basic_001.phpt",
    "tests/output/ob_start_error_001.phpt",
    "tests/output/ob_implicit_flush_basic_002.phpt",
    "tests/output/ob_020.phpt",
    "tests/output/ob_get_length_basic_001.phpt",
    "tests/output/ob_start_error_002.phpt",
    "tests/output/ob_clean_basic_001.phpt",
    "tests/output/ob_start_basic_006.phpt",
    "tests/output/ob_002.phpt",
    "tests/output/ob_start_basic_001.phpt",
    "tests/output/ob_flush_basic_001.phpt",
    "tests/output/ob_clean_error_001.phpt",
    "tests/output/ob_get_clean_error_001.phpt",
    "tests/output/ob_start_basic_unerasable_002.phpt",
    "tests/output/ob_start_basic_003.phpt",
    "tests/output/flush_basic_001.phpt",
    "tests/output/ob_start_basic_unerasable_001.phpt",
    "tests/output/ob_end_clean_error_001.phpt",
    "tests/output/ob_005.phpt",
    "tests/output/ob_get_level_error_001.phpt",
    "tests/output/ob_start_error_003.phpt",
    "tests/output/ob_015.phpt",
    "tests/output/ob_flush_error_001.phpt",
    "tests/output/ob_get_clean_basic_001.phpt",
    "tests/output/ob_start_basic_unerasable_005.phpt",
    "tests/output/ob_implicit_flush_basic_001.phpt",
    "tests/output/ob_start_basic_005.phpt",
    "tests/output/ob_010.phpt",
    "tests/output/ob_001.phpt",
    "tests/output/ob_003.phpt",
    "tests/output/ob_start_error_005.phpt",
    "tests/output/ob_get_status.phpt",
    "tests/output/ob_014.phpt",
    "tests/output/ob_start_basic_002.phpt",
    "tests/output/ob_start_basic_unerasable_004.phpt",
    "tests/output/ob_get_clean_basic_002.phpt",
    "tests/output/ob_004.phpt",
    "tests/output/ob_start_error_004.phpt",
    "tests/output/ob_018.phpt",
    "tests/output/ob_start_basic_004.phpt",
    "tests/strings/bug22592.phpt",
    "tests/strings/001.phpt",
    "tests/strings/004.phpt",
    "tests/strings/bug26703.phpt",
    "tests/strings/002.phpt",
    "tests/classes/interfaces_001.phpt",
    "tests/classes/inheritance_005.phpt",
    "tests/classes/final.phpt",
    "tests/classes/iterators_006.phpt",
    "tests/classes/constants_scope_001.phpt",
    "tests/classes/private_006b.phpt",
    "tests/classes/autoload_018.phpt",
    "tests/classes/property_recreate_private.phpt",
    "tests/classes/destructor_and_exceptions.phpt",
    "tests/classes/protected_002.phpt",
    "tests/classes/constants_basic_003.phpt",
    "tests/classes/type_hinting_005d.phpt",
    "tests/classes/factory_and_singleton_004.phpt",
    "tests/classes/final_abstract.phpt",
    "tests/classes/factory_and_singleton_009.phpt",
    "tests/classes/autoload_006.phpt",
    "tests/classes/static_this.phpt",
    "tests/classes/new_001.phpt",
    "tests/classes/ctor_failure.phpt",
    "tests/classes/property_override_protectedStatic_protected.phpt",
    "tests/classes/final_redeclare.phpt",
    "tests/classes/clone_003.phpt",
    "tests/classes/tostring_004.phpt",
    "tests/classes/iterators_003.phpt",
    "tests/classes/interface_and_extends.phpt",
    "tests/classes/visibility_002b.phpt",
    "tests/classes/interface_method_private.phpt",
    "tests/classes/property_override_protectedStatic_privateStatic.phpt",
    "tests/classes/visibility_000a.phpt",
    "tests/classes/property_override_private_privateStatic.phpt",
    "tests/classes/interface_instantiate.phpt",
    "tests/classes/inheritance.phpt",
    "tests/classes/factory_and_singleton_005.phpt",
    "tests/classes/factory_001.phpt",
    "tests/classes/iterators_005.phpt",
    "tests/classes/static_properties_undeclared_assign.phpt",
    "tests/classes/class_stdclass.phpt",
    "tests/classes/__set__get_001.phpt",
    "tests/classes/abstract_derived.phpt",
    "tests/classes/property_override_privateStatic_publicStatic.phpt",
    "tests/classes/property_override_private_public.phpt",
    "tests/classes/destructor_inheritance.phpt",
    "tests/classes/abstract_not_declared.phpt",
    "tests/classes/autoload_016.phpt",
    "tests/classes/__set_data_corrupt.phpt",
    "tests/classes/destructor_and_globals.phpt",
    "tests/classes/visibility_001b.phpt",
    "tests/classes/array_access_001.phpt",
    "tests/classes/static_properties_003_error3.phpt",
    "tests/classes/autoload_007.phpt",
    "tests/classes/private_003.phpt",
    "tests/classes/final_ctor2.phpt",
    "tests/classes/property_override_protected_private.phpt",
    "tests/classes/interface_must_be_implemented.phpt",
    "tests/classes/static_properties_003_error2.phpt",
    "tests/classes/private_004b.phpt",
    "tests/classes/method_override_optional_arg_001.phpt",
    "tests/classes/__call_003.phpt",
    "tests/classes/constants_basic_004.phpt",
    "tests/classes/factory_and_singleton_007.phpt",
    "tests/classes/property_override_private_publicStatic.phpt",
    "tests/classes/property_recreate_protected.phpt",
    "tests/classes/iterators_007.phpt",
    "tests/classes/bug29446.phpt",
    "tests/classes/array_access_012.phpt",
    "tests/classes/private_members.phpt",
    "tests/classes/interface_method_final.phpt",
    "tests/classes/autoload_004.phpt",
    "tests/classes/autoload_014.phpt",
    "tests/classes/private_006.phpt",
    "tests/classes/type_hinting_002.phpt",
    "tests/classes/interfaces_002.phpt",
    "tests/classes/property_override_protected_privateStatic.phpt",
    "tests/classes/private_003b.phpt",
    "tests/classes/tostring_001.phpt",
    "tests/classes/factory_and_singleton_003.phpt",
    "tests/classes/ctor_dtor.phpt",
    "tests/classes/interface_optional_arg.phpt",
    "tests/classes/static_properties_undeclared_assignRef.phpt",
    "tests/classes/clone_006.phpt",
    "tests/classes/static_properties_003_error1.phpt",
    "tests/classes/interface_optional_arg_002.phpt",
    "tests/classes/interface_member.phpt",
    "tests/classes/property_override_public_privateStatic.phpt",
    "tests/classes/type_hinting_005c.phpt",
    "tests/classes/visibility_004a.phpt",
    "tests/classes/class_abstract.phpt",
    "tests/classes/visibility_004b.phpt",
    "tests/classes/type_hinting_005a.phpt",
    "tests/classes/protected_001.phpt",
    "tests/classes/private_005.phpt",
    "tests/classes/static_properties_undeclared_inc.phpt",
    "tests/classes/__call_006.phpt",
    "tests/classes/iterators_004.phpt",
    "tests/classes/visibility_000c.phpt",
    "tests/classes/this.phpt",
    "tests/classes/inheritance_004.phpt",
    "tests/classes/incdec_property_003.phpt",
    "tests/classes/interface_optional_arg_003.phpt",
    "tests/classes/array_access_008.phpt",
    "tests/classes/private_002.phpt",
    "tests/classes/autoload_008.phpt",
    "tests/classes/visibility_002a.phpt",
    "tests/classes/autoload_002.phpt",
    "tests/classes/interface_constant_inheritance_001.phpt",
    "tests/classes/factory_and_singleton_010.phpt",
    "tests/classes/visibility_003c.phpt",
    "tests/classes/array_access_011.phpt",
    "tests/classes/bug23951.phpt",
    "tests/classes/interface_constant_inheritance_002.phpt",
    "tests/classes/property_override_privateStatic_protected.phpt",
    "tests/classes/type_hinting_001.phpt",
    "tests/classes/destructor_and_echo.phpt",
    "tests/classes/implicit_instantiation_001.phpt",
    "tests/classes/private_004.phpt",
    "tests/classes/clone_004.phpt",
    "tests/classes/final_ctor1.phpt",
    "tests/classes/__set__get_005.phpt",
    "tests/classes/visibility_001a.phpt",
    "tests/classes/property_override_privateStatic_privateStatic.phpt",
    "tests/classes/bug27504.phpt",
    "tests/classes/ctor_in_interface_01.phpt",
    "tests/classes/clone_005.phpt",
    "tests/classes/interface_doubled.phpt",
    "tests/classes/bug24399.phpt",
    "tests/classes/property_override_publicStatic_privateStatic.phpt",
    "tests/classes/interface_constant_inheritance_003.phpt",
    "tests/classes/autoload_013.phpt",
    "tests/classes/property_override_public_public.phpt",
    "tests/classes/type_hinting_003.phpt",
    "tests/classes/abstract_final.phpt",
    "tests/classes/ctor_name_clash.phpt",
    "tests/classes/factory_and_singleton_002.phpt",
    "tests/classes/interface_implemented.phpt",
    "tests/classes/array_access_010.phpt",
    "tests/classes/autoload_017.phpt",
    "tests/classes/autoload_010.phpt",
    "tests/classes/autoload_011.phpt",
    "tests/classes/interfaces_003.phpt",
    "tests/classes/inheritance_003.phpt",
    "tests/classes/interface_constant_inheritance_004.phpt",
    "tests/classes/__set__get_002.phpt",
    "tests/classes/array_access_009.phpt",
    "tests/classes/abstract_by_interface_002.phpt",
    "tests/classes/class_example.phpt",
    "tests/classes/bug27468.phpt",
    "tests/classes/property_override_publicStatic_public.phpt",
    "tests/classes/property_override_protectedStatic_private.phpt",
    "tests/classes/inheritance_006.phpt",
    "tests/classes/abstract_inherit.phpt",
    "tests/classes/dereferencing_001.phpt",
    "tests/classes/destructor_and_references.phpt",
    "tests/classes/final_ctor3.phpt",
    "tests/classes/autoload_001.phpt",
    "tests/classes/autoload_003.phpt",
    "tests/classes/tostring_002.phpt",
    "tests/classes/arrayobject_001.phpt",
    "tests/classes/destructor_visibility_001.phpt",
    "tests/classes/visibility_003a.phpt",
    "tests/classes/__call_002.phpt",
    "tests/classes/property_override_public_publicStatic.phpt",
    "tests/classes/property_override_protected_public.phpt",
    "tests/classes/visibility_002c.phpt",
    "tests/classes/abstract_class.phpt",
    "tests/classes/ctor_visibility.phpt",
    "tests/classes/array_access_013.phpt",
    "tests/classes/array_access_002.phpt",
    "tests/classes/property_override_publicStatic_publicStatic.phpt",
    "tests/classes/incdec_property_004.phpt",
    "tests/classes/array_access_005.phpt",
    "tests/classes/__call_007.phpt",
    "tests/classes/inheritance_007.phpt",
    "tests/classes/constants_error_006.phpt",
    "tests/classes/clone_002.phpt",
    "tests/classes/autoload_015.phpt",
    "tests/classes/ctor_in_interface_02.phpt",
    "tests/classes/property_override_public_protected.phpt",
    "tests/classes/interface_class.phpt",
    "tests/classes/static_properties_undeclared_isset.phpt",
    "tests/classes/method_override_optional_arg_002.phpt",
    "tests/classes/factory_and_singleton_008.phpt",
    "tests/classes/property_override_privateStatic_protectedStatic.phpt",
    "tests/classes/abstract_redeclare.phpt",
    "tests/classes/array_access_004.phpt",
    "tests/classes/visibility_003b.phpt",
    "tests/classes/factory_and_singleton_001.phpt",
    "tests/classes/abstract_user_call.phpt",
    "tests/classes/constants_error_004.phpt",
    "tests/classes/array_access_006.phpt",
    "tests/classes/inheritance_002.phpt",
    "tests/classes/constants_error_005.phpt",
    "tests/classes/property_override_protectedStatic_publicStatic.phpt",
    "tests/classes/static_properties_004.phpt",
    "tests/classes/__call_004.phpt",
    "tests/classes/constants_basic_006.phpt",
    "tests/classes/static_properties_003_error4.phpt",
    "tests/classes/property_override_publicStatic_private.phpt",
    "tests/classes/factory_and_singleton_006.phpt",
    "tests/classes/property_override_publicStatic_protectedStatic.phpt",
    "tests/classes/property_override_protected_protectedStatic.phpt",
    "tests/classes/iterators_002.phpt",
    "tests/classes/bug24445.phpt",
    "tests/classes/class_final.phpt",
    "tests/classes/iterators_001.phpt",
    "tests/classes/static_mix_2.phpt",
    "tests/classes/property_override_private_protected.phpt",
    "tests/classes/autoload_009.phpt",
    "tests/classes/private_007.phpt",
    "tests/classes/object_reference_001.phpt",
    "tests/classes/property_override_private_protectedStatic.phpt",
    "tests/classes/autoload_005.phpt",
    "tests/classes/property_override_protectedStatic_protectedStatic.phpt",
    "tests/classes/__call_005.phpt",
    "tests/classes/constants_error_007.phpt",
    "tests/classes/property_override_public_private.phpt",
    "tests/classes/constants_basic_005.phpt",
    "tests/classes/__set__get_003.phpt",
    "tests/classes/protected_001b.phpt",
    "tests/classes/constants_error_003.phpt",
    "tests/classes/autoload_019.phpt",
    "tests/classes/static_mix_1.phpt",
    "tests/classes/destructor_visibility_003.phpt",
    "tests/classes/method_call_variation_001.phpt",
    "tests/classes/static_properties_undeclared_read.phpt",
    "tests/classes/static_properties_001.phpt",
    "tests/classes/visibility_005.phpt",
    "tests/classes/autoload_020.phpt",
    "tests/classes/array_access_007.phpt",
    "tests/classes/property_override_protected_protected.phpt",
    "tests/classes/static_properties_undeclared_assignInc.phpt",
    "tests/classes/destructor_visibility_002.phpt",
    "tests/classes/ctor_in_interface_03.phpt",
    "tests/classes/property_override_privateStatic_private.phpt",
    "tests/classes/abstract_by_interface_001.phpt",
    "tests/classes/ctor_in_interface_04.phpt",
    "tests/classes/ctor_dtor_inheritance.phpt",
    "tests/classes/serialize_001.phpt",
    "tests/classes/tostring_003.phpt",
    "tests/classes/private_007b.phpt",
    "tests/classes/property_override_protectedStatic_public.phpt",
    "tests/classes/__set__get_004.phpt",
    "tests/classes/property_override_private_private.phpt",
    "tests/classes/visibility_000b.phpt",
    "tests/classes/incdec_property_001.phpt",
    "tests/classes/abstract_static.phpt",
    "tests/classes/property_override_privateStatic_public.phpt",
    "tests/classes/private_005b.phpt",
    "tests/classes/type_hinting_004.phpt",
    "tests/classes/visibility_004c.phpt",
    "tests/classes/abstract.phpt",
    "tests/classes/singleton_001.phpt",
    "tests/classes/constants_basic_001.phpt",
    "tests/classes/bug26737.phpt",
    "tests/classes/iterators_008.phpt",
    "tests/classes/constants_basic_002.phpt",
    "tests/classes/autoload_012.phpt",
    "tests/classes/property_override_publicStatic_protected.phpt",
    "tests/classes/property_override_protected_publicStatic.phpt",
    "tests/classes/visibility_001c.phpt",
    "tests/classes/assign_op_property_001.phpt",
    "tests/classes/static_properties_003.phpt",
    "tests/classes/type_hinting_005b.phpt",
    "tests/classes/private_001.phpt",
    "tests/classes/property_override_public_protectedStatic.phpt",
    "tests/classes/__call_001.phpt",
    "tests/classes/clone_001.phpt",
    "tests/classes/constants_error_002.phpt",
    "tests/classes/interface_method.phpt",
    "tests/classes/constants_error_001.phpt",
    "tests/classes/incdec_property_002.phpt",
    "tests/classes/array_access_003.phpt",
    "tests/classes/private_redeclare.phpt",
    "tests/lang/027.phpt",
    "tests/lang/passByReference_005.phpt",
    "tests/lang/returnByReference.007.phpt",
    "tests/lang/foreachLoopObjects.001.phpt",
    "tests/lang/engine_assignExecutionOrder_007.phpt",
    "tests/lang/bug35382.phpt",
    "tests/lang/bug7515.phpt",
    "tests/lang/bug30862.phpt",
    "tests/lang/033.phpt",
    "tests/lang/foreachLoop.003.phpt",
    "tests/lang/010.phpt",
    "tests/lang/foreachLoop.005.phpt",
    "tests/lang/engine_assignExecutionOrder_008.phpt",
    "tests/lang/engine_assignExecutionOrder_003.phpt",
    "tests/lang/foreachLoop.008.phpt",
    "tests/lang/005.phpt",
    "tests/lang/bug21669.phpt",
    "tests/lang/returnByReference.002.phpt",
    "tests/lang/foreachLoop.009.phpt",
    "tests/lang/bug24396.phpt",
    "tests/lang/014.phpt",
    "tests/lang/044.phpt",
    "tests/lang/bug22592.phpt",
    "tests/lang/func_get_args.002.phpt",
    "tests/lang/func_get_arg.002.phpt",
    "tests/lang/bug17115.phpt",
    "tests/lang/error_2_exception_001.phpt",
    "tests/lang/func_get_args.004.phpt",
    "tests/lang/type_hints_002.phpt",
    "tests/lang/bug28213.phpt",
    "tests/lang/foreachLoopIteratorAggregate.002.phpt",
    "tests/lang/011.phpt",
    "tests/lang/script_tag.phpt",
    "tests/lang/bug30578.phpt",
    "tests/lang/017.phpt",
    "tests/lang/passByReference_007.phpt",
    "tests/lang/foreachLoop.011.phpt",
    "tests/lang/bug24403.phpt",
    "tests/lang/036.phpt",
    "tests/lang/func_num_args.004.phpt",
    "tests/lang/func_num_args.001.phpt",
    "tests/lang/bug24652.phpt",
    "tests/lang/bug24054.phpt",
    "tests/lang/foreachLoopIteratorAggregate.003.phpt",
    "tests/lang/func_get_arg.003.phpt",
    "tests/lang/bug28800.phpt",
    "tests/lang/engine_assignExecutionOrder_006.phpt",
    "tests/lang/returnByReference.004.phpt",
    "tests/lang/bug26696.phpt",
    "tests/lang/043.phpt",
    "tests/lang/bug25652.phpt",
    "tests/lang/028.phpt",
    "tests/lang/foreachLoopIterator.001.phpt",
    "tests/lang/001.phpt",
    "tests/lang/041.phpt",
    "tests/lang/015.phpt",
    "tests/lang/bug22690.phpt",
    "tests/lang/026.phpt",
    "tests/lang/foreachLoopObjects.005.phpt",
    "tests/lang/004.phpt",
    "tests/lang/returnByReference.003.phpt",
    "tests/lang/bug24499.phpt",
    "tests/lang/passByReference_012.phpt",
    "tests/lang/static_variation_002.phpt",
    "tests/lang/short_tags.002.phpt",
    "tests/lang/engine_assignExecutionOrder_002.phpt",
    "tests/lang/bug26869.phpt",
    "tests/lang/bug23922.phpt",
    "tests/lang/bug25922.phpt",
    "tests/lang/short_tags.003.phpt",
    "tests/lang/include_variation3.phpt",
    "tests/lang/bug26182.phpt",
    "tests/lang/bug18872.phpt",
    "tests/lang/short_tags.004.phpt",
    "tests/lang/bug26866.phpt",
    "tests/lang/passByReference_002.phpt",
    "tests/lang/string_decimals_001.phpt",
    "tests/lang/operators/operator_equals_variation_64bit.phpt",
    "tests/lang/operators/operator_lt_or_equal_variation.phpt",
    "tests/lang/operators/subtract_basiclong_64bit.phpt",
    "tests/lang/operators/bitwiseOr_variationStr.phpt",
    "tests/lang/operators/bitwiseShiftLeft_variationStr.phpt",
    "tests/lang/operators/bitwiseShiftRight_variationStr.phpt",
    "tests/lang/operators/operator_gt_variation.phpt",
    "tests/lang/operators/add_variationStr.phpt",
    "tests/lang/operators/operator_equals_basic.phpt",
    "tests/lang/operators/operator_lt_or_equal_variation_64bit.phpt",
    "tests/lang/operators/operator_lt_variation.phpt",
    "tests/lang/operators/predec_basiclong_64bit.phpt",
    "tests/lang/operators/negate_basiclong_64bit.phpt",
    "tests/lang/operators/bitwiseAnd_variationStr.phpt",
    "tests/lang/operators/negate_variationStr.phpt",
    "tests/lang/operators/operator_gt_variation_64bit.phpt",
    "tests/lang/operators/modulus_variationStr.phpt",
    "tests/lang/operators/predec_variationStr.phpt",
    "tests/lang/operators/preinc_basiclong_64bit.phpt",
    "tests/lang/operators/operator_lt_or_equal_basic.phpt",
    "tests/lang/operators/operator_equals_variation.phpt",
    "tests/lang/operators/bitwiseXor_variationStr.phpt",
    "tests/lang/operators/operator_gt_or_equal_variation_64bit.phpt",
    "tests/lang/operators/operator_notequals_variation_64bit.phpt",
    "tests/lang/operators/operator_gt_or_equal_variation.phpt",
    "tests/lang/operators/postinc_variationStr.phpt",
    "tests/lang/operators/operator_identical_variation_64bit.phpt",
    "tests/lang/operators/operator_gt_basic.phpt",
    "tests/lang/operators/preinc_variationStr.phpt",
    "tests/lang/operators/operator_notidentical_variation_64bit.phpt",
    "tests/lang/operators/bitwiseNot_basiclong_64bit.phpt",
    "tests/lang/operators/operator_notequals_variation.phpt",
    "tests/lang/operators/multiply_variationStr.phpt",
    "tests/lang/operators/operator_notidentical_basic.phpt",
    "tests/lang/operators/add_basiclong_64bit.phpt",
    "tests/lang/operators/bitwiseShiftLeft_basiclong_64bit.phpt",
    "tests/lang/operators/bitwiseShiftLeft_variationStr_64bit.phpt",
    "tests/lang/operators/bitwiseOr_basiclong_64bit.phpt",
    "tests/lang/operators/bitwiseXor_basiclong_64bit.phpt",
    "tests/lang/operators/bitwiseNot_variationStr.phpt",
    "tests/lang/operators/operator_lt_variation_64bit.phpt",
    "tests/lang/operators/operator_notidentical_variation.phpt",
    "tests/lang/operators/operator_notequals_basic.phpt",
    "tests/lang/operators/divide_basiclong_64bit.phpt",
    "tests/lang/operators/postinc_basiclong_64bit.phpt",
    "tests/lang/operators/divide_variationStr.phpt",
    "tests/lang/operators/operator_lt_basic.phpt",
    "tests/lang/operators/operator_gt_or_equal_basic.phpt",
    "tests/lang/operators/postdec_basiclong_64bit.phpt",
    "tests/lang/operators/multiply_basiclong_64bit.phpt",
    "tests/lang/operators/operator_identical_basic.phpt",
    "tests/lang/operators/bitwiseShiftRight_basiclong_64bit.phpt",
    "tests/lang/operators/modulus_basiclong_64bit.phpt",
    "tests/lang/operators/bitwiseAnd_basiclong_64bit.phpt",
    "tests/lang/operators/postdec_variationStr.phpt",
    "tests/lang/operators/operator_identical_variation.phpt",
    "tests/lang/operators/subtract_variationStr.phpt",
    "tests/lang/comments.phpt",
    "tests/lang/compare_objects_basic2.phpt",
    "tests/lang/engine_assignExecutionOrder_001.phpt",
    "tests/lang/003.phpt",
    "tests/lang/bug45392.phpt",
    "tests/lang/engine_assignExecutionOrder_005.phpt",
    "tests/lang/bug32828.phpt",
    "tests/lang/foreachLoop.014.phpt",
    "tests/lang/passByReference_006.phpt",
    "tests/lang/foreachLoopObjects.002.phpt",
    "tests/lang/func_get_args.003.phpt",
    "tests/lang/throw_variation_001.phpt",
    "tests/lang/bug43958.phpt",
    "tests/lang/bug24951.phpt",
    "tests/lang/bug32924.phpt",
    "tests/lang/compare_objects_basic1.phpt",
    "tests/lang/returnByReference.006.phpt",
    "tests/lang/020.phpt",
    "tests/lang/type_hints_001.phpt",
    "tests/lang/022.phpt",
    "tests/lang/006.phpt",
    "tests/lang/031.phpt",
    "tests/lang/bug23624.phpt",
    "tests/lang/bug27535.phpt",
    "tests/lang/bug30726.phpt",
    "tests/lang/comments2.phpt",
    "tests/lang/023.phpt",
    "tests/lang/passByReference_004.phpt",
    "tests/lang/foreachLoop.006.phpt",
    "tests/lang/this_assignment.phpt",
    "tests/lang/static_basic_001.phpt",
    "tests/lang/func_num_args.002.phpt",
    "tests/lang/returnByReference.001.phpt",
    "tests/lang/bug24573.phpt",
    "tests/lang/catchable_error_002.phpt",
    "tests/lang/foreachLoopObjects.004.phpt",
    "tests/lang/foreachLoop.004.phpt",
    "tests/lang/bug22231.phpt",
    "tests/lang/each_binary_safety.phpt",
    "tests/lang/bug24658.phpt",
    "tests/lang/007.phpt",
    "tests/lang/returnByReference.008.phpt",
    "tests/lang/bug19943.phpt",
    "tests/lang/bug29944.phpt",
    "tests/lang/bug21820.phpt",
    "tests/lang/bug23584.phpt",
    "tests/lang/bug29566.phpt",
    "tests/lang/bug25547.phpt",
    "tests/lang/type_hints_003.phpt",
    "tests/lang/func_get_args.001.phpt",
    "tests/lang/002.phpt",
    "tests/lang/empty_variation.phpt",
    "tests/lang/bug23489.phpt",
    "tests/lang/008.phpt",
    "tests/lang/passByReference_008.phpt",
    "tests/lang/018.phpt",
    "tests/lang/040.phpt",
    "tests/lang/passByReference_009.phpt",
    "tests/lang/catchable_error_001.phpt",
    "tests/lang/bug25145.phpt",
    "tests/lang/returnByReference.009.phpt",
    "tests/lang/foreachLoop.016.phpt",
    "tests/lang/foreachLoop.010.phpt",
    "tests/lang/034.phpt",
    "tests/lang/039.phpt",
    "tests/lang/035.phpt",
    "tests/lang/foreachLoopObjects.006.phpt",
    "tests/lang/bug20175.phpt",
    "tests/lang/013.phpt",
    "tests/lang/037.phpt",
    "tests/lang/016.phpt",
    "tests/lang/engine_assignExecutionOrder_004.phpt",
    "tests/lang/static_basic_002.phpt",
    "tests/lang/bug19566.phpt",
    "tests/lang/bug21849.phpt",
    "tests/lang/bison1.phpt",
    "tests/lang/func_get_arg.004.phpt",
    "tests/lang/025.phpt",
    "tests/lang/func_get_arg_variation.phpt",
    "tests/lang/bug44654.phpt",
    "tests/lang/bug21961.phpt",
    "tests/lang/foreachLoopIterator.002.phpt",
    "tests/lang/foreachLoopObjects.003.phpt",
    "tests/lang/foreachLoop.013.phpt",
    "tests/lang/012.phpt",
    "tests/lang/038.phpt",
    "tests/lang/bug23524.phpt",
    "tests/lang/bug21600.phpt",
    "tests/lang/bug35176.phpt",
    "tests/lang/foreachLoopIteratorAggregate.001.phpt",
    "tests/lang/021.phpt",
    "tests/lang/bug27443.phpt",
    "tests/lang/returnByReference.005.phpt",
    "tests/lang/passByReference_001.phpt",
    "tests/lang/include_variation2.phpt",
    "tests/lang/foreachLoopIteratorAggregate.004.phpt",
    "tests/lang/bug24640.phpt",
    "tests/lang/bug24783.phpt",
    "tests/lang/foreach_with_object_001.phpt",
    "tests/lang/passByReference_010.phpt",
    "tests/lang/bug38579.phpt",
    "tests/lang/foreach_with_references_001.phpt",
    "tests/lang/func_get_arg.005.phpt",
    "tests/lang/bug24908.phpt",
    "tests/lang/030.phpt",
    "tests/lang/foreachLoop.002.phpt",
    "tests/lang/foreachLoop.007.phpt",
    "tests/lang/bug24926.phpt",
    "tests/lang/032.phpt",
    "tests/lang/bug22510.phpt",
    "tests/lang/passByReference_003.phpt",
    "tests/lang/019.phpt",
    "tests/lang/func_get_arg.001.phpt",
    "tests/lang/bug23384.phpt",
    "tests/lang/include_variation1.phpt",
    "tests/lang/bug44827.phpt",
    "tests/lang/bug27354.phpt",
    "tests/lang/foreachLoop.012.phpt",
    "tests/lang/024.phpt",
    "tests/lang/bug23279.phpt",
    "tests/lang/bug29893.phpt",
    "tests/lang/foreachLoop.001.phpt",
    "tests/lang/009.phpt",
    "tests/lang/042.phpt",
    "tests/lang/engine_assignExecutionOrder_009.phpt",
    "tests/lang/foreachLoop.017.phpt",
    "tests/lang/func_num_args.003.phpt",
    "tests/lang/short_tags.001.phpt",
    "tests/lang/foreachLoop.015.phpt",
    "tests/lang/static_variation_001.phpt",
    "tests/lang/bug21094.phpt",
    "tests/lang/bug30638.phpt",
    "tests/lang/execution_order.phpt",
    "tests/lang/bug24436.phpt",
    "tests/lang/bug27439.phpt",
    "tests/run-test/test010.phpt",
    "tests/run-test/test009.phpt",
    "tests/run-test/test004.phpt",
    "tests/run-test/test007.phpt",
    "tests/run-test/test008a.phpt",
    "tests/run-test/test001.phpt",
    "tests/run-test/test005.phpt",
    "tests/run-test/test002.phpt",
    "tests/run-test/test008.phpt",
    "tests/run-test/test006.phpt",
    "tests/run-test/test003.phpt",
    "Zend/tests/027.phpt",
    "Zend/tests/catch_004.phpt",
    "Zend/tests/ns_064.phpt",
    "Zend/tests/bug54265.phpt",
    "Zend/tests/instanceof_001.phpt",
    "Zend/tests/bug54804.phpt",
    "Zend/tests/nowdoc_016.phpt",
    "Zend/tests/gc_012.phpt",
    "Zend/tests/bug31683.phpt",
    "Zend/tests/bug43344_2.phpt",
    "Zend/tests/ns_044.phpt",
    "Zend/tests/dereference_009.phpt",
    "Zend/tests/closure_025.phpt",
    "Zend/tests/closure_015.phpt",
    "Zend/tests/bug32290.phpt",
    "Zend/tests/anonymous_func_001.phpt",
    "Zend/tests/bug42817.phpt",
    "Zend/tests/exception_handler_004.phpt",
    "Zend/tests/gc_016.phpt",
    "Zend/tests/bug53347.phpt",
    "Zend/tests/objects_030.phpt",
    "Zend/tests/bug36214.phpt",
    "Zend/tests/bug39445.phpt",
    "Zend/tests/bug26802.phpt",
    "Zend/tests/bug48770_2.phpt",
    "Zend/tests/isset_001.phpt",
    "Zend/tests/objects_015.phpt",
    "Zend/tests/033.phpt",
    "Zend/tests/error_reporting08.phpt",
    "Zend/tests/errmsg_032.phpt",
    "Zend/tests/bug38234.phpt",
    "Zend/tests/bug39602.phpt",
    "Zend/tests/bug34199.phpt",
    "Zend/tests/assign_to_var_002.phpt",
    "Zend/tests/bug44184.phpt",
    "Zend/tests/bug32252.phpt",
    "Zend/tests/bug53748.phpt",
    "Zend/tests/call_static.phpt",
    "Zend/tests/bug33116.phpt",
    "Zend/tests/heredoc_002.phpt",
    "Zend/tests/gc_010.phpt",
    "Zend/tests/ns_077_3.phpt",
    "Zend/tests/closure_033.phpt",
    "Zend/tests/unset_cv08.phpt",
    "Zend/tests/ns_082.phpt",
    "Zend/tests/010.phpt",
    "Zend/tests/mul_001.phpt",
    "Zend/tests/lsb_019.phpt",
    "Zend/tests/objects_001.phpt",
    "Zend/tests/bug39017.phpt",
    "Zend/tests/add_003.phpt",
    "Zend/tests/gc_011.phpt",
    "Zend/tests/closure_004.phpt",
    "Zend/tests/bug30889.phpt",
    "Zend/tests/closure_011.phpt",
    "Zend/tests/bug54358.phpt",
    "Zend/tests/bug44141.phpt",
    "Zend/tests/bug40236.phpt",
    "Zend/tests/ns_011.phpt",
    "Zend/tests/bug52237.phpt",
    "Zend/tests/bug51822.phpt",
    "Zend/tests/bug26229.phpt",
    "Zend/tests/anonymous_func_002.phpt",
    "Zend/tests/bug33710.phpt",
    "Zend/tests/bug34358.phpt",
    "Zend/tests/ns_039.phpt",
    "Zend/tests/nowdoc_013.phpt",
    "Zend/tests/ns_065.phpt",
    "Zend/tests/access_modifiers_004.phpt",
    "Zend/tests/foreach_undefined.phpt",
    "Zend/tests/bug41026.phpt",
    "Zend/tests/bug44899_2.phpt",
    "Zend/tests/jump05.phpt",
    "Zend/tests/errmsg_044.phpt",
    "Zend/tests/gc_008.phpt",
    "Zend/tests/errmsg_041.phpt",
    "Zend/tests/005.phpt",
    "Zend/tests/dynamic_call_004.phpt",
    "Zend/tests/bug49893.phpt",
    "Zend/tests/ns_026.phpt",
    "Zend/tests/globals_001.phpt",
    "Zend/tests/foreach.phpt",
    "Zend/tests/trait_exists_003.phpt",
    "Zend/tests/class_alias_006.phpt",
    "Zend/tests/bug54367.phpt",
    "Zend/tests/bug27669.phpt",
    "Zend/tests/bug36037.phpt",
    "Zend/tests/bug32833.phpt",
    "Zend/tests/class_exists_001.phpt",
    "Zend/tests/class_alias_010.phpt",
    "Zend/tests/bug43344_12.phpt",
    "Zend/tests/get_parent_class_001.phpt",
    "Zend/tests/bug54013.phpt",
    "Zend/tests/bug45805.phpt",
    "Zend/tests/heredoc_011.phpt",
    "Zend/tests/clone_003.phpt",
    "Zend/tests/bug40757.phpt",
    "Zend/tests/bug32993.phpt",
    "Zend/tests/014.phpt",
    "Zend/tests/bug35239.phpt",
    "Zend/tests/list_006.phpt",
    "Zend/tests/objects_026.phpt",
    "Zend/tests/bug40899.phpt",
    "Zend/tests/bug41209.phpt",
    "Zend/tests/bug50394.phpt",
    "Zend/tests/bug51176.phpt",
    "Zend/tests/bug40833.phpt",
    "Zend/tests/029.phpt",
    "Zend/tests/bug47699.phpt",
    "Zend/tests/bug52484.phpt",
    "Zend/tests/globals_003.phpt",
    "Zend/tests/bug52160.phpt",
    "Zend/tests/settype_string.phpt",
    "Zend/tests/bug46246.phpt",
    "Zend/tests/exception_006.phpt",
    "Zend/tests/ns_047.phpt",
    "Zend/tests/bug43344_7.phpt",
    "Zend/tests/bug35437.phpt",
    "Zend/tests/bug20242.phpt",
    "Zend/tests/jump11.phpt",
    "Zend/tests/isset_003_2_4.phpt",
    "Zend/tests/bug38623.phpt",
    "Zend/tests/errmsg_018.phpt",
    "Zend/tests/bug30451.phpt",
    "Zend/tests/bug41813.phpt",
    "Zend/tests/unset_cv09.phpt",
    "Zend/tests/objects_019.phpt",
    "Zend/tests/heredoc_009.phpt",
    "Zend/tests/bug41421.phpt",
    "Zend/tests/exception_handler_006.phpt",
    "Zend/tests/list_001.phpt",
    "Zend/tests/bug44913.phpt",
    "Zend/tests/errmsg_023.phpt",
    "Zend/tests/bug43323.phpt",
    "Zend/tests/011.phpt",
    "Zend/tests/call_user_func_002.phpt",
    "Zend/tests/bug43343.phpt",
    "Zend/tests/bug36303.phpt",
    "Zend/tests/bug31102.phpt",
    "Zend/tests/heredoc_014.phpt",
    "Zend/tests/bug53511.phpt",
    "Zend/tests/heredoc_010.phpt",
    "Zend/tests/selfParent_002.phpt",
    "Zend/tests/bug29210.phpt",
    "Zend/tests/jump06.phpt",
    "Zend/tests/bug26698.phpt",
    "Zend/tests/objects_021.phpt",
    "Zend/tests/access_modifiers_001.phpt",
    "Zend/tests/bug32660.phpt",
    "Zend/tests/bug43344_13.phpt",
    "Zend/tests/ns_084.phpt",
    "Zend/tests/get_defined_vars.phpt",
    "Zend/tests/bug38942.phpt",
    "Zend/tests/bug49866.phpt",
    "Zend/tests/property_exists.phpt",
    "Zend/tests/get_class_vars_006.phpt",
    "Zend/tests/list_003.phpt",
    "Zend/tests/bug29883.phpt",
    "Zend/tests/bug48004.phpt",
    "Zend/tests/class_alias_012.phpt",
    "Zend/tests/ns_045.phpt",
    "Zend/tests/bug30346.phpt",
    "Zend/tests/bug20240.phpt",
    "Zend/tests/get_required_files.phpt",
    "Zend/tests/gc_003.phpt",
    "Zend/tests/017.phpt",
    "Zend/tests/heredoc_015.phpt",
    "Zend/tests/settype_bool.phpt",
    "Zend/tests/dereference_012.phpt",
    "Zend/tests/bug39542.phpt",
    "Zend/tests/bug43918.phpt",
    "Zend/tests/halt01.phpt",
    "Zend/tests/gc_032.phpt",
    "Zend/tests/objects_007.phpt",
    "Zend/tests/lsb_014.phpt",
    "Zend/tests/closure_026.phpt",
    "Zend/tests/bug30828.phpt",
    "Zend/tests/bug43332_1.phpt",
    "Zend/tests/offset_object.phpt",
    "Zend/tests/bug24635.phpt",
    "Zend/tests/decrement_001.phpt",
    "Zend/tests/bug36568.phpt",
    "Zend/tests/objects_008.phpt",
    "Zend/tests/bug54039.phpt",
    "Zend/tests/exception_001.phpt",
    "Zend/tests/036.phpt",
    "Zend/tests/each_001.phpt",
    "Zend/tests/class_alias_007.phpt",
    "Zend/tests/bug28072.phpt",
    "Zend/tests/class_alias_017.phpt",
    "Zend/tests/bug47880.phpt",
    "Zend/tests/ns_067.phpt",
    "Zend/tests/decrement_001_64bit.phpt",
    "Zend/tests/bug33999.phpt",
    "Zend/tests/error_reporting04.phpt",
    "Zend/tests/offset_long.phpt",
    "Zend/tests/class_exists_002.phpt",
    "Zend/tests/heredoc_007.phpt",
    "Zend/tests/function_exists_error.phpt",
    "Zend/tests/bug48215_2.phpt",
    "Zend/tests/ns_002.phpt",
    "Zend/tests/bug46241.phpt",
    "Zend/tests/bug41351.phpt",
    "Zend/tests/increment_001.phpt",
    "Zend/tests/objects_011.phpt",
    "Zend/tests/objects_006.phpt",
    "Zend/tests/cast_to_int.phpt",
    "Zend/tests/heredoc_013.phpt",
    "Zend/tests/offset_bool.phpt",
    "Zend/tests/constants_003.phpt",
    "Zend/tests/cast_to_string.phpt",
    "Zend/tests/dereference_007.phpt",
    "Zend/tests/call_static_007.phpt",
    "Zend/tests/constants_007.phpt",
    "Zend/tests/lsb_020.phpt",
    "Zend/tests/lsb_001.phpt",
    "Zend/tests/closure_007.phpt",
    "Zend/tests/function_exists_basic.phpt",
    "Zend/tests/objects_024.phpt",
    "Zend/tests/dynamic_call_002.phpt",
    "Zend/tests/ns_009.phpt",
    "Zend/tests/ns_070.phpt",
    "Zend/tests/div_001.phpt",
    "Zend/tests/ns_081.phpt",
    "Zend/tests/anonymous_func_003.phpt",
    "Zend/tests/bug32429.phpt",
    "Zend/tests/bug29104.phpt",
    "Zend/tests/nowdoc_008.phpt",
    "Zend/tests/class_alias_001.phpt",
    "Zend/tests/ns_017.phpt",
    "Zend/tests/bug52940.phpt",
    "Zend/tests/ns_074.phpt",
    "Zend/tests/gc_020.phpt",
    "Zend/tests/class_constants_002.phpt",
    "Zend/tests/ns_006.phpt",
    "Zend/tests/bug26696.phpt",
    "Zend/tests/bug37138.phpt",
    "Zend/tests/bug31525.phpt",
    "Zend/tests/get_class_vars_004.phpt",
    "Zend/tests/bug43483.phpt",
    "Zend/tests/ns_003.phpt",
    "Zend/tests/bug47353.phpt",
    "Zend/tests/bug43344_8.phpt",
    "Zend/tests/bug53958.phpt",
    "Zend/tests/bug52508.phpt",
    "Zend/tests/error_reporting09.phpt",
    "Zend/tests/bug33732.phpt",
    "Zend/tests/bug34065.phpt",
    "Zend/tests/028.phpt",
    "Zend/tests/bug30162.phpt",
    "Zend/tests/bug42143.phpt",
    "Zend/tests/magic_by_ref_005.phpt",
    "Zend/tests/class_exists_003.phpt",
    "Zend/tests/bug45877.phpt",
    "Zend/tests/globals_004.phpt",
    "Zend/tests/001.phpt",
    "Zend/tests/error_reporting10.phpt",
    "Zend/tests/errmsg_040.phpt",
    "Zend/tests/interface_exists_001.phpt",
    "Zend/tests/gc_025.phpt",
    "Zend/tests/exception_handler_005.phpt",
    "Zend/tests/dereference_008.phpt",
    "Zend/tests/class_alias_018.phpt",
    "Zend/tests/errmsg_011.phpt",
    "Zend/tests/bug52361.phpt",
    "Zend/tests/gc_030.phpt",
    "Zend/tests/mod_001.phpt",
    "Zend/tests/bug54372.phpt",
    "Zend/tests/lsb_009.phpt",
    "Zend/tests/access_modifiers_005.phpt",
    "Zend/tests/gc_006.phpt",
    "Zend/tests/bug29890.phpt",
    "Zend/tests/gc_031.phpt",
    "Zend/tests/add_004.phpt",
    "Zend/tests/ns_010.phpt",
    "Zend/tests/exception_003.phpt",
    "Zend/tests/bug27798.phpt",
    "Zend/tests/bug27268.phpt",
    "Zend/tests/bug42798.phpt",
    "Zend/tests/015.phpt",
    "Zend/tests/closure_019.phpt",
    "Zend/tests/bug26697.phpt",
    "Zend/tests/dereference_003.phpt",
    "Zend/tests/gc_007.phpt",
    "Zend/tests/unset_cv05.phpt",
    "Zend/tests/bug40509.phpt",
    "Zend/tests/026.phpt",
    "Zend/tests/jump01.phpt",
    "Zend/tests/jump09.phpt",
    "Zend/tests/oct_overflow_32bit.phpt",
    "Zend/tests/dynamic_call_003.phpt",
    "Zend/tests/bug35393.phpt",
    "Zend/tests/bug26077.phpt",
    "Zend/tests/access_modifiers_002.phpt",
    "Zend/tests/closure_005.phpt",
    "Zend/tests/bug39036.phpt",
    "Zend/tests/constants_008.phpt",
    "Zend/tests/004.phpt",
    "Zend/tests/int_underflow_64bit.phpt",
    "Zend/tests/call_static_006.phpt",
    "Zend/tests/errmsg_007.phpt",
    "Zend/tests/call_user_func_001.phpt",
    "Zend/tests/bug43344_1.phpt",
    "Zend/tests/bug39449.phpt",
    "Zend/tests/compare_003_64bit.phpt",
    "Zend/tests/xor_001.phpt",
    "Zend/tests/compare_005_64bit.phpt",
    "Zend/tests/strict_002.phpt",
    "Zend/tests/bug29015.phpt",
    "Zend/tests/bug46106.phpt",
    "Zend/tests/bug43053.phpt",
    "Zend/tests/dereference_010.phpt",
    "Zend/tests/ns_036.phpt",
    "Zend/tests/bug48428.phpt",
    "Zend/tests/add_005.phpt",
    "Zend/tests/object_handlers.phpt",
    "Zend/tests/bug39304.phpt",
    "Zend/tests/settype_null.phpt",
    "Zend/tests/closure_021.phpt",
    "Zend/tests/bug36071.phpt",
    "Zend/tests/bug38772.phpt",
    "Zend/tests/bug43651.phpt",
    "Zend/tests/bug43703.phpt",
    "Zend/tests/closure_003.phpt",
    "Zend/tests/ns_029.phpt",
    "Zend/tests/bug24699.phpt",
    "Zend/tests/offset_string.phpt",
    "Zend/tests/ns_008.phpt",
    "Zend/tests/gc_026.phpt",
    "Zend/tests/ns_028.phpt",
    "Zend/tests/bug41351_2.phpt",
    "Zend/tests/compare_002.phpt",
    "Zend/tests/ns_068.phpt",
    "Zend/tests/ns_040.phpt",
    "Zend/tests/bug30922.phpt",
    "Zend/tests/bug38146.phpt",
    "Zend/tests/bug54268.phpt",
    "Zend/tests/bug37811.phpt",
    "Zend/tests/bug28444.phpt",
    "Zend/tests/bug36513.phpt",
    "Zend/tests/class_alias_011.phpt",
    "Zend/tests/errmsg_043.phpt",
    "Zend/tests/bug43344_6.phpt",
    "Zend/tests/lsb_010.phpt",
    "Zend/tests/bug32596.phpt",
    "Zend/tests/class_alias_014.phpt",
    "Zend/tests/lsb_005.phpt",
    "Zend/tests/closure_024.phpt",
    "Zend/tests/error_reporting06.phpt",
    "Zend/tests/errmsg_028.phpt",
    "Zend/tests/ns_077_7.phpt",
    "Zend/tests/bug50383.phpt",
    "Zend/tests/halt02.phpt",
    "Zend/tests/get_class_vars_001.phpt",
    "Zend/tests/bug21888.phpt",
    "Zend/tests/object-null.phpt",
    "Zend/tests/heredoc_018.phpt",
    "Zend/tests/array_append_COW.phpt",
    "Zend/tests/heredoc_001.phpt",
    "Zend/tests/heredoc_017.phpt",
    "Zend/tests/bug29689.phpt",
    "Zend/tests/heredoc_006.phpt",
    "Zend/tests/gc_014.phpt",
    "Zend/tests/bug30407.phpt",
    "Zend/tests/bug22836.phpt",
    "Zend/tests/bug50146.phpt",
    "Zend/tests/ns_023.phpt",
    "Zend/tests/ns_043.phpt",
    "Zend/tests/bug47516.phpt",
    "Zend/tests/nowdoc_002.phpt",
    "Zend/tests/exception_008.phpt",
    "Zend/tests/declare_002.phpt",
    "Zend/tests/ns_042.phpt",
    "Zend/tests/errmsg_030.phpt",
    "Zend/tests/inter_03.phpt",
    "Zend/tests/isset_002.phpt",
    "Zend/tests/bug41117_1.phpt",
    "Zend/tests/bug41929.phpt",
    "Zend/tests/lsb_012.phpt",
    "Zend/tests/inter_06.phpt",
    "Zend/tests/errmsg_016.phpt",
    "Zend/tests/closure_032.phpt",
    "Zend/tests/bug40809.phpt",
    "Zend/tests/bug52939.phpt",
    "Zend/tests/exception_handler_002.phpt",
    "Zend/tests/bug46238.phpt",
    "Zend/tests/bug32296.phpt",
    "Zend/tests/bug44653.phpt",
    "Zend/tests/bug47836.phpt",
    "Zend/tests/ns_086.phpt",
    "Zend/tests/get_called_class_001.phpt",
    "Zend/tests/bug38047.phpt",
    "Zend/tests/unset_cv01.phpt",
    "Zend/tests/bug48215.phpt",
    "Zend/tests/bug33802.phpt",
    "Zend/tests/ns_035.phpt",
    "Zend/tests/bug30820.phpt",
    "Zend/tests/compare_003.phpt",
    "Zend/tests/nowdoc_010.phpt",
    "Zend/tests/bug43344_9.phpt",
    "Zend/tests/objects_029.phpt",
    "Zend/tests/bug33558.phpt",
    "Zend/tests/debug_backtrace_options.phpt",
    "Zend/tests/class_alias_019.phpt",
    "Zend/tests/bug30702.phpt",
    "Zend/tests/bug35411.phpt",
    "Zend/tests/ns_030.phpt",
    "Zend/tests/access_modifiers_003.phpt",
    "Zend/tests/closure_012.phpt",
    "Zend/tests/bug34062.phpt",
    "Zend/tests/magic_methods_006.phpt",
    "Zend/tests/compare_001.phpt",
    "Zend/tests/objects_013.phpt",
    "Zend/tests/bug45744.phpt",
    "Zend/tests/bug43175.phpt",
    "Zend/tests/instanceof_002.phpt",
    "Zend/tests/nowdoc_005.phpt",
    "Zend/tests/bug38779_1.phpt",
    "Zend/tests/bug35509.phpt",
    "Zend/tests/bug52879.phpt",
    "Zend/tests/ns_079.phpt",
    "Zend/tests/003.phpt",
    "Zend/tests/bug46381.phpt",
    "Zend/tests/bug38220.phpt",
    "Zend/tests/lsb_007.phpt",
    "Zend/tests/ns_012.phpt",
    "Zend/tests/bug45180.phpt",
    "Zend/tests/compare_004.phpt",
    "Zend/tests/exception_005.phpt",
    "Zend/tests/closure_029.phpt",
    "Zend/tests/unset_cv12.phpt",
    "Zend/tests/ns_037.phpt",
    "Zend/tests/bug39003.phpt",
    "Zend/tests/list_007.phpt",
    "Zend/tests/bug45910.phpt",
    "Zend/tests/objects_014.phpt",
    "Zend/tests/bug41919.phpt",
    "Zend/tests/ns_053.phpt",
    "Zend/tests/jump08.phpt",
    "Zend/tests/cast_to_bool.phpt",
    "Zend/tests/bug40770.phpt",
    "Zend/tests/bug35017.phpt",
    "Zend/tests/class_alias_005.phpt",
    "Zend/tests/bug38469.phpt",
    "Zend/tests/bug43201.phpt",
    "Zend/tests/ns_062.phpt",
    "Zend/tests/nowdoc_015.phpt",
    "Zend/tests/bug51394.phpt",
    "Zend/tests/bug38287.phpt",
    "Zend/tests/call_static_004.phpt",
    "Zend/tests/bug41075.phpt",
    "Zend/tests/bug41633_4.phpt",
    "Zend/tests/bug43426.phpt",
    "Zend/tests/ns_061.phpt",
    "Zend/tests/errmsg_034.phpt",
    "Zend/tests/bug31177-2.phpt",
    "Zend/tests/int_overflow_32bit.phpt",
    "Zend/tests/bug28442.phpt",
    "Zend/tests/closure_037.phpt",
    "Zend/tests/errmsg_013.phpt",
    "Zend/tests/objects_009.phpt",
    "Zend/tests/halt_compiler4.phpt",
    "Zend/tests/bug48930.phpt",
    "Zend/tests/errmsg_009.phpt",
    "Zend/tests/bug43027.phpt",
    "Zend/tests/ns_077_5.phpt",
    "Zend/tests/heredoc_004.phpt",
    "Zend/tests/ns_022.phpt",
    "Zend/tests/020.phpt",
    "Zend/tests/ns_076.phpt",
    "Zend/tests/022.phpt",
    "Zend/tests/clone_004.phpt",
    "Zend/tests/bug42818.phpt",
    "Zend/tests/bug34467.phpt",
    "Zend/tests/bug42211.phpt",
    "Zend/tests/errmsg_033.phpt",
    "Zend/tests/006.phpt",
    "Zend/tests/031.phpt",
    "Zend/tests/magic_methods_005.phpt",
    "Zend/tests/exception_handler_003.phpt",
    "Zend/tests/bug34064.phpt",
    "Zend/tests/objects_010.phpt",
    "Zend/tests/bug31720.phpt",
    "Zend/tests/indexing_001.phpt",
    "Zend/tests/objects_032.phpt",
    "Zend/tests/constants_004.phpt",
    "Zend/tests/traits/error_013.phpt",
    "Zend/tests/traits/static_get_called_class.phpt",
    "Zend/tests/traits/methods_002.phpt",
    "Zend/tests/traits/property002.phpt",
    "Zend/tests/traits/get_declared_traits_001.phpt",
    "Zend/tests/traits/conflict003.phpt",
    "Zend/tests/traits/static_002.phpt",
    "Zend/tests/traits/error_001.phpt",
    "Zend/tests/traits/methods_003.phpt",
    "Zend/tests/traits/language008b.phpt",
    "Zend/tests/traits/methods_001.phpt",
    "Zend/tests/traits/interface_002.phpt",
    "Zend/tests/traits/conflict002.phpt",
    "Zend/tests/traits/bugs/case-sensitive.phpt",
    "Zend/tests/traits/bugs/abstract-methods01.phpt",
    "Zend/tests/traits/bugs/alias-semantics.phpt",
    "Zend/tests/traits/bugs/alias01.phpt",
    "Zend/tests/traits/bugs/missing-trait.phpt",
    "Zend/tests/traits/bugs/abstract-methods04.phpt",
    "Zend/tests/traits/bugs/interfaces.phpt",
    "Zend/tests/traits/bugs/alias-semantics02.phpt",
    "Zend/tests/traits/bugs/abstract-methods02.phpt",
    "Zend/tests/traits/bugs/overridding-conflicting-methods.phpt",
    "Zend/tests/traits/bugs/abstract-methods03.phpt",
    "Zend/tests/traits/error_016.phpt",
    "Zend/tests/traits/language001.phpt",
    "Zend/tests/traits/static_forward_static_call.phpt",
    "Zend/tests/traits/interface_003.phpt",
    "Zend/tests/traits/error_012.phpt",
    "Zend/tests/traits/inheritance003.phpt",
    "Zend/tests/traits/static_001.phpt",
    "Zend/tests/traits/error_004.phpt",
    "Zend/tests/traits/static_003.phpt",
    "Zend/tests/traits/property004.phpt",
    "Zend/tests/traits/language007.phpt",
    "Zend/tests/traits/error_002.phpt",
    "Zend/tests/traits/flattening002.phpt",
    "Zend/tests/traits/language010.phpt",
    "Zend/tests/traits/error_015.phpt",
    "Zend/tests/traits/get_declared_traits_003.phpt",
    "Zend/tests/traits/language002.phpt",
    "Zend/tests/traits/error_005.phpt",
    "Zend/tests/traits/error_009.phpt",
    "Zend/tests/traits/language003.phpt",
    "Zend/tests/traits/error_008.phpt",
    "Zend/tests/traits/language009.phpt",
    "Zend/tests/traits/property001.phpt",
    "Zend/tests/traits/language005.phpt",
    "Zend/tests/traits/error_014.phpt",
    "Zend/tests/traits/language004.phpt",
    "Zend/tests/traits/language011.phpt",
    "Zend/tests/traits/language013.phpt",
    "Zend/tests/traits/error_003.phpt",
    "Zend/tests/traits/error_010.phpt",
    "Zend/tests/traits/conflict001.phpt",
    "Zend/tests/traits/noctor001.phpt",
    "Zend/tests/traits/language008a.phpt",
    "Zend/tests/traits/error_011.phpt",
    "Zend/tests/traits/inheritance001.phpt",
    "Zend/tests/traits/static_004.phpt",
    "Zend/tests/traits/error_006.phpt",
    "Zend/tests/traits/flattening003.phpt",
    "Zend/tests/traits/interface_001.phpt",
    "Zend/tests/traits/get_declared_traits_002.phpt",
    "Zend/tests/traits/inheritance002.phpt",
    "Zend/tests/traits/error_007.phpt",
    "Zend/tests/traits/property005.phpt",
    "Zend/tests/traits/property003.phpt",
    "Zend/tests/traits/flattening001.phpt",
    "Zend/tests/traits/language012.phpt",
    "Zend/tests/traits/language006.phpt",
    "Zend/tests/unset_cv10.phpt",
    "Zend/tests/ns_048.phpt",
    "Zend/tests/bug40621.phpt",
    "Zend/tests/023.phpt",
    "Zend/tests/heredoc_008.phpt",
    "Zend/tests/exception_007.phpt",
    "Zend/tests/halt_compiler3.phpt",
    "Zend/tests/ns_077_1.phpt",
    "Zend/tests/bug37715.phpt",
    "Zend/tests/ns_031.phpt",
    "Zend/tests/gc_019.phpt",
    "Zend/tests/dereference_006.phpt",
    "Zend/tests/heredoc_005.phpt",
    "Zend/tests/bug47343.phpt",
    "Zend/tests/catch.phpt",
    "Zend/tests/jump10.phpt",
    "Zend/tests/bug39775.phpt",
    "Zend/tests/bug52484_3.phpt",
    "Zend/tests/or_001.phpt",
    "Zend/tests/ns_077_2.phpt",
    "Zend/tests/bug48770.phpt",
    "Zend/tests/error_reporting03.phpt",
    "Zend/tests/bug47320.phpt",
    "Zend/tests/magic_by_ref_004.phpt",
    "Zend/tests/inter_02.phpt",
    "Zend/tests/bug48408.phpt",
    "Zend/tests/bug38808.phpt",
    "Zend/tests/bug33282.phpt",
    "Zend/tests/div_002.phpt",
    "Zend/tests/errmsg_036.phpt",
    "Zend/tests/ns_051.phpt",
    "Zend/tests/gc_021.phpt",
    "Zend/tests/bug46308.phpt",
    "Zend/tests/errmsg_014.phpt",
    "Zend/tests/bug46196.phpt",
    "Zend/tests/errmsg_029.phpt",
    "Zend/tests/bug36759.phpt",
    "Zend/tests/method_exists_002.phpt",
    "Zend/tests/jump12.phpt",
    "Zend/tests/dereference_002.phpt",
    "Zend/tests/errmsg_038.phpt",
    "Zend/tests/bug30707.phpt",
    "Zend/tests/closure_022.phpt",
    "Zend/tests/ns_013.phpt",
    "Zend/tests/cast_to_array.phpt",
    "Zend/tests/ns_024.phpt",
    "Zend/tests/closure_034.phpt",
    "Zend/tests/get_class_methods_001.phpt",
    "Zend/tests/bug34617.phpt",
    "Zend/tests/ns_038.phpt",
    "Zend/tests/class_constants_003.phpt",
    "Zend/tests/bug47593.phpt",
    "Zend/tests/offset_null.phpt",
    "Zend/tests/ns_046.phpt",
    "Zend/tests/add_007.phpt",
    "Zend/tests/bug32674.phpt",
    "Zend/tests/class_alias_020.phpt",
    "Zend/tests/bug34873.phpt",
    "Zend/tests/bug48533.phpt",
    "Zend/tests/add_006.phpt",
    "Zend/tests/errmsg_037.phpt",
    "Zend/tests/bug39297.phpt",
    "Zend/tests/ns_001.phpt",
    "Zend/tests/get_class_vars_003.phpt",
    "Zend/tests/nowdoc_011.phpt",
    "Zend/tests/magic_by_ref_001.phpt",
    "Zend/tests/offset_assign.phpt",
    "Zend/tests/trait_exists_001.phpt",
    "Zend/tests/ns_078.phpt",
    "Zend/tests/compare_002_64bit.phpt",
    "Zend/tests/get_defined_functions_basic.phpt",
    "Zend/tests/bug26801.phpt",
    "Zend/tests/closure_006.phpt",
    "Zend/tests/ns_087.phpt",
    "Zend/tests/bug39944.phpt",
    "Zend/tests/bug22725.phpt",
    "Zend/tests/bug34879.phpt",
    "Zend/tests/lsb_013.phpt",
    "Zend/tests/exception_009.phpt",
    "Zend/tests/ns_077_4.phpt",
    "Zend/tests/bug40815.phpt",
    "Zend/tests/errmsg_003.phpt",
    "Zend/tests/constants_009.phpt",
    "Zend/tests/indirect_reference_this.phpt",
    "Zend/tests/bug35163_2.phpt",
    "Zend/tests/call_user_func_004.phpt",
    "Zend/tests/ns_054.phpt",
    "Zend/tests/closure_018.phpt",
    "Zend/tests/bug46665.phpt",
    "Zend/tests/call_static_005.phpt",
    "Zend/tests/bug33996.phpt",
    "Zend/tests/bug28377.phpt",
    "Zend/tests/bug32799.phpt",
    "Zend/tests/bug33318.phpt",
    "Zend/tests/closure_036.phpt",
    "Zend/tests/bug27598.phpt",
    "Zend/tests/exception_002.phpt",
    "Zend/tests/multibyte/multibyte_encoding_004.phpt",
    "Zend/tests/multibyte/multibyte_encoding_005.phpt",
    "Zend/tests/multibyte/multibyte_encoding_003.phpt",
    "Zend/tests/multibyte/multibyte_encoding_006.phpt",
    "Zend/tests/multibyte/multibyte_encoding_002.phpt",
    "Zend/tests/multibyte/multibyte_encoding_001.phpt",
    "Zend/tests/get_class_vars_007.phpt",
    "Zend/tests/jump13.phpt",
    "Zend/tests/007.phpt",
    "Zend/tests/bug31341.phpt",
    "Zend/tests/unset_cv06.phpt",
    "Zend/tests/bug48899.phpt",
    "Zend/tests/bug33257.phpt",
    "Zend/tests/str_offset_001.phpt",
    "Zend/tests/ns_019.phpt",
    "Zend/tests/and_001.phpt",
    "Zend/tests/magic_by_ref_007.phpt",
    "Zend/tests/function_exists_variation1.phpt",
    "Zend/tests/error_reporting05.phpt",
    "Zend/tests/bug44660.phpt",
    "Zend/tests/float_prec_001.phpt",
    "Zend/tests/nowdoc_009.phpt",
    "Zend/tests/nowdoc_004.phpt",
    "Zend/tests/objects_027.phpt",
    "Zend/tests/bug32427.phpt",
    "Zend/tests/array_type_hint_001.phpt",
    "Zend/tests/int_underflow_32bit.phpt",
    "Zend/tests/dereference_005.phpt",
    "Zend/tests/objects_005.phpt",
    "Zend/tests/bug29944.phpt",
    "Zend/tests/double_to_string_64bit.phpt",
    "Zend/tests/settype_resource.phpt",
    "Zend/tests/get_class_vars_002.phpt",
    "Zend/tests/bug44899.phpt",
    "Zend/tests/bug47704.phpt",
    "Zend/tests/is_a.phpt",
    "Zend/tests/errmsg_020.phpt",
    "Zend/tests/bug37251.phpt",
    "Zend/tests/unset_cv02.phpt",
    "Zend/tests/bug42819.phpt",
    "Zend/tests/bug34893.phpt",
    "Zend/tests/ns_049.phpt",
    "Zend/tests/bug48667_2.phpt",
    "Zend/tests/bug52193.phpt",
    "Zend/tests/objects_023.phpt",
    "Zend/tests/bug43183.phpt",
    "Zend/tests/bug54585.phpt",
    "Zend/tests/bug41633_3.phpt",
    "Zend/tests/errmsg_024.phpt",
    "Zend/tests/settype_int.phpt",
    "Zend/tests/closure_001.phpt",
    "Zend/tests/ns_075.phpt",
    "Zend/tests/dereference_004.phpt",
    "Zend/tests/catch_003.phpt",
    "Zend/tests/bug32226.phpt",
    "Zend/tests/errmsg_021.phpt",
    "Zend/tests/sub_001.phpt",
    "Zend/tests/shift_002.phpt",
    "Zend/tests/closure_030.phpt",
    "Zend/tests/dtor_scope.phpt",
    "Zend/tests/errmsg_004.phpt",
    "Zend/tests/bug33171.phpt",
    "Zend/tests/inter_01.phpt",
    "Zend/tests/call_with_refs.phpt",
    "Zend/tests/access_modifiers_012.phpt",
    "Zend/tests/002.phpt",
    "Zend/tests/gc_004.phpt",
    "Zend/tests/bug49269.phpt",
    "Zend/tests/declare_001.phpt",
    "Zend/tests/bug48693.phpt",
    "Zend/tests/ns_057.phpt",
    "Zend/tests/objects_004.phpt",
    "Zend/tests/008.phpt",
    "Zend/tests/bug49908.phpt",
    "Zend/tests/method_exists.phpt",
    "Zend/tests/gc_023.phpt",
    "Zend/tests/bug47801.phpt",
    "Zend/tests/bug52614.phpt",
    "Zend/tests/closure_002.phpt",
    "Zend/tests/constants_005.phpt",
    "Zend/tests/lsb_008.phpt",
    "Zend/tests/018.phpt",
    "Zend/tests/magic_methods_002.phpt",
    "Zend/tests/class_alias_008.phpt",
    "Zend/tests/bug39018.phpt",
    "Zend/tests/exception_004.phpt",
    "Zend/tests/errmsg_015.phpt",
    "Zend/tests/closure_031.phpt",
    "Zend/tests/bug52060.phpt",
    "Zend/tests/ns_072.phpt",
    "Zend/tests/ns_071.phpt",
    "Zend/tests/bug38461.phpt",
    "Zend/tests/bug30140.phpt",
    "Zend/tests/bug43344_5.phpt",
    "Zend/tests/jump02.phpt",
    "Zend/tests/bug43851.phpt",
    "Zend/tests/debug_print_backtrace_limit.phpt",
    "Zend/tests/bug24773.phpt",
    "Zend/tests/error_reporting07.phpt",
    "Zend/tests/bug45186.phpt",
    "Zend/tests/bug48228.phpt",
    "Zend/tests/objects_018.phpt",
    "Zend/tests/closure_014.phpt",
    "Zend/tests/clone_002.phpt",
    "Zend/tests/034.phpt",
    "Zend/tests/ns_083.phpt",
    "Zend/tests/gc_002.phpt",
    "Zend/tests/objects_022.phpt",
    "Zend/tests/lsb_017.phpt",
    "Zend/tests/bug49472.phpt",
    "Zend/tests/bug38211.phpt",
    "Zend/tests/035.phpt",
    "Zend/tests/compare_005.phpt",
    "Zend/tests/errmsg_022.phpt",
    "Zend/tests/closure_035.phpt",
    "Zend/tests/bug42820.phpt",
    "Zend/tests/bug47981.phpt",
    "Zend/tests/bug30791.phpt",
    "Zend/tests/access_modifiers_010.phpt",
    "Zend/tests/bug24884.phpt",
    "Zend/tests/ns_004.phpt",
    "Zend/tests/compare_006_64bit.phpt",
    "Zend/tests/closure_017.phpt",
    "Zend/tests/class_alias_009.phpt",
    "Zend/tests/lsb_022.phpt",
    "Zend/tests/bug47572.phpt",
    "Zend/tests/bug42802.phpt",
    "Zend/tests/ns_034.phpt",
    "Zend/tests/bug27731.phpt",
    "Zend/tests/exception_010.phpt",
    "Zend/tests/nowdoc_007.phpt",
    "Zend/tests/bug34137.phpt",
    "Zend/tests/unexpected_ref_bug.phpt",
    "Zend/tests/bug40784.phpt",
    "Zend/tests/013.phpt",
    "Zend/tests/lsb_002.phpt",
    "Zend/tests/ns_085.phpt",
    "Zend/tests/037.phpt",
    "Zend/tests/016.phpt",
    "Zend/tests/bug48667_1.phpt",
    "Zend/tests/list_004.phpt",
    "Zend/tests/bug52484_2.phpt",
    "Zend/tests/ns_041.phpt",
    "Zend/tests/bug43344_3.phpt",
    "Zend/tests/xor_002.phpt",
    "Zend/tests/nowdoc_003.phpt",
    "Zend/tests/errmsg_019.phpt",
    "Zend/tests/bug37632.phpt",
    "Zend/tests/bug29505.phpt",
    "Zend/tests/bug50005.phpt",
    "Zend/tests/errmsg_012.phpt",
    "Zend/tests/gc_001.phpt",
    "Zend/tests/bug39304_2_4.phpt",
    "Zend/tests/zend_strtod.phpt",
    "Zend/tests/bug44414.phpt",
    "Zend/tests/bug31177.phpt",
    "Zend/tests/bug53305.phpt",
    "Zend/tests/heredoc_016.phpt",
    "Zend/tests/constants_001.phpt",
    "Zend/tests/errmsg_001.phpt",
    "Zend/tests/errmsg_006.phpt",
    "Zend/tests/bug48248.phpt",
    "Zend/tests/class_constants_004.phpt",
    "Zend/tests/bug51421.phpt",
    "Zend/tests/xor_003.phpt",
    "Zend/tests/bug29896.phpt",
    "Zend/tests/bug26010.phpt",
    "Zend/tests/bug_debug_backtrace.phpt",
    "Zend/tests/bug43344_10.phpt",
    "Zend/tests/call_user_func_005.phpt",
    "Zend/tests/get_class_methods_002.phpt",
    "Zend/tests/gc_024.phpt",
    "Zend/tests/bug36006.phpt",
    "Zend/tests/bug40705.phpt",
    "Zend/tests/bug29674.phpt",
    "Zend/tests/objects_002.phpt",
    "Zend/tests/constants_002.phpt",
    "Zend/tests/access_modifiers_007.phpt",
    "Zend/tests/bug34518.phpt",
    "Zend/tests/ns_016.phpt",
    "Zend/tests/bug37667.phpt",
    "Zend/tests/exception_handler_001.phpt",
    "Zend/tests/magic_by_ref_002.phpt",
    "Zend/tests/025.phpt",
    "Zend/tests/heredoc_012.phpt",
    "Zend/tests/foreach_unset_globals.phpt",
    "Zend/tests/dereference_001.phpt",
    "Zend/tests/bug47714.phpt",
    "Zend/tests/unset_cv07.phpt",
    "Zend/tests/gc_013.phpt",
    "Zend/tests/constants/dir-constant-eval.phpt",
    "Zend/tests/constants/dir-constant-nested_includes.phpt",
    "Zend/tests/constants/dir-constant-includes.phpt",
    "Zend/tests/constants/dir-constant-normal.phpt",
    "Zend/tests/bug41372.phpt",
    "Zend/tests/lsb_003.phpt",
    "Zend/tests/bug38779.phpt",
    "Zend/tests/bug45910_2.phpt",
    "Zend/tests/012.phpt",
    "Zend/tests/objects_028.phpt",
    "Zend/tests/bug30080.phpt",
    "Zend/tests/038.phpt",
    "Zend/tests/call_static_003.phpt",
    "Zend/tests/bug32428.phpt",
    "Zend/tests/bug51791.phpt",
    "Zend/tests/bug34310.phpt",
    "Zend/tests/ns_069.phpt",
    "Zend/tests/bug38624.phpt",
    "Zend/tests/bug43332_2.phpt",
    "Zend/tests/not_001.phpt",
    "Zend/tests/errmsg_025.phpt",
    "Zend/tests/str_offset_002.phpt",
    "Zend/tests/errmsg_035.phpt",
    "Zend/tests/dereference_014.phpt",
    "Zend/tests/class_alias_002.phpt",
    "Zend/tests/bug53629.phpt",
    "Zend/tests/abstract-static.phpt",
    "Zend/tests/closure_016.phpt",
    "Zend/tests/closure_008.phpt",
    "Zend/tests/errmsg_027.phpt",
    "Zend/tests/unset_cv04.phpt",
    "Zend/tests/ns_025.phpt",
    "Zend/tests/ns_055.phpt",
    "Zend/tests/021.phpt",
    "Zend/tests/bug23104.phpt",
    "Zend/tests/bug34260.phpt",
    "Zend/tests/dereference_013.phpt",
    "Zend/tests/debug_backtrace_limit.phpt",
    "Zend/tests/bug30161.phpt",
    "Zend/tests/halt_compiler2.phpt",
    "Zend/tests/class_alias_003.phpt",
    "Zend/tests/class_alias_016.phpt",
    "Zend/tests/ns_027.phpt",
    "Zend/tests/ns_066.phpt",
    "Zend/tests/add_001.phpt",
    "Zend/tests/ns_060.phpt",
    "Zend/tests/call_static_002.phpt",
    "Zend/tests/jump07.phpt",
    "Zend/tests/closure_028.phpt",
    "Zend/tests/interface_exists_002.phpt",
    "Zend/tests/bug34678.phpt",
    "Zend/tests/bug53632.phpt",
    "Zend/tests/bug35106.phpt",
    "Zend/tests/array_unshift_COW.phpt",
    "Zend/tests/bug45742.phpt",
    "Zend/tests/bug37144.phpt",
    "Zend/tests/gc_017.phpt",
    "Zend/tests/heredoc_003.phpt",
    "Zend/tests/ns_052.phpt",
    "Zend/tests/halt_compiler1.phpt",
    "Zend/tests/bug47596.phpt",
    "Zend/tests/bug46811.phpt",
    "Zend/tests/assign_to_var_003.phpt",
    "Zend/tests/ns_077_8.phpt",
    "Zend/tests/closure_013.phpt",
    "Zend/tests/ns_056.phpt",
    "Zend/tests/bug43344_11.phpt",
    "Zend/tests/bug41351_3.phpt",
    "Zend/tests/int_overflow_64bit.phpt",
    "Zend/tests/bug53971.phpt",
    "Zend/tests/zend_operators.phpt",
    "Zend/tests/list_005.phpt",
    "Zend/tests/lsb_004.phpt",
    "Zend/tests/bug50810.phpt",
    "Zend/tests/settype_double.phpt",
    "Zend/tests/errmsg_005.phpt",
    "Zend/tests/errmsg_031.phpt",
    "Zend/tests/gc_022.phpt",
    "Zend/tests/ns_007.phpt",
    "Zend/tests/access_modifiers_009.phpt",
    "Zend/tests/objects_003.phpt",
    "Zend/tests/offset_array.phpt",
    "Zend/tests/assign_to_var_001.phpt",
    "Zend/tests/class_constants_001.phpt",
    "Zend/tests/bug40261.phpt",
    "Zend/tests/assign_to_var_004.phpt",
    "Zend/tests/lsb_006.phpt",
    "Zend/tests/settype_array.phpt",
    "Zend/tests/closure_020.phpt",
    "Zend/tests/class_alias_013.phpt",
    "Zend/tests/shift_001.phpt",
    "Zend/tests/ns_032.phpt",
    "Zend/tests/class_alias_015.phpt",
    "Zend/tests/bug47054.phpt",
    "Zend/tests/constants_006.phpt",
    "Zend/tests/bug41961.phpt",
    "Zend/tests/bug50261.phpt",
    "Zend/tests/gc_009.phpt",
    "Zend/tests/gc_005.phpt",
    "Zend/tests/ns_077_6.phpt",
    "Zend/tests/bug52041.phpt",
    "Zend/tests/closure_009.phpt",
    "Zend/tests/compare_004_64bit.phpt",
    "Zend/tests/ns_059.phpt",
    "Zend/tests/lsb_016.phpt",
    "Zend/tests/get_class_methods_003.phpt",
    "Zend/tests/access_modifiers_006.phpt",
    "Zend/tests/bug26281.phpt",
    "Zend/tests/bug39990.phpt",
    "Zend/tests/bug36268.phpt",
    "Zend/tests/bug41633_2.phpt",
    "Zend/tests/magic_methods_008.phpt",
    "Zend/tests/objects_020.phpt",
    "Zend/tests/bug35634.phpt",
    "Zend/tests/bug42772.phpt",
    "Zend/tests/closure_023.phpt",
    "Zend/tests/bug27304.phpt",
    "Zend/tests/bug44830.phpt",
    "Zend/tests/bug19859.phpt",
    "Zend/tests/jump14.phpt",
    "Zend/tests/bug54262.phpt",
    "Zend/tests/bug26166.phpt",
    "Zend/tests/bug30725.phpt",
    "Zend/tests/030.phpt",
    "Zend/tests/jump03.phpt",
    "Zend/tests/bug43128.phpt",
    "Zend/tests/bug44069.phpt",
    "Zend/tests/bug30394.phpt",
    "Zend/tests/compare_006.phpt",
    "Zend/tests/nowdoc_017.phpt",
    "Zend/tests/double_to_string.phpt",
    "Zend/tests/bug37046.phpt",
    "Zend/tests/lsb_018.phpt",
    "Zend/tests/bug52051.phpt",
    "Zend/tests/lsb_021.phpt",
    "Zend/tests/bug41640.phpt",
    "Zend/tests/jump04.phpt",
    "Zend/tests/access_modifiers_008.phpt",
    "Zend/tests/bug46304.phpt",
    "Zend/tests/gc_018.phpt",
    "Zend/tests/magic_by_ref_003.phpt",
    "Zend/tests/bug34045.phpt",
    "Zend/tests/bug50255.phpt",
    "Zend/tests/ns_020.phpt",
    "Zend/tests/gc_027.phpt",
    "Zend/tests/bug32322.phpt",
    "Zend/tests/bug45186_2.phpt",
    "Zend/tests/bug39438.phpt",
    "Zend/tests/gc_029.phpt",
    "Zend/tests/ns_014.phpt",
    "Zend/tests/add_002.phpt",
    "Zend/tests/bug52001.phpt",
    "Zend/tests/bug48770_3.phpt",
    "Zend/tests/concat_001.phpt",
    "Zend/tests/each_002.phpt",
    "Zend/tests/bug39018_2.phpt",
    "Zend/tests/bug42767.phpt",
    "Zend/tests/catch_002.phpt",
    "Zend/tests/magic_methods_003.phpt",
    "Zend/tests/errmsg_017.phpt",
    "Zend/tests/cast_to_object.phpt",
    "Zend/tests/foreach_002.phpt",
    "Zend/tests/bug33512.phpt",
    "Zend/tests/bug35655.phpt",
    "Zend/tests/bug48912.phpt",
    "Zend/tests/bug45147.phpt",
    "Zend/tests/032.phpt",
    "Zend/tests/bug31098.phpt",
    "Zend/tests/bug29368.phpt",
    "Zend/tests/ns_058.phpt",
    "Zend/tests/bug47165.phpt",
    "Zend/tests/objects_017.phpt",
    "Zend/tests/closure_027.phpt",
    "Zend/tests/ns_063.phpt",
    "Zend/tests/bug50174.phpt",
    "Zend/tests/ns_018.phpt",
    "Zend/tests/error_reporting01.phpt",
    "Zend/tests/settype_object.phpt",
    "Zend/tests/errmsg_039.phpt",
    "Zend/tests/constructor_args.phpt",
    "Zend/tests/errmsg_026.phpt",
    "Zend/tests/bug35163.phpt",
    "Zend/tests/access_modifiers_011.phpt",
    "Zend/tests/magic_methods_004.phpt",
    "Zend/tests/ns_021.phpt",
    "Zend/tests/019.phpt",
    "Zend/tests/nowdoc_014.phpt",
    "Zend/tests/unset_cv03.phpt",
    "Zend/tests/list_002.phpt",
    "Zend/tests/lsb_015.phpt",
    "Zend/tests/bug47771.phpt",
    "Zend/tests/magic_methods_010.phpt",
    "Zend/tests/bug44827.phpt",
    "Zend/tests/bug33771.phpt",
    "Zend/tests/magic_methods_009.phpt",
    "Zend/tests/bug37212.phpt",
    "Zend/tests/inter_007.phpt",
    "Zend/tests/bug41401.phpt",
    "Zend/tests/gc_015.phpt",
    "Zend/tests/bug35470.phpt",
    "Zend/tests/inter_05.phpt",
    "Zend/tests/selfParent_001.phpt",
    "Zend/tests/bug43450.phpt",
    "Zend/tests/halt03.phpt",
    "Zend/tests/each_003.phpt",
    "Zend/tests/lsb_011.phpt",
    "Zend/tests/compare_001_64bit.phpt",
    "Zend/tests/errmsg_042.phpt",
    "Zend/tests/bug39127.phpt",
    "Zend/tests/bug30998.phpt",
    "Zend/tests/globals_002.phpt",
    "Zend/tests/magic_methods_001.phpt",
    "Zend/tests/bug48409.phpt",
    "Zend/tests/objects_031.phpt",
    "Zend/tests/closure_010.phpt",
    "Zend/tests/nowdoc_012.phpt",
    "Zend/tests/bug34786.phpt",
    "Zend/tests/not_002.phpt",
    "Zend/tests/class_alias_021.phpt",
    "Zend/tests/ns_005.phpt",
    "Zend/tests/ns_015.phpt",
    "Zend/tests/bug46409.phpt",
    "Zend/tests/024.phpt",
    "Zend/tests/bug33277.phpt",
    "Zend/tests/bug45178.phpt",
    "Zend/tests/inter_04.phpt",
    "Zend/tests/error_reporting02.phpt",
    "Zend/tests/bug43344_4.phpt",
    "Zend/tests/nowdoc_001.phpt",
    "Zend/tests/dynamic_call_001.phpt",
    "Zend/tests/instanceof.phpt",
    "Zend/tests/bug35163_3.phpt",
    "Zend/tests/strict_001.phpt",
    "Zend/tests/bug37707.phpt",
    "Zend/tests/009.phpt",
    "Zend/tests/unset_cv11.phpt",
    "Zend/tests/bug45862.phpt",
    "Zend/tests/bug21478.phpt",
    "Zend/tests/bug46701.phpt",
    "Zend/tests/bug39346.phpt",
    "Zend/tests/bug46813.phpt",
    "Zend/tests/get_defined_functions_error.phpt",
    "Zend/tests/dereference_011.phpt",
    "Zend/tests/ns_050.phpt",
    "Zend/tests/call_user_func_003.phpt",
    "Zend/tests/get_class_vars_005.phpt",
    "Zend/tests/result_unused.phpt",
    "Zend/tests/ns_080.phpt",
    "Zend/tests/bug42859.phpt",
    "Zend/tests/errmsg_008.phpt",
    "Zend/tests/objects_012.phpt",
    "Zend/tests/bug42937.phpt",
    "Zend/tests/ns_033.phpt",
    "Zend/tests/class_alias_004.phpt",
    "Zend/tests/magic_by_ref_006.phpt",
    "Zend/tests/bug30519.phpt",
    "Zend/tests/objects_025.phpt",
    "Zend/tests/bug41633_1.phpt",
    "Zend/tests/clone_001.phpt",
    "Zend/tests/bug47109.phpt",
    "Zend/tests/bug39721.phpt",
    "Zend/tests/ns_073.phpt",
    "Zend/tests/magic_methods_007.phpt",
    "Zend/tests/gc_028.phpt",
    "Zend/tests/bug39825.phpt",
    "Zend/tests/errmsg_010.phpt",
    "Zend/tests/nowdoc_006.phpt",
    "Zend/tests/trait_exists_002.phpt",
    "Zend/tests/cast_to_double.phpt",
    "Zend/tests/hex_overflow_32bit.phpt",
    "Zend/tests/bug24436.phpt",
    "Zend/tests/errmsg_002.phpt",
    "Zend/tests/isset_003.phpt",
    "sapi/cgi/tests/010.phpt",
    "sapi/cgi/tests/005.phpt",
    "sapi/cgi/tests/011.phpt",
    "sapi/cgi/tests/001.phpt",
    "sapi/cgi/tests/004.phpt",
    "sapi/cgi/tests/003.phpt",
    "sapi/cgi/tests/006.phpt",
    "sapi/cgi/tests/007.phpt",
    "sapi/cgi/tests/002.phpt",
    "sapi/cgi/tests/008.phpt",
    "sapi/cgi/tests/009.phpt",
    "sapi/tests/test004.phpt",
    "sapi/tests/test007.phpt",
    "sapi/tests/test001.phpt",
    "sapi/tests/test005.phpt",
    "sapi/tests/test002.phpt",
    "sapi/tests/test006.phpt",
    "sapi/tests/test003.phpt",
    "sapi/cli/tests/002-win32.phpt",
    "sapi/cli/tests/010.phpt",
    "sapi/cli/tests/005.phpt",
    "sapi/cli/tests/014.phpt",
    "sapi/cli/tests/011.phpt",
    "sapi/cli/tests/003-2.phpt",
    "sapi/cli/tests/017.phpt",
    "sapi/cli/tests/010-2.phpt",
    "sapi/cli/tests/001.phpt",
    "sapi/cli/tests/015.phpt",
    "sapi/cli/tests/004.phpt",
    "sapi/cli/tests/003.phpt",
    "sapi/cli/tests/020.phpt",
    "sapi/cli/tests/022.phpt",
    "sapi/cli/tests/006.phpt",
    "sapi/cli/tests/007.phpt",
    "sapi/cli/tests/002.phpt",
    "sapi/cli/tests/008.phpt",
    "sapi/cli/tests/018.phpt",
    "sapi/cli/tests/013.phpt",
    "sapi/cli/tests/016.phpt",
    "sapi/cli/tests/012.phpt",
    "sapi/cli/tests/021.phpt",
    "sapi/cli/tests/bug44564.phpt",
    "sapi/cli/tests/019.phpt",
    "sapi/cli/tests/009.phpt",
};

int run_test(char* name) {
    char buffer[1024];
    sprintf(buffer, "./sapi/cli/php ../php-helper.php -p ./sapi/cli/php -q %s", name);
    int killed = system("killall php &> /dev/null");
    if (killed == 0) {
        system("echo A php process was killed > A-PHP-Process-Was-Killed");    }
    int res = system(buffer);

    if (res == 0) {
        printf("PASS: %s\n", name);
        return 0;
    } else {
        printf("FAIL: %s\n", name);
        return 1;
    }
}

/*
 * Usage:
 * ./php-run-tests TESTNUM [True]
 * ./php-run-tests length [True]
 *     TESTNUM is an integer specifying the test to run. If 'length' is given,
 *       we return the total number of tests to stdout then exit.
 *     The optional third argument should be True if we want to run
 *       abreviated tests.
*/
int main(int argc, char** argv) {
    // If we wanted the length, atoi will return 0 which is just fine
    int num = atoi(argv[1]) - 1;

    char** tests;
    int length;
    tests = TESTS;
    length = sizeof(TESTS)/sizeof(char*);
    // Check if we just wanted the number of tests, if so give it and die
    if (argv[1][0] == 'l') {
        printf("%d", length);
        exit(0);
    }

    if (num < 0 || num >= length) {
        fprintf(stderr, "Invalid test number: %d\n", num);
        exit(255);
    }
    char* name = tests[num];
    return run_test(name);
}
